package uk.ac.ox.cs.pdq.algebra;

import java.util.List;
import java.util.Objects;

import uk.ac.ox.cs.pdq.db.AccessMethod;
import uk.ac.ox.cs.pdq.db.AccessMethod.AccessMethodTypes;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.plan.AccessOperator;
import uk.ac.ox.cs.pdq.util.TupleType;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;

/**
 * Logical operator representation of an scan (input free access).
 *
 * @author Julien Leblay
 */
public class Scan extends RelationalOperator implements AccessOperator {

	/** The input table of the access. */
	private final Relation relation;

	/** The free access method associated with this scan. */
	private final AccessMethod binding;

	/** The columns. */
	protected List<Term> columns;

	/**
	 * Instantiates a new projection.
	 *
	 * @param relation Relation
	 */
	public Scan(Relation relation) {
		this(relation, findFreeAccessMethod(relation));
	}

	/**
	 * @param relation Relation
	 * @return AccessMethod
	 */
	private static AccessMethod findFreeAccessMethod(Relation relation) {
		for (AccessMethod b: relation.getAccessMethods()) {
			if (b.getType() == AccessMethodTypes.FREE) {
				return b;
			}
		}
		Preconditions.checkState(false, "Attempted to the instantiate a Scan operator from a relation without free access.");
		return null;
	}

	/**
	 * Instantiates a new projection.
	 *
	 * @param relation Relation
	 * @param accessMethod AccessMethod
	 */
	protected Scan(Relation relation, AccessMethod accessMethod) {
		super(TupleType.DefaultFactory.createFromTyped(relation.getAttributes()));
		Preconditions.checkArgument(accessMethod != null);
		Preconditions.checkArgument(accessMethod.getType() == AccessMethodTypes.FREE);
		Preconditions.checkArgument(relation != null);
		this.relation = relation;
		this.binding = accessMethod;
		this.columns = Utility.typedToTerms(relation.getAttributes());
	}

	/**
	 * @return the relation scanned by the operator
	 * @see uk.ac.ox.cs.pdq.plan.AccessOperator#getRelation()
	 */
	@Override
	public Relation getRelation() {
		return this.relation;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.plan.relational.logical.LogicalOperator#deepCopy()
	 */
	@Override
	public Scan deepCopy() throws RelationalOperatorException {
		return new Scan(this.relation);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName()).append('(');
		result.append(this.relation.getName()).append(')');
		return result.toString();
	}

	/**
	 * @return Integer
	 */
	@Override
	public Integer getDepth() {
		return 1;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.plan.relational.logical.LogicalOperator#getColumn(int)
	 */
	@Override
	public Term getColumn(int i) {
		return this.columns.get(i);
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.plan.relational.logical.LogicalOperator#getColumns()
	 */
	@Override
	public List<Term> getColumns() {
		return this.columns;
	}

	/**
	 * @return List<Term>
	 */
	@Override
	public List<Term> getInputTerms() {
		return Lists.newArrayList();
	}

	/**
	 * @param o Object
	 * @return boolean
	 */
	@Override
	public boolean equals(Object o) {
		return super.equals(o)
				&& this.getClass().isInstance(o)
				&& this.relation.equals(((Scan) o).relation)
				&& this.columns.equals(((Scan) o).columns);

	}

	/**
	 * @return int
	 */
	@Override
	public int hashCode() {
		return Objects.hash(this.outputType, this.inputType, this.columns,
				this.relation, this.metadata);
	}

	/**
	 * @return AccessMethod
	 * @see uk.ac.ox.cs.pdq.plan.AccessOperator#getBinding()
	 */
	@Override
	public AccessMethod getBinding() {
		return this.binding;
	}

	/**
	 * @return boolean
	 */
	@Override
	public boolean isClosed() {
		return true;
	}

	/**
	 * @return boolean
	 */
	@Override
	public boolean isQuasiLeaf() {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.plan.relational.logical.LogicalOperator#isLeftDeep()
	 */
	@Override
	public boolean isLeftDeep() {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.plan.relational.logical.LogicalOperator#isRightDeep()
	 */
	@Override
	public boolean isRightDeep() {
		return true;
	}
}
