package uk.ac.ox.cs.pdq.sql;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.builder.BuilderException;
import uk.ac.ox.cs.pdq.db.Attribute;
import uk.ac.ox.cs.pdq.db.LinearGuarded;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.db.View;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Skolem;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.sql.AbstractSQLSchemaDiscoverer;
import uk.ac.ox.cs.pdq.sql.PostgresqlRelationWrapper;
import uk.ac.ox.cs.pdq.sql.PostgresqlSchemaDiscoverer;
import uk.ac.ox.cs.pdq.sql.PostgresqlViewWrapper;
import uk.ac.ox.cs.pdq.util.Triple;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.Lists;

/**
 * PostgresqlTranslator-specific class for automatically discovering a schema.
 *  
 * @author Julien Leblay
 *
 */
public class PostgresqlSchemaDiscoverer extends AbstractSQLSchemaDiscoverer {

	/** The logger */
	public static Logger log = Logger.getLogger(PostgresqlSchemaDiscoverer.class);

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.discovery.sql.AbstractSQLSchemaDiscoverer#getRelationsDiscoveryStatement(java.lang.String)
	 */
	@Override
	protected String getRelationsDiscoveryStatement(String databaseName) {
		return "SELECT table_name "
				+ "FROM information_schema.tables "
				+ "WHERE table_schema = 'public' "
				+ "AND table_catalog = '" + databaseName + "' "
				+ "ORDER BY table_name";
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.discovery.sql.AbstractSQLSchemaDiscoverer#getForeignKeyDiscoveryStatement(java.lang.String)
	 */
	@Override
	protected String getForeignKeyDiscoveryStatement(String relationName) {
		return "select c.constraint_name "
				+ "    , x.column_name "
				+ "    , y.table_name as REFERENCED_TABLE_NAME "
				+ "    , y.column_name as REFERENCED_COLUMN_NAME "
				+ "from information_schema.referential_constraints c "
				+ "join information_schema.key_column_usage x "
				+ "    on x.constraint_name = c.constraint_name "
				+ "join information_schema.key_column_usage y "
				+ "    on y.ordinal_position = x.position_in_unique_constraint "
				+ "    and y.constraint_name = c.unique_constraint_name "
				+ "where x.table_name = '" + relationName + "'"
				+ "order by c.constraint_name, x.ordinal_position";
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.discovery.sql.AbstractSQLSchemaDiscoverer#getViewsDiscoveryStatement(java.lang.String)
	 */
	@Override
	protected String getViewsDiscoveryStatement(String databaseName) {
		return "SELECT table_name "
				+ "FROM information_schema.views "
				+ "WHERE table_schema = 'public' "
				+ "AND table_catalog = '" + databaseName + "' "
				+ "ORDER BY table_name";
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.builder.discovery.sql.AbstractSQLSchemaDiscoverer#getRelationInstance(java.util.Properties, java.lang.String, java.util.List)
	 */
	@Override
	protected Relation getRelationInstance(Properties properties,
			String relationName, List<Attribute> attributes) {
		return new PostgresqlRelationWrapper(properties, relationName, attributes);
	}

	/**
	 * @param properties Properties
	 * @param viewName String
	 * @param relationMap Map<String,Relation>
	 * @return View
	 */
	@Override
	protected View getViewInstance(Properties properties, String viewName, Map<String, Relation> relationMap) {
		String definition = this.getViewDefinition(viewName);
		return new PostgresqlViewWrapper(properties, this.parseViewDefinition(viewName, definition, relationMap));
	}

	/**
	 * @param viewName String
	 * @return String
	 */
	@Override
	protected String getViewDefinition(String viewName) {
		try(Connection connection = getConnection(this.properties);
			Statement stmt = connection.createStatement();
			ResultSet rs = stmt.executeQuery("SELECT view_definition "
						+ "FROM information_schema.views "
						+ "WHERE table_schema = 'public' "
						+ "AND table_name = '" + viewName + "' "
						+ "ORDER BY table_name")) {
			if (rs.next()) {
				return rs.getString(1);
			}
		} catch (SQLException e) {
			log.error(e);
			throw new BuilderException(e.getMessage(), e);
		}
		return null;
	}
	
	private static final String TOP_PATTERN = "select (?<select>.*) from (?<from>.*) where (?<where>.*)";
	private static final String ALIAS_PATTERN = "((?<alias>\\w*)\\.)?(?<attribute>\\w*)(\\s*as\\s*(?<renamed>\\w*))?";
	private static final String NESTED_CONJUNCTION_PATTERN = "\\((?<condition>.*)(\\s*and\\s*(?<rest>.*))+?\\)";
	private static final String CONDITION_PATTERN = "\\((?<condition>.*)\\)";

	/**
	 * Parse a SQL statement featuring a conjunctive view definition and returns
	 * the corresponding linear guarded dependency.
	 * @param viewDef
	 * @param viewName String
	 * @param relationMap Map<String,Relation>
	 * @return the corresponding linear guarded dependency representation of the
	 * given SQL view definition.
	 */
	public LinearGuarded parseViewDefinition(String viewName, String viewDef, Map<String, Relation> relationMap) {
		Preconditions.checkArgument(viewDef != null && !viewDef.isEmpty());
		
		String crFreeViewDef = viewDef.replace("\n", " ");
		Matcher m = Pattern.compile(TOP_PATTERN, Pattern.CASE_INSENSITIVE).matcher(crFreeViewDef);
		String select = null;
		String from = null;
		String where = null;
		if (m.find()) {
			select = m.group("select").trim();
			from = m.group("from").trim();
			where = m.group("where").trim();
		}
		if (Strings.isNullOrEmpty(select) || Strings.isNullOrEmpty(from) || Strings.isNullOrEmpty(where)) {
			throw new IllegalArgumentException("Not a valid view definition " + viewDef);
		}
		BiMap<String, PredicateFormula> predicates = this.makePredicate(from, relationMap);
		this.makeJoins(where, predicates);
		Pair<List<Term>, List<Attribute>> freeTermsAndAttributes = this.makeFreeTerms(select, predicates);
		List<PredicateFormula> right = new ArrayList<>(predicates.values());
		List<Variable> boundTerms = new ArrayList<>(Utility.getVariables(right));
		boundTerms.removeAll(freeTermsAndAttributes.getLeft());
		return new LinearGuarded(
				//this.toVariable(freeTermsAndAttributes.getLeft()),
				new PredicateFormula(
						new Relation(viewName, freeTermsAndAttributes.getRight()) {},
						freeTermsAndAttributes.getLeft()), 
				//boundTerms,
				Conjunction.of(right));
	}
	
	/**
	 * Parse a SQL from clause and returns a map from relation aliases to 
	 * predicates featuring fresh variables.
	 * @param fromClause
	 * @param relationMap
	 * @return a map from relation aliases to predicates with fresh variables
	 */
	private BiMap<String, PredicateFormula> makePredicate(String fromClause, Map<String, Relation> relationMap) {
		BiMap<String, PredicateFormula> result = HashBiMap.create();
		for (String token: fromClause.trim().split(",")) {
			String[] aliased = token.trim().split("(AS|\\s)");
			Relation r = relationMap.get(aliased[0].trim());
			PredicateFormula pred = new PredicateFormula(r, Utility.generateVariables(r));
			if (aliased.length == 1) {
				result.put(aliased[0].trim(), pred);
			} else {
				result.put(aliased[1].trim(), pred);
			}
		}
		return result;
	}
	
	/**
	 * Parse a SQL where clause and returns a map from relation aliases to 
	 * predicates featuring fresh variables.
	 * @param whereClause
	 * @param predMap BiMap<String,PredicateFormula>
	 */
	private void makeJoins(String whereClause, BiMap<String, PredicateFormula> predMap) {
		Matcher m = Pattern.compile(NESTED_CONJUNCTION_PATTERN, Pattern.CASE_INSENSITIVE).matcher(whereClause.trim());
		String nested = null;
		String rest = null;
		if (m.find()) {
			do {
				nested = m.group("condition");
				rest = m.group("rest");
				if (rest != null) {
					this.makeJoins(rest.trim(), predMap);
				}
				this.makeJoins(nested.trim(), predMap);
			} while (m.find());
		} else {
			String clause = null;
			Matcher m2 = Pattern.compile(CONDITION_PATTERN, Pattern.CASE_INSENSITIVE).matcher(whereClause.trim());
			if (m2.find()) {
				clause = m2.group("condition").trim();
			} else {
				clause = whereClause;
			}
			String[] operands = clause.split("=");
			if (operands != null && operands.length == 2) {
				Triple<PredicateFormula, Integer, Attribute> t1 = this.parseAlias(operands[0], predMap);
				if (t1 == null) {
					return;
				}
				Triple<PredicateFormula, Integer, Attribute> t2 = this.parseAlias(operands[1], predMap);
				if (t2 == null) {
					return;
				}
				PredicateFormula p2 = t2.getFirst();
				List<Term> terms = Lists.newArrayList(p2.getTerms());
				terms.set(t2.getSecond(), t1.getFirst().getTerm(t1.getSecond()));
				String key = predMap.inverse().get(p2);
				predMap.forcePut(key, new PredicateFormula(p2.getSignature(), terms));
			}
		}
	}

	/**
	 * Parse a SQL select clause and returns a list of corresponding free 
	 * variables from the given predicates collection.
	 * @param selectClause
	 * @param predMap Map<String,PredicateFormula>
	 * @return a map from relation aliases to predicates with fresh variables
	 */
	private Pair<List<Term>, List<Attribute>> makeFreeTerms(String selectClause, Map<String, PredicateFormula> predMap) {
		List<Term> terms = new ArrayList<>();
		List<Attribute> attributes = new ArrayList<>();
		for (String token: selectClause.trim().split(",")) {
			Triple<PredicateFormula, Integer, Attribute> triple = this.parseAlias(token.trim(), predMap);
			if (triple == null) {
				Pair<Term, Attribute> pair = this.parseConstant(token, predMap.values());
				terms.add(pair.getLeft());
				attributes.add(pair.getRight());
			} else {
				terms.add(triple.getFirst().getTerm(triple.getSecond()));
				attributes.add(triple.getThird());
			}
		}
		return Pair.of(terms, attributes);
	}
	
	/**
	 * @param token String
	 * @param predMap Map<String,PredicateFormula>
	 * @return Triple<PredicateFormula,Integer,Attribute>
	 */
	private Triple<PredicateFormula, Integer, Attribute> parseAlias(String token, Map<String, PredicateFormula> predMap) {
		String alias = null;
		String attribute = null;
		String renamed = null;
		Matcher m = Pattern.compile(ALIAS_PATTERN, Pattern.CASE_INSENSITIVE).matcher(token.trim());
		if (m.find()) {
			alias = m.group("alias");
			attribute = m.group("attribute");
			renamed = m.group("renamed");
		}
		alias = alias != null ? alias.trim() : null;
		attribute = attribute != null ? attribute.trim() : null;
		renamed = renamed != null ? renamed.trim() : null;
		if (Strings.isNullOrEmpty(attribute)) {
			throw new IllegalArgumentException("Not a valid alias clause in view definition " + token);
		}
		
		PredicateFormula pred = null;
		Integer index = null;
		Attribute att = null;
		if (alias == null || alias.isEmpty()){
			Iterator<PredicateFormula> it = predMap.values().iterator();
			while (it.hasNext()) {
				pred = it.next();
				index = ((Relation) pred.getSignature()).getAttributeIndex(attribute);
				if (index >= 0) {
					break;
				}
			}
		} else {
			pred = predMap.get(alias);
			if (pred != null) {
				index = ((Relation) pred.getSignature()).getAttributeIndex(attribute);
			}
		}
		if (renamed != null && !renamed.isEmpty()){
			att = new Attribute(((Relation) pred.getSignature()).getAttribute(index).getType(), renamed);
		}
		return new Triple<>(pred, index, att);
	}
	
	/**
	 * @param predicates
	 * @param token String
	 * @return the first predicate found in the given collection featuring the 
	 * given attribute and its position.
	 */
	private Pair<Term, Attribute> parseConstant(String token, Collection<PredicateFormula> predicates) {
//		String alias = null;
		String attribute = null;
		String renamed = null;
		Matcher m = Pattern.compile(ALIAS_PATTERN, Pattern.CASE_INSENSITIVE).matcher(token);
		if (m.find()) {
//			alias = m.group("alias").trim();
			attribute = m.group("attribute").trim();
			renamed = m.group("renamed").trim();
		}
		if (Strings.isNullOrEmpty(attribute)) {
			throw new IllegalArgumentException("Not a valid select clause in view definition " + token);
		}
		
		Term term = null;
		Attribute att = null;
		if (attribute != null
				&& ((attribute.startsWith("\"") && attribute.endsWith("\""))
				|| (attribute.startsWith("'") && attribute.endsWith("'")))) {
			term = new TypedConstant<>(attribute.substring(1, attribute.length() - 1));
		}
		if (renamed != null && !renamed.isEmpty()) {
			att = new Attribute(String.class, String.valueOf(Skolem.getFreshConstant()));
		}
		return Pair.of(term, att);
	}
}
