package uk.ac.ox.cs.pdq.cost.estimators;
import static uk.ac.ox.cs.pdq.logging.performance.StatKeys.COST_ESTIMATION_COUNT;
import static uk.ac.ox.cs.pdq.logging.performance.StatKeys.COST_ESTIMATION_TIME;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import uk.ac.ox.cs.pdq.cost.CostParameters.BlackBoxQueryTypes;
import uk.ac.ox.cs.pdq.logging.performance.StatisticsCollector;
import uk.ac.ox.cs.pdq.plan.Cost;
import uk.ac.ox.cs.pdq.plan.DoubleCost;
import uk.ac.ox.cs.pdq.plan.Plan;
import uk.ac.ox.cs.pdq.rewrite.RewriterException;
import uk.ac.ox.cs.pdq.rewrite.sql.SQLTranslator;
import uk.ac.ox.cs.pdq.rewrite.sql.SQLTranslator.SupportedDialect;

import com.google.common.eventbus.EventBus;

/**
 * Blackbox cost estimator relying on PostgresqlTranslator internal query cost 
 * estimator.
 *
 * @author Julien Leblay
 *
 */
public class PostgresqlBlackBoxEstimator<P extends Plan> implements BlackBoxCostEstimator<P> {

	protected final StatisticsCollector stats;
	
	/**
	 * The regular expression used to retrieve the costs from the answer of
	 * an EXPLAIN ANALYZE query to postgresql.
	 */
	private static final String COST_REGEXP_PATTERN =
			"\\(cost=\\d+\\.\\d+\\.\\.(?<cost>\\d+\\.\\d+)\\s.*\\)";

	/** Properties featuring database connection details. */
	protected final Properties properties;

	/** The query type use by this estimator */
	protected final BlackBoxQueryTypes queryType;

	/**
	 * Default constructor.
	 * @param prop
	 * @param eventBus EventBus
	 * @param collectStats boolean
	 */
	public PostgresqlBlackBoxEstimator(EventBus eventBus, boolean collectStats, Properties prop) {
		this(eventBus, collectStats, prop, BlackBoxQueryTypes.DEFAULT);
	}

	/**
	 * Default constructor.
	 * @param prop
	 * @param eventBus EventBus
	 * @param collectStats boolean
	 * @param type BlackBoxQueryTypes
	 */
	public PostgresqlBlackBoxEstimator(EventBus eventBus, boolean collectStats, Properties prop, BlackBoxQueryTypes type) {
		this(new StatisticsCollector(collectStats, eventBus), prop, type);
	}

	/**
	 * Default constructor.
	 * @param prop
	 * @param type
	 * @param stats StatisticsCollector
	 */
	public PostgresqlBlackBoxEstimator(StatisticsCollector stats, Properties prop, BlackBoxQueryTypes type) {
		this.stats = stats;
		this.properties = prop;
		this.queryType = type;
	}

	/**
	 * @return PostgresqlBlackBoxEstimator<P,S>
	 * @see uk.ac.ox.cs.pdq.plan.cost.DAGConfigurationCostEstimator#clone()
	 */
	@Override
	public PostgresqlBlackBoxEstimator<P> clone() {
		return new PostgresqlBlackBoxEstimator<>(this.stats.clone(), this.properties, this.queryType);
	}

	/**
	 * @param plan P
	 * @return DoubleCost
	 * @see uk.ac.ox.cs.pdq.cost.estimators.CostEstimator#cost(P)
	 */
	@Override
	public Cost cost(P plan) {
		plan.setCost(this.estimateCost(plan));
		return plan.getCost();
	}

	/**
	 * @return a connection database connection for the given properties.
	 * @throws SQLException
	 */
	public Connection getConnection() throws SQLException {
		String url = this.properties.getProperty("url");
		String database = this.properties.getProperty("database");
		String username = this.properties.getProperty("username");
		String password = this.properties.getProperty("password");
		return DriverManager.getConnection(url + database, username, password);
	}

	@Override
	public Cost estimateCost(P plan) {
		double result = Double.POSITIVE_INFINITY;
		this.stats.start(COST_ESTIMATION_TIME);
		if (plan.isClosed()) {
			StringBuilder sql = new StringBuilder("EXPLAIN ");
			try {
				switch (this.queryType) {
				case SQL_WITH:
					sql.append(SQLTranslator.target(SupportedDialect.POSTGRESQL)
							.toSQLWith(plan.getEffectiveOperator()));
					break;
				default:
					sql.append(SQLTranslator.target(SupportedDialect.POSTGRESQL)
							.toSQL(plan.getEffectiveOperator()));
					break;
				}
			} catch (RewriterException e) {
				throw new IllegalStateException(e);
			}
			try(Connection conn = this.getConnection();
				Statement stmt = conn.createStatement();
				ResultSet rs = stmt.executeQuery(sql.toString())) {
				if (rs.next()) {
					String s = rs.getString(1);
					Pattern p = Pattern.compile(COST_REGEXP_PATTERN);
					Matcher m = p.matcher(s);
					if (m.find()) {
						String cost = m.group("cost");
						result = Double.valueOf(cost);
					}
				}
			} catch (SQLException e) {
				throw new IllegalStateException(e.getMessage() + " for " + sql + "\n" + plan);
			}
		}
		this.stats.stop(COST_ESTIMATION_TIME);
		this.stats.increase(COST_ESTIMATION_COUNT, 1);
		return new DoubleCost(result);
	}

}

