package uk.ac.ox.cs.pdq.planner.dag;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibilityAxiom;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessibleRelation;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.InferredAccessibleRelation;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;
import com.google.common.collect.Sets;

/**
 * Configuration post pruning. Removes redundant access commands and follow-up joins
 *
 * @author Efthymia Tsamoura
 */
public class ConfigurationPostPruning<S extends ChaseState> {

	/** A successful configuration*/
	private final DAGChaseConfiguration<S> configuration;

	/** Facts that match the query */
	private final Collection<PredicateFormula> queryFacts;
	/**
	 * True if the input configuration is pruned
	 */
	private Boolean isPruned = false;

	/**
	 *
	 * @param configuration
	 * 		Input configuration
	 * @param queryFacts
	 * 		Facts that match the query. 
	 * An exception is thrown if the input facts are not accessible or inferred accessible ones
	 *
	 */
	public ConfigurationPostPruning(DAGChaseConfiguration<S> configuration, Collection<PredicateFormula> queryFacts) {
		Preconditions.checkNotNull(configuration);
		Preconditions.checkArgument(configuration.isClosed());
		Preconditions.checkArgument(configuration.isSuccessful());
		Preconditions.checkNotNull(queryFacts);
		Collection<PredicateFormula> qF = new LinkedHashSet<>();
		for(PredicateFormula queryFact: queryFacts) {
			if(queryFact.getSignature() instanceof InferredAccessibleRelation) {
				qF.add(queryFact);
			} else if(!(queryFact.getSignature() instanceof AccessibleRelation)) {
				throw new java.lang.IllegalArgumentException();
			}
		}
		this.configuration = configuration;
		this.queryFacts = qF;
	}


	/**
	 * Prunes the input configuration
	 * @return the post-pruned configuration
	 * @throws PlannerException
	 */
	public DAGChaseConfiguration<S> prune() throws PlannerException {
		Collection<PredicateFormula> infAccFacts
		= this.getAccessibilityAxioms(this.queryFacts, this.configuration.getState());
		DAGChaseConfiguration<S> output = this.pruneRecursive(this.configuration, infAccFacts);
		if(this.isPruned) {
			return output;
		}
		return null;
	}


	/**
	 * 
	 * @param configuration
	 * @param queryFacts
	 * @return
	 * 		a configuration that performs only the accesses required to match the input query facts
	 * @throws PlannerException
	 */
	private DAGChaseConfiguration<S> pruneRecursive(DAGChaseConfiguration<S> configuration, Collection<PredicateFormula> queryFacts) throws PlannerException {
		
		/** If the configuration is an ApplyRule one, then find the facts that after chasing lead to the derivation of the input query facts  */
		if(configuration instanceof ApplyRule) {
			Set<PredicateFormula> firings = ((ApplyRule<S>) configuration).getFiringsThatExposeFacts(queryFacts);
			if(!firings.equals(((ApplyRule<S>) configuration).getFacts())) {
				this.isPruned = true;
				return ((ApplyRule<S>) configuration).prune(firings);
			}
			return configuration;
		}
		/** Otherwise, prune first the right subconfiguration, and then, the right subconfiguration*/
		DAGChaseConfiguration<S> c1 = ((BinaryConfiguration<S>)configuration).getLeft();
		DAGChaseConfiguration<S> c2 = ((BinaryConfiguration<S>)configuration).getRight();

		Set<PredicateFormula> atoms1 = Sets.newLinkedHashSet(c1.getDerivedInferred());
		Set<PredicateFormula> atoms2 = Sets.newLinkedHashSet(c2.getDerivedInferred());
		if(atoms1.containsAll(atoms2)) {
			this.isPruned = true;
			return this.pruneRecursive(c1, queryFacts);
		}
		DAGChaseConfiguration<S> ret = null;
		DAGChaseConfiguration<S> r = this.pruneRecursive(c2, queryFacts);
		Collection<Constant> input2 = r.getInput();
		Collection<Set<PredicateFormula>> sets = ConfigurationUtility.getMinimalSetThatExposesConstants(c1, input2);
		for(Set<PredicateFormula> set:sets) {
			DAGChaseConfiguration<S> l = this.pruneRecursive(c1, set);
			DAGChaseConfiguration<S> output = new BinaryConfiguration<>(
					c1.getAccessibleSchema(),
					c1.getQuery(),
					c1.getChaser(),
					null,
					c1.getDominanceDetectors(),
					c1.getSuccessDominanceDetector(),
					c1.getCostEstimator(),
					l, r,
					true
					);
			if(ret == null || ret.getPlan().getCost().greaterThan(output.getPlan().getCost())) {
				ret = output;
			}
		}
		return ret;
	}
	
	/**
	 * 
	 * @param input
	 * @param state
	 * @return the accessed facts that lead to the derivation of the input facts
	 */
	protected Collection<PredicateFormula> getAccessibilityAxioms(Collection<PredicateFormula> input, ChaseState state) {
		Collection<PredicateFormula> facts = new LinkedHashSet<>();
		for(PredicateFormula fact:input) {
			Pair<Constraint, Collection<PredicateFormula>> pair = state.getFactProvenance(fact);
			if(pair!= null) {
				if(pair.getLeft() instanceof AccessibilityAxiom) {
					facts.add(fact);
				}
				else {
					facts.addAll(this.getAccessibilityAxioms(pair.getRight(), state));
				}
			}
			else {
				throw new java.lang.IllegalArgumentException();
			}
		}
		return facts;
	}


	/**
	 * 
	 * @return true if the input configuration is post-pruned
	 */
	public Boolean getIsPruned() {
		return this.isPruned;
	}
}
