package uk.ac.ox.cs.pdq.planner.dag;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import uk.ac.ox.cs.pdq.algebra.Access;
import uk.ac.ox.cs.pdq.algebra.DependentAccess;
import uk.ac.ox.cs.pdq.algebra.DependentJoin;
import uk.ac.ox.cs.pdq.algebra.Join;
import uk.ac.ox.cs.pdq.algebra.Join.Variants;
import uk.ac.ox.cs.pdq.algebra.Operators;
import uk.ac.ox.cs.pdq.algebra.Projection;
import uk.ac.ox.cs.pdq.algebra.RelationalOperator;
import uk.ac.ox.cs.pdq.algebra.Scan;
import uk.ac.ox.cs.pdq.algebra.Selection;
import uk.ac.ox.cs.pdq.algebra.StaticInput;
import uk.ac.ox.cs.pdq.algebra.SubPlanAlias;
import uk.ac.ox.cs.pdq.algebra.predicates.Predicate;
import uk.ac.ox.cs.pdq.db.AccessMethod;
import uk.ac.ox.cs.pdq.db.AccessMethod.AccessMethodTypes;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.plan.DAGPlan;
import uk.ac.ox.cs.pdq.plan.Plan.ControlFlows;
import uk.ac.ox.cs.pdq.planner.dag.BinaryConfiguration.BinaryConfigurationTypes;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;
import uk.ac.ox.cs.pdq.util.TupleType;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;

/**
 * Implement algorithms 1 and 2 from the dependent join notes.
 *
 * @author Julien Leblay
 *
 */
public class PlanGenerator {

	/**
	 * @param config DAGConfiguration<S>
	 * @param cf ControlFlows
	 * @return DAGPlan
	 */
	public static <S extends ChaseState> DAGPlan toPlan(DAGChaseConfiguration<S> config, ControlFlows cf) {
		switch (cf) {
		case BOTTOM_UP:
			return toBottomUpPlan(config);
		default:
			return toPlan(config);
		}
	}

	/**
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @param type BinaryConfigurationTypes
	 * @param cf ControlFlows
	 * @return DAGPlan
	 */
	public static <S extends ChaseState> DAGPlan toPlan(DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right, BinaryConfigurationTypes type, ControlFlows cf) {
		Set<Constant> inputs = Sets.newLinkedHashSet(right.getInput());
		inputs.removeAll(left.getOutput());
		inputs.addAll(left.getInput());
		return toPlan(inputs, left, right, type, cf);
	}

	/**
	 * @param inputs Collection<? extends Term>
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @param type BinaryConfigurationTypes
	 * @param cf ControlFlows
	 * @return DAGPlan
	 */
	public static <S extends ChaseState> DAGPlan toPlan(Collection<? extends Term> inputs, DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right, BinaryConfigurationTypes type, ControlFlows cf) {
		switch (cf) {
		case BOTTOM_UP:
			return toBottomUpPlan(inputs, left, right);
		default:
			return toPlan(inputs, left, right, type);
		}
	}

	/**
	 * Creates a top-down physical plan from a DAG plan.
	 *
	 * @param config BinaryConfiguration<S>
	 * @return a relational expression equivalent to the given linear plan
	 */
	private static <S extends ChaseState> DAGPlan toPlan(BinaryConfiguration<S> config) {
		return toPlan(config, config.getInput(), config.getLeft(), config.getRight(), config.getType());
	}

	/**
	 * Creates a top-down physical plan from a DAG plan.
	 *
	 * @param inputs Collection<? extends Term>
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @param type BinaryConfigurationTypes
	 * @return a relational expression equivalent to the given linear plan
	 */
	private static <S extends ChaseState> DAGPlan toPlan(Collection<? extends Term> inputs, DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right, BinaryConfigurationTypes type) {
		return toPlan(null, inputs, left, right, type);
	}

	/**
	 * Creates a top-down physical plan from a DAG plan.
	 *
	 * @param config DAGConfiguration<S>
	 * @param inputs Collection<? extends Term>
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @param type BinaryConfigurationTypes
	 * @return a relational expression equivalent to the given linear plan
	 */
	private static <S extends ChaseState> DAGPlan toPlan(DAGChaseConfiguration<S> config, Collection<? extends Term> inputs, DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right, BinaryConfigurationTypes type) {
		ControlFlows cf = left.getPlan().getControlFlow();
		RelationalOperator lOp = left.getPlan().getOperator();
		RelationalOperator rOp = right.getPlan().getOperator();
		RelationalOperator operator = null;

		switch(type) {
		case PCOMPOSE:
			operator = new DependentJoin(lOp, rOp);
			break;
		case MERGE:
			operator = new Join(lOp, rOp);
			break;
		case JCOMPOSE:
			operator = new DependentJoin(lOp, rOp);
			break;
		case GENCOMPOSE:
			operator = new DependentJoin(lOp, rOp);
			break;
		default:
			throw new java.lang.IllegalArgumentException();
		}

		TreePlan plan = new TreePlan(inputs, operator, cf);
		plan.addChild(left.getPlan());
		plan.addChild(right.getPlan());
		return plan;
	}

	/**
	 * Creates a top-down physical plan from a DAG plan.
	 *
	 * @param config DAGConfiguration<S>
	 * @return a relational expression equivalent to the given linear plan
	 */
	private static <S extends ChaseState> DAGPlan toPlan(DAGChaseConfiguration<S> config) {
		if (config instanceof ApplyRule) {
			return toPlan((ApplyRule<S>) config);
		} else if (config instanceof BinaryConfiguration) {
			return toPlan((BinaryConfiguration<S>) config);
		}
		throw new IllegalStateException("DAGConfiguration type " + config + " not supported.");
	}

	/**
	 * Creates a top-down physical plan from a DAG plan.
	 *
	 * @param config ApplyRule<S>
	 * @return a relational expression equivalent to the given linear plan
	 */
	private static <S extends ChaseState> DAGPlan toPlan(ApplyRule<S> config) {
		Relation relation = config.getRelation();
		AccessMethod binding = config.getBindingPositions();
		Collection<PredicateFormula> facts = config.getFacts();
		if (facts.isEmpty()) {
			return new DAGPlan(new DependentAccess(relation, binding), ControlFlows.TOP_DOWN);
		}
		RelationalOperator op = null;
		for (PredicateFormula fact: facts) {
			RelationalOperator access = new DependentAccess(relation, binding, fact.getTerms());
			Predicate selectPredicates = Operators.createSelectPredicates(fact.getTerms());
			if (selectPredicates != null) {
				access = new Selection(selectPredicates, access);
			}
			access = new Projection(access, renameAllVariables(access.getColumns(), fact.getTerms()));
			if (op == null) {
				op = access;
			} else {
				op = new Join(op, access);
			}
		}
		return new DAGPlan(config.getInput(), op, ControlFlows.TOP_DOWN);
	}

	/**
	 * @param from List<Term>
	 * @param to List<Term>
	 * @return Map<Integer,Term>
	 */
	private static Map<Integer, Term> renameAllVariables(List<Term> from, List<Term> to) {
		Preconditions.checkArgument(from.size() == to.size());
		Map<Integer, Term> result = new LinkedHashMap<>();
		for (int i = 0, l = from.size(); i < l; i++) {
			if (to.get(i).isVariable() || to.get(i).isSkolem()) {
				result.put(i, to.get(i));
			}
		}
		return result;
	}

	/**
	 * Creates a DAGPlan for bottom-up evaluation, based on the given left and
	 * right configurations.
	 *
	 * @param inputs Collection<? extends Term>
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @return a DAGPlan whose operator uses a bottom-up control flow.
	 */
	public static <S extends ChaseState> DAGPlan toBottomUpPlan(Collection<? extends Term> inputs, DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right) {
		return toBottomUpPlan(null, inputs, left, right);
	}

	/**
	 * Creates a DAGPlan for bottom-up evaluation, based on the given left and
	 * right configurations.
	 *
	 * @param config BinaryConfiguration<S>
	 * @param inputs Collection<? extends Term>
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @return a DAGPlan whose operator uses a bottom-up control flow.
	 */
	private static <S extends ChaseState> DAGPlan toBottomUpPlan(BinaryConfiguration<S> config, Collection<? extends Term> inputs, DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right) {
		Join logOp = new Join(left.getPlan().getOperator(), right.getPlan().getOperator());
		logOp.setVariant(Variants.SYMMETRIC_HASH);
		return new DAGPlan(inputs, logOp, ControlFlows.BOTTOM_UP, Lists.<DAGPlan>newArrayList(), Lists.newArrayList(left.getPlan(), right.getPlan()));
	}

	/**
	 * Creates a DAGPlan for bottom-up evaluation, based on the given left and
	 * right configurations.
	 *
	 * @param config BinaryConfiguration<S>
	 * @return a DAGPlan whose operator uses a bottom-up control flow.
	 */
	private static <S extends ChaseState> DAGPlan toBottomUpPlan(BinaryConfiguration<S> config) {
		return toBottomUpPlan(config, config.getInput(), config.getLeft(), config.getRight());
	}

	/**
	 * Creates a DAGPlan for bottom-up evaluation.
	 *
	 * @param config ApplyRule<S>
	 * @return a DAGPlan whose operator has a bottom-up control flow.
	 */
	private static <S extends ChaseState> DAGPlan toBottomUpPlan(ApplyRule<S> config) {
		Relation relation = config.getRelation();
		AccessMethod binding = config.getBindingPositions();
		Collection<PredicateFormula> facts = config.getFacts();
		Preconditions.checkState(!facts.isEmpty());
		if (facts.isEmpty()) {
			return new DAGPlan(
					config.getInput(),
					new Access(
							((ApplyRule<?>) config).getRelation(),
							((ApplyRule<?>) config).getBindingPositions()),
							ControlFlows.BOTTOM_UP);
		}
		RelationalOperator op1 = null;
		RelationalOperator access = null;
		for (PredicateFormula fact: facts) {
			if (access == null) {
				if (binding.getType() == AccessMethodTypes.FREE) {
					access = new Scan(relation);
				} else if (config.isClosed()) {
					StaticInput input = new StaticInput(makeStaticInputs(config));
					access = new Access(relation, binding, input);
				} else {
					access = new Access(relation, binding);
					List<Term> inputTerms = Utility.extract(fact.getTerms(), binding.getZeroBasedInputs());
					((Access) access).setChild(
							new Projection(new SubPlanAlias(inputTerms,
									TupleType.DefaultFactory.createFromTyped(relation.getInputAttributes(binding))),
									inputTerms));
				}
			}
			RelationalOperator op2 = access;
			Predicate selectPredicates = Operators.createSelectPredicates(fact.getTerms());
			if (selectPredicates != null) {
				op2 = new Selection(selectPredicates, op2);
			}
			op2 = new Projection(op2, renameAllVariables(access.getColumns(), fact.getTerms()));
			if (op1 == null) {
				op1 = op2;
			} else {
				op1 = new Join(op1, op2);
			}
		}
		return new DAGPlan(config.getInput(), op1, ControlFlows.BOTTOM_UP);
	}

	/**
	 * Creates a DAGPlan for bottom-up evaluation.
	 *
	 * @param config DAGConfiguration<S>
	 * @return a DAGPlan whose operator has a bottom-up control flow.
	 */
	public static <S extends ChaseState> DAGPlan toBottomUpPlan(DAGChaseConfiguration<S> config) {
		if (config instanceof ApplyRule) {
			return toBottomUpPlan((ApplyRule<S>) config);
		} else if (config instanceof BinaryConfiguration) {
			return toBottomUpPlan((BinaryConfiguration<S>) config);
		}
		throw new IllegalStateException("DAGConfiguration type " + config + " not supported.");
	}

	/**
	 * @param rhs
	 * @return a list of constant corresponding to input schema constant present in the RHS facts.
	 */
	private static List<TypedConstant<?>> makeStaticInputs(ApplyRule<?> rhs) {
		Collection<TypedConstant<?>> result = new LinkedHashSet<>(rhs.getBindingPositions().getInputs().size());
		for (PredicateFormula fact: rhs.getFacts()) {
			for (Term t: fact.getTerms()) {
				if (!t.isVariable() && !t.isSkolem()) {
					result.add(((TypedConstant) t));
				}
			}
		}
		Preconditions.checkState(result.size() == rhs.getBindingPositions().getInputs().size());
		return Lists.newArrayList(result);
	}
}
