package uk.ac.ox.cs.pdq.planner.dag.equivalence;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import uk.ac.ox.cs.pdq.plan.Plan;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;


/**
 * A collection of structurally equivalent classes that supports multi-threading.
 * According to this implementation different threads can add, remove or perform domination detection inside each class concurrently.
 *
 * @author Efthymia Tsamoura
 *
 * @param <S>
 */
public class SynchronizedEquivalenceClasses<S extends ChaseState> implements DAGEquivalenceClasses<S>{

	/** Maps each configuration to its class*/
	private final Map<DAGChaseConfiguration<S>, SynchronizedEquivalenceClass<S>> configurationToEquivalenceClass;

	public SynchronizedEquivalenceClasses() {
		this.configurationToEquivalenceClass = new ConcurrentHashMap<>(100, 10, 100);
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#addEntry(DAGChaseConfiguration<S>)
	 */
	@Override
	public void addEntry(DAGChaseConfiguration<S> configuration) {
		SynchronizedEquivalenceClass<S> e;
		DAGChaseConfiguration<S> equivalent = this.structurallyEquivalentTo(configuration);
		if(equivalent != null) {
			e = (SynchronizedEquivalenceClass<S>) equivalent.getEquivalenceClass();
			e.addEntry(configuration);
		}
		else {
			e = new SynchronizedEquivalenceClass<>(configuration);
		}
		this.configurationToEquivalenceClass.put(configuration, e);
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#removeEntry(DAGChaseConfiguration<S>)
	 */
	@Override
	public void removeEntry(DAGChaseConfiguration<S> configuration) {
		this.removeAll(Lists.newArrayList(configuration));
	}

	/**
	 * @return Collection<DAGChaseConfiguration<S>>
	 * @see uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses#getConfigurations()
	 */
	@Override
	public Collection<DAGChaseConfiguration<S>> getConfigurations() {
		return this.configurationToEquivalenceClass.keySet();
	}

	/**
	 * @param configurations Collection<DAGChaseConfiguration<S>>
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#removeAll(Collection<DAGChaseConfiguration<S>>)
	 */
	@Override
	public void removeAll(Collection<DAGChaseConfiguration<S>> configurations) {
		for(DAGChaseConfiguration<S> configuration: configurations) {
			SynchronizedEquivalenceClass<S> e = this.configurationToEquivalenceClass.get(configuration);
			if (e != null) {
				this.configurationToEquivalenceClass.remove(configuration);
				e.removeEntry(configuration);
			}
		}
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 * @return Collection<DAGChaseConfiguration<S>>
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#dominatedBy(DAGChaseConfiguration<S>)
	 */
	@Override
	public Collection<DAGChaseConfiguration<S>> dominatedBy(DAGChaseConfiguration<S> configuration) {
		Collection<DAGChaseConfiguration<S>> dominated = new LinkedHashSet<>();
		for(SynchronizedEquivalenceClass<S> c: this.configurationToEquivalenceClass.values()) {
			dominated.addAll(c.dominatedBy(configuration));
		}
		return dominated;
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 * @return DAGChaseConfiguration<S>
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#structurallyEquivalentTo(DAGChaseConfiguration<S>)
	 */
	@Override
	public DAGChaseConfiguration<S> structurallyEquivalentTo(DAGChaseConfiguration<S> configuration) {
		for(SynchronizedEquivalenceClass<S> c: this.configurationToEquivalenceClass.values()) {
			if(c.structurallyEquivalentTo(configuration)) {
				return c.getRepresentative();
			}
		}
		return null;
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 * @return DAGChaseConfiguration<S>
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#dominate(DAGChaseConfiguration<S>)
	 */
	@Override
	public DAGChaseConfiguration<S> dominate(DAGChaseConfiguration<S> configuration) {
		for(SynchronizedEquivalenceClass<S> c: this.configurationToEquivalenceClass.values()) {
			DAGChaseConfiguration<S> dominating = c.dominate(configuration);
			if(dominating != null) {
				return dominating;
			}
		}
		return null;
	}

	/**
	 * @return Collection<DAGEquivalenceClass<S>>
	 * @see uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses#getEquivalenceClasses()
	 */
	@Override
	public Collection<DAGEquivalenceClass<S>> getEquivalenceClasses() {
		return Sets.<DAGEquivalenceClass<S>>newHashSet(this.configurationToEquivalenceClass.values());
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 * @return DAGEquivalenceClass<S>
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#getEquivalenceClass(DAGChaseConfiguration<S>)
	 */
	@Override
	public DAGEquivalenceClass<S> getEquivalenceClass(DAGChaseConfiguration<S> configuration) {
		return this.configurationToEquivalenceClass.get(configuration);
	}

	/**
	 * @return boolean
	 * @see uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses#isEmpty()
	 */
	@Override
	public boolean isEmpty() {
		return this.configurationToEquivalenceClass.isEmpty();
	}

	/**
	 * @param plan Plan
	 */
	public void wakeupSleep(Plan plan) {
		for(SynchronizedEquivalenceClass<S> c: this.configurationToEquivalenceClass.values()) {
			c.wakeupSleep(plan);
		}
	}

	/**
	 * @return int
	 * @see uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses#size()
	 */
	@Override
	public int size() {
		return this.configurationToEquivalenceClass.size();
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 * @return boolean
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#contains(DAGChaseConfiguration<S>)
	 */
	@Override
	public boolean contains(DAGChaseConfiguration<S> configuration) {
		return this.configurationToEquivalenceClass.containsKey(configuration);
	}

	/**
	 * @return Iterator<Entry<DAGChaseConfiguration<S>,SynchronizedEquivalenceClass<S>>>
	 * @see uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses#iterator()
	 */
	@Override
	public Iterator<Entry<DAGChaseConfiguration<S>, SynchronizedEquivalenceClass<S>>> iterator() {
		return this.configurationToEquivalenceClass.entrySet().iterator();
	}

	/**
	 * @see uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses#clear()
	 */
	@Override
	public void clear() {
		this.configurationToEquivalenceClass.clear();
	}

	/**
	 * @return double
	 * @see uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses#averageClassSize()
	 */
	@Override
	public double averageClassSize() {
		double total = 0.0;
		for(SynchronizedEquivalenceClass<S> c: this.configurationToEquivalenceClass.values()) {
			total += c.size();
		}
		return total / this.configurationToEquivalenceClass.size();
	}

	/**
	 * @return int
	 * @see uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses#medianClassSize()
	 */
	@Override
	public int medianClassSize() {
		int[] totals = new int[this.configurationToEquivalenceClass.size()];
		if (totals.length == 0) {
			return -1;
		}
		Iterator<SynchronizedEquivalenceClass<S>> it = this.configurationToEquivalenceClass.values().iterator();
		for (int i = 0, l = totals.length; i < l; i ++) {
			totals[i] = it.next().size();
		}
		Arrays.sort(totals);
		return totals[totals.length / 2];
	}

	/**
	 * @param configurations Collection<DAGChaseConfiguration<S>>
	 * @see uk.ac.ox.cs.pdq.equivalence.dag.DAGEquivalenceClasses#retainAll(Collection<DAGChaseConfiguration<S>>)
	 */
	@Override
	public void retainAll(Collection<DAGChaseConfiguration<S>> configurations) {
		for(DAGChaseConfiguration<S> configuration: this.configurationToEquivalenceClass.keySet()) {
			if(!configurations.contains(configuration)) {
				SynchronizedEquivalenceClass<S> e = this.configurationToEquivalenceClass.get(configuration);
				Preconditions.checkNotNull(e);
				this.configurationToEquivalenceClass.remove(configuration);
				e.removeEntry(configuration);
			}
		}
	}
}