package uk.ac.ox.cs.pdq.planner.dag.explorer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.LimitReachedException;
import uk.ac.ox.cs.pdq.logging.performance.StatKeys;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.dag.ApplyRule;
import uk.ac.ox.cs.pdq.planner.dag.BinaryConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.explorer.filters.Filter;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.StrictInclusionValidator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.Validator;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;
import com.google.common.collect.Sets;
import com.google.common.eventbus.EventBus;

/**
 * Simple dag explorer. It searches the space of binary configurations exhaustively
 *
 * @author Efthymia Tsamoura
 *
 */
public class GenericExplorer<S extends ChaseState> extends DAGExplorer<S> {

	/**
	 * The maximum depth we can explore. The exploration ends when
	 * there does not exist any configuration with depth < maxDepth
	 */
	protected final int maxDepth;

	/** Filters out configurations at the end of each iteration*/
	private final Filter<S> filter;

	/** Validates pairs of configurations to be composed*/
	private final Validator<S> validator;

	private final List<DAGChaseConfiguration<S>> left;
	private final List<DAGChaseConfiguration<S>> right;

	/** The current exploration depth */
	protected int depth = 1;

	/** True if pair selection is order aware */
	protected boolean orderAware;

	/** Returns pairs of configurations to combine */
	protected PairSelector<S> selector;

	/**
	 * 
	 * @param eventBus
	 * @param collectStats
	 * @param initialConfigurations
	 * 		ApplyRule configurations to initilise the explorer
	 * @param filter
	 * 		Filters out configurations at the end of each iteration
	 * @param validator
	 * 		Validates pairs of configurations to be composed
	 * @param maxDepth
	 * 		The maximum depth to explore
	 * @param orderAware True if pair selection is order aware
	 * @throws PlannerException
	 */
	public GenericExplorer(
			EventBus eventBus, boolean collectStats,
			List<DAGChaseConfiguration<S>> initialConfigurations,
			Filter<S> filter,
			Validator<S> validator,
			int maxDepth,
			boolean orderAware) throws PlannerException {
		super(eventBus, collectStats);
		this.checkFilterAndValidator(filter, validator);
		this.filter = filter;
		this.validator = validator;
		this.orderAware = orderAware;
		this.maxDepth = maxDepth;

		if(this.filter != null) {
			Collection<DAGChaseConfiguration<S>> toDelete = this.filter.filter(initialConfigurations);
			initialConfigurations.removeAll(toDelete);
		}
		this.left = new ArrayList<>();
		this.right = new ArrayList<>();
		this.left.addAll(initialConfigurations);
		this.right.addAll(initialConfigurations);
		this.selector = new PairSelector<>(this.left, this.right, this.validator, this.orderAware);
	}

	/**
	 * @throws PlannerException
	 */
	@Override
	protected void _explore() throws PlannerException, LimitReachedException {
		//if the current depth exceeds the threshold return
		if (this.depth > this.maxDepth) {
			this.forcedTermination = true;
			return;
		}
		//Check the ApplyRule configurations for success
		if (this.depth == 1) {
			for (DAGChaseConfiguration<S> configuration:this.right) {
				if (configuration.isClosed()
						&& (this.bestPlan == null
						|| configuration.getPlan().getCost().lessThan(this.bestPlan.getCost()))
						&& configuration.isSuccessful()) {
					this.setBestPlan(configuration);
				}
			}
			this.stats.set(StatKeys.CONFIGURATIONS, this.right.size());
			this.stats.set(StatKeys.CANDIDATES, this.right.size());
		} else if (this.depth > 1) {
			//Create all binary configurations of depth up to this.depth
			Collection<DAGChaseConfiguration<S>> last = this.mainLoop();
			//Stop if we cannot create any new configuration
			if (last.isEmpty()) {
				this.forcedTermination = true;
				return;
			}
			//Filter out configurations
			if (this.filter != null) {
				Collection<DAGChaseConfiguration<S>> toDelete;
				toDelete = this.filter.filter(CollectionUtils.union(last,this.right));
				this.right.removeAll(toDelete);
				last.removeAll(toDelete);
			}

			this.left.clear();
			this.left.addAll(last);
			this.selector = new PairSelector<>(this.left, this.right, this.validator, this.orderAware);

			this.stats.set(StatKeys.CONFIGURATIONS, this.right.size());
			this.stats.set(StatKeys.CANDIDATES, this.left.size());
		} else {
			throw new IllegalStateException("Search depth cannot be < 1");
		}
		this.depth++;
	}

	/**
	 * @return Collection<DAGChaseConfiguration<S>>
	 * @throws PlannerException
	 */
	protected Collection<DAGChaseConfiguration<S>> mainLoop() throws PlannerException, LimitReachedException {
		Map<Pair<DAGChaseConfiguration<S>, DAGChaseConfiguration<S>>, DAGChaseConfiguration<S>> last = new HashMap<>();
		Pair<DAGChaseConfiguration<S>, DAGChaseConfiguration<S>> pair = null;
		//Get the next pair of configurations to combine
		while ((pair = this.selector.getNext(this.depth)) != null) {
			if(!last.containsKey(pair)) {
				//Create a new binary configuration
				BinaryConfiguration<S> configuration = new BinaryConfiguration<>(
						pair.getLeft().getAccessibleSchema(),
						pair.getLeft().getQuery(),
						pair.getLeft().getChaser(),
						null,
						pair.getLeft().getDominanceDetectors(),
						pair.getLeft().getSuccessDominanceDetector(),
						pair.getLeft().getCostEstimator(),
						pair.getLeft(),
						pair.getRight(),
						true);
				//If the newly created binary configuration has the potential to lead to the optimal plan
				if (ExplorerUtils.getPotential(configuration, this.bestPlan)) {
					//If it is closed and has a match, update the best configuration
					if (configuration.isClosed() && configuration.isSuccessful()) {
						this.setBestPlan(configuration);
					} else {
						last.put(pair, configuration);
					}
				}
			}
			if (this.checkLimitReached()) {
				this.forcedTermination = true;
				break;
			}
		}
		return last.values();
	}


	/**
	 * Returns pairs of configurations to combine
	 *
	 * @author Efthymia Tsamoura
	 */
	protected static class PairSelector<S extends ChaseState> {
		/** Configurations to consider on the left*/
		private List<DAGChaseConfiguration<S>> left;

		/** Configurations to consider on the right*/
		private List<DAGChaseConfiguration<S>> right;

		/** Validates pairs of configurations to be composed*/
		private final Validator<S> validator;
		private final boolean orderAware;

		private final Set<Set<Integer>> cache = Sets.newLinkedHashSet();

		private Pair<DAGChaseConfiguration<S>, DAGChaseConfiguration<S>> reverse = null;
		private int i = 0;
		private int j = 0;
		private boolean sendReverse = false;

		/**
		 * Constructor for PairSelector.
		 * @param left List<DAGChaseConfiguration<S>>
		 * @param right List<DAGChaseConfiguration<S>>
		 * @param validator Validator<S>
		 */
		public PairSelector(List<DAGChaseConfiguration<S>> left,
				List<DAGChaseConfiguration<S>> right, Validator<S> validator) {
			this(left, right, validator, true);
		}

		/**
		 * Constructor for PairSelector.
		 * @param left List<DAGChaseConfiguration<S>>
		 * @param right List<DAGChaseConfiguration<S>>
		 * @param validator Validator<S>
		 * @param orderAware
		 */
		public PairSelector(
				List<DAGChaseConfiguration<S>> left,
				List<DAGChaseConfiguration<S>> right,
				Validator<S> validator,
				boolean orderAware) {
			Preconditions.checkNotNull(left);
			Preconditions.checkNotNull(right);
			Preconditions.checkNotNull(validator);
			this.left = left;
			this.right = right;
			this.validator = validator;
			this.orderAware = orderAware;
		}


		/**
		 * @param depth
		 * @return the next pair of configurations of the given combined depth
		 */
		public Pair<DAGChaseConfiguration<S>, DAGChaseConfiguration<S>> getNext(int depth) {
			if(!this.sendReverse) {
				if(this.i >= this.left.size() || this.j >= this.right.size()) {
					return null;
				}
				DAGChaseConfiguration<S> l = null;
				DAGChaseConfiguration<S> r = null;
				boolean validLR = false;
				boolean validRL = false;
				do {
					l = this.left.get(this.i);
					r = this.right.get(this.j);
					if (this.orderAware || !this.cache.contains(this.makeCacheKey(l, r))) {
						validLR = this.validator.validate(l, r, depth);
						validRL = this.validator.validate(r, l, depth);
						if (validLR || validRL) {
							break;
						}
					}
					this.forward();
					if (this.i >= this.left.size()) {
						return null;
					}
				} while(this.j < this.right.size());

				if (validLR) {
					this.sendReverse = false;
					if (this.orderAware && validRL) {
						this.reverse = Pair.of(r, l);
						this.sendReverse = true;
					}
				} else if (validRL) {
					this.sendReverse = false;
				}
				if (!this.orderAware) {
					this.cache.add(this.makeCacheKey(l, r));
				}
				this.forward();
				return Pair.of(l, r);
			}
			this.sendReverse = false;
			return this.reverse;
		}

		private void forward() {
			this.j++;
			if (this.j >= this.right.size()) {
				this.i++;
				this.j = 0;
			}
		}

		private Set<Integer> makeCacheKey(DAGChaseConfiguration<S>... configs) {
			Set<Integer> result = new HashSet<>();
			for (DAGChaseConfiguration<S> config: configs) {
				for (ApplyRule<S> applyRule: config.getApplyRules()) {
					result.add(applyRule.getId());
				}
			}
			return result;
		}
	}

	/**
	 * @return List<DAGChaseConfiguration<S>>
	 */
	public List<DAGChaseConfiguration<S>> getRight() {
		return this.right;
	}

	/**
	 * @param filter
	 * @param validator
	 * @return true if both the input filter and validator satisfy given type restrictions
	 */
	protected void checkFilterAndValidator(Filter<S> filter, Validator<S> validator) {
		Preconditions.checkArgument(validator != null);
		Preconditions.checkState(!(validator instanceof StrictInclusionValidator));
	}

}
