package uk.ac.ox.cs.pdq.planner.dag.explorer;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.TimeUnit;

import org.apache.commons.collections4.CollectionUtils;

import uk.ac.ox.cs.pdq.LimitReachedException;
import uk.ac.ox.cs.pdq.logging.performance.StatKeys;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.DAGConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses;
import uk.ac.ox.cs.pdq.planner.dag.equivalence.SynchronizedEquivalenceClasses;
import uk.ac.ox.cs.pdq.planner.dag.explorer.filters.Filter;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.DefaultValidator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.LinearValidator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.ReachabilityValidator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.Validator;
import uk.ac.ox.cs.pdq.planner.dag.priority.PriorityAssessor;
import uk.ac.ox.cs.pdq.planner.parallel.FinalIterationThreadResults;
import uk.ac.ox.cs.pdq.planner.parallel.IterativeExecutor;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.SuccessDominance;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.eventbus.EventBus;

/**
 * ILP explorer.
 * In the first round, it creates all left-deep configurations, either open or closed.
 * In the second round it creates new binary configurations by considering the sub-configurations of the successful configurations returned during round 1.
 * The next rounds proceed similarly.
 *
 * @author Efthymia Tsamoura
 *
 */
public class ILPExplorer<S extends ChaseState> extends DAGExplorer<S> {

	/**
	 * The maximum depth we can explore. The exploration ends when
	 * there does not exist any configuration with depth < maxDepth
	 */
	protected int maxDepth;

	/** The current exploration depth */
	protected int depth = 1;

	/** Performs parallel chasing */
	protected final IterativeExecutor<S> firstPhaseExecutor;

	/** Iterates over all newly created configurations in parallel and returns the best configuration*/
	protected final IterativeExecutor<S> secondPhaseExecutor;

	/** Filters out configurations at the end of each iteration*/
	protected final Filter<S> filter;

	/** Prioritises configurations */
	protected final PriorityAssessor<S> priority;

	/** Configuration to consider on the left*/
	protected final Queue<DAGChaseConfiguration<S>> left;
	/** Configuration to consider on the right*/

	protected final Collection<DAGChaseConfiguration<S>> right;

	/** Classes of structurally equivalent configurations*/
	protected final DAGEquivalenceClasses<S> equivalenceClasses;

	/** Map of binary configurations to the set of their constituting ApplyRule configurations.
	 * Used to further reduce the chasing overhead.
	 * If a configuration to be created c consist of the same ApplyRule configurations with another configuration c'
	 * and c' has already been chased, then c has the same state as c'.
	 * */
	protected final Template<S> templates;

	/** Successful configurations */
	protected final Set<DAGChaseConfiguration<S>> successfulConfigurations;

	/** ApplyRule configurations */
	protected final List<DAGChaseConfiguration<S>> applyRuleConfigurations;

	/** Maximum number of iterations */
	protected final int maxIterations;
	protected int iterations = 1;


	/**
	 *
	 * @param eventBus
	 * @param collectStats
	 * @param applyRuleConfigurations
	 * @param left
	 * 		Configuration to consider on the left
	 * @param right
	 * 		Configuration to consider on the right
	 * @param equivalenceClasses
	 * 		Classes of structurally equivalent configurations
	 * @param templates
	 * 		Map of binary configurations to the set of their constituting ApplyRule configurations.
	 * @param filter
	 * 		Filters out configurations at the end of each iteration
	 * @param priority
	 * 		Prioritises configurations
	 * @param firstPhaseExecutor
	 * 		Performs parallel chasing
	 * @param secondPhaseExecutor
	 * 		Iterates over all newly created configurations in parallel and returns the best configuration
	 * @param minDepth
	 * 		Minimum exploration depth
	 * @param maxDepth
	 * 		Maximum exploration depth
	 * @param maxIterations
	 * 		Maximum number of rounds
	 * @throws PlannerException
	 */
	public ILPExplorer(
			EventBus eventBus, boolean collectStats,
			List<DAGChaseConfiguration<S>> applyRuleConfigurations,
			Queue<DAGChaseConfiguration<S>> left,
			Collection<DAGChaseConfiguration<S>> right,
			DAGEquivalenceClasses<S> equivalenceClasses,
			Template<S> templates,
			Filter<S> filter,
			PriorityAssessor<S> priority,
			IterativeExecutor<S> firstPhaseExecutor,
			IterativeExecutor<S> secondPhaseExecutor,
			int minDepth,
			int maxDepth,
			int maxIterations) throws PlannerException {
		super(eventBus, collectStats);
		Preconditions.checkArgument(firstPhaseExecutor != null);
		Preconditions.checkArgument(secondPhaseExecutor != null);
		Preconditions.checkArgument(priority != null);
		Preconditions.checkArgument(applyRuleConfigurations != null);
		Preconditions.checkArgument(left != null);
		Preconditions.checkArgument(right != null);
		Preconditions.checkArgument(equivalenceClasses != null);
		Preconditions.checkArgument(templates != null);

		this.checkFilterAndValidator(filter, priority.getValidators());

		this.filter = filter;
		this.priority = priority;
		this.firstPhaseExecutor = firstPhaseExecutor;
		this.secondPhaseExecutor = secondPhaseExecutor;
		this.depth = minDepth;
		this.maxDepth = maxDepth;
		this.maxIterations = maxIterations;
		this.applyRuleConfigurations = applyRuleConfigurations;
		this.successfulConfigurations = new LinkedHashSet<>();

		this.left = left;
		this.right = right;
		this.templates = templates;
		this.equivalenceClasses = equivalenceClasses;

		this.preprocessInput();
	}


	/**
	 * 
	 * @param eventBus
	 * @param collectStats
	 * @param initialConfigurations
	 * 		ApplyRule configurations to initilise the explorer
	 * @param filter
	 * 		Filters out configurations at the end of each iteration
	 * @param priority Prioritises configurations
	 * 
	 * @param firstPhaseExecutor
	 * 		Performs parallel chasing
	 * @param secondPhaseExecutor
	 * 		Iterates over all newly created configurations in parallel and returns the best configuration
	 * @param maxDepth
	 * 		The maximum depth to explore
	 * @param maxIterations
	 * @throws PlannerException
	 */
	public ILPExplorer(
			EventBus eventBus, boolean collectStats,
			List<DAGChaseConfiguration<S>> initialConfigurations,
			Filter<S> filter,
			PriorityAssessor<S> priority,
			IterativeExecutor<S> firstPhaseExecutor,
			IterativeExecutor<S> secondPhaseExecutor,
			int maxDepth,
			int maxIterations) throws PlannerException {
		super(eventBus, collectStats);
		Preconditions.checkArgument(initialConfigurations != null);
		Preconditions.checkArgument(firstPhaseExecutor != null);
		Preconditions.checkArgument(secondPhaseExecutor != null);
		this.checkFilterAndValidator(filter, priority.getValidators());

		this.filter = filter;
		this.priority = priority;
		this.firstPhaseExecutor = firstPhaseExecutor;
		this.secondPhaseExecutor = secondPhaseExecutor;
		this.maxDepth = maxDepth;
		this.maxIterations = maxIterations;
		this.applyRuleConfigurations = Lists.newArrayList(initialConfigurations);
		this.successfulConfigurations = new LinkedHashSet<>();

		this.left = new ConcurrentLinkedQueue<>(this.applyRuleConfigurations);
		this.right = new LinkedHashSet<>(this.applyRuleConfigurations);
		this.templates = new Template<>();
		this.equivalenceClasses = new SynchronizedEquivalenceClasses<>();
		for (DAGChaseConfiguration<S> configuration: this.applyRuleConfigurations) {
			this.equivalenceClasses.addEntry(configuration);
		}
		this.preprocessInput();
	}


	/**
	 * @throws PlannerException
	 */
	@Override
	protected void _explore() throws PlannerException, LimitReachedException {
		if(this.iterations > this.maxIterations) {
			this.forcedTermination = true;
			return;
		}
		boolean bestPlanFound = false;
		this.successfulConfigurations.clear();
		//During each round
		for(; this.depth < this.maxDepth; ++this.depth) {
			if (this.depth == 1) {
				for (DAGChaseConfiguration<S> configuration: this.left) {
					if (this.bestPlan == null
							|| (configuration.isClosed() && configuration.getPlan().getCost().lessThan(this.bestPlan.getCost()))) {
						if (configuration.isClosed() && configuration.isSuccessful()) {
							this.setBestPlan(configuration);
						}
					}
				}
				this.stats.set(StatKeys.CONFIGURATIONS, this.left.size());
			} else if (this.depth > 1) {
				this.checkLimitReached();
				//Perform chasing in parallel
				Collection<DAGChaseConfiguration<S>> configurations = this.runChaseOrPropagatePhase();
				if (configurations == null) {
					this.forcedTermination = true;
					return;
				}

				//Stop if no new configuration is being found
				if(configurations.isEmpty()) {
					int currentDepth = this.depth;
					this.depth = this.maxDepth;
					if(this.iterations == 1) {
						this.maxDepth = currentDepth + 1;
					}
					break;
				}
				SuccessDominance successDominance = configurations.iterator().next().getSuccessDominanceDetector();
				this.checkLimitReached();
				//Iterate over all newly created configurations in parallel and return the best configuration
				FinalIterationThreadResults<S> results = this.secondPhaseExecutor.finalIteration(
						new ConcurrentLinkedQueue<>(configurations),
						this.equivalenceClasses,
						this.bestConfiguration,
						successDominance,
						Double.valueOf((this.maxElapsedTime - (this.elapsedTime/1e6))).longValue(),
						TimeUnit.MILLISECONDS);

				if (results == null) {
					this.forcedTermination = true;
					return;
				}

				//Update the best configuration
				List<DAGChaseConfiguration<S>> output = results.getOutput();
				this.successfulConfigurations.addAll(results.getSuccessful());
				if(results.getBest()!= null) {
					if(this.setBestPlan(results.getBest())) {
						bestPlanFound = true;
					}
				}

				if(output.isEmpty()) {
					int currentDepth = this.depth;
					this.depth = this.maxDepth;
					if(this.iterations == 1) {
						this.maxDepth = currentDepth + 1;
					}
					break;
				}
				//Update the configurations to consider on the left and right, respectively.
				this.left.clear();
				this.left.addAll(CollectionUtils.intersection(output, this.equivalenceClasses.getConfigurations()));
				this.right.retainAll(this.equivalenceClasses.getConfigurations());
				this.templates.put(output);
			}
		}

		if(this.bestPlan == null && this.iterations == 1) {
			this.forcedTermination = true;
			return;
		}

		if(this.bestPlan != null && !bestPlanFound) {
			this.forcedTermination = true;
			return;
		}

		Set<DAGChaseConfiguration<S>> subconfigurations = new LinkedHashSet<>();
		for(DAGChaseConfiguration<S> configuration:this.successfulConfigurations) {
			for(DAGConfiguration subconfiguration:configuration.getSubconfigurations()) {
				subconfigurations.add((DAGChaseConfiguration<S>) subconfiguration);
			}
		}
		subconfigurations.removeAll(this.successfulConfigurations);

		this.left.clear();
		this.left.addAll(this.applyRuleConfigurations);
		this.right.clear();
		this.right.addAll(CollectionUtils.intersection(subconfigurations, this.equivalenceClasses.getConfigurations()));

		this.stats.set(StatKeys.CONFIGURATIONS, this.equivalenceClasses.size());
		this.stats.set(StatKeys.CANDIDATES, this.left.size());

		this.depth = 2;
		this.iterations++;
	}

	/**
	 * @return Collection<DAGChaseConfiguration<S>>
	 * @throws PlannerException
	 */
	protected Collection<DAGChaseConfiguration<S>> runChaseOrPropagatePhase() throws PlannerException, LimitReachedException {
		return this.firstPhaseExecutor.chaseOrPropagate(
				this.depth, this.left, this.right,
				this.priority, this.templates, this.bestConfiguration, false,
				Double.valueOf((this.maxElapsedTime - (this.elapsedTime/1e6))).longValue(),
				TimeUnit.MILLISECONDS);
	}

	/**
	 * @param filter
	 * @param validators List<Validator<S>>
	 * @return true if both the input filter and validators satisfy given type restrictions
	 */
	protected boolean checkFilterAndValidator(Filter<S> filter, List<Validator<S>> validators) {
		Preconditions.checkArgument(filter == null);
		Preconditions.checkArgument(validators != null);
		Preconditions.checkArgument(!validators.isEmpty());
		for(Validator<S> validator: validators) {
			Preconditions.checkState(validator instanceof DefaultValidator
					|| validator instanceof ReachabilityValidator
					|| validator instanceof LinearValidator,
					"Validator type " + validator.getClass() + " not allowed here.");
		}
		return true;
	}

	/**
	 * Pre-processes the input ApplyRule configurations by removing the non-reachable ones.
	 */
	protected void preprocessInput() {
		for(Validator<S> validator: this.priority.getValidators()) {
			if(validator instanceof ReachabilityValidator) {
				Collection<DAGChaseConfiguration<S>> forbidden = ((ReachabilityValidator<S>) validator).getForbidden();
				this.applyRuleConfigurations.removeAll(forbidden);
				this.left.removeAll(forbidden);
				this.right.removeAll(forbidden);
				this.equivalenceClasses.removeAll(forbidden);
				break;
			}
		}
	}

	/**
	 * @return Set<DAGChaseConfiguration<S>>
	 */
	public Set<DAGChaseConfiguration<S>> getSuccessfulConfigurations() {
		return this.successfulConfigurations;
	}

//	/**
//	 * @return Map<Collection<ApplyRule<S>>,DAGChaseConfiguration<S>>
//	 */
//	public Map<Collection<ApplyRule>,DAGChaseConfiguration> getTemplates() {
//		return this.templates;
//	}

//	/**
//	 * @param configurations List<DAGChaseConfiguration<S>>
//	 */
//	private void updateTemplates(List<DAGChaseConfiguration<S>> configurations) {
//		for(DAGChaseConfiguration<S> configuration:configurations) {
//			Set<ApplyRule> key = Sets.<ApplyRule>newHashSet(configuration.getApplyRules());
//			this.templates.put(key, configuration);
//		}
//	}

	/**
	 * @return DAGEquivalenceClasses<S>
	 */
	public DAGEquivalenceClasses<S> getEquivalenceClasses() {
		return this.equivalenceClasses;
	}
}
