package uk.ac.ox.cs.pdq.planner.dag.explorer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.TimeUnit;

import uk.ac.ox.cs.pdq.LimitReachedException;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.explorer.filters.Filter;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.ReachabilityValidator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.StrictInclusionValidator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.Validator;
import uk.ac.ox.cs.pdq.planner.dag.priority.PriorityAssessor;
import uk.ac.ox.cs.pdq.planner.parallel.IterativeExecutor;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;
import com.google.common.eventbus.EventBus;

/**
 * ILP variation that during the first round works just as ILP, but during the next rounds it considers only configurations
 * that are strictly included within the closed and successful configurations produced during the first round.
 *
 *  A configuration c which comprises the sequence of ApplyRule
	configurations <c_1,c_2,…c_M> (sorted in order of appearance) is
	strictly included in a configuration c’ which comprises the
	sequence of ApplyRule configurations <c’_1,c’_2,…c’_N>  (sorted again in
	order of appearance) with N>=M, if <c_1,c_2,…c_M> is a
	subsequence of <c’_1,c’_2,…c’_N>.
 *
 * @author Efthymia Tsamoura
 *
 */
public class ILPSIExplorer<S extends ChaseState> extends ILPExplorer<S> {

	/** Best k successful and closed configurations*/
	protected int topk;
	
	protected List<DAGChaseConfiguration<S>> topConfigurations;


	/**
	 * 
	 * @param eventBus
	 * @param collectStats
	 * @param initialConfigurations
	 * 		ApplyRule configurations to initilise the explorer
	 * @param filter
	 * 		Filters out configurations at the end of each iteration
	 * @param priority Prioritises configurations
	 * 
	 * @param firstPhaseExecutor
	 * 		Performs parallel chasing
	 * @param secondPhaseExecutor
	 * 		Iterates over all newly created configurations in parallel and returns the best configuration
	 * @param maxDepth
	 * 		The maximum depth to explore
	 * @param maxIterations
	 * @param topk
	 * 		Best k successful and closed configurations
	 * @throws PlannerException
	 */
	public ILPSIExplorer(
			EventBus eventBus, boolean collectStats,
			List<DAGChaseConfiguration<S>> initialConfigurations,
			Filter<S> filter,
			PriorityAssessor<S> priority,
			IterativeExecutor<S> firstPhaseExecutor,
			IterativeExecutor<S> secondPhaseExecutor,
			int maxDepth, int maxIterations, int topk) throws PlannerException {
		super(eventBus,collectStats,
				initialConfigurations,
				filter,
				priority,
				firstPhaseExecutor,
				secondPhaseExecutor,
				maxDepth, maxIterations);
		this.topk = topk;
		this.topConfigurations = new ArrayList<>();
	}

	/**
	 * @throws PlannerException
	 */
	@Override
	protected void _explore() throws PlannerException, LimitReachedException {
		for(Validator<S> validator: this.priority.getValidators()) {
			if(validator instanceof StrictInclusionValidator) {
				((StrictInclusionValidator<S>)validator).setTemplates(this.iterations > 1 && this.topConfigurations.size() == this.topk ? this.topConfigurations : null);
			}
		}
		super._explore();
		if(this.topConfigurations.size() != this.topk) {
			this.topk = this.topConfigurations.size();
		}
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 * @return boolean
	 */
	@Override
	public boolean setBestPlan(DAGChaseConfiguration<S> configuration) {
		boolean ret = false;
		if(this.bestPlan == null ||
				this.bestPlan != null && this.bestPlan.getCost().greaterThan(configuration.getPlan().getCost())) {
			this.bestConfiguration = configuration;
			this.bestPlan = configuration.getPlan();
			this.eventBus.post(this);
			this.eventBus.post(this.getBestPlan());
			ret = true;
		}
		if(this.topConfigurations.size() < this.topk) {
			this.topConfigurations.add(configuration);
		}
		else {
			if(this.topConfigurations.get(this.topConfigurations.size()-1).getPlan().getCost().greaterThan(configuration.getPlan().getCost())) {
				this.topConfigurations.remove(this.topConfigurations.size()-1);
				this.topConfigurations.add(configuration);
				Collections.sort(this.topConfigurations);
			}
		}
		return ret;
	}

	/**
	 * @return Collection<DAGChaseConfiguration<S>>
	 * @throws PlannerException
	 */
	@Override
	protected Collection<DAGChaseConfiguration<S>> runChaseOrPropagatePhase() throws PlannerException, LimitReachedException {
		return this.firstPhaseExecutor.chaseOrPropagate(this.depth,
				this.left,
				this.right,
				this.priority,
				this.templates,
				this.topConfigurations.size() < this.topk ? null : this.topConfigurations.get(this.topConfigurations.size()-1),
						false,
						Double.valueOf((this.maxElapsedTime - (this.elapsedTime/1e6))).longValue(),
						TimeUnit.MILLISECONDS);
	}

	/**
	 * @param filter Filter<S>
	 * @param validators List<Validator<S>>
	 * @return boolean
	 */
	@Override
	protected boolean checkFilterAndValidator(Filter<S> filter, List<Validator<S>> validators) {
		Preconditions.checkArgument(filter == null);
		Preconditions.checkArgument(validators != null);
		Preconditions.checkArgument(!validators.isEmpty());
		for(Validator<S> validator: validators) {
			Preconditions.checkState(validator instanceof StrictInclusionValidator || validator instanceof ReachabilityValidator);
		}
		return true;
	}
}
