package uk.ac.ox.cs.pdq.planner.dag.explorer;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.TimeUnit;

import uk.ac.ox.cs.pdq.LimitReachedException;
import uk.ac.ox.cs.pdq.plan.DAGPlan;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.PlannerParameters.DominanceTypes;
import uk.ac.ox.cs.pdq.planner.PlannerParameters.SuccessDominanceTypes;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.DAGConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationCostEstimator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.filters.Filter;
import uk.ac.ox.cs.pdq.planner.dag.priority.Distance2CoreTopK;
import uk.ac.ox.cs.pdq.planner.dag.priority.Distance2CoreTopKHardCoded;
import uk.ac.ox.cs.pdq.planner.dag.priority.PriorityAssessor;
import uk.ac.ox.cs.pdq.planner.parallel.FinalIterationThreadResults;
import uk.ac.ox.cs.pdq.planner.parallel.IterativeExecutor;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.DominanceFactory;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.SuccessDominance;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.SuccessDominanceFactory;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.eventbus.EventBus;

/**
 * Mixed ILP explorer.
 * Every odd number of rounds it calls the ILP explorer.
 * Every even number of rounds it creates binary configurations
 * by considering sub-configurations of the successful configurations found during the previous round.
 *
 * @author Efthymia Tsamoura
 *
 */
public class MixedILPExplorer<S extends ChaseState> extends ILPExplorer<S> {

	/** ILP explorer */
	private ILPExplorer<S> ilp;

	private int bushiness = 0;
	private final int maxBushiness;
	private final PriorityAssessor<S> distance2core;
	private final ConfigurationCostEstimator costEstimator;


	/**
	 * 
	 * @param eventBus
	 * @param collectStats
	 * @param initialConfigurations
	 * 		ApplyRule configurations to initilise the explorer
	 * @param filter
	 * 		Filters out configurations at the end of each iteration
	 * @param priority Prioritises configurations
	 * 
	 * @param firstPhaseExecutor
	 * 		Performs parallel chasing
	 * @param secondPhaseExecutor
	 * 		Iterates over all newly created configurations in parallel and returns the best configuration
	 * @param maxDepth
	 * 		The maximum depth to explore
	 * @param maxIterations
	 * @param maxBushiness
	 * 		The maximum number of bushiness to explore
	 * @throws PlannerException
	 */
	public MixedILPExplorer(
			EventBus eventBus, boolean collectStats,
			List<DAGChaseConfiguration<S>> initialConfigurations,
			Filter<S> filter,
			PriorityAssessor<S> priority,
			IterativeExecutor<S> firstPhaseExecutor,
			IterativeExecutor<S> secondPhaseExecutor,
			int maxDepth, int maxIterations,
			int maxBushiness) throws PlannerException {
		super(eventBus, collectStats,
				initialConfigurations,
				filter,
				priority,
				firstPhaseExecutor,
				secondPhaseExecutor,
				maxDepth,
				maxIterations);

		if(priority.getClass().equals(Distance2CoreTopK.class)) {
			this.distance2core = new Distance2CoreTopKHardCoded<>(priority.getValidators(),
					priority.getPotential(),
					((Distance2CoreTopK<S>) priority).getTopk(),
					((Distance2CoreTopK<S>) priority).getRanking(),
					((Distance2CoreTopK<S>) priority).getSeed());
		}
		else {
			this.distance2core = null;
		}
		this.maxBushiness = maxBushiness;
		this.costEstimator = initialConfigurations.get(0).getCostEstimator();

		//Initialise an ILP explorer
		this.ilp = new ILPExplorer<>(eventBus, collectStats,
				this.applyRuleConfigurations,
				new ConcurrentLinkedQueue<>(this.applyRuleConfigurations),
				new LinkedHashSet<>(this.applyRuleConfigurations),
				this.equivalenceClasses,
				this.templates,
				filter,
				this.distance2core != null ? this.distance2core : this.priority,
						firstPhaseExecutor,
						secondPhaseExecutor,
						1,
						maxDepth,
						1);
	}


	/**
	 * @throws PlannerException
	 */
	@Override
	protected void _explore() throws PlannerException, LimitReachedException {
		if(this.iterations > this.maxIterations || this.bushiness > this.maxBushiness) {
			this.forcedTermination = true;
			return;
		}

		this.checkLimitReached();
		//Run the ILP explorer
		this.ilp.explore();
		this.iterations++;
		this.bushiness++;

		if(this.ilp.getBestConfiguration() == null && this.iterations == 1) {
			this.forcedTermination = true;
		}
		else if(this.iterations > 1){
			if(this.ilp.getBestConfiguration() != null) {
				this.setBestPlan(this.ilp.getBestConfiguration());
			}

			// Find out all sub-configurations of the successful configurations
			// returned during the previous round
			Set<DAGChaseConfiguration<S>> subconfigurations = new LinkedHashSet<>();
			for(DAGChaseConfiguration<S> configuration:this.ilp.getSuccessfulConfigurations()) {
				for(DAGConfiguration subconfiguration:configuration.getSubconfigurations()) {
					subconfigurations.add((DAGChaseConfiguration<S>) subconfiguration);
				}
			}
			subconfigurations.removeAll(this.ilp.getSuccessfulConfigurations());
			subconfigurations.retainAll(this.equivalenceClasses.getConfigurations());

			// During the first call of ILP we used closed (success) domination, in order not to miss the optimal found by
			// when using the Linear validator (it produces closed left-deep configurations in each round).
			// For the subsequent iterations, we only use open (success) domination
			if(this.iterations == 2) {
				DominanceFactory df = new DominanceFactory(DominanceTypes.OPEN, true);
				SuccessDominanceFactory<DAGPlan> sdf = new SuccessDominanceFactory<>(this.costEstimator, SuccessDominanceTypes.OPEN);
				this.priority.getPotential().update(sdf.getInstance());

				for(DAGChaseConfiguration<S> configuration:subconfigurations) {
					configuration.setDominance(df.getInstance());
					configuration.setSuccessDominance(sdf.getInstance());
				}
				for(DAGChaseConfiguration<S> configuration:this.applyRuleConfigurations) {
					configuration.setDominance(df.getInstance());
					configuration.setSuccessDominance(sdf.getInstance());
				}
				for(DAGChaseConfiguration<S> configuration:this.equivalenceClasses.getConfigurations()) {
					configuration.setDominance(df.getInstance());
					configuration.setSuccessDominance(sdf.getInstance());
				}
			}

			//Create binary configurations by considering sub-configurations of the successful configurations
			//found during the previous round.
			this.checkLimitReached();
			Collection<DAGChaseConfiguration<S>> configurations =
					this.firstPhaseExecutor.chaseOrPropagate(
							-1,
							new ConcurrentLinkedQueue<>(subconfigurations),
							new LinkedHashSet<>(subconfigurations),
							this.priority,
							this.templates,
							this.bestConfiguration,
							false,
							Double.valueOf((this.maxElapsedTime - (this.elapsedTime/1e6))).longValue(),
							TimeUnit.MILLISECONDS);


			if(configurations == null || configurations.isEmpty()) {
				this.forcedTermination = true;
				return;
			}
			SuccessDominance successDominance = configurations.iterator().next().getSuccessDominanceDetector();

			this.checkLimitReached();
			//Iterate over all newly created configurations in parallel and return the best configuration
			FinalIterationThreadResults<S> results = this.secondPhaseExecutor.finalIteration(
					new ConcurrentLinkedQueue<>(configurations),
					this.equivalenceClasses,
					this.bestConfiguration,
					successDominance,
					Double.valueOf((this.maxElapsedTime - (this.elapsedTime/1e6))).longValue(),
					TimeUnit.MILLISECONDS);


			if (results == null) {
				this.forcedTermination = true;
				return;
			}
			List<DAGChaseConfiguration<S>> output = results.getOutput();
			if(results.getBest()!= null) {
				this.setBestPlan(results.getBest());
			}

			if (output.isEmpty()) {
				this.forcedTermination = true;
				return;
			}

			this.iterations++;

			output.retainAll(this.equivalenceClasses.getConfigurations());

			//Instantiate ILP using as input the output of the previous round
			this.priority.getPotential().update(this.bestConfiguration == null ? null : this.bestConfiguration.getPlan());
			this.ilp = new ILPExplorer<>(this.eventBus,
					true,
					this.applyRuleConfigurations,
					new ConcurrentLinkedQueue<>(output),
					new LinkedHashSet<>(this.applyRuleConfigurations),
					this.equivalenceClasses,
					this.templates,
					this.filter,
					this.priority,
					this.firstPhaseExecutor,
					this.secondPhaseExecutor,
					2,
					this.maxDepth,
					1);

			if(this.bestConfiguration != null) {
				this.ilp.setBestPlan(this.bestConfiguration);
			}
		}
	}
}
