package uk.ac.ox.cs.pdq.planner.dag.explorer;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.LimitReachedException;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.dag.BinaryConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.explorer.filters.Filter;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.Validator;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.eventbus.EventBus;

/**
 * Simple dag explorer. The exploration proceeds similarly to the GenericExplorer.
 * First, it creates a new binary configuration and checks whether or not it could lead to the optimal solution.
 * Each newly created configuration is only further considered
 * if it is not dominated by an existing configuration.
 *
 * @author Efthymia Tsamoura
 *
 */
public class SimpleDPExplorer<S extends ChaseState> extends GenericExplorer<S> {

	/**
	 * @param eventBus
	 * @param collectStats
	 * @param schema
	 * 		The input schema
	 * @param accessibleSchema
	 *      The accessible version of the input schema
	 * @param query
	 *      The input query
	 * @param accessibleQuery
	 * @param factory
	 * 		Returns the ApplyRule configurations
	 * @param maxDepth
	 *      The maximum depth in the dag we can explore
	 * @param filter Filter<S>
	 * @param validator Validator<S>
	 * @throws PlannerException
	 */
	public SimpleDPExplorer(
			EventBus eventBus, boolean collectStats,
			List<DAGChaseConfiguration<S>> initialConfigurations,
			Filter<S> filter, Validator<S> validator,
			int maxDepth, boolean orderAware) throws PlannerException {
		super(eventBus, collectStats,
				initialConfigurations, filter, validator, maxDepth, orderAware);
	}

	/**
	 * @return Collection<DAGChaseConfiguration<S>>
	 * @throws PlannerException
	 */
	@Override
	protected Collection<DAGChaseConfiguration<S>> mainLoop() throws PlannerException, LimitReachedException {
		Map<Pair<DAGChaseConfiguration<S>,DAGChaseConfiguration<S>>,DAGChaseConfiguration<S>> last = new HashMap<>();
		Pair<DAGChaseConfiguration<S>, DAGChaseConfiguration<S>> pair = null;
		while ((pair = this.selector.getNext(this.depth)) != null) {
			if(!last.containsKey(pair)) {
				BinaryConfiguration<S> configuration = new BinaryConfiguration<>(
//						pair.getLeft().getSchema(),
						pair.getLeft().getAccessibleSchema(),
						pair.getLeft().getQuery(),
						pair.getLeft().getChaser(),
						null,
						pair.getLeft().getDominanceDetectors(),
						pair.getLeft().getSuccessDominanceDetector(),
						pair.getLeft().getCostEstimator(),
						pair.getLeft(),
						pair.getRight(),
						true);
				if (ExplorerUtils.isDominated(this.getRight(), configuration) == null &&
						ExplorerUtils.isDominated(last.values(), configuration) == null &&
						ExplorerUtils.getPotential(configuration, this.bestPlan)) {
					if (configuration.isClosed()
							&& configuration.isSuccessful()) {
						this.setBestPlan(configuration);
					} else {
						last.put(pair, configuration);
					}
				}
			}
			if (this.checkLimitReached()) {
				this.forcedTermination = true;
				break;
			}
		}
		return last.values();
	}
}
