package uk.ac.ox.cs.pdq.planner.dag.explorer.validators;

import java.util.Collection;
import java.util.HashSet;

import org.apache.commons.collections4.CollectionUtils;

import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.planner.dag.ApplyRule;
import uk.ac.ox.cs.pdq.planner.dag.ConfigurationUtility;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.explorer.filters.Filter;
import uk.ac.ox.cs.pdq.planner.dag.explorer.filters.ReachabilityFilter;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;


/**
 * Requires the input configurations to be non trivial and to be either neighbouring in the chase graph or the left configuration to expose constants
 * that are input for the right configuration
 * @author Efthymia Tsamoura
 *
 * @param <S>
 */
public class ReachabilityValidator<S extends ChaseState> implements Validator<S>{

	/** Currently not used*/
	private final Collection<DAGChaseConfiguration<S>> forbidden;
	/** Maintains information w.r.t. chasing plus answers proximity queries*/
	private final FiringGraph<S> firingGraph;
	/** The database core*/
	private final Collection<PredicateFormula> core;

	/**
	 * Constructor for ReachabilityValidator.
	 * @param query Query
	 * @param input Collection<DAGConfiguration<S>>
	 * @param firingGraph FiringGraph<S>
	 */
	public ReachabilityValidator(Query<?> query, Collection<DAGChaseConfiguration<S>> input, FiringGraph<S> firingGraph) {
		Filter<S> reachabilityFilter = new ReachabilityFilter<>(query);
		this.firingGraph = firingGraph;
		this.forbidden = reachabilityFilter.filter(input);
		this.core = query.getCanonicalQuery().getPredicates();
	}

	/**
	 * Constructor for ReachabilityValidator.
	 * @param forbidden Collection<DAGConfiguration<S>>
	 * @param core Collection<PredicateFormula>
	 * @param firingGraph FiringGraph<S>
	 */
	private ReachabilityValidator(Collection<DAGChaseConfiguration<S>> forbidden, Collection<PredicateFormula> core, FiringGraph<S> firingGraph) {
		this.forbidden = forbidden;
		this.core = core;
		this.firingGraph = firingGraph;
	}

	/**
	 *
	 * @param right
	 * @return true if the right configuration contains at least one core fact
	 */
	private boolean isCoreConfiguration(DAGChaseConfiguration<S> right) {
		for(ApplyRule<S> applyRule:right.getApplyRules()) {
			if(CollectionUtils.containsAny(applyRule.getFacts(), this.core)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @return boolean
	 * @see uk.ac.ox.cs.pdq.dag.explorer.validators.Validator#validate(DAGConfiguration<S>, DAGConfiguration<S>)
	 */
	@Override
	public boolean validate(DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right) {
		return ConfigurationUtility.isNonTrivial(left, right)
				&& (CollectionUtils.containsAny(left.getProperOutput(), right.getInput())
						|| this.isCoreConfiguration(right)
						|| this.areNeighbors(this.firingGraph, left, right));
	}

	/**
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @param depth int
	 * @return boolean
	 * @see uk.ac.ox.cs.pdq.dag.explorer.validators.Validator#validate(DAGConfiguration<S>, DAGConfiguration<S>, int)
	 */
	@Override
	public boolean validate(DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right, int depth) {
		return left.getHigh() + right.getHigh() == depth && this.validate(left, right);
	}

	/**
	 * @return Validator<S>
	 * @see uk.ac.ox.cs.pdq.planner.dag.explorer.validators.Validator#clone()
	 */
	@Override
	public Validator<S> clone() {
		return new ReachabilityValidator<>(this.forbidden, this.core, this.firingGraph);
	}

	/**
	 * @return Collection<DAGConfiguration<S>>
	 */
	public Collection<DAGChaseConfiguration<S>> getForbidden() {
		return this.forbidden;
	}
	
	
	/**
	 *
	 * @param source
	 * 		Input dag configuration
	 * @param target
	 * 		Input dag configuration
	 * @return
	 * 		true if the input dag configurations are neighbours.
	 * 		Two dag configurations are neighbours if they corresponding ApplyRule sub-configurations consists of facts that are neighbouring in the chase graph. * @see uk.ac.ox.cs.pdq.chase.FiringGraph#areNeighbors(DAGConfiguration<S>, DAGConfiguration<S>)
	 */
	public boolean areNeighbors(FiringGraph<S> firingGraph, DAGChaseConfiguration<S> source, DAGChaseConfiguration<S> target) {

		//Get the ApplyRule configurations of the source configuration
		Collection<PredicateFormula> sources = new HashSet<>();
		for(ApplyRule<S> applyRule:source.getApplyRules()) {
			sources.addAll(applyRule.getFacts());
		}

		//Get the ApplyRule configurations of the target configuration
		Collection<PredicateFormula> targets = new HashSet<>();
		for(ApplyRule<S> applyRule:target.getApplyRules()) {
			targets.addAll(applyRule.getFacts());
		}

		for(Collection<PredicateFormula> key:this.firingGraph.getPreconditions()) {
			//For each ApplyRule in the source configuration,
			//if any of its facts was used to produce any of the facts in any ApplyRule configuration of the target configuration
			//then the corresponding configurations are neighbours
			if(CollectionUtils.containsAny(sources, key)) {
				for(Collection<PredicateFormula> facts:this.firingGraph.getConsequences(key)) {
					if(CollectionUtils.containsAny(facts, targets)) {
						return true;
					}
				}
			}
			if(CollectionUtils.containsAny(targets, key)) {
				for(Collection<PredicateFormula> facts:this.firingGraph.getConsequences(key)) {
					if(CollectionUtils.containsAny(facts, sources)) {
						return true;
					}
				}
			}
		}
		return false;
	}

}
