package uk.ac.ox.cs.pdq.planner.dag.explorer.validators;

import java.util.Collection;
import java.util.Iterator;

import uk.ac.ox.cs.pdq.planner.dag.ConfigurationUtility;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.collect.Sets;


/**
 * Requires the input pair of configurations to be non trivial
 * and their combination to be strictly included in one of the templates.
 *
 *	A configuration c which comprises the sequence of ApplyRule
	configurations <c_1,c_2,…c_M> (sorted in order of appearance) is
	strictly included in a configuration c’ which comprises the
	sequence of ApplyRule configurations <c’_1,c’_2,…c’_N>  (sorted again in
	order of appearance) with N>=M, if <c_1,c_2,…c_M> is a
	subsequence of <c’_1,c’_2,…c’_N>.

	Only used by the ILPSIExplorer
 *
 * @author Efthymia Tsamoura
 *
 * @param <S>
 */
public class StrictInclusionValidator<S extends ChaseState> implements Validator<S>{

	private Collection<DAGChaseConfiguration<S>> templates = null;

	public StrictInclusionValidator() {
	}

	/**
	 * Constructor for StrictInclusionValidator.
	 * @param templates Collection<DAGConfiguration<S>>
	 */
	private StrictInclusionValidator(Collection<DAGChaseConfiguration<S>> templates) {
		this.templates = templates;
	}

	/**
	 * @param templates Collection<DAGConfiguration<S>>
	 */
	public void setTemplates(Collection<DAGChaseConfiguration<S>> templates) {
		if(templates != null) {
			this.templates = Sets.newLinkedHashSet(templates);
		}
	}

	/**
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @return boolean
	 * @see uk.ac.ox.cs.pdq.dag.explorer.validators.Validator#validate(DAGConfiguration<S>, DAGConfiguration<S>)
	 */
	@Override
	public boolean validate(DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right) {
		if(ConfigurationUtility.isNonTrivial(left, right)) {
			if(this.templates != null) {
				Iterator<DAGChaseConfiguration<S>> iterator = this.templates.iterator();
				while(iterator.hasNext()) {
					DAGChaseConfiguration<S> top = iterator.next();
					if(top.strictlyIncludes(left, right)) {
						return true;
					}
				}
				return false;
			}
			return true;
		}
		return false;
	}

	/**
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @param depth int
	 * @return boolean
	 * @see uk.ac.ox.cs.pdq.dag.explorer.validators.Validator#validate(DAGConfiguration<S>, DAGConfiguration<S>, int)
	 */
	@Override
	public boolean validate(DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right, int depth) {
		return left.getHigh() + right.getHigh() == depth && this.validate(left, right);
	}

	/**
	 * @return Validator<S>
	 * @see uk.ac.ox.cs.pdq.planner.dag.explorer.validators.Validator#clone()
	 */
	@Override
	public Validator<S> clone() {
		return new StrictInclusionValidator<>(this.templates);
	}

}
