package uk.ac.ox.cs.pdq.planner.dag.explorer.validators;

import uk.ac.ox.cs.pdq.planner.PlannerParameters.ValidatorTypes;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

/**
 * Creates validators (shape restrictions).
 * The available options are
 * 		DEFAULT: No shape restrictions
 * 		FACT_THROTTLING: At least one of the configurations that will be composed must be an ApplyRule
 * 		DEPTH_THROTTLING: The combined depth of the configurations that will be composed must be < a depth threshold
 * 		RIGHT_DEPTH_THROTTLING: The right's configuration depth must be < a depth threshold
 * 		FACT_DEPTH_THROTTLING: Combination of the FACT_THROTTLING and DEPTH_THROTTLING
 * 		STRICT_INCLUSION: The composition of the input configurations must be strictly included in an input template
 *
 * @author Efthymia Tsamoura
 *
 */
public class ValidatorFactory<S extends ChaseState> {

	private final ValidatorTypes type;

	private final int depthThreshold;

	/**
	 * Constructor for ValidatorFactory.
	 * @param type ValidatorTypes
	 */
	public ValidatorFactory(ValidatorTypes type) {
		this.type = type;
		this.depthThreshold = 3;
	}

	/**
	 * Constructor for ValidatorFactory.
	 * @param type ValidatorTypes
	 * @param depthThreshold int
	 */
	public ValidatorFactory(ValidatorTypes type, int depthThreshold) {
		this.type = type;
		this.depthThreshold = depthThreshold;
	}

	/**
	 * @return Validator<S>
	 */
	public Validator<S> getInstance() {
		switch(this.type) {
		case DEFAULT_VALIDATOR:
			return new DefaultValidator<>();
		case APPLYRULE_VALIDATOR:
			return new ApplyRuleValidator<>();
		case DEPTH_VALIDATOR:
			return new DepthValidator<>(this.depthThreshold);
		case RIGHT_DEPTH_VALIDATOR:
			return new RightDepthValidator<>(this.depthThreshold);
		case APPLYRULE_DEPTH_VALIDATOR:
			return new ApplyRuleDepthValidator<>(this.depthThreshold);
		case LINEAR_VALIDATOR:
			return new LinearValidator<>();
		case STRICT_INCLUSION:
			return new StrictInclusionValidator<>();
		default:
			break;
		}
		return null;
	}
}
