package uk.ac.ox.cs.pdq.planner.dag.potential;

import uk.ac.ox.cs.pdq.plan.Plan;
import uk.ac.ox.cs.pdq.planner.dag.ConfigurationUtility;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationCostEstimator;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.SuccessDominance;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

/**
 * A plan has the potential to lead to the minimum-cost configuration if it is not success dominated by the best plan found so far
 *
 * @author Efthymia Tsamoura
 *
 * @param <S>
 */
public class DefaultPotentialAssessor<S extends ChaseState> implements PotentialAssessor<S>{

	private Plan bestPlan;
	private final ConfigurationCostEstimator costEstimator;
	private SuccessDominance successDominance;

	/**
	 * Constructor for DefaultPotentialAssessor.
	 * @param bestPlan Plan
	 * @param costEstimator CostEstimator<DAGPlan>
	 * @param successDominance SuccessDominance
	 */
	public DefaultPotentialAssessor(Plan bestPlan,
			ConfigurationCostEstimator costEstimator,
			SuccessDominance successDominance) {
		this.bestPlan = bestPlan;
		this.costEstimator = costEstimator;
		this.successDominance = successDominance;
	}

	/**
	 * @param configuration DAGConfiguration<S>
	 * @return Boolean
	 * @see uk.ac.ox.cs.pdq.potential.PotentialAssessor#getPotential(DAGConfiguration<S>)
	 */
	@Override
	public Boolean getPotential(DAGChaseConfiguration<S> configuration) {
		return ConfigurationUtility.getPotential(configuration, this.bestPlan, this.successDominance);
	}

	/**
	 * @param left DAGConfiguration<S>
	 * @param right DAGConfiguration<S>
	 * @return Boolean
	 * @see uk.ac.ox.cs.pdq.potential.PotentialAssessor#getPotential(DAGConfiguration<S>, DAGConfiguration<S>)
	 */
	@Override
	public Boolean getPotential(DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right) {
		return ConfigurationUtility.getPotential(left, right, this.bestPlan, this.costEstimator, this.successDominance);
	}

	/**
	 * @param bestPlan Plan
	 * @see uk.ac.ox.cs.pdq.planner.dag.potential.PotentialAssessor#update(Plan)
	 */
	@Override
	public void update(Plan bestPlan) {
		this.bestPlan = bestPlan;
	}

	/**
	 * @param sd SuccessDominance
	 * @see uk.ac.ox.cs.pdq.planner.dag.potential.PotentialAssessor#update(SuccessDominance)
	 */
	@Override
	public void update(SuccessDominance sd) {
		this.successDominance = sd;
	}

	/**
	 * @return DefaultPotentialAssessor<S>
	 * @see uk.ac.ox.cs.pdq.planner.dag.potential.PotentialAssessor#clone()
	 */
	@Override
	public DefaultPotentialAssessor<S> clone() {
		return new DefaultPotentialAssessor<>(this.bestPlan, this.costEstimator.clone(), this.successDominance.clone());
	}
}
