package uk.ac.ox.cs.pdq.planner.dag.priority;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.DefaultValidator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.Validator;
import uk.ac.ox.cs.pdq.planner.dag.potential.PotentialAssessor;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;
import com.google.common.collect.MinMaxPriorityQueue;


/**
 * Returns all configurations that are within a specific distance from the core
 * @author Efthymia Tsamoura
 *
 * @param <S>
 */
public class Distance2CoreRange<S extends ChaseState> extends Distance2CoreTopK<S>{

	/** The maximum distance of a configuration from the core */
	private final Integer range;

	/**
	 * @param validators Validate pairs of configurations to be composed
	 * @param potential Assesses the potential of a configuration to lead to the minimum-cost configuration
	 * @param range The maximum distance of a configuration from the core
	 * @param query The input query
	 * @param firingGraph Maintains the chase graph
	 * @param seed
	 */
	public Distance2CoreRange(List<Validator<S>> validators,
			PotentialAssessor<S> potential,
			Integer range,
			Query<?> query,
			FiringGraph<S> firingGraph,
			Integer seed) {
		super(validators, potential, Integer.MAX_VALUE, query, firingGraph, seed);
		this.range = range;
	}

	/**
	 * @param validators Validate pairs of configurations to be composed
	 * @param potential Assesses the potential of a configuration to lead to the minimum-cost configuration
	 * @param range The maximum distance of a configuration from the core
	 * @param ranking Keeps for each chase fact its distance from the core (the canonical database)
	 * @param seed
	 */
	public Distance2CoreRange(
			List<Validator<S>> validators,
			PotentialAssessor<S> potential,
			Integer range,
			Map<PredicateFormula, Double> ranking,
			Integer seed) {
		super(validators, potential, Integer.MAX_VALUE, ranking, seed);
		this.range = range;
	}

	/**
	 *
	 * @param left
	 * @param right Candidate configurations to combine with the left input one
	 * @param depth The target depth of the composition
	 * @return the configurations to combine with the left input one that are within a specific distance from the core
	 */
	@Override
	public Collection<DAGChaseConfiguration<S>> select(DAGChaseConfiguration<S> left, Collection<DAGChaseConfiguration<S>> right, int depth) {
		Set<DAGChaseConfiguration<S>> pool = new TreeSet<>(new DAGConfigurationComparator<S>());
		pool.addAll(right);
		return this.hiddenSelect(left, pool, depth);
	}

	/**
	 * @param left DAGConfiguration<S>
	 * @param right Collection<DAGConfiguration<S>>
	 * @param depth int
	 * @return Collection<DAGConfiguration<S>>
	 */
	private Collection<DAGChaseConfiguration<S>> hiddenSelect(DAGChaseConfiguration<S> left, Collection<DAGChaseConfiguration<S>> right, int depth) {
		MinMaxPriorityQueue<DAGChaseConfiguration<S>> selected = MinMaxPriorityQueue.orderedBy(this.getComparator()).maximumSize(this.getTopk()).create();
		for(DAGChaseConfiguration<S> configuration:right) {
			Preconditions.checkNotNull(configuration.getEquivalenceClass());
			Preconditions.checkState(!configuration.getEquivalenceClass().isEmpty());
			if(!configuration.getEquivalenceClass().isSleeping() &&
					this.validate(left, configuration, depth) &&
					this.getPotential().getPotential(left, configuration) ){
				this.getRanking(configuration);
				if(configuration.getRanking() <= this.range) {
					selected.add(configuration);
				}
			}
		}
		return selected;
	}

	/**
	 * @return Distance2CoreRange<S>
	 */
	@Override
	public Distance2CoreRange<S> clone() {
		return new Distance2CoreRange<>(DefaultValidator.deepCopy(this.getValidators()),
				this.getPotential().clone(), this.range, this.getRanking(), this.getSeed());
	}
}
