package uk.ac.ox.cs.pdq.planner.db.access;

import java.util.ArrayList;
import java.util.List;

import uk.ac.ox.cs.pdq.db.AccessMethod;
import uk.ac.ox.cs.pdq.db.AccessMethod.AccessMethodTypes;
import uk.ac.ox.cs.pdq.db.GuardedDependency;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.TGD;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessedRelation;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessibleRelation;

/**
 *
 * An accessibility axiom
 * @author Efthymia Tsamoura
 * @author Julien Leblay
 */
public class AccessibilityAxiom extends TGD implements GuardedDependency {

	/** The accessed relation of this axiom*/
	private final AccessedRelation accessedRelation;
	/** The relation's access method */
	private final AccessMethod method;

	/**
	 * Creates an accessibility axiom for the input relation and binding
	 * @param accessedRelation
	 * @param method
	 */
	public AccessibilityAxiom(AccessedRelation accessedRelation, AccessMethod method) {
		super(createLeft(accessedRelation, method), createRight(accessedRelation, method));
		this.method = method;
		this.accessedRelation = accessedRelation;
	}

	/**
	 * Creates the left-hand side of an accessibility axiom for the input relation and binding
	 * @param accessedRelation
	 * @param method
	 * @return a conjunction of predicates corresponding to the LHS of an
	 * accessibility axiom
	 */
	private static Conjunction<PredicateFormula> createLeft(AccessedRelation accessedRelation, AccessMethod method) {
		List<PredicateFormula> leftAtoms = new ArrayList<>();
		List<Integer> bindingPositions = method.getInputs();
		Relation r = AccessibleRelation.getInstance();
		PredicateFormula f = accessedRelation.getBaseRelation().createAtoms();
		List<Term> terms = f.getTerms();
		for (int bindingPos: bindingPositions) {
			if (method.getType() != AccessMethodTypes.FREE) {
				leftAtoms.add(new PredicateFormula(r, terms.get(bindingPos - 1)));
			}
		}
		leftAtoms.add(f);
		return Conjunction.of(leftAtoms);
	}

	/**
	 * Creates the right-hand side of an accessibility axiom for the input relation and binding
	 * @param accessedRelation
	 * @param method
	 * @return a conjunction of predicates corresponding to the RHS of an
	 * accessibility axiom
	 */
	private static Conjunction<PredicateFormula> createRight(AccessedRelation accessedRelation, AccessMethod method) {
		Relation relation = accessedRelation.getBaseRelation();
		List<PredicateFormula> rightAtoms = new ArrayList<>();
		List<Integer> bindingPositions = method.getInputs();
		Relation accessible = AccessibleRelation.getInstance();
		PredicateFormula f = accessedRelation.createAtoms();
		List<Term> terms = f.getTerms();
		for (int i = 1; i <= relation.getArity(); ++i) {
			if (!bindingPositions.contains(i)) {
				rightAtoms.add(new PredicateFormula(accessible, terms.get(i - 1)));
			}
		}
		rightAtoms.add(f);
		return Conjunction.of(rightAtoms);
	}

	/**
	 * @return the accessed relation of the accessibility axiom.
	 */
	public AccessedRelation getAccessedRelation() {
		return this.accessedRelation;
	}

	/**
	 * @return the base relation of the accessibility axiom.
	 */
	public Relation getBaseRelation() {
		return this.accessedRelation.getBaseRelation();
	}

	/**
	 * @return the access method of the accessibility axiom.
	 */
	public AccessMethod getAccessMethod() {
		return this.method;
	}

	/**
	 * @return PredicateFormula
	 * @see uk.ac.ox.cs.pdq.db.GuardedDependency#getGuard()
	 */
	@Override
	public PredicateFormula getGuard() {
		return this.left.getPredicates().get(this.left.size()-1);
	}
}