package uk.ac.ox.cs.pdq.planner.db.access;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.db.AccessMethod;
import uk.ac.ox.cs.pdq.db.Attribute;
import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.Schema;
import uk.ac.ox.cs.pdq.db.TGD;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.db.View;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.ConjunctiveQuery;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableListMultimap;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;

/**
 * Accessible schema
 *
 * @author Efthymia Tsamoura
 * @author Julien Leblay
 *
 */
public class AccessibleSchema extends Schema {

	/** The accessible relations*/
	private final List<AccessibleRelation> accessibleRelation;

	/** Mapping from schema relations to the corresponding accessed relations*/
	private final Map<String, AccessedRelation> accessedRelations;

	/** Mapping from schema relations to the corresponding inferred accessed relations*/
	private final Map<String, InferredAccessibleRelation> infAccessibleRelations;

	/** Mapping from schema relations to the corresponding defining axioms*/
	private final Multimap<String, DefiningAxiom> definingAxioms;

	/** Mapping from a relation binding pair to an accessibility axiom*/
	private final Map<Pair<? extends Relation, AccessMethod>, AccessibilityAxiom> accessibilityAxioms;

	/** Mapping from a relation binding pair to an accessibility axioms, not relying on AccessedRelations */
	private final Map<Pair<? extends Relation, AccessMethod>, FastAccessibilityAxiom> fastAccessibilityAxioms;

	/** Mapping from a relation to an axiom of the form "AccessedR->InferredAccessedR"*/
	private final Map<String, InferredAccessibleAxiom> inferredAccessibilityAxioms;

	/** Mapping from a dependency to its inferred accessible version*/
	private final Map<Constraint, InferredAccessibleAxiom> shortInferredAccessibilityAxioms;

	private final List<InferredAccessibleAxiom> inferredAccessibleViews = new ArrayList<>();

	/**
	 *
	 * @param relations
	 * 		List of schema relations
	 * @param ics
	 * 		list if schema ICs
	 */
	public AccessibleSchema(List<Relation> relations, List<Constraint> ics, Map<String, TypedConstant<?>> constantsMap) {
		super(relations, ics);
		this.constantsMap = constantsMap;
		ImmutableMap.Builder<String, AccessedRelation> b1 = ImmutableMap.builder();
		ImmutableMap.Builder<String, InferredAccessibleRelation> b2 = ImmutableMap.builder();
		ImmutableListMultimap.Builder<String, DefiningAxiom> b3 = ImmutableListMultimap.builder();
		ImmutableMap.Builder<Pair<? extends Relation, AccessMethod>, AccessibilityAxiom> b4 = ImmutableMap.builder();
		ImmutableMap.Builder<Pair<? extends Relation, AccessMethod>, FastAccessibilityAxiom> f4 = ImmutableMap.builder();
		ImmutableMap.Builder<String, uk.ac.ox.cs.pdq.planner.db.access.InferredAccessibleAxiom> b5 = ImmutableMap.builder();
		for (Relation relation:relations) {
			AccessedRelation acc = new AccessedRelation(relation);
			InferredAccessibleRelation infAcc = new InferredAccessibleRelation(relation);
			b1.put(relation.getName(), acc);
			b2.put(relation.getName(), infAcc);

			boolean isAccessible = false;
			// Defining axioms/
			for (Attribute attribute : relation.getAttributes()) {
				b3.put(relation.getName(), new DefiningAxiom(acc, attribute));
			}
			// Accessibility axioms/
			for (AccessMethod bindingMethod:relation.getAccessMethods()) {
				b4.put(Pair.of(relation, bindingMethod), new AccessibilityAxiom(acc, bindingMethod));
				f4.put(Pair.of(relation, bindingMethod), new FastAccessibilityAxiom(infAcc, bindingMethod));
				isAccessible = true;
			}
			// Inferred accessible axioms/
			if (isAccessible) {
				b5.put(relation.getName(), new InferredAccessibleAxiom(infAcc, acc));
			}
		}

		ImmutableMap.Builder<Constraint, InferredAccessibleAxiom> b6 = ImmutableMap.builder();

		this.accessibleRelation = ImmutableList.of(AccessibleRelation.getInstance());
		this.accessedRelations = b1.build();
		this.infAccessibleRelations = b2.build();

		// Inferred accessible axioms the schema ICs
		for (Constraint<?, ?> ic: ics) {
			Map<PredicateFormula, InferredAccessibleRelation> predicateToInfAccessibleRelation = new LinkedHashMap<>();
			for (PredicateFormula p: ic.getPredicates()) {
				predicateToInfAccessibleRelation.put(p, this.infAccessibleRelations.get(p.getPredicateName()));
			}
			if (ic instanceof TGD) {
				InferredAccessibleAxiom infAcc = new InferredAccessibleAxiom((TGD) ic, predicateToInfAccessibleRelation);
				b6.put(ic, infAcc);
				if(this.views.contains(ic)) {
					this.inferredAccessibleViews.add(infAcc);
				}
			}
		}
		try {
			this.definingAxioms = b3.build();
			this.accessibilityAxioms = b4.build();
			this.fastAccessibilityAxioms = f4.build();
			this.inferredAccessibilityAxioms = b5.build();
			this.shortInferredAccessibilityAxioms = b6.build();
		} catch (Exception e) {
			throw e;
		}
	}

	/**
	 * Constructor for AccessibleSchema.
	 * @param schema Schema
	 */
	public AccessibleSchema(Schema schema) {
		this(schema.getRelations(), schema.getDependencies(), schema.getConstantsMap());
	}

	/**
	 * @return both the inferred accessible versions of the schema dependencies and the dependencies of the form "AccessedR->InferredAccessedR"
	 */
	public Iterable<InferredAccessibleAxiom> getInferredAccessibilityAxioms() {
		return Iterables.concat(this.inferredAccessibilityAxioms.values(), this.shortInferredAccessibilityAxioms.values());
	}


	/**
	 * @return the inferred accessible versions of the schema dependencies
	 */
	public Collection<InferredAccessibleAxiom> getShortInferredAccessibilityAxioms() {
		return this.shortInferredAccessibilityAxioms.values();
	}


	/**
	 * @return List<InferredAccessibleAxiom>
	 */
	public List<InferredAccessibleAxiom> getInferredAccessibleViews() {
		return this.inferredAccessibleViews;
	}

	/**
	 * @return Collection<AccessibilityAxiom>
	 */
	public Collection<AccessibilityAxiom> getAccessibilityAxioms() {
		return this.accessibilityAxioms.values();
	}

	/**
	 * @return Collection<FastAccessibilityAxiom>
	 */
	public Collection<FastAccessibilityAxiom> getFastAccessibilityAxioms() {
		return this.fastAccessibilityAxioms.values();
	}

	/**
	 * @return Collection<DefiningAxiom>
	 */
	public Collection<DefiningAxiom> getDefiningAxioms() {
		return this.definingAxioms.values();
	}

	/**
	 * @return Map<String,Relation>
	 */
	@Override
	public Map<String,Relation> getRelationsMap() {
		Map<String, Relation> result = new LinkedHashMap<>();
		AccessibleRelation acc = this.accessibleRelation.get(0);
		result.put(acc.getName(), acc);
		for (Relation r: this.accessedRelations.values()) {
			result.put(r.getName(), r);
		}
		for (Relation r: this.infAccessibleRelations.values()) {
			result.put(r.getName(), r);
		}
		return result;
	}

	/**
	 * @return List<Relation>
	 */
	@Override
	public List<Relation> getRelations() {
		return Lists.newArrayList(
				Iterables.concat(
						this.relations,
						this.accessedRelations.values(),
						this.infAccessibleRelations.values(),
						this.accessibleRelation));
	}

	/**
	 * @param relationName String
	 * @return Relation
	 */
	@Override
	public Relation getRelation(String relationName) {
		Relation result = super.getRelation(relationName);
		if (result == null && this.accessibleRelation.get(0).getName().equals(relationName)) {
			result = this.accessibleRelation.get(0);
		}
		if (result == null) {
			result = this.accessedRelations.get(relationName);
		}
		if (result == null) {
			result = this.infAccessibleRelations.get(relationName);
		}
		return result;
	}

	/**
	 * @return AccessibleRelation
	 */
	public AccessibleRelation getAccessibleRelation() {
		return this.accessibleRelation.get(0);
	}

	/**
	 * @param r Relation
	 * @return AccessedRelation
	 */
	public AccessedRelation getAccessedRelation(Relation r) {
		return this.accessedRelations.get(r.getName());
	}

	/**
	 * @param r Relation
	 * @return InferredAccessibleRelation
	 */
	public InferredAccessibleRelation getInferredAccessibleRelation(Relation r) {
		return this.infAccessibleRelations.get(r.getName());
	}

	/**
	 * @param relation Relation
	 * @return boolean
	 */
	@Override
	public boolean contains(Relation relation) {
		boolean result = super.contains(relation);
		if (!result) {
			result = this.accessibleRelation.contains(relation)
					|| this.accessedRelations.containsValue(relation)
					|| this.infAccessibleRelations.containsValue(relation);
		}
		return result;
	}

	/**
	 * @param ic IC
	 * @return boolean
	 */
	@Override
	public boolean contains(Constraint ic) {
		boolean result = super.contains(ic);
		if (!result) {
			result = this.accessibilityAxioms.containsValue(ic)
					|| this.definingAxioms.containsValue(ic)
					|| this.inferredAccessibilityAxioms.containsValue(ic)
					|| this.shortInferredAccessibilityAxioms.containsValue(ic);
		}
		return result;
	}

	/**
	 * @return all the accessible schema dependencies including the ones of the original schema
	 */
	@Override
	public List<Constraint> getDependencies() {
		return Lists.newArrayList(
				Iterables.concat(
						this.schemaDependencies,
						this.accessibilityAxioms.values(),
						this.inferredAccessibilityAxioms.values(),
						this.shortInferredAccessibilityAxioms.values()));
	}

	/**
	 * @param r Input relation
	 * @return The defining axioms that correspond to the relation
	 */
	public Collection<DefiningAxiom> getDefiningAxioms(Relation r) {
		return this.definingAxioms.get(r.getName());
	}

	/**
	 * @param r Input relation
	 * @param b Input binding
	 * @return The accessibility axiom that corresponds to the relation-binding pair
	 */
	public AccessibilityAxiom getAccessibilityAxiom(Relation r, AccessMethod b) {
		return this.accessibilityAxioms.get(Pair.of(r, b));
	}

	/**
	 * @param r Input relation
	 * @return a dependency of the form "AccessedR->InferredAccessedR"
	 */
	public InferredAccessibleAxiom getInferredAccAxiom(Relation r) {
		return this.inferredAccessibilityAxioms.get(r.getName());
	}

	/**
	 * @param ic IC
	 * @return the inferred accessible version of the input query
	 */
	public InferredAccessibleAxiom getInferredAccAxiom(Constraint ic) {
		return this.shortInferredAccessibilityAxioms.get(ic);
	}

	/**
	 * @param n int
	 * @return List<Integer>
	 */
	static List<Integer> bindingPositions(int n) {
		List<Integer> result = new ArrayList<>();
		for (int i = 1; i <= n; i++) {
			result.add(i);
		}
		return result;
	}

	/**
	 * @param accessibleSchema AccessibleSchema
	 * @return the accessible query
	 * @see uk.ac.ox.cs.pdq.fol.Query#accessible(AccessibleSchema)
	 */
	public <Q extends Query<?>> Q accessible(Q query) {
		if (!(query instanceof ConjunctiveQuery)) {
			throw new UnsupportedOperationException("Non CQ not supported yet.");
		}
		List<PredicateFormula> atomicFormulas = new ArrayList<>();
		for (PredicateFormula af: query.getBody().getPredicates()) {
			atomicFormulas.add(
					new PredicateFormula(this.getInferredAccessibleRelation((Relation) af.getSignature()), af.getTerms()));
		}
		return (Q) new ConjunctiveQuery(query.getHead(), Conjunction.of(atomicFormulas));
	}
	
	/**
	 * @param accessibleSchema AccessibleSchema
	 * @return the accessible query
	 * @see uk.ac.ox.cs.pdq.fol.Query#accessible(AccessibleSchema)
	 */
	public <Q extends Query<?>> Q accessible(Q query, Map<Variable, Constant> canonicalMapping) {
		if (!(query instanceof ConjunctiveQuery)) {
			throw new UnsupportedOperationException("Non CQ not supported yet.");
		}
		List<PredicateFormula> atomicFormulas = new ArrayList<>();
		for (PredicateFormula af: query.getBody().getPredicates()) {
			atomicFormulas.add(
					new PredicateFormula(this.getInferredAccessibleRelation((Relation) af.getSignature()), af.getTerms()));
		}
		return (Q) new ConjunctiveQuery(query.getHead(), Conjunction.of(atomicFormulas), canonicalMapping);
	}


	/**
	 *
	 * Accessed relations implementation
	 * @author Efthymia Tsamoura
	 *
	 */
	public static class AccessedRelation extends Relation {

		/** */
		private static final long serialVersionUID = 4850080089128282906L;
		public static final String PREFIX = "Accessed";
		private final Relation baseRelation;

		/**
		 * Constructor with input method.
		 * @param r Relation
		 */
		public AccessedRelation(Relation r) {
			super(PREFIX + r.getName(), r.getAttributes(),
					Lists.newArrayList(
							new AccessMethod(
									AccessMethod.DEFAULT_PREFIX +
									AccessMethod.AccessMethodTypes.FREE + r.getName(),
									AccessMethod.AccessMethodTypes.FREE,
									Lists.<Integer>newArrayList())));
			this.baseRelation = r;
		}

		/**
		 * Constructor for AccessedRelation.
		 * @param view View
		 */
		public AccessedRelation(View view) {
			super(PREFIX + view.getName(),
					Utility.canonicalAttributes(view.getAttributes()),
					Lists.newArrayList(
							new AccessMethod(
									AccessMethod.DEFAULT_PREFIX +
									AccessMethod.AccessMethodTypes.FREE + view.getName(),
									AccessMethod.AccessMethodTypes.FREE,
									Lists.<Integer>newArrayList())));
			this.baseRelation = view;
		}

		/**
		 * @return Relation
		 */
		public Relation getBaseRelation() {
			return this.baseRelation;
		}
	}


	/**
	 *
	 * Accessible relations implementation
	 * @author Efthymia Tsamoura
	 *
	 */
	public static class AccessibleRelation extends Relation {

		/** */
		private static final long serialVersionUID = 5762653825140737301L;
		public static final String PREFIX = "Access";
		private static AccessibleRelation singleton = null;

		private AccessibleRelation() {
			super(PREFIX, asList(new Attribute(String.class, "x0")),
					Lists.newArrayList(
							new AccessMethod(
									AccessMethod.DEFAULT_PREFIX +
									AccessMethod.AccessMethodTypes.FREE + PREFIX,
									AccessMethod.AccessMethodTypes.FREE,
									Lists.<Integer>newArrayList())));

		}

		/**
		 * @param constant TypedConstant<?>
		 * @return PredicateFormula
		 */
		public static PredicateFormula getAccessibleFact(TypedConstant<?> constant) {
			return new PredicateFormula(AccessibleRelation.getInstance(), new TypedConstant<>(constant));
		}

		/**
		 * @param a Attribute
		 * @return List<Attribute>
		 */
		private static List<Attribute> asList(Attribute a) {
			List<Attribute> result = new ArrayList<>(1);
			result.add(a);
			return result;
		}

		/**
		 * @return AccessibleRelation
		 */
		public static AccessibleRelation getInstance() {
			if (singleton == null) {
				singleton = new AccessibleRelation();
			}
			return singleton;
		}
	}

	/**
	 *
	 * Inferred accessible relations implementation
	 * @author Efthymia Tsamoura
	 *
	 */
	public static class InferredAccessibleRelation extends Relation {

		/** */
		private static final long serialVersionUID = 5612488076416617646L;
		public static final String PREFIX = "InferredAccessible";
		private final Relation baseRelation;

		/**
		 * Constructor for InferredAccessibleRelation.
		 * @param r Relation
		 */
		public InferredAccessibleRelation(Relation r) {
			super(PREFIX + r.getName(), r.getAttributes(),
					Lists.newArrayList(
							new AccessMethod(
									AccessMethod.DEFAULT_PREFIX +
									AccessMethod.AccessMethodTypes.FREE + r.getName(),
									AccessMethod.AccessMethodTypes.FREE,
									Lists.<Integer>newArrayList())));
			this.baseRelation = r;
		}

		/**
		 * Constructor for InferredAccessibleRelation.
		 * @param view View
		 */
		public InferredAccessibleRelation(View view) {
			super(PREFIX + view.getName(),
					Utility.canonicalAttributes(view.getAttributes()),
					Lists.newArrayList(
							new AccessMethod(
									AccessMethod.DEFAULT_PREFIX +
									AccessMethod.AccessMethodTypes.FREE + view.getName(),
									AccessMethod.AccessMethodTypes.FREE,
									Lists.<Integer>newArrayList())));
			this.baseRelation = view;
		}

		/**
		 * @return Relation
		 */
		public Relation getBaseRelation() {
			return this.baseRelation;
		}
	}
}
