package uk.ac.ox.cs.pdq.planner.db.access;

import java.util.ArrayList;
import java.util.List;

import uk.ac.ox.cs.pdq.db.AccessMethod;
import uk.ac.ox.cs.pdq.db.AccessMethod.AccessMethodTypes;
import uk.ac.ox.cs.pdq.db.GuardedDependency;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.TGD;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessibleRelation;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.InferredAccessibleRelation;

/**
 * An accessibility axiom
 *
 * @author Efthymia Tsamoura
 */
public class FastAccessibilityAxiom extends TGD implements GuardedDependency {

	private final InferredAccessibleRelation infAccRelation;
	private final AccessMethod binding;

	/**
	 * @param infAccRel
	 * @param binding
	 */
	public FastAccessibilityAxiom(InferredAccessibleRelation infAccRel, AccessMethod binding) {
		super(createLeft(infAccRel.getBaseRelation(), binding), createRight(infAccRel, binding));
		this.binding = binding;
		this.infAccRelation = infAccRel;
	}

	/**
	 * @param relation
	 * @param binding
	 * @return Conjunction<PredicateFormula>
	 */
	private static Conjunction<PredicateFormula> createLeft(Relation relation, AccessMethod binding) {
		List<PredicateFormula> leftAtoms = new ArrayList<>();
		List<Integer> bindingPositions = binding.getInputs();
		Relation r = AccessibleRelation.getInstance();
		PredicateFormula f = relation.createAtoms();
		List<Term> terms = f.getTerms();
		for (int bindingPos: bindingPositions) {
			if (binding.getType() != AccessMethodTypes.FREE) {
				leftAtoms.add(new PredicateFormula(r, terms.get(bindingPos - 1)));
			}
		}
		leftAtoms.add(f);
		return Conjunction.of(leftAtoms);
	}

	/**
	 * @param infAccRel
	 * @param binding
	 * @return Conjunction<PredicateFormula>
	 */
	private static Conjunction<PredicateFormula> createRight(InferredAccessibleRelation infAccRel, AccessMethod binding) {
		Relation relation = infAccRel.getBaseRelation();
		List<PredicateFormula> rightAtoms = new ArrayList<>();
		List<Integer> bindingPositions = binding.getInputs();
		Relation accessible = AccessibleRelation.getInstance();
		PredicateFormula f = infAccRel.createAtoms();
		List<Term> terms = f.getTerms();
		for (int i = 1; i <= relation.getArity(); ++i) {
			if (!bindingPositions.contains(i)) {
				rightAtoms.add(new PredicateFormula(accessible, terms.get(i - 1)));
			}
		}
		rightAtoms.add(infAccRel.createAtoms());
		return Conjunction.of(rightAtoms);
	}

	/**
	 * @return the inferred accessible relation of the accessibility axiom.
	 */
	public InferredAccessibleRelation getInferredAccessibleRelation() {
		return this.infAccRelation;
	}

	/**
	 * @return the base relation of the accessibility axiom.
	 */
	public Relation getBaseRelation() {
		return this.infAccRelation.getBaseRelation();
	}

	/**
	 * @return the access method of the accessibility axiom.
	 */
	public AccessMethod getBinding() {
		return this.binding;
	}

	/**
	 * @return PredicateFormula
	 * @see uk.ac.ox.cs.pdq.db.GuardedDependency#getGuard()
	 */
	@Override
	public PredicateFormula getGuard() {
		return this.left.getPredicates().get(this.left.size()-1);
	}
}