package uk.ac.ox.cs.pdq.planner.db.access;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.TGD;
import uk.ac.ox.cs.pdq.db.View;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.Disjunction;
import uk.ac.ox.cs.pdq.fol.Formula;
import uk.ac.ox.cs.pdq.fol.Negation;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Signature;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessedRelation;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.InferredAccessibleRelation;
import uk.ac.ox.cs.pdq.rewrite.Rewriter;

import com.google.common.base.Preconditions;

/**
 * An inferred accessible axiom
 *
 * @author Efthymia Tsamoura
 * @author Julien Leblay
 */
public class InferredAccessibleAxiom extends TGD {

	/** True if the axiom is accessed to inferred accessible one */
	private final Boolean isAccessedToInferredAccessedAxiom;

	/**
	 * Creates an accessed to inferred accessible axiom
	 * @param infAccRelation
	 * @param accRelation
	 */
	public InferredAccessibleAxiom(InferredAccessibleRelation infAccRelation, AccessedRelation accRelation) {
		super(Conjunction.of(accRelation.createAtoms()),
				Conjunction.of(infAccRelation.createAtoms()));
		Relation relation = infAccRelation.getBaseRelation();
		this.isAccessedToInferredAccessedAxiom = true;
		Preconditions.checkArgument(relation != null && relation.equals(accRelation.getBaseRelation()), "Invalid instantiation of InferredAccessibleAxiom");
	}

	/**
	 * Creates an inferred accessible axiom
	 * @param dependency
	 * @param predToInfAcc
	 */
	public InferredAccessibleAxiom(TGD dependency, Map<PredicateFormula, InferredAccessibleRelation> predToInfAcc) {
		super(substitute(dependency.getLeft(), predToInfAcc),
				substitute(dependency.getRight(), predToInfAcc));
		this.isAccessedToInferredAccessedAxiom = false;
	}

	/**
	 * Creates the left-hand side of an inferred accessible axiom for the input dependency
	 * @param f Formula
	 * @param predToInfAcc Map<PredicateFormula,InferredAccessibleRelation>
	 * @return a conjunction of predicates corresponding to the LHS of an
	 * inferred accessible axiom
	 */
	private static <T extends Formula> T substitute(T f, Map<PredicateFormula, InferredAccessibleRelation> predToInfAcc) {
		return new InferredAccessibleRelationSubstituter<T>(predToInfAcc).rewrite(f);
	}

	/**
	 * @return View
	 * @see uk.ac.ox.cs.pdq.db.Constraint#definesView()
	 */
	@Override
	public View definesView() {
		if (this.left.getPredicates().size() != 1) {
			return null;
		}
		Signature result = this.left.getPredicates().get(0).getSignature();
		if (result instanceof InferredAccessibleRelation) {
			Relation base = ((InferredAccessibleRelation) result).getBaseRelation();
			if (base instanceof View) {
				return (View) base;
			}
		}
		return null;
	}

	/**
	 * @return Boolean
	 */
	public Boolean isAccessedToInferredAccessedAxiom() {
		return this.isAccessedToInferredAccessedAxiom;
	}

	/**
	 */
	private static class InferredAccessibleRelationSubstituter<T extends Formula> implements Rewriter<T, T> {
		final Map<PredicateFormula, InferredAccessibleRelation> predToInfAcc;

		/**
		 * Constructor for InferredAccessibleRelationSubstituter.
		 * @param predToInfAcc Map<PredicateFormula,InferredAccessibleRelation>
		 */
		public InferredAccessibleRelationSubstituter(Map<PredicateFormula, InferredAccessibleRelation> predToInfAcc) {
			this.predToInfAcc = predToInfAcc;
		}

		/**
		 * @param input T
		 * @return T
		 */
		@Override
		public T rewrite(T input) {
			return (T) this.substitute(input);
		}

		/**
		 * @param f Formula
		 * @return Formula
		 */
		private Formula substitute(Formula f) {
			if (f instanceof Conjunction) {
				return this.substitute((Conjunction<Formula>) f);
			}
			if (f instanceof Disjunction) {
				return this.substitute((Disjunction<Formula>) f);
			}
			if (f instanceof Negation) {
				return this.substitute((Negation<Formula>) f);
			}
			if (f instanceof PredicateFormula) {
				return this.substitute((PredicateFormula) f);
			}
			return f;
		}

		/**
		 * @param conjunction Conjunction<Formula>
		 * @return Conjunction<Formula>
		 */
		private Conjunction<Formula> substitute(Conjunction<Formula> conjunction) {
			List<Formula> result = new LinkedList<>();
			for (Formula f: conjunction) {
				result.add(this.substitute(f));
			}
			return Conjunction.of(result);
		}

		/**
		 * @param disjunction Disjunction<Formula>
		 * @return Disjunction<Formula>
		 */
		private Disjunction<Formula> substitute(Disjunction<Formula> disjunction) {
			List<Formula> result = new LinkedList<>();
			for (Formula f: disjunction) {
				result.add(this.substitute(f));
			}
			return Disjunction.of(result);
		}

		/**
		 * @param neg Negation<Formula>
		 * @return Negation<Formula>
		 */
		private Negation<Formula> substitute(Negation<Formula> neg) {
			return Negation.of(this.substitute(neg.getSubFormula()));
		}

		/**
		 * @param pred PredicateFormula
		 * @return PredicateFormula
		 */
		private PredicateFormula substitute(PredicateFormula pred) {
			return new PredicateFormula(this.predToInfAcc.get(pred), pred.getTerms());
		}
	}
}
