package uk.ac.ox.cs.pdq.planner.explorer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.cost.estimators.CostEstimator;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.plan.LinearPlan;
import uk.ac.ox.cs.pdq.plan.Plan;
import uk.ac.ox.cs.pdq.plan.Plan.ControlFlows;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.PlannerParameters;
import uk.ac.ox.cs.pdq.planner.PlannerParameters.FollowUpHandling;
import uk.ac.ox.cs.pdq.planner.PlannerParameters.PlannerTypes;
import uk.ac.ox.cs.pdq.planner.dag.ApplyRule;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationCostEstimator;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibilityAxiom;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.linear.LinearChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.linear.LinearConfiguration;
import uk.ac.ox.cs.pdq.planner.linear.explorer.Candidate;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Chaser;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.DominanceFactory;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.SuccessDominanceFactory;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.DatabaseTreeState;
import uk.ac.ox.cs.pdq.reasoning.Reasoner;

import com.google.common.base.Preconditions;
import com.google.common.collect.Sets;
import com.google.common.eventbus.EventBus;


/**
 * Creates configurations based on the input arguments
 *
 * @author Efthymia Tsamoura
 */
public final class ConfigurationFactory<S extends ChaseState, P extends Plan> {

	private final AccessibleSchema accessibleSchema;
	private final Query<?> query;
	/** Performs reasoning*/
	private final Reasoner reasoner;
	/** The state of a configuration */
	private final S state; 
	/** Defines a plan's control flow*/
	private final ControlFlows controlFlow;
	/** Factory of dominance detectors*/
	private final DominanceFactory df;
	/** Factory of success dominance detectors*/
	private final SuccessDominanceFactory<P> sdf;
	/** Estimates a plan's cost*/
	private final CostEstimator<P> costEstimator;

	private final EventBus eventBus;
	private final boolean collectStats;	
	private final Random random;

	/** Defines whether or not the ApplyRule configurations will be minimal or not*/
	private final FollowUpHandling followUps;
	private final List<DAGChaseConfiguration<S>> dagConfigurations;

	/**
	 * @param eventBus
	 * @param collectStats
	 * @param schema
	 * @param accessibleSchema
	 * @param state
	 * 		The state of a configuration
	 * @param reasoner
	 * 		Performs reasoning
	 * @param costEstimator
	 * 		Estimates a plan's cost
	 * @param dominanceFactory
	 * 		Factory of dominance detectors
	 * @param successDominanceFactory
	 * 		Factory of success dominance detectors
	 * @param params
	 * @throws PlannerException
	 */
	public ConfigurationFactory(
			EventBus eventBus,
			boolean collectStats,
			AccessibleSchema accessibleSchema,
			Query<?> query,
			S state,
			Reasoner reasoner,
			CostEstimator<P> costEstimator,
			DominanceFactory dominanceFactory,
			SuccessDominanceFactory<P> successDominanceFactory,
			PlannerParameters params) throws PlannerException {
		this(eventBus, collectStats,
				params.getPlannerType(),
				accessibleSchema, 
				query,
				state,
				reasoner,
				costEstimator,
				dominanceFactory,
				successDominanceFactory,
				new Random(params.getSeed()),
				params.getControlFlow(),
				params.getFollowUpHandling(),
				params.getPlannerType());
	}

	/**
	 *
	 * @param eventBus
	 * @param collectStats
	 * @param type
	 * @param schema
	 * @param accessibleSchema
	 * @param state
	 * 		The state of a configuration
	 * @param reasoner
	 * 		Performs reasoning
	 * @param costEstimator
	 * 		Estimates a plan's cost
	 * @param dominanceFactory
	 * 		Factory of dominance detectors
	 * @param successDominanceFactory
	 * 		Factory of success dominance detectors
	 * @param random
	 * 		Random generator
	 * @param cf
	 * 		Defines a plan's control flow
	 * @param fu
	 * 		Defines whether or not the ApplyRule configurations will be minimal or not
	 * @param plannerType
	 * 		The planner's type
	 * @throws PlannerException
	 */
	protected ConfigurationFactory(
			EventBus eventBus,
			boolean collectStats,
			PlannerTypes type,
			AccessibleSchema accessibleSchema,
			Query<?> query,
			S state,
			Reasoner reasoner,
			CostEstimator<P> costEstimator,
			DominanceFactory dominanceFactory,
			SuccessDominanceFactory<P> successDominanceFactory,
			Random random,
			ControlFlows cf,
			FollowUpHandling fu,
			PlannerTypes plannerType) throws PlannerException {
		Preconditions.checkNotNull(type);
		Preconditions.checkNotNull(accessibleSchema);
		Preconditions.checkNotNull(state);
		Preconditions.checkNotNull(reasoner);
		Preconditions.checkNotNull(costEstimator);
		Preconditions.checkNotNull(random );
		Preconditions.checkNotNull(costEstimator);
		Preconditions.checkNotNull(dominanceFactory);
		Preconditions.checkNotNull(successDominanceFactory);
		Preconditions.checkNotNull(query);
		this.eventBus = eventBus;
		this.collectStats = collectStats;
		this.accessibleSchema = accessibleSchema;
		this.state = state;
		this.reasoner = reasoner;
		this.random = random;
		this.controlFlow = cf;
		this.df = dominanceFactory;
		this.sdf = successDominanceFactory;
		this.followUps = fu;
		this.costEstimator = costEstimator;
		this.query = query;
		if(plannerType.equals(PlannerTypes.LINEAR_GENERIC) || 
				plannerType.equals(PlannerTypes.LINEAR_OPTIMIZED) ||
				plannerType.equals(PlannerTypes.LINEAR_KCHASE)) {
			this.dagConfigurations = null;
		}
		else {
			this.dagConfigurations = this.createDAGInstances();
		}
	}

	/**
	 * @return a list of ApplyRule configurations based on the facts derived after chasing the input schema with the canonical database of the query
	 * @throws PlannerException
	 */
	private List<DAGChaseConfiguration<S>> createDAGInstances() throws PlannerException {
		List<DAGChaseConfiguration<S>> collection = new ArrayList<>();
		Collection<Pair<AccessibilityAxiom,Collection<PredicateFormula>>> pairs = this.state.groupByBinding(this.accessibleSchema.getAccessibilityAxioms());
		for (Pair<AccessibilityAxiom, Collection<PredicateFormula>> pair: pairs) {
			ApplyRule<S> applyRule = null;
			Collection<Collection<PredicateFormula>> bindings = new LinkedHashSet<>();
			switch (this.followUps) {
			case MINIMAL:
				for (PredicateFormula p: pair.getRight()) {
					bindings.add(Sets.newHashSet(p));
				}
				break;
			default:
				bindings.add(pair.getRight());
				break;
			}
			for (Collection<PredicateFormula> binding: bindings) {
				S newState = null;
				if (this.state instanceof DatabaseTreeState) {
					newState = (S) ((DatabaseTreeState) this.state).replicate();
				} else {
					newState = (S) this.state.clone();
				}
				applyRule = new ApplyRule<S>(
						this.accessibleSchema,
						this.query,
						((Chaser<S>) this.reasoner),
						newState,
						this.controlFlow,
						this.df.getInstance(),
						this.sdf.getInstance(),
						(ConfigurationCostEstimator) this.costEstimator,
						pair.getLeft(),
						Sets.newHashSet(binding)
						);
				applyRule.generate();
				collection.add(applyRule);
			}
		}
		return collection;
	}

	/**
	 * @return List<DAGConfiguration<S>>
	 * @throws PlannerException
	 */
	public List<DAGChaseConfiguration<S>> getDAGInstances() throws PlannerException {
		return this.dagConfigurations;
	}


	/**
	 * @param parent
	 * @param exposedCandidates
	 * @return a linear configuration given the input parent configuration and the exposed set of facts
	 * @throws PlannerException
	 */
	public LinearChaseConfiguration getLinearInstance(LinearConfiguration parent, Set<Candidate> exposedCandidates) throws PlannerException {
		return new LinearChaseConfiguration<>(
				this.eventBus,
				this.collectStats,
				this.accessibleSchema,
				this.query,
				(Chaser<S>) this.reasoner,
				this.controlFlow,
				this.df.getInstance(),
				this.sdf.getInstance(),
				(CostEstimator<LinearPlan>) this.costEstimator,
				(LinearChaseConfiguration<S>) parent,
				exposedCandidates,
				this.random);
	}

	/**
	 * @return a linear configurations based on the facts derived after chasing the input schema with the canonical database of the query
	 * @throws PlannerException
	 */
	public LinearChaseConfiguration getLinearInstance() throws PlannerException {
		return new LinearChaseConfiguration<>(
				this.eventBus,
				this.collectStats,
				this.accessibleSchema,
				this.query,
				(Chaser<S>) this.reasoner,
				this.state,
				this.controlFlow,
				this.df.getInstance(),
				this.sdf.getInstance(),
				(CostEstimator<LinearPlan>) this.costEstimator,
				this.random);
	}

	/**
	 * @return Reasoner
	 */
	public Reasoner getReasoner() {
		return this.reasoner;
	}

	public CostEstimator getCostEstimator() {
		return this.costEstimator;
	}
}
