package uk.ac.ox.cs.pdq.planner.explorer;


import java.sql.SQLException;
import java.util.Properties;

import uk.ac.ox.cs.pdq.cost.CostParameters;
import uk.ac.ox.cs.pdq.cost.estimators.CostEstimator;
import uk.ac.ox.cs.pdq.cost.statistics.Catalog;
import uk.ac.ox.cs.pdq.cost.statistics.SimpleCatalog;
import uk.ac.ox.cs.pdq.cost.statistics.estimators.CardinalityEstimator;
import uk.ac.ox.cs.pdq.cost.statistics.estimators.CardinalityEstimatorFactory;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.Schema;
import uk.ac.ox.cs.pdq.plan.Plan;
import uk.ac.ox.cs.pdq.planner.PlannerParameters;
import uk.ac.ox.cs.pdq.planner.PlannerParameters.PlannerTypes;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationAccessCountCostEstimator;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationLengthBasedCostEstimator;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationPerInputCostEstimator;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationTotalERSPICostEstimator;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationTotalResponeTimeCostEstimator;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationWhiteBoxCostEstimator;
import uk.ac.ox.cs.pdq.planner.linear.cost.LinearAccessCountCostEstimator;
import uk.ac.ox.cs.pdq.planner.linear.cost.LinearLengthBasedCostEstimator;
import uk.ac.ox.cs.pdq.planner.linear.cost.LinearPerInputCostEstimator;
import uk.ac.ox.cs.pdq.planner.linear.cost.LinearPostgresqlBlackBoxEstimator;
import uk.ac.ox.cs.pdq.planner.linear.cost.LinearTotalERSPICostEstimator;
import uk.ac.ox.cs.pdq.planner.linear.cost.LinearTotalResponseTimeCostEstimator;
import uk.ac.ox.cs.pdq.planner.linear.cost.LinearWhiteBoxCostEstimator;

import com.google.common.base.Preconditions;
import com.google.common.eventbus.EventBus;

/**
 * A factory of cost estimation objects
 *
 * @author Julien Leblay
 * @author Efthymia Tsamoura 
 *
 */
public class CostEstimatorFactory {

	/**
	 * @param params
	 * @param schema
	 * @return a cost estimator implementation corresponding to that specified
	 * in the initialConfig.
	 * @throws SQLException
	 */
	public static <P extends Plan> CostEstimator<P> getEstimator(PlannerParameters planParams, CostParameters costParams, Schema schema) throws SQLException {
		return getInstance(null, false, planParams, costParams, schema);
	}

	/**
	 * @param planParams
	 * @param schema
	 * @param eventBus EventBus
	 * @param collectStats boolean
	 * @return a cost estimator implementation corresponding to that specified
	 * in the initialConfig.
	 * @throws SQLException
	 */
	public static <P extends Plan> CostEstimator<P> getInstance(
			EventBus eventBus, 
			boolean collectStats,
			PlannerParameters planParams, 
			CostParameters costParams, 
			Schema schema) throws SQLException {
		Preconditions.checkArgument(costParams.getCostType() != null, "Cost type param must is not defined.");
		Properties properties = null;
		boolean linearPlanner = planParams.getPlannerType().equals(PlannerTypes.LINEAR_GENERIC) ||
				planParams.getPlannerType().equals(PlannerTypes.LINEAR_KCHASE) ||
				planParams.getPlannerType().equals(PlannerTypes.LINEAR_OPTIMIZED);
		switch (costParams.getCostType()) {
		case BLACKBOX:
			CardinalityEstimator card = CardinalityEstimatorFactory.getInstance(costParams.getCostType(), costParams.getCardinalityEstimationType(), schema);
			if(linearPlanner) {
				return (CostEstimator<P>) new LinearWhiteBoxCostEstimator(eventBus, collectStats, card);
			}
			return (CostEstimator<P>) new ConfigurationWhiteBoxCostEstimator(eventBus, collectStats, card);
		case BLACKBOX_DB:
			if ((properties = findRelationalProperties(schema)) != null) {
				String url = properties.getProperty("url");
				String bbUrl = costParams.getBlackBoxConnectionUrl();
				Properties prop = new Properties();
				if (url != null && bbUrl != null) {
					if (url.contains("postgresql") && bbUrl.contains("postgresql")) {
						prop.put("url", costParams.getBlackBoxConnectionUrl());
						prop.put("database", costParams.getBlackBoxDatabaseName());
						prop.put("username", costParams.getBlackBoxDatabaseUser());
						prop.put("password", costParams.getBlackBoxDatabasePassword());
					}
				}
				if(linearPlanner) {
					return (CostEstimator<P>) new LinearPostgresqlBlackBoxEstimator(eventBus, collectStats, properties, costParams.getBlackBoxQueryType()); 
				}
				return (CostEstimator<P>) new uk.ac.ox.cs.pdq.planner.dag.cost.PostgresqlBlackBoxEstimator(eventBus, collectStats, properties, costParams.getBlackBoxQueryType());
			}
			throw new UnsupportedOperationException("BLACKBOX_DB cost estimator is supported for the provided schema.");
		case SIMPLE_ERSPI:
			Preconditions.checkNotNull(costParams.getCatalog());
			Catalog catalog = new SimpleCatalog(schema, costParams.getCatalog());
			if(linearPlanner) {
				return (CostEstimator<P>) new LinearTotalERSPICostEstimator(eventBus, collectStats, catalog);
			}
			return (CostEstimator<P>) new ConfigurationTotalERSPICostEstimator(eventBus, collectStats, catalog);
		case SIMPLE_RESPONSETIME:
			Preconditions.checkNotNull(costParams.getCatalog());
			Catalog catalog2 = new SimpleCatalog(schema, costParams.getCatalog());
			if(linearPlanner) {
				return (CostEstimator<P>) new LinearTotalResponseTimeCostEstimator(eventBus, collectStats, catalog2);
			}
			return (CostEstimator<P>) new ConfigurationTotalResponeTimeCostEstimator(eventBus, collectStats, catalog2);
		case SIMPLE_CONSTANT:
		case SIMPLE_GIVEN:
			if(linearPlanner) {
				return (CostEstimator<P>) new LinearPerInputCostEstimator(eventBus, collectStats); 
			}
			return (CostEstimator<P>) new ConfigurationPerInputCostEstimator(eventBus, collectStats);
		case INVERSE_LENGTH:
			if(linearPlanner) {
				return (CostEstimator<P>) new LinearLengthBasedCostEstimator(eventBus, collectStats); 
			}
			return (CostEstimator<P>) new ConfigurationLengthBasedCostEstimator(eventBus, collectStats);
		default:
			if(linearPlanner) {
				return (CostEstimator<P>) new LinearAccessCountCostEstimator(eventBus, collectStats); 
			}
			return (CostEstimator<P>) new ConfigurationAccessCountCostEstimator(eventBus, collectStats);
		}
	}

	/**
	 * @param schema
	 * @return the schema underlying relational's properties if all of the
	 * relations in the schema have the same properties, null otherwise.
	 */
	private static Properties findRelationalProperties(Schema schema) {
		Properties result = null;
		for (Relation relation: schema.getRelations()) {
			Properties properties = relation.getProperties();
			if (result == null) {
				result = properties;
			} else if (!(properties.get("url").equals(result.get("url"))
					&& properties.get("database").equals(result.get("database")))) {
				return null;
			}
		}
		return result;
	}
}
