package uk.ac.ox.cs.pdq.planner.explorer;

import java.util.ArrayList;
import java.util.List;

import uk.ac.ox.cs.pdq.cost.estimators.CostEstimator;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.plan.LinearPlan;
import uk.ac.ox.cs.pdq.plan.Plan;
import uk.ac.ox.cs.pdq.planner.PlannerParameters;
import uk.ac.ox.cs.pdq.planner.PlannerParameters.PlannerTypes;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationCostEstimator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.ILPExplorer;
import uk.ac.ox.cs.pdq.planner.dag.explorer.ILPSIExplorer;
import uk.ac.ox.cs.pdq.planner.dag.explorer.MixedILPExplorer;
import uk.ac.ox.cs.pdq.planner.dag.explorer.VeryChaseFriendlyDPExplorer;
import uk.ac.ox.cs.pdq.planner.dag.explorer.filters.FilterFactory;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.ReachabilityValidator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.Validator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.validators.ValidatorFactory;
import uk.ac.ox.cs.pdq.planner.dag.potential.DefaultPotentialAssessor;
import uk.ac.ox.cs.pdq.planner.dag.priority.PriorityAssessor;
import uk.ac.ox.cs.pdq.planner.dag.priority.PriorityAssessorFactory;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismDetector;
import uk.ac.ox.cs.pdq.planner.linear.cost.Propagatable;
import uk.ac.ox.cs.pdq.planner.linear.explorer.GenericExplorer;
import uk.ac.ox.cs.pdq.planner.linear.explorer.KChaseExplorer;
import uk.ac.ox.cs.pdq.planner.linear.explorer.OptimizedExplorer;
import uk.ac.ox.cs.pdq.planner.linear.node.NodeFactory;
import uk.ac.ox.cs.pdq.planner.linear.node.SearchNode;
import uk.ac.ox.cs.pdq.planner.linear.pruning.PostPruningFactory;
import uk.ac.ox.cs.pdq.planner.parallel.IterativeExecutor;
import uk.ac.ox.cs.pdq.planner.parallel.IterativeExecutorFactory;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.ReasonerFactory;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.SuccessDominanceFactory;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.eventbus.EventBus;

/**
 * Creates an explorer given the input arguments
 * @author Efthymia Tsamoura
 *
 */
public class ExplorerFactory {

	/**
	 *
	 * @param eventBus
	 * @param collectStats
	 * @param schema
	 * @param accessibleSchema
	 * @param query
	 * @param accessibleQuery
	 * @param state
	 * 		The state of a configuration
	 * @param reasonerFactory
	 * 		Creates reasoners
	 * @param detector
	 * 		Detects homomorphisms
	 * @param costEstimator
	 * 		Estimates a plan's cost
	 * @param configurationFactory
	 * 		Creates configurations
	 * @param successDominanceFactory
	 * 		Creates success dominance detectors
	 * @param parameters
	 * @return Explorer<P>
	 * @throws Exception
	 */
	public static <P extends Plan, S extends ChaseState> Explorer<P> createExplorer(
			EventBus eventBus, boolean collectStats,
			AccessibleSchema accessibleSchema,
			Query<?> query,
			S state,
			ReasonerFactory reasonerFactory,
			HomomorphismDetector detector,
			CostEstimator<P> costEstimator,
			ConfigurationFactory<S,P> configurationFactory,
			SuccessDominanceFactory<P> successDominanceFactory,
			PlannerParameters parameters) throws Exception {

		NodeFactory<S,SearchNode> nf = null;
		PostPruningFactory<S,SearchNode> ppf = null;
		ValidatorFactory<S> vf = null;
		FilterFactory<S> ff = null;
		IterativeExecutor<S> executor0 = null;
		IterativeExecutor<S> executor1 = null;
		PriorityAssessor<S> pra = null;

		if (parameters.getPlannerType().equals(PlannerTypes.LINEAR_GENERIC)
				|| parameters.getPlannerType().equals(PlannerTypes.LINEAR_KCHASE)
				|| parameters.getPlannerType().equals(PlannerTypes.LINEAR_OPTIMIZED)) {
			nf = new NodeFactory<>((ConfigurationFactory<S, LinearPlan>) configurationFactory);
			ppf = new PostPruningFactory<>(parameters.getPostPruningType(), nf, accessibleSchema);
		}
		else {
			vf = new ValidatorFactory<>(parameters.getValidatorType(), parameters.getDepthThreshold());
			ff = new FilterFactory<>(parameters.getFilterType());

			List<Validator<S>> validators = new ArrayList<>();
			validators.add(vf.getInstance());

			if(parameters.getReachabilityFiltering()) {
				ReachabilityValidator<S> rv = new ReachabilityValidator<>(query,
						configurationFactory.getDAGInstances(),
						(FiringGraph<S>) state.getFiringGraph());
				validators.add(rv);
			}
			
			DefaultPotentialAssessor dpa = new DefaultPotentialAssessor(null, (ConfigurationCostEstimator) costEstimator, successDominanceFactory.getInstance());

			pra = PriorityAssessorFactory.createPriorityAssessor(parameters,
					validators,
					dpa,
					query,
					state,
					parameters.getSeed());

			executor0 = IterativeExecutorFactory.createIterativeExecutor(
					parameters.getIterativeExecutorType(),
					parameters.getFirstPhaseThreads(),
					reasonerFactory,
					detector,
					(ConfigurationCostEstimator) costEstimator,
					successDominanceFactory.getInstance(),
					eventBus, collectStats);

			executor1 = IterativeExecutorFactory.createIterativeExecutor(
					parameters.getIterativeExecutorType(),
					parameters.getSecondPhaseThreads(),
					reasonerFactory,
					detector,
					(ConfigurationCostEstimator) costEstimator,
					successDominanceFactory.getInstance(),
					eventBus, collectStats);
		}

		switch(parameters.getPlannerType()) {
		case LINEAR_GENERIC:
			return (Explorer<P>) new GenericExplorer<>(
					eventBus, collectStats,
					(Propagatable<LinearPlan>) costEstimator,
					configurationFactory.getLinearInstance(),
					nf,
					parameters.getMaxDepth());
		case LINEAR_KCHASE:
			return (Explorer<P>) new KChaseExplorer<>(
					eventBus, collectStats,
					(Propagatable<LinearPlan>) costEstimator,
					configurationFactory.getLinearInstance(),
					nf,
					parameters.getMaxDepth(),
					parameters.getChaseInterval(),
					ppf.getInstance());

		case DAG_GENERIC:
			return (Explorer<P>) new uk.ac.ox.cs.pdq.planner.dag.explorer.GenericExplorer<>(
					eventBus, collectStats,
					configurationFactory.getDAGInstances(),
					ff.getInstance(),
					vf.getInstance(),
					parameters.getMaxDepth(),
					parameters.getOrderAware());

		case DAG_SIMPLEDP:
			return (Explorer<P>) new uk.ac.ox.cs.pdq.planner.dag.explorer.SimpleDPExplorer<>(
					eventBus, collectStats,
					configurationFactory.getDAGInstances(),
					ff.getInstance(),
					vf.getInstance(),
					parameters.getMaxDepth(),
					parameters.getOrderAware());

		case DAG_CHASEFRIENDLYDP:
			return (Explorer<P>) new uk.ac.ox.cs.pdq.planner.dag.explorer.ChaseFriendlyDPExplorer<>(
					eventBus, collectStats,
					configurationFactory.getDAGInstances(),
					ff.getInstance(), vf.getInstance(),
					parameters.getMaxDepth(),
					parameters.getOrderAware());
			
		case DAG_OPTIMIZED:
			return (Explorer<P>) new VeryChaseFriendlyDPExplorer<>(
					eventBus, collectStats,
					configurationFactory.getDAGInstances(),
					ff.getInstance(),
					pra, executor0, executor1,
					parameters.getMaxDepth());

		case DAG_ILP:
			return (Explorer<P>) new ILPExplorer<>(
					eventBus, collectStats,
					configurationFactory.getDAGInstances(),
					ff.getInstance(),
					pra,
					executor0,
					executor1,
					parameters.getMaxDepth(),
					parameters.getMaxIterations());

		case DAG_ILP_SI:
			return (Explorer<P>) new ILPSIExplorer<>(
					eventBus, collectStats,
					configurationFactory.getDAGInstances(),
					ff.getInstance(),
					pra,
					executor0,
					executor1,
					parameters.getMaxDepth(),
					parameters.getMaxIterations(),
					parameters.getTopk());

		case DAG_MIXED_ILP:
			return (Explorer<P>) new MixedILPExplorer<>(
					eventBus, collectStats,
					configurationFactory.getDAGInstances(),
					ff.getInstance(),
					pra,
					executor0,
					executor1,
					parameters.getMaxDepth(),
					parameters.getMaxIterations(),
					parameters.getMaxBushiness());

		case LINEAR_OPTIMIZED:
			return (Explorer<P>) new OptimizedExplorer<>(
					eventBus, collectStats,
					(Propagatable<LinearPlan>) costEstimator,
					configurationFactory.getLinearInstance(),
					nf,
					parameters.getMaxDepth(),
					parameters.getQueryMatchInterval(),
					ppf.getInstance(),
					parameters.getZombification());

		default:
			throw new IllegalStateException("Unsupported planner type " + parameters.getPlannerType());
		}
	}
}
