package uk.ac.ox.cs.pdq.planner.homomorphism;

import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.db.Attribute;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.Evaluatable;
import uk.ac.ox.cs.pdq.fol.Formula;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessibleRelation;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismConstraint.AtomCollectionScope;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismConstraint.BagScope;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismConstraint.SuperMap;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismConstraint.TopK;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Bag;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Joiner;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;

/**
 * Creates SQL statements for relation database-backed homomorphism detectors.
 *
 * @author Efthymia Tsamoura
 */
public abstract class AbstractHomomorphismStatementBuilder {

	protected List<String> fromPredicates = new ArrayList<>();
	protected List<String> selectPredicates = new ArrayList<>();
	protected List<String> wherePredicates = new ArrayList<>();
	protected List<String> constraintPredicates = new ArrayList<>();
	protected List<String> bagPredicates = new ArrayList<>();
	protected Map<PredicateFormula, String> atomAliases = new LinkedHashMap<>();

	/**
	 * @param databaseName
	 * @return the complete list of SQL statements required to set up the fact database
	 */
	public abstract Collection<String> setupStatements(String databaseName);

	/**
	 * @param databaseName String
	 * @return the complete list of SQL statements required to clean up the fact database
	 */
	public abstract Collection<String> cleanupStatements(String databaseName);

	/**
	 *
	 * @param relation the table to drop
	 * @return a SQL statement for dropping the facts table of the given relation
	 */
	protected String dropTableStatement(Relation relation) {
		return "DROP TABLE " + relation.getName();
	}

	/**
	 * @param relation 
	 * @return a new database relation with attributes x0,x1,...,x_{N-1}, where
	 *         x_i corresponds to the i-th input relation's attribute
	 */
	protected DBRelation createTable(Relation relation) {
		List<Attribute> attributes = new ArrayList<>();
		for (int index = 0, l = relation.getAttributes().size(); index < l; ++index) {
			attributes.add(new Attribute(relation.getAttribute(index).getType(), "x" + index));
		}
		return new DBRelation(this.encodeName(relation.getName()), attributes);
	}

	/**
	 * @param name
	 * @return a new String that is a copy of the given name modified such that
	 * it is acceptable for the underlying system
	 */
	public abstract String encodeName(String name);

	/**
	 * @param name
	 * @return a new String the decode the given using a mapping inverse from
	 * that of the encode method.
	 */
	public abstract String decodeName(String name);

	/**
	 * @param relation the table to create
	 * @return a SQL statement that creates the fact table of the given relation
	 */
	protected String createTableStatement(Relation relation) {
		StringBuilder result = new StringBuilder();
		result.append("CREATE TABLE  ").append(relation.getName()).append('(');
		for (int index = 0; index < relation.getAttributes().size(); ++index) {
			result.append(' ').append(relation.getAttributes().get(index).getName()).append(" VARCHAR(500),");
		}
		result.append(" BAG int, ");
		result.append(" FACT int, ");
		result.append(" PRIMARY KEY ").append('(').append("BAG, FACT").append(')');
		result.append(')');
		return result.toString();
	}

	/**
	 * @param relation
	 * @param columns
	 * @return a SQL statement that creates an index for the columns of the input relation
	 */
	protected String createTableIndex(Relation relation, Integer... columns) {
		StringBuilder indexName = new StringBuilder();
		StringBuilder indexColumns = new StringBuilder();
		String sep1 = "", sep2 = "";
		for (Integer i: columns) {
			indexName.append(sep1).append(relation.getAttribute(i).getName());
			indexColumns.append(sep2).append(relation.getAttribute(i).getName());
			sep1 = "_";
			sep2 = ",";
		}
		return "CREATE INDEX idx_" + relation.getName() + "_" + indexName +
				" ON " + relation.getName() + "(" + indexColumns + ")";
	}

	/**
	 * @param relation
	 * @return a SQL statement that creates an index for the bag and fact attributes of the database tables
	 */
	protected Collection<String> createTableNonJoinIndexes(Relation relation) {
		Collection<String> result = Sets.newLinkedHashSet();
		result.add("CREATE INDEX idx_" + relation.getName() + "_bag ON " + relation.getName() + "(BAG)");
		result.add("CREATE INDEX idx_" + relation.getName() + "_fact ON " + relation.getName() + "(FACT)");
		return result;
	}


	/**
	 * 
	 * @param relationMap
	 * @param rule
	 * @return
	 */
	protected Collection<String> createTableIndexes(Map<String, DBRelation> relationMap, Evaluatable rule) {
		Conjunction<?> body = null;
		if (rule.getBody() instanceof PredicateFormula) {
			body = Conjunction.of((PredicateFormula) rule.getBody());
		} else if (rule.getBody() instanceof Conjunction<?>) {
			body = (Conjunction) rule.getBody();
		} else {
			throw new UnsupportedOperationException("Homomorphism check only supported on conjunction of atomic predicate formulas for now.");
		}
		Set<String> result = new LinkedHashSet<>();
		Multimap<Variable, PredicateFormula> clusters = LinkedHashMultimap.create();
		for (Formula subFormula: body) {
			if (subFormula instanceof PredicateFormula) {
				for (Term t: subFormula.getTerms()) {
					if (t instanceof Variable) {
						clusters.put((Variable) t, (PredicateFormula) subFormula);
					}
				}
			} else {
				throw new UnsupportedOperationException("Homomorphism check only supported on conjunction of atomic predicate formulas for now.");
			}
		}
		for (Variable t: clusters.keys()) {
			Collection<PredicateFormula> atoms = clusters.get(t);
			if (atoms.size() > 1) {
				for (PredicateFormula atom: atoms) {
					for (int i = 0, l = atom.getTermCount(); i < l; i++) {
						if (atom.getTerm(i).equals(t)) {
							result.add(this.createTableIndex(relationMap.get(atom.getPredicateName()), i));
						}
					}
				}
			}
		}
		return result;
	}

	/**
	 *
	 * @param constraints
	 * 		Constraints that should be satisfied by the homomorphism found
	 * @return the bags within which we search for homomorphisms
	 */
	BagScope findBagConstraint(HomomorphismConstraint... constraints) {
		for (HomomorphismConstraint c: constraints) {
			if (c instanceof BagScope) {
				return (BagScope) c;
			}
		}
		return null;
	}

	/**
	 * @param constraints 
	 * 		Constraints that should be satisfied by the homomorphism found
	 * @return the facts considered during homomorphism detection
	 */
	Conjunction<PredicateFormula> findAtomScopeConstraint(HomomorphismConstraint... constraints) {
		for (HomomorphismConstraint c: constraints) {
			if (c instanceof AtomCollectionScope) {
				return ((AtomCollectionScope) c).atoms;
			}
		}
		return null;
	}

	/**
	 * @param constraints 
	 * 		Constraints that should be satisfied by the homomorphism found
	 * @return a variable to constant constraint that the homomorphism found should satisfy
	 */
	Map<Variable, Constant> findSuperMapConstraint(HomomorphismConstraint... constraints) {
		for (HomomorphismConstraint c: constraints) {
			if (c instanceof SuperMap) {
				return ((SuperMap) c).map;
			}
		}
		return null;
	}

	/**
	 * @param constraints 
	 * 		Constraints that should be satisfied by the homomorphism found
	 * @return the number of homomorphisms to detect, a negative value if none is specified.
	 */
	Integer findTopKConstraint(HomomorphismConstraint... constraints) {
		for (HomomorphismConstraint c: constraints) {
			if (c instanceof TopK) {
				return ((TopK) c).k;
			}
		}
		return null;
	}

	/**
	 * @param alias 
	 * 		The alias of a database relation
	 * @return a bag attribute for this alias
	 */
	public abstract String getBagAttribute(String alias);

	/**
	 * Creates an SQL statement for detecting homomorphisms
	 * @param relationMap
	 * @param source
	 * @param constraints
	 * @return a SQL query for detecting homomorphisms.
	 */
	public abstract Pair<String, String> createStatement(Map<String, DBRelation> relationMap, Evaluatable source, HomomorphismConstraint... constraints);

	/**
	 * @param targetBags
	 * @return a string of bag identifiers
	 */
	protected String createBagIdsInterval(Collection<Bag> targetBags) {
		List<Integer> bagIds = new ArrayList<>();
		if (targetBags.isEmpty()) {
			bagIds.add(-1);
		} else {
			for (Bag bag : targetBags) {
				bagIds.add(bag.getId());
			}
		}
		return "(" + Joiner.on(",").join(bagIds) + ")";
	}

	/**
	 * @param target
	 * @return a string of fact identifiers
	 */
	protected String createAtomIdsInterval(Iterable<PredicateFormula> target) {
		if (target != null) {
			List<Integer> atomIds = new ArrayList<>();
			for (PredicateFormula atom : target) {
				atomIds.add(atom.getId());
			}
			return "(" + Joiner.on(",").join(atomIds) + ")";
		}
		return null;
	}

	/**
	 * Makes a FROM clause
	 * @param source
	 * @param relationMap
	 */
	protected void createFromClause(Evaluatable source, Map<String, DBRelation> relationMap) {
		String aliasPrefix = "A";
		int aliasCounter = 0;
		for (PredicateFormula atom : source.getBody().getPredicates()) {
			String predicateName = relationMap.get(atom.getPredicateName()).getName();
			if (!this.atomAliases.containsValue(predicateName)) {
				String aliasName = aliasPrefix + aliasCounter;
				this.fromPredicates.add(predicateName + " " + aliasName);
				this.atomAliases.put((atom), aliasName);
				aliasCounter++;
			}
		}

	}

	/**
	 * Makes a SELECT clause
	 * @param source
	 * @param relationMap
	 */
	protected void createSelectClause(Evaluatable source, Map<String, DBRelation> relationMap) {
		List<String> attributes = new ArrayList<>();
		for (PredicateFormula atom : source.getBody().getPredicates()) {
			DBRelation dbr = relationMap.get(atom.getPredicateName());
			if (!(atom.getSignature() instanceof AccessibleRelation)) {
				String aliasName = this.atomAliases.get(atom);
				List<Term> terms = atom.getTerms();
				for (int index = 0; index < terms.size(); ++index) {
					Term term = terms.get(index);
					if (term instanceof Variable && !attributes.contains(((Variable) term).getName())) {
						this.selectPredicates.add(aliasName + '.' + dbr.getAttribute(index).getName());
						attributes.add(((Variable) term).getName());
					}
				}
			}
		}
	}

	/**
	 * Makes a WHERE clause
	 * @param source
	 * @param relationMap
	 */
	protected void createWhereClause(Evaluatable source, Map<String, DBRelation> relationMap) {
		Collection<Term> terms = Utility.getTerms(source.getBody().getPredicates());
		terms = Utility.removeDuplicates(terms);
		for (Term term : terms) {
			List<String> equalPredicates = new ArrayList<>();
			for (PredicateFormula atom : source.getBody().getPredicates()) {
				List<Integer> positions = atom.getTermPositions(term);
				for (Integer it : positions) {
					String aliasName = this.atomAliases.get(atom);
					DBRelation dbr = relationMap.get(atom.getPredicateName());
					equalPredicates.add(aliasName + '.' + dbr.getAttribute(it).getName());
				}
			}
			if (equalPredicates.size() > 1) {
				for (int index = 0; index < equalPredicates.size() - 1; ++index) {
					this.wherePredicates.add(equalPredicates.get(index) + '=' + equalPredicates.get(index + 1));
				}
			}
		}
	}

	/**
	 * Makes WHERE predicate clauses
	 * @param source
	 * @param relationMap
	 * @param constraint Map<Variable,Term>
	 */
	protected void createConstraintClause(
			Evaluatable source,
			Map<Variable, Constant> constraint,
			Map<String, DBRelation> relationMap) {
		for (PredicateFormula atom : source.getBody().getPredicates()) {
			if (!(atom.getSignature() instanceof AccessibleRelation)) {
				DBRelation dbr = relationMap.get(atom.getPredicateName());
				String aliasName = this.atomAliases.get(atom);
				List<Term> terms = atom.getTerms();
				for (int index = 0; index < terms.size(); ++index) {
					Term term = terms.get(index);
					Term constraintTerm = null;
					if (constraint != null) {
						constraintTerm = constraint.get(term);
					}
					if (constraintTerm != null && constraintTerm.isVariable()) {
						throw new IllegalStateException();
					} else if (constraintTerm != null) {
						if (term instanceof Variable) {
							this.constraintPredicates.add(aliasName + '.' + dbr.getAttribute(index).getName() + " = "
									+ "'" + constraintTerm.toString() + "'");
						}
					}
				}
			}
		}
	}

	/**
	 * Makes WHERE predicate clauses related to schema constants
	 * @param source
	 * @param relationMap
	 */
	protected void createConstraintClauseForSchemaConstants(Evaluatable source, Map<String, DBRelation> relationMap) {
		for (PredicateFormula atom : source.getBody().getPredicates()) {
			DBRelation dbr = relationMap.get(atom.getPredicateName());
			if (!(atom.getSignature() instanceof AccessibleRelation)) {
				String aliasName = this.atomAliases.get(atom);
				List<Term> terms = atom.getTerms();
				for (int index = 0; index < terms.size(); ++index) {
					Term term = terms.get(index);
					if (!term.isVariable() && !term.isSkolem()) {
						this.constraintPredicates.add(aliasName + '.' +
								dbr.getAttribute(index).getName() + " = " +
								this.equalityTerm(((TypedConstant) term)));
					}
				}
			}
		}
	}

	/**
	 * @param constant
	 * @return an equality predicate
	 */
	protected String equalityTerm(TypedConstant<?> constant) {
		return "'" + constant.getValue() + "'";
	}

	/**
	 *
	 */
	protected void createBagPredicates() {
		for (String predicateLeft : this.atomAliases.values()) {
			for (String predicateRight : this.atomAliases.values()) {
				if (!predicateLeft.equals(predicateRight)) {
					String conjunct = predicateLeft + '.' + "BAG" + '=' + predicateRight + '.' + "BAG ";
					this.bagPredicates.add(conjunct);
				}
			}
		}
	}

	/**
	 * The fact relation built for a schema relation
	 * @author Efthymia Tsamoura
	 *
	 */
	static class DBRelation extends Relation {
		/** */
		private static final long serialVersionUID = 3503553786085749666L;

		/**
		 * Constructor for DBRelation.
		 * @param name String
		 * @param attributes List<Attribute>
		 */
		public DBRelation(String name, List<Attribute> attributes) {
			super(name, attributes);
		}
	}

	/**
	 * @param schema
	 * @param source
	 * @param dbRelationMap
	 * @param resultSet
	 * @return Map<Variable,Term>
	 */
	public abstract Map<Variable, Constant> createMappingAlias(
			Map<String, TypedConstant<?>> constantsMap,
			Evaluatable source, Map<String, DBRelation> dbRelationMap, ResultSet resultSet);

	/**
	 * @param facts
	 * @return insert statements that add the input fact to the fact database.
	 */
	protected abstract Collection<String> makeInserts(Collection<? extends PredicateFormula> facts, Map<String, DBRelation> dbRelationMap);

	/**
	 * @return Map<PredicateFormula,String>
	 */
	public Map<PredicateFormula, String> getAtomsAliases() {
		return this.atomAliases;
	}

	/**
	 * @return List<String>
	 */
	public List<String> getBagPredicates() {
		return this.bagPredicates;
	}

	/**
	 * @return AbstractHomomorphismStatementBuilder
	 */
	@Override
	public abstract AbstractHomomorphismStatementBuilder clone();
}
