package uk.ac.ox.cs.pdq.planner.homomorphism;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.Evaluatable;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.homomorphism.AbstractHomomorphismStatementBuilder.DBRelation;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Bag;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.BagMatch;
import uk.ac.ox.cs.pdq.reasoning.Match;

import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.google.common.collect.Sets;

/**
 * Detects homomorphisms from a conjunction to a set of facts (e.g., facts produced during chasing).
 * For each schema/accessible schema relation of N attributes it creates a table
 * with attributes x0, x1, ..., x_{N-1}, where x_i corresponds to the
 * i-th relation attribute and an additional attribute named BAG which
 * corresponds to the bag (if it exists) where this fact is placed in.
 *
 * @author Efthymia Tsamoura
 * @author Julien Leblay
 *
 */
public class DBHomomorphismManager implements HomomorphismManager {

	/** Logger. */
	private static Logger log = Logger.getLogger(DBHomomorphismManager.class);
	
	/** Collection of expected queries. */
	protected Set<Evaluatable> queries;
	protected final List<Relation> relations;
	/** A map of the string representation of a constant to the constant*/
	protected final Map<String, TypedConstant<?>> constants;
	/** Statement builder */
	protected final AbstractHomomorphismStatementBuilder builder;
	/**
	 * A map from the schema relation names to the created relations (the ones that
	 * correspond to the created database tables)
	 */
	protected final Map<String, DBRelation> relationMap;
	/** Connection to the database */
	protected Connection connection;
	/** Chase database name */
	protected final String database;
	/** Information to connect to the facts database*/
	protected final String driver;
	protected final String url;
	protected final String username;
	protected final String password;
	protected boolean isInitialized = false;

	protected List<Connection> clones = new ArrayList<>();

	/**
	 * 
	 * @param driver
	 * 		Database driver
	 * @param url
	 * 		Database url
	 * @param database
	 * 		Database name
	 * @param username
	 * 		Database user
	 * @param password
	 * 		Database pass
	 * @param builder
	 * 		Builds SQL queries that detect homomorphisms
	 * @param accessibleSchema
	 * 		Input schema
	 * @param accessibleQuery
	 * 		Input query
	 * @throws SQLException
	 */
	public DBHomomorphismManager(
			String driver, String url, String database,
			String username, String password,
			AbstractHomomorphismStatementBuilder builder,
			AccessibleSchema accessibleSchema, 
			Query<?> accessibleQuery
			) throws SQLException {
		this.connection = getConnection(driver, url, database, username, password);
		this.driver = driver;
		this.url = url;
		this.username = username;
		this.password = password;
		this.database = database;
		this.builder = builder;
		
		this.queries = makeQueries(accessibleQuery, accessibleSchema);
		this.constants = accessibleSchema.getConstantsMap();
		this.relations = accessibleSchema.getRelations();
		this.relationMap = new LinkedHashMap<>();
	}
	
	/**
	 * 
	 * @param driver
	 * 		Database driver
	 * @param url
	 * 		Database url
	 * @param database
	 * 		Database name
	 * @param username
	 * 		Database user
	 * @param password
	 * 		Database pass
	 * @param builder
	 * 		Builds SQL queries that detect homomorphisms
	 * @param relations
	 * 		Database relations
	 * @param constantsMap
	 * 		Schema constants
	 * @param schemaToDBRelationMap
	 * 		A map from the schema relation names to the created relations (the ones that
	 * 		correspond to the created database tables)
	 * @param queries
	 * 		Cached queries of interest
	 * @throws SQLException
	 */
	private DBHomomorphismManager(
			String driver, String url, String database,
			String username, String password,
			AbstractHomomorphismStatementBuilder builder,
			List<Relation> relations,
			Map<String, TypedConstant<?>> constantsMap,
			Map<String, DBRelation> schemaToDBRelationMap,
			Set<Evaluatable> queries) throws SQLException {
		this.connection = getConnection(driver, url, database, username, password);
		this.driver = driver;
		this.url = url;
		this.username = username;
		this.password = password;
		this.database = database;
		this.builder = builder;
		this.relationMap = schemaToDBRelationMap;
		this.queries =queries;
		this.constants = constantsMap;
		this.relations = relations;
	}

	/**
	 * 
	 * @param accessibleQuery
	 * @param accessibleSchema
	 * @return
	 * 		queries of interest
	 */
	private static Set<Evaluatable> makeQueries(
			Query<?> accessibleQuery, 
			AccessibleSchema accessibleSchema
			) {
		Set<Evaluatable> result = Sets.newLinkedHashSet();
		result.add(accessibleQuery);
		for (Constraint ic: accessibleSchema.getDependencies()) {
			result.add(ic);
		}
		return result;
	}

	/**
	 * @see uk.ac.ox.cs.pdq.homomorphism.HomomorphismManager#initialize()
	 */
	@Override
	public void initialize() {
		if (!this.isInitialized) {
			this.initialize(this.queries);
			this.isInitialized = true;
		}
	}

	/**
	 * @param queries Collection<Evaluatable>
	 * @see uk.ac.ox.cs.pdq.homomorphism.HomomorphismManager#initialize(Collection<Evaluatable>)
	 */
	@Override
	public void initialize(Collection<Evaluatable> queries) {
		if (!this.isInitialized) {
			this.setup();
		}
	}

	/**
	 * Sets up the database that will store the facts
	 */
	private void setup() {
		try(Statement sqlStatement = this.connection.createStatement()) {
			try {
				for (String sql: this.builder.setupStatements(this.database)) {
					sqlStatement.addBatch(sql);
				}
				this.createBaseTables(this.relations, sqlStatement);
				this.createJoinIndexes(sqlStatement);
				sqlStatement.executeBatch();
			} catch (SQLException ex) {
				throw new IllegalStateException(ex.getMessage(), ex);
			}
		} catch (SQLException ex) {
			throw new IllegalStateException(ex.getMessage(), ex);
		}
	}

	/**
	 * @param relations List<Relation>
	 * @param stmt Statement
	 * @throws SQLException
	 */
	private void createBaseTables(List<Relation> relations, Statement stmt) throws SQLException {
		DBRelation dbRelation = null;
		for (Relation relation : relations) {
			dbRelation = this.builder.createTable(relation);
			this.relationMap.put(relation.getName(), dbRelation);
			stmt.addBatch(this.builder.createTableStatement(dbRelation));
			for (String batch : this.builder.createTableNonJoinIndexes(dbRelation)) {
				stmt.addBatch(batch);
			}
		}
	}

	/**
	 * @param stmt Statement
	 * @throws SQLException
	 */
	private void createJoinIndexes(Statement stmt) throws SQLException {
		Set<String> joinIndexes = Sets.newLinkedHashSet();
		for (Evaluatable query: this.queries) {
			joinIndexes.addAll(this.builder.createTableIndexes(this.relationMap, query));
		}
		for (String b: joinIndexes) {
			stmt.addBatch(b);
		}
	}

	/**
	 * Cleans up the database
	 * @throws PlannerException
	 */
	protected void cleanupDB() throws PlannerException {
		try (Statement sqlStatement = this.connection.createStatement();) {
			for (String sql: this.builder.cleanupStatements(this.database)) {
				sqlStatement.addBatch(sql);
			}
			sqlStatement.executeBatch();
		} catch (SQLException ex) {
			throw new PlannerException(ex.getMessage(), ex);
		}
	}

	/**
	 * @param source Q
	 * @param constraints HomomorphismConstraint[]
	 * @return List<Matching>
	 * @see uk.ac.ox.cs.pdq.homomorphism.HomomorphismDetector#getMatches(Q, HomomorphismConstraint[])
	 */
	@Override
	public <Q extends Evaluatable> List<Match> getMatches(Q source, HomomorphismConstraint... constraints) {
		Preconditions.checkArgument(source != null);
		List<Match> result = new LinkedList<>();
		Pair<String, String> q = this.builder.createStatement(this.relationMap, source, constraints);
		Bag[] targetBags = null;
		if(this.builder.findBagConstraint(constraints)!= null) {
			targetBags = this.builder.findBagConstraint(constraints).bags;
		}

		/*
		 * For each returned homomorphism (each homomorphism corresponds to each
		 * returned tuple) we create a mapping of variables to constants.
		 * We also maintain the bag where this homomorphism is found
		 */
		List<Pair<?, Map<Variable, Constant>>> match = null;
		try(Statement sqlStatement = this.connection.createStatement();
				ResultSet resultSet = sqlStatement.executeQuery(q.getValue())) {
			while (resultSet.next()) {
				Map<Variable, Constant> mapping = this.builder.createMappingAlias(
						this.constants,source, this.relationMap, resultSet);
				if (match == null) {
					match = new ArrayList<>();
				}
				if (targetBags != null && targetBags.length > 0) {
					int bagId = resultSet.getInt(this.builder.getBagAttribute(q.getKey()));
					result.add(new BagMatch(source, mapping, bagId));
				} else {
					result.add(new Match(source, mapping));
				}
			}
		} catch (SQLException ex) {
			log.debug(q);
			throw new IllegalStateException(ex.getMessage(), ex);
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.AutoCloseable#close()
	 */
	@Override
	public void close() throws Exception {
		this.cleanupDB();
		this.connection.close();
		for(Connection con:this.clones) {
			con.close();
		}
	}

	/**
	 * @param url
	 * @param database
	 * @param username
	 * @param password
	 * @param driver String
	 * @return a connection database connection for the given properties.
	 * @throws SQLException
	 */
	public static Connection getConnection(String driver, String url, String database, String username, String password) throws SQLException {
		if (!Strings.isNullOrEmpty(driver)) {
			try {
				Class.forName(driver);
			} catch (ClassNotFoundException e) {
				throw new IllegalStateException("Could not load chase database driver '" + driver + "'");
			}
		}
		String u = null;
		if (url.contains("{1}")) {
			u = url.replace("{1}", database);
		} else {
			u = url + database;
		}
		try {
			Connection result = DriverManager.getConnection(u, username, password);
			result.setAutoCommit(true);
			return result;
		} catch (SQLException e) {
			log.debug(e.getMessage());
		}
		Connection result = DriverManager.getConnection(url, username, password);
		result.setAutoCommit(true);
		return result;
	}

	/**
	 * @param facts Collection<? extends PredicateFormula>
	 * @see uk.ac.ox.cs.pdq.homomorphism.HomomorphismManager#addFacts(Collection<? extends PredicateFormula>)
	 */
	@Override
	public void addFacts(Collection<? extends PredicateFormula> facts) {
		try (Statement sqlStatement = this.connection.createStatement()) {
			for (String stmt : this.builder.makeInserts(facts, this.relationMap)) {
				sqlStatement.addBatch(stmt);
			}
			sqlStatement.executeBatch();
		} catch (SQLException ex) {
			throw new IllegalStateException(ex.getMessage(), ex);
		}
	}
	
	/**
	 * @param queries Collection<Q>
	 * @param constraints HomomorphismConstraint[]
	 * @return Map<Q,List<Matching>>
	 * @see uk.ac.ox.cs.pdq.homomorphism.HomomorphismDetector#getMatches(Collection<Q>, HomomorphismConstraint[])
	 */
	@Override
	public <Q extends Evaluatable> Map<Q, List<Match>> getMatches(Collection<Q> queries, HomomorphismConstraint... constraints) {
		Map<Q, List<Match>> result = new LinkedHashMap<>();
		for (Q q: queries) {
			List<Match> matches = this.getMatches(q, constraints);
			if (!matches.isEmpty()) {
				result.put(q, matches);
			}
		}
		return result;
	}

	/**
	 * @return DBHomomorphismManager
	 * @see uk.ac.ox.cs.pdq.homomorphism.HomomorphismDetector#clone()
	 */
	@Override
	public DBHomomorphismManager clone() {
		try {
			DBHomomorphismManager ret = new DBHomomorphismManager(
					this.driver, this.url, this.database, this.username, this.password,
					this.builder.clone(), 
					this.relations, 
					this.constants,
					this.relationMap, 
					this.queries);
			ret.isInitialized = this.isInitialized;
			this.clones.add(ret.connection);
			return ret;
		} catch (SQLException e) {
			e.printStackTrace();
			return null;
		}
	}
}
