package uk.ac.ox.cs.pdq.planner.homomorphism;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.Evaluatable;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Skolem;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismConstraint.BagScope;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.BagBoundPredicate;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Joiner;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.Sets;

/**
 * Builds queries for detecting homomorphisms in Derby
 *
 * @author Efthymia Tsamoura
 * @author Julien leblay
 */
public class DerbyHomomorphismStatementBuilder extends AbstractHomomorphismStatementBuilder {

	private static BiMap<String, String> nameEncodings = HashBiMap.create();
	static {
		nameEncodings.put(":", "cl_");
		nameEncodings.put("$", "dl_");
		nameEncodings.put("#", "hs_");
		nameEncodings.put("%", "pc_");
		nameEncodings.put("-", "hp_");
		nameEncodings.put("{", "lbc_");
		nameEncodings.put("}", "rbc_");
		nameEncodings.put("[", "lbk_");
		nameEncodings.put("]", "rbk_");
		nameEncodings.put("(", "lpr_");
		nameEncodings.put(")", "rpr_");
		nameEncodings.put("int", "int_");
		nameEncodings.put("float", "flt_");
		nameEncodings.put("string", "str_");
		nameEncodings.put("boolean", "bool_");
		nameEncodings.put("decimal", "dec_");
		nameEncodings.put("datetime", "date_");
		nameEncodings.put("both", "both_");
		nameEncodings.put("none", "none_");
		nameEncodings.put("check", "chk_");
	}

	/**
	 * Default constructor
	 */
	public DerbyHomomorphismStatementBuilder() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.homomorphism.AbstractHomomorphismStatementBuilder#setupStatements(java.lang.String)
	 */
	@Override
	public Collection<String> setupStatements(String databaseName) {
		return new LinkedList<>();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.homomorphism.AbstractHomomorphismStatementBuilder#cleanupStatements(java.lang.String)
	 */
	@Override
	public Collection<String> cleanupStatements(String databaseName) {
		return new LinkedList<>();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.homomorphism.AbstractHomomorphismStatementBuilder#getBagAttribute(java.lang.String)
	 */
	@Override
	public String getBagAttribute(String alias) {
		return "BAG";
	}

	/**
	 * @param relationMap Map<String,DBRelation>
	 * @param source Evaluatable
	 * @param constraints HomomorphismConstraint[]
	 * @return Pair<String,String>
	 */
	@Override
	public Pair<String, String> createStatement(Map<String, DBRelation> relationMap, Evaluatable source, HomomorphismConstraint... constraints) {

		this.fromPredicates.clear();
		this.selectPredicates.clear();
		this.wherePredicates.clear();
		this.constraintPredicates.clear();
		this.bagPredicates.clear();
		this.atomAliases.clear();

		this.createFromClause(source, relationMap);
		this.createSelectClause(source, relationMap);
		this.createWhereClause(source, relationMap);
		this.createConstraintClause(source, this.findSuperMapConstraint(constraints), relationMap);
		this.createConstraintClauseForSchemaConstants(source, relationMap);
		this.createBagPredicates();

		String aliasName = this.fromPredicates.get(0).split(" ")[1];
		Collection<String> aliasNames = this.atomAliases.values();
		String queryString = null;
		String selectStatement = null;

		/*
		 * If the facts that satisfy the homomorphism must all be within  a
		 * single bag, then the SQL query should return the bag
		 * identifier where the homomorphism holds.
		 */
		BagScope bagConstraint = this.findBagConstraint(constraints);
		if (!this.selectPredicates.isEmpty()) {
			selectStatement = " DISTINCT " + Joiner.on(",").join(this.selectPredicates);
			if (bagConstraint != null) {
				selectStatement += ", " + aliasName + ".BAG ";
			}
		}

		queryString = "SELECT  " + selectStatement + " FROM " + Joiner.on(",").join(this.fromPredicates);
		if (this.wherePredicates.size() > 0) {
			queryString += " WHERE " + Joiner.on(" AND ").join(this.wherePredicates);
		}

		/*
		 * if the target set of facts is not null), we
		 * add in the WHERE statement a predicate which limits the identifiers
		 * of the facts that satisfy any homomorphism to the
		 * identifiers of these facts
		 */
		Conjunction<PredicateFormula> target = this.findAtomScopeConstraint(constraints);
		if (this.createAtomIdsInterval(target) != null) {
			List<String> targetAtomsString = new ArrayList<>();
			for (String _pair : this.fromPredicates) {
				String _trimmed = _pair.trim();
				String[] trimmed = _trimmed.split(" ");
				String alias = trimmed[1];
				targetAtomsString.add(alias + ".FACT " + "IN " + this.createAtomIdsInterval(target));
			}
			if (queryString.contains("WHERE")) {
				queryString += " AND " + Joiner.on(" AND ").join(targetAtomsString);
			} else {
				queryString += " WHERE " + Joiner.on(" AND ").join(targetAtomsString);
			}
		}

		if (queryString.contains("WHERE") && !this.constraintPredicates.isEmpty()) {
			queryString += " AND " + Joiner.on(" AND ").join(this.constraintPredicates);
		} else if (!queryString.contains("WHERE") && !this.constraintPredicates.isEmpty()) {
			queryString += " WHERE " + Joiner.on(" AND ").join(this.constraintPredicates);
		}

		/*
		 * if the target set of bags is not null,
		 * we add in the WHERE statement a predicate which limits the
		 * identifiers of the bags within which we search for homomorphisms to
		 * the identifiers of these bags
		 */
		if (bagConstraint != null && bagConstraint.bags.length > 0) {
			if(bagConstraint.singleBagMatch) {
				List<String> bagPredicates = this.getBagPredicates();
				if (queryString.contains("WHERE") && bagPredicates.size() > 1) {
					queryString += " AND " + Joiner.on(" AND ").join(bagPredicates);
				} else if (bagPredicates.size() > 1) {
					queryString += " WHERE " + Joiner.on(" AND ").join(bagPredicates);
				}
			}

			String targetBagsString = "";
			if(bagConstraint.singleBagMatch) {
				targetBagsString = aliasName + ".BAG " + "IN " + this.createBagIdsInterval(Sets.newLinkedHashSet(Sets.newHashSet(bagConstraint.bags)));
			}
			else {
				List<String> bagStrings = new ArrayList<>();
				for(String alias:aliasNames) {
					bagStrings.add(alias + ".BAG " + "IN " + this.createBagIdsInterval(Sets.newLinkedHashSet(Sets.newHashSet(bagConstraint.bags))));
				}
				targetBagsString = Joiner.on(" AND ").join(bagStrings);
			}
			if (queryString.contains("WHERE")) {
				queryString += " AND " + targetBagsString;
			} else {
				queryString += " WHERE " + targetBagsString;
			}

			queryString += " ORDER BY " + aliasName + ".BAG";
		}

		// Limit the number of returned homomorphisms
		Integer topk = this.findTopKConstraint(constraints);
		if (topk != null) {
			queryString += " FETCH NEXT " + topk + " ROWS ONLY  ";
		}
		return Pair.of(aliasName, queryString);
	}

	/**
	 * @param atoms Collection<? extends PredicateFormula>
	 * @return Collection<String>
	 */
	@Override
	protected Collection<String> makeInserts(Collection<? extends PredicateFormula> atoms, Map<String, DBRelation> dbRelationMap) {
		Collection<String> result = new LinkedList<>();
		for (PredicateFormula atom : atoms) {
			Relation dbr = dbRelationMap.get(atom.getPredicateName());
			List<Term> terms = atom.getTerms();
			String insertInto = "INSERT INTO " + dbr.getName() + " " + "VALUES ( ";
			for (Term term : terms) {
				if (!term.isVariable()) {
					insertInto += "'" + term + "'" + ",";
				}
			}
			if (atom instanceof BagBoundPredicate) {
				insertInto += ((BagBoundPredicate) atom).getBag() + ",";
			} else {
				insertInto += 0 + ",";
			}
			insertInto += atom.getId();
			insertInto += ")";
			result.add(insertInto);
		}
		return result;
	}

	/**
	 * @param schema Schema
	 * @param source Evaluatable
	 * @param dbRelationMap Map<String,DBRelation>
	 * @param resultSet ResultSet
	 * @return Map<Variable,Term>
	 */
	@Override
	public Map<Variable, Constant> createMappingAlias(
			Map<String, TypedConstant<?>> constants,
			Evaluatable source, Map<String, DBRelation> dbRelationMap, ResultSet resultSet) {
		Set<String> projected = Sets.newLinkedHashSet(this.selectPredicates);
		Map<Variable, Constant> mapping = new LinkedHashMap<>();
		for (PredicateFormula atom : source.getBody().getPredicates()) {
			PredicateFormula predicate = atom;

			DBRelation dbr = dbRelationMap.get(predicate.getPredicateName());

			try {
				List<Variable> chaseTerms = Utility.typedToVariableTerms(dbr.getAttributes());;

				for (int i = 0, l = chaseTerms.size(); i < l; i++) {
					String columnName = this.atomAliases.get(predicate) + '.' + chaseTerms.get(i);
					if (projected.contains(columnName)) {
						int columnIndex = getColumnIndex(
								resultSet.getMetaData(),
								dbr.getName(),
								String.valueOf(chaseTerms.get(i)));
						String assigned = resultSet.getString(columnIndex);
//						TypedConstant<?> constant = schema.getConstants(assigned);
						TypedConstant<?> constant = constants.get(assigned);
						Constant constantTerm = null;
						if (constant != null) {
							constantTerm = constant;
						} else {
							constantTerm = new Skolem(assigned);
						}

						mapping.put((Variable) atom.getTerm(i), constantTerm);

					}
				}
			} catch (SQLException e) {
				throw new IllegalStateException(e.getMessage(), e);
			}
		}

		return mapping;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.homomorphism.AbstractHomomorphismStatementBuilder#createChaseTableStatement(uk.ac.ox.cs.pdq.schema.Relation)
	 */
	@Override
	protected String createTableStatement(Relation relation) {
		StringBuilder result = new StringBuilder();
		result.append("CREATE TABLE  ").append(relation.getName()).append('(');
		for (int index = 0; index < relation.getAttributes().size(); ++index) {
			result.append(' ').append(relation.getAttributes().get(index).getName()).append(" VARCHAR(500),");
		}
		result.append(" BAG int, ");
		result.append(" FACT int)");
		return result.toString();
	}

	/**
	 * @param rsmd
	 * @param table
	 * @param column
	 * @return the column index in the result set meta data, matching the given
	 * tablename and column names (case insensitive)
	 * @throws SQLException
	 */
	private static int getColumnIndex(ResultSetMetaData rsmd, String table, String column) throws SQLException {
		for (int i = 1, l = rsmd.getColumnCount(); i <= l; i++) {
			if (rsmd.getTableName(i).equalsIgnoreCase(table)
					&& rsmd.getColumnName(i).equalsIgnoreCase(column)) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * @return DerbyHomomorphismStatementBuilder
	 */
	@Override
	public DerbyHomomorphismStatementBuilder clone() {
		return new DerbyHomomorphismStatementBuilder();
	}

	@Override
	public String encodeName(String name) {
		String result = name;
		for (Map.Entry<String, String> entry: nameEncodings.entrySet()) {
			result = result.replace(entry.getKey(), entry.getValue());
		}
		return result;
	}

	@Override
	public String decodeName(String name) {
		String result = name;
		for (Map.Entry<String, String> entry: nameEncodings.entrySet()) {
			result = result.replace(entry.getValue(), entry.getKey());
		}
		return result;
	}
}