package uk.ac.ox.cs.pdq.planner.homomorphism;

import java.util.Map;

import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Bag;

/**
 * Super class for all constraints that can be passed for homomorphism detection
 *
 * @author Julien Leblay
 *
 */
public abstract class HomomorphismConstraint {

	/**
	 * Limits the number of matches to k
	 * @author Julien Leblay
	 */
	static class TopK extends HomomorphismConstraint {
		final int k;
		/**
		 * Constructor for TopK.
		 * @param k int
		 */
		private TopK(int k) {
			this.k = k;
		}
	}

	/**
	 * Limits the matches to a subset of facts.
	 * @author Julien Leblay
	 */
	static class AtomCollectionScope extends HomomorphismConstraint {
		final Conjunction<PredicateFormula> atoms;
		/**
		 * Constructor for AtomCollectionScope.
		 * @param atoms Conjunction<PredicateFormula>
		 */
		private AtomCollectionScope(Conjunction<PredicateFormula> atoms) {
			this.atoms = atoms;
		}
	}

	/**
	 * Limits the matches to a subset of bags.
	 * @author Julien Leblay
	 */
	static class BagScope extends HomomorphismConstraint {
		final Bag[] bags;
		final boolean singleBagMatch;
		/**
		 * Constructor for BagScope.
		 * @param singleBagMatch boolean
		 * @param bags Bag[]
		 */
		private BagScope(boolean singleBagMatch, Bag... bags) {
			this.bags = bags;
			this.singleBagMatch = singleBagMatch;
		}
	}

	/**
	 * Limits the matches to those subsuming the given map.
	 * @author Julien Leblay
	 */
	static class SuperMap extends HomomorphismConstraint {
		final Map<Variable, Constant> map;
		/**
		 * Constructor for SuperMap.
		 * @param map Map<Variable,Term>
		 */
		private SuperMap(Map<Variable, Constant> map) {
			this.map = map;
		}
	}

	/**
	 * @param k
	 * @return a fresh top k constraint
	 */
	public static TopK topK(int k) {
		return new TopK(k);
	}

	/**
	 * @param atoms
	 * @return a fresh atom collection scope constraint
	 */
	public static AtomCollectionScope scope(Conjunction<PredicateFormula> atoms) {
		return new AtomCollectionScope(atoms);
	}

	/**
	 * @param bags
	 * @param singleBagMatch boolean
	 * @return a fresh bag collection scope constraint
	 */
	public static BagScope scope(boolean singleBagMatch, Bag... bags) {
		return new BagScope(singleBagMatch, bags);
	}

	/**
	 * @param map
	 * @return a fresh super map constraint
	 */
	public static SuperMap satisfy(Map<Variable, Constant> map) {
		return new SuperMap(map);
	}
}
