package uk.ac.ox.cs.pdq.planner.homomorphism;

import java.sql.SQLException;

import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.PlannerParameters;
import uk.ac.ox.cs.pdq.planner.PlannerParameters.HomomorphismDetectorTypes;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;

import com.google.common.base.Strings;

/**
 * Provide the proper instance of homomorphism homoChecker depending and
 * contextual information
 *
 * @author Julien Leblay
 * @author Efthymia Tsamoura
 */
public class HomomorphismManagerFactory {

	/** Logger. */
	private static Logger log = Logger.getLogger(HomomorphismManagerFactory.class);

	/**
	 * Counter used to create distinct database schema for distinct method
	 * calls.
	 */
	private static Integer counter = 0;

	/**
	 * 
	 * @param schema
	 * @param accessibleQuery
	 * @param parameters
	 * @return an instance of homomorphism homoChecker instantiated according to
	 *         contextual information
	 * @throws PlannerException
	 */
	public static synchronized HomomorphismManager getInstance(
			AccessibleSchema schema, 
			Query<?> accessibleQuery, 
			PlannerParameters parameters)
					throws PlannerException {
		return getInstance(schema, 
				accessibleQuery, 
				parameters.getHomomorphismDetectorType(), 
				parameters.getDatabaseDriver(), 
				parameters.getConnectionUrl(),
				parameters.getDatabaseName(), 
				parameters.getDatabaseUser(),
				parameters.getDatabasePassword()
				);
	}
	
	/**
	 * 
	 * @param schema
	 * @param accessibleQuery
	 * @param type
	 * @param driver
	 * @param url
	 * @param database
	 * @param username
	 * @param password
	 * @return an instance of homomorphism homoChecker instantiated according to
	 *         contextual information
	 * @throws PlannerException
	 */
	public static synchronized HomomorphismManager getInstance(
			AccessibleSchema schema, 
			Query<?> accessibleQuery,
			HomomorphismDetectorTypes type,
			String driver,
			String url,
			String database,
			String username,			
			String password
			) throws PlannerException {
		HomomorphismManager result = null;
		try {
			if (type != null) {
				switch (type) {
				case DATABASE:
					AbstractHomomorphismStatementBuilder builder = null;
					if (url != null && url.contains("mysql")) {
						builder = new MySQLHomomorphismStatementBuilder();
					} else {
						if (Strings.isNullOrEmpty(driver)) {
							driver = "org.apache.derby.jdbc.EmbeddedDriver";
						}
						if (Strings.isNullOrEmpty(url)) {
							url = "jdbc:derby:memory:{1};create=true";
						}
						if (Strings.isNullOrEmpty(database)) {
							database = "chase";
						}
						database +=  "_" + System.currentTimeMillis() + "_" + counter++;
						synchronized (counter) {
							username = "APP_" + (counter++);
						}
						password = "";
						builder = new DerbyHomomorphismStatementBuilder();
					}
					result = new DBHomomorphismManager(
							driver, url, database, username, password, builder,
							schema, accessibleQuery);
					result.initialize();
					return result;
				}
			}
		} catch (SQLException e) {
			log.warn("Could not load " + database + ". Falling back to default database.");
		}
		synchronized (counter) {
			username = "APP_" + (counter++);
		}
		// Fail safe is in-memory derby
		try {
			result = new DBHomomorphismManager("org.apache.derby.jdbc.EmbeddedDriver",
					"jdbc:derby:memory:{1};create=true", "chase", username, "", new DerbyHomomorphismStatementBuilder(),
					schema, accessibleQuery);
			result.initialize();
			return result;
		} catch (SQLException e) {
			throw new IllegalStateException("Could not load default database.");
		}
	}
}