package uk.ac.ox.cs.pdq.planner.linear;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.pdq.algebra.Access;
import uk.ac.ox.cs.pdq.algebra.DependentAccess;
import uk.ac.ox.cs.pdq.algebra.DependentJoin;
import uk.ac.ox.cs.pdq.algebra.Join;
import uk.ac.ox.cs.pdq.algebra.Operators;
import uk.ac.ox.cs.pdq.algebra.Projection;
import uk.ac.ox.cs.pdq.algebra.RelationalOperator;
import uk.ac.ox.cs.pdq.algebra.Scan;
import uk.ac.ox.cs.pdq.algebra.Selection;
import uk.ac.ox.cs.pdq.algebra.StaticInput;
import uk.ac.ox.cs.pdq.algebra.SubPlanAlias;
import uk.ac.ox.cs.pdq.algebra.predicates.Predicate;
import uk.ac.ox.cs.pdq.db.AccessMethod.AccessMethodTypes;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.plan.AccessOperator;
import uk.ac.ox.cs.pdq.plan.LinearPlan;
import uk.ac.ox.cs.pdq.plan.Plan.ControlFlows;
import uk.ac.ox.cs.pdq.planner.linear.explorer.Candidate;
import uk.ac.ox.cs.pdq.planner.linear.node.SearchNode;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;

/**
 * Creates linear plans
 *
 * @author Efthymia Tsamoura
 */
public class LinearPlanGenerator {

	/**
	 * @param configuration LinearConfiguration
	 * @param parentPlan LinearPlan
	 * @param cf ControlFlows
	 * @return LinearPlan
	 */
	public static <S extends ChaseState> LinearPlan createLinearPlan(LinearChaseConfiguration<S> configuration, LinearPlan parentPlan, ControlFlows cf) {
		Preconditions.checkArgument(cf != null);
		switch(cf) {
		case TOP_DOWN:
			return createTopDownLinearPlan(configuration, parentPlan, inferOutputChaseConstants(configuration));
		default:
			return createBottomUpLinearPlan(configuration, parentPlan, inferOutputChaseConstants(configuration));
		}
	}

	/**
	 * @param nodes List<T>
	 * @return LinearPlan
	 */
	public static<T extends SearchNode> LinearPlan createLinearPlan(List<T> nodes) {
		LinearPlan parentPlan = null;
		for (T node: nodes) {
			parentPlan = node.getConfiguration().createPlan(parentPlan);
		}
		return parentPlan;
	}

	/**
	 * @param configuration LinearConfiguration
	 * @param predecessor LinearPlan
	 * @param toProject List<Term>
	 * @return LinearPlan
	 */
	private static LinearPlan createTopDownLinearPlan(LinearConfiguration configuration,
			LinearPlan predecessor,
			List<Term> toProject) {
		Preconditions.checkArgument(configuration.getExposedCandidates() != null);
		RelationalOperator op1 = null;
		AccessOperator access = null;
		RelationalOperator predAlias = null;
		if (predecessor != null) {
			predAlias = new SubPlanAlias(predecessor);
		}

		for (Candidate candidate: configuration.getExposedCandidates()) {
			if (access == null) {
				if (candidate.getBinding().getType() == AccessMethodTypes.FREE) {
					access = new Scan(candidate.getRelation());
				} else {
					access = new DependentAccess(candidate.getRelation(), candidate.getBinding(), candidate.getFact().getTerms());
				}
			}
			RelationalOperator op2 = (RelationalOperator) access;
			Predicate selectPredicates = Operators.createSelectPredicates(candidate.getFact().getTerms());
			if (selectPredicates != null) {
				op2 = new Selection(selectPredicates, op2);
			}
			LinkedHashMap<Integer, Term> renaming = getOutputMap(candidate, toProject);
			if (renaming.size() != op2.getColumns().size()
					|| !renaming.values().containsAll(op2.getColumns())
					|| !op2.getColumns().containsAll(renaming.values())) {
				op2 = new Projection(op2, renaming);
			}

			if (op1 == null) {
				op1 = op2;
			} else {
				op1 = new Join(op1, op2);
			}
		}
		if (predecessor != null) {
			if (access instanceof DependentAccess) {
				op1 = new DependentJoin(predAlias, op1);
			} else {
				op1 = new Join(predAlias, op1);
			}
		}
		LinearPlan lp = new LinearPlan(op1, access, ControlFlows.TOP_DOWN);
		if (predecessor != null) {
			lp.addPrefix(predecessor);
			predecessor.addSuffix(lp);
		}
		return lp;
	}

	/**
	 * @param configuration LinearConfiguration
	 * @param predecessor LinearPlan
	 * @param toProject List<Term>
	 * @return LinearPlan
	 */
	private static LinearPlan createBottomUpLinearPlan(LinearConfiguration configuration,
			LinearPlan predecessor,
			List<Term> toProject) {
		Preconditions.checkArgument(configuration.getExposedCandidates() != null);
		RelationalOperator op1 = null;
		AccessOperator access = null;
		RelationalOperator project = null;

		RelationalOperator predAlias = null;
		if (predecessor != null) {
			predAlias = new SubPlanAlias(predecessor);
		}
		Candidate exposedCandidate = configuration.getExposedCandidates().iterator().next();
		List<Constant> inputs = exposedCandidate.getInput();
		if (!inputs.isEmpty()) {
			List<TypedConstant<?>> constants = getSchemaConstants(inputs);
			if (constants.size() == inputs.size()) {
				project = new StaticInput(constants);
			} else if (predecessor != null){
				project = new Projection(predAlias, getInputList(exposedCandidate, inputs));
			} else {
				throw new java.lang.IllegalStateException("Attempting to create open plan: " + constants + ": " + inputs);
			}
		}

		for (Candidate candidate: configuration.getExposedCandidates()) {
			if (access == null) {
				if (candidate.getBinding().getType() == AccessMethodTypes.FREE) {
					access = new Scan(candidate.getRelation());
				} else if (project != null) {
					access = new Access(candidate.getRelation(), candidate.getBinding(), project);
				} else {
					throw new IllegalStateException("No child defined as access inputs.");
				}
			}
			RelationalOperator op2 = (RelationalOperator) access;
			Predicate selectPredicates = Operators.createSelectPredicates(candidate.getFact().getTerms());
			if (selectPredicates != null) {
				op2 = new Selection(selectPredicates, op2);
			}
			LinkedHashMap<Integer, Term> renaming = getOutputMap(candidate, toProject);
			if (renaming.size() != op2.getColumns().size()
					|| !renaming.values().containsAll(op2.getColumns())
					|| !op2.getColumns().containsAll(renaming.values())) {
				op2 = new Projection(op2, renaming);
			}

			if (op1 == null) {
				op1 = op2;
			} else {
				op1 = new Join(op1, op2);
			}
		}
		if (predecessor != null) {
			op1 = new Join(predAlias, op1);
		}
		LinearPlan lp = new LinearPlan(op1, access, ControlFlows.BOTTOM_UP);
		if (predecessor != null) {
			lp.addPrefix(predecessor);
			predecessor.addSuffix(lp);
		}
		return lp;
	}
	

	/**
	 * @param configuration LinearConfiguration
	 * @return List<Term>
	 */
	public static <S extends ChaseState> List<Term> inferOutputChaseConstants(LinearChaseConfiguration<S> configuration) {
		Collection<Term> result = new LinkedHashSet<>();
		for (PredicateFormula fact: configuration.getOutputFacts()) {
			for (Term t: fact.getTerms()) {
				if (t.isVariable() || t.isSkolem()) {
					result.add(t);
				}
			}
		}
		return Lists.newArrayList(result);
	}

	/**
	 * @param constants Collection<Constant>
	 * @return List<TypedConstant<?>>
	 */
	private static List<TypedConstant<?>> getSchemaConstants(Collection<Constant> constants) {
		List<TypedConstant<?>> ret = new ArrayList<>();
		for (Constant o: constants) {
			if (o instanceof TypedConstant) {
				ret.add((TypedConstant) o);
			}
		}
		return ret;
	}

	/**
	 * @param candidate Candidate
	 * @param inputs Collection<? extends Term>
	 * @return List<Term>
	 */
	private static List<Term> getInputList(Candidate candidate, Collection<? extends Term> inputs) {
		List<Term> result = new ArrayList<>();
		List<Term> factTerms = candidate.getFact().getTerms();
		for (Integer i: candidate.getBinding().getInputs()) {
			Term t = factTerms.get(i - 1);
			result.add(t);
		}
		return result;
	}

	/**
	 * @param candidate Candidate
	 * @param toProject List<? extends Term>
	 * @return LinkedHashMap<Integer,Term>
	 */
	private static LinkedHashMap<Integer, Term> getOutputMap(Candidate candidate, List<? extends Term> toProject) {
		LinkedHashMap<Integer, Term> ret = new LinkedHashMap<>();
		for (Map.Entry<Integer, Term> entry: candidate.getOutputChaseConstants().entrySet()) {
			if (toProject.contains(entry.getValue())) {
				ret.put(entry.getKey(), entry.getValue());
			}
		}
		return ret;
	}
}
