package uk.ac.ox.cs.pdq.planner.linear.pruning;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.jgrapht.graph.DefaultEdge;

import uk.ac.ox.cs.pdq.LimitReachedException;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.plan.LinearPlan;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessibleRelation;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.InferredAccessibleRelation;
import uk.ac.ox.cs.pdq.planner.linear.explorer.Candidate;
import uk.ac.ox.cs.pdq.planner.linear.node.NodeFactory;
import uk.ac.ox.cs.pdq.planner.linear.node.PlanTree;
import uk.ac.ox.cs.pdq.planner.linear.node.SearchNode;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;

/**
 * Plan post pruning class
 *
 * @author Efthymia Tsamoura
 */
public abstract class PostPruning<S extends ChaseState, N extends SearchNode> {

	/** Creates tree nodes */
	protected final NodeFactory<S,N> nodeFactory;
	protected final AccessibleSchema accessibleSchema;
	/** True if the input path is pruned */
	protected Boolean isPruned = false;
	/** The pruned path and its corresponding plan */
	protected List<N> path = null;
	protected LinearPlan plan = null;

	/**
	 * @param nodeFactory
	 * @param accessibleSchema
	 */
	public PostPruning(NodeFactory<S,N> nodeFactory, AccessibleSchema accessibleSchema) {
		Preconditions.checkArgument(nodeFactory != null);
		Preconditions.checkArgument(accessibleSchema != null);
		this.nodeFactory = nodeFactory;
		this.accessibleSchema = accessibleSchema;
	}

	/**
	 * Post-prunes the input path and adds it to the input plan tree
	 * @param planTree
	 * @param path
	 * @param queryFacts facts of a query match
	 * @return true if the input path is postpruned
	 * @throws PlannerException
	 */
	public boolean prune(N root, List<N> path, Collection<PredicateFormula> queryFacts) throws PlannerException, LimitReachedException {
		Preconditions.checkArgument(path != null);
		Preconditions.checkArgument(queryFacts != null);
		this.isPruned = false;
		this.path = null;
		this.plan = null;
		Collection<PredicateFormula> qF = new LinkedHashSet<>();
		for(PredicateFormula queryFact: queryFacts) {
			if (queryFact.getSignature() instanceof InferredAccessibleRelation) {
				qF.add(queryFact);
			} else {
				Preconditions.checkState(queryFact.getSignature() instanceof AccessibleRelation);
			}
		}
		Collection<PredicateFormula> factsToExpose = this.findFactsToExpose(path, qF);
		if(this.isPruned) {
			this.createPath(root, path, factsToExpose);
		}
		return this.isPruned;
	}

	/**
	 * @param candidates Exposed candidates
	 * @param minimalFacts
	 * @return the candidates that correspond to the input facts
	 */
	protected static Set<Candidate> getUtilisedCandidates(Collection<Candidate> candidates, Collection<PredicateFormula> minimalFacts) {
		Set<Candidate> useful = new HashSet<>();
		for(Candidate candidate: candidates) {
			PredicateFormula inferredAccessibleFact = candidate.getInferredAccessibleFact();
			if (minimalFacts.contains(inferredAccessibleFact)) {
				useful.add(candidate);
			}
		}
		return useful;
	}

	/**
	 * @param path
	 * 		A path that must be added to the plan tree
	 * 		If the root of this path is not the root of the plan tree, then an exception is thrown
	 * @param planTree DirectedGraph<N,DefaultEdge>
	 * @param parentNode N
	 */
	public void addPrunedPathToTree(PlanTree<N> planTree, N parentNode, List<N> path) {
		Preconditions.checkArgument(path != null);
		Preconditions.checkArgument(!path.isEmpty());
		
		for(int j = 0; j < path.size(); ++j) {
			try {
				planTree.addVertex(path.get(j));
				if(j == 0) {
					planTree.addEdge(parentNode, path.get(j), new DefaultEdge());
				}
				else {
					planTree.addEdge(path.get(j-1), path.get(j), new DefaultEdge());
				}
			} catch(Exception ex) {
				throw new IllegalStateException(ex);
			}
		}
	}
	
	public List<N> getPath() {
		return this.path;
	}
	
	public LinearPlan getPlan() {
		return this.plan;
	}

	/**
	 * Method findFactsToExpose.
	 * @param path List<N>
	 * @param queryFacts Collection<PredicateFormula>
	 * @return Collection<PredicateFormula>
	 */
	protected abstract Collection<PredicateFormula> findFactsToExpose(List<N> path, Collection<PredicateFormula> queryFacts);
	
	/**
	 * Creates a post-pruned path given the input exposed facts
	 * @param path Input path
	 * @param factsToExpose The facts that will be exposed
	 * @throws PlannerException
	 */
	protected abstract void createPath(N root, List<N> path, Collection<PredicateFormula> factsToExpose) throws PlannerException, LimitReachedException;
}
