package uk.ac.ox.cs.pdq.planner.linear.pruning;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.LimitReachedException;
import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibilityAxiom;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.db.access.InferredAccessibleAxiom;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessibleRelation;
import uk.ac.ox.cs.pdq.planner.linear.explorer.Candidate;
import uk.ac.ox.cs.pdq.planner.linear.node.NodeFactory;
import uk.ac.ox.cs.pdq.planner.linear.node.SearchNode;
import uk.ac.ox.cs.pdq.planner.linear.node.SearchNode.NodeStatus;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;
import uk.ac.ox.cs.pdq.reasoning.Match;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;

/**
 * An improved path post pruning algorithm.
 * The latter aims at removing not only the redundant accesses but also the redundant follow-up joins.
 *
 * @author Efthymia Tsamoura
 */
public final class PostPruningRemoveFollowUps<S extends ChaseState, N extends SearchNode> extends PostPruning<S,N> {


	/**
	 * Constructor for PlanPostPruningImprovedImplementation.
	 * @param nodeFactory NodeFactory<S,N>
	 * @param accessibleSchema AccessibleSchema
	 */
	public PostPruningRemoveFollowUps(NodeFactory<S,N> nodeFactory, AccessibleSchema accessibleSchema) {
		super(nodeFactory, accessibleSchema);
	}

	/**
	 * Creates a post-pruned path given the input exposed facts
	 * @param path Input path
	 * @param factsToExpose The facts that will be exposed
	 * @throws PlannerException
	 */
	protected void createPath(N root, List<N> path, Collection<PredicateFormula> factsToExpose) throws PlannerException, LimitReachedException {
		if(this.isPruned) {
			this.path = new ArrayList<>();
			List<Integer> nodeIds = new ArrayList<>();

			for(int i = 0; i < path.size(); ++i) {
				N node = path.get(i);
				Preconditions.checkNotNull(node.getConfiguration().getExposedCandidates());
				Collection<Candidate> alreadyExposed = node.getConfiguration().getExposedCandidates();
				/*
				 * Get the candidate facts C of the current node that are part of the minimalFacts collection.
				 * The list C corresponds to the minimal set of candidate facts that must be exposed by this node in order to answer the query.
				 * If C is empty, then the current node has to be eliminated from the path
				 */
				Set<Candidate> mustBeExposed = getUtilisedCandidates(alreadyExposed, factsToExpose);
				if(!mustBeExposed.isEmpty()) {
					N freshNode;
					if(i > 0) {
						N parentNode = path.get(i-1);
						//freshNode = this.nodeFactory.getInstance(parentNode, mustBeExposed);
						freshNode = this.nodeFactory.getInstance(parentNode, Sets.newHashSet(alreadyExposed));
					}
					else {
						freshNode = this.nodeFactory.getInstance(root, Sets.newHashSet(alreadyExposed));
					}
					freshNode.setStatus(NodeStatus.FAKE_TERMINAL);
					freshNode.close();
					this.path.add(freshNode);
					nodeIds.add(freshNode.getId());
				}
			}

			int n;
			for(n = 0; n < this.path.size() - 1; ++n) {
				this.path.get(n).setPathToSuccess(Lists.<Integer>newArrayList(nodeIds.subList(n+1, nodeIds.size())));
			}
			this.path.get(n).setPathToSuccess(Lists.<Integer>newArrayList());
			this.path.get(n).setStatus(NodeStatus.SUCCESSFUL);
			List<Match> matches = this.path.get(n).matchesQuery();
			Preconditions.checkArgument(!matches.isEmpty());
			this.plan = this.path.get(n).getConfiguration().getPlan();
		}
	}

	/**
	 * @param path List<N>
	 * @param queryFacts Collection<PredicateFormula>
	 * @return Collection<PredicateFormula>
	 */
	@Override
	protected Collection<PredicateFormula> findFactsToExpose(List<N> path, Collection<PredicateFormula> queryFacts) {
		N successNode = path.get(path.size() - 1);
		// Find the minimal set of facts that have to be exposed in order to make each fact in the query match accessible
		Collection<PredicateFormula> minimalFacts = this.getMinimalFacts(queryFacts, successNode);

		// For each node in the path
		int i = 0;
		while(i < path.size()) {
			N node = path.get(i);
			Collection<Candidate> alreadyExposed = node.getConfiguration().getExposedCandidates();
			if(alreadyExposed != null) {
				/*
				 * Get the list of candidates C of the current node that are part of the minimalFacts collection.
				 * The list C corresponds to the minimal set of candidate facts that must be exposed in this node.
				 * If C is empty, then the current node has to be eliminated from the path
				 */
				Collection<Candidate> mustBeExposed = getUtilisedCandidates(alreadyExposed, minimalFacts);
				if (mustBeExposed.isEmpty() || !Sets.newHashSet(alreadyExposed).equals(Sets.newHashSet(mustBeExposed))) {
					this.isPruned = true;
					i = path.size();
					break;
				}
			}
			++i;
		}
		return minimalFacts;
	}

	/**
	 *
	 * @param queryMatch The facts of a query match
	 * @param successNode A success node (i.e., a node where the query match is being found)
	 * @return a minimal set of facts that have to be exposed in order to make each input fact accessible */
	private Collection<PredicateFormula> getMinimalFacts(Collection<PredicateFormula> queryMatch, SearchNode successNode) {
		Preconditions.checkArgument(queryMatch != null);
		Preconditions.checkArgument(successNode != null);
		Preconditions.checkArgument(successNode.getStatus() == NodeStatus.SUCCESSFUL);
		Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> factProvenance = successNode.getConfiguration().getFactProvenance();
		Collection<PredicateFormula> minimalFacts = this.getMinimalFactsRecursive(queryMatch, Utility.getConstants(queryMatch), new LinkedHashSet<Constant>(), factProvenance);
		return minimalFacts;
	}

	/**
	 * This function aims at minimising the set of facts that have to be exposed in order to make each input fact accessible
	 *
	 * @param facts
	 * 		A list of accessible, inferred accessible and accessed atoms
	 * @param inputTerms
	 * 		The input constants of the input atoms (i.e., the constants that are required to expose the input atoms)
	 * @param outputTerms
	 * 		The output constants of the input atoms
	 * @param factProvenance
	 * 		Mapping of atoms F to pairs of dependencies and atoms.
	 * 		Each pair corresponds to the dependency and the facts that were used to fire the former dependency and derive a fact in F.
	 * @return A (minimal) list of inferred accessible facts.
	 */
	private Collection<PredicateFormula> getMinimalFactsRecursive(
			Collection<PredicateFormula> facts,
			Collection<Constant> inputTerms,
			Collection<Constant> outputTerms,
			Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> factProvenance) {

		// The minimal set of atoms
		Collection<PredicateFormula> parentFacts = new LinkedHashSet<>();
		// For each input fact
		for(PredicateFormula fact:facts) {
			// Get its provenance
			Pair<Constraint, Collection<PredicateFormula>> icToFacts = factProvenance.get(fact);
			if(icToFacts == null) {
				return new LinkedHashSet<>();
			}
			Constraint ic = icToFacts.getLeft();
			/*
			 * If the dependency (query) that leads to the derivation of the current atom is an accessed atom -> inferred accessible atom dependency
			 * then add the current atom in the output list of facts and continue the recursion with the accessed fact.
			 * The recursion continues in order to find the accesses that must be performed in order to expose the current fact
			 */
			if(ic instanceof InferredAccessibleAxiom && ((InferredAccessibleAxiom)ic).isAccessedToInferredAccessedAxiom()) {
				parentFacts.add(fact);
				parentFacts.addAll(this.getMinimalFactsRecursive(icToFacts.getRight(), inputTerms, outputTerms, factProvenance));
			}
			/*
			 * If the dependency (query) that leads to the derivation of the current atom is the inferred accessible version of a schema dependency
			 * then add the current atom in the output list of facts and continue the recursion with the facts that were used to fire the considered dependency.
			 * The recursion continues in order to reduce to the case where all of the returned inferred accessible facts come from accessibility
			 */
			else if(ic instanceof InferredAccessibleAxiom && !((InferredAccessibleAxiom)ic).isAccessedToInferredAccessedAxiom()) {
				parentFacts.addAll(this.getMinimalFactsRecursive(icToFacts.getRight(), inputTerms, outputTerms, factProvenance));
			}
			// If the dependency is an accessibility axiom
			else if(ic instanceof AccessibilityAxiom) {
				if(!outputTerms.containsAll(inputTerms)) {
					/*
					 * Find the accessible facts I that were used to fire the considered dependency (query),
					 * the accessible facts O that were derived from this firing and
					 * the inferred accessible fact that corresponds to the dependency's accessed fact.
					 * Update the list of input and output constants and continue the recursion for each accessible fact in I.
					 * The recursion continues in order to find the accesses that must be performed in order to expose the current fact
					 */
					Element elements = this.getElement(icToFacts.getRight());
					inputTerms.addAll(elements.getInputTerms());
					outputTerms.addAll(elements.getOutputTerms());
					outputTerms.removeAll(elements.getInputTerms());  // Julien-bugfix: if a terms previously thought to be accessible turns out to be a required input earlier in the plan, remove it from the output.
					parentFacts.add(elements.getInferredAccessibleFact());
					for(PredicateFormula accessibleFact:elements.getInputAccessibleFacts()) {
						parentFacts.addAll(this.getMinimalFactsRecursive(Sets.newLinkedHashSet(Sets.newHashSet(accessibleFact)), inputTerms, outputTerms, factProvenance));
					}
				}
			}
		}
		return parentFacts;
	}

	/**
	 * @param facts Facts that were used to fire an accessibility axiom
	 * @return A structure of
	 * 				the accessible facts that were used to fire an accessibility axiom
	 * 				the accessible facts O that were derived from this firing and
	 * 				the inferred accessible fact that corresponds to the axiom's accessed fact.
	 */
	private Element getElement(Collection<PredicateFormula> facts) {
		Collection<Constant> inputTerms = new LinkedHashSet<>();
		Collection<Constant> outputTerms = new LinkedHashSet<>();
		Collection<PredicateFormula> inputAccessibleFacts = new LinkedHashSet<>();
		PredicateFormula inferredAccessibleFact = null;
		for(PredicateFormula fact:facts) {
			if(fact.getSignature() instanceof AccessibleRelation) {
				Set<Constant> constants = Utility.getNonSchemaConstants(fact);
				if(!constants.isEmpty()) {
					inputTerms.addAll(constants);
					inputAccessibleFacts.add(fact);
				}
			}
			else if(fact.getSignature() instanceof Relation) {
				outputTerms.addAll(Utility.getConstants(fact));
				Relation relation = (Relation) fact.getSignature();
				inferredAccessibleFact = new PredicateFormula(this.accessibleSchema.getInferredAccessibleRelation(relation), fact.getTerms() );
			}
			else {
				throw new java.lang.IllegalArgumentException();
			}
		}
		outputTerms.removeAll(inputTerms);
		return new Element(inputTerms, outputTerms, inputAccessibleFacts, inferredAccessibleFact);
	}

	/**
	 */
	private class Element {
		private final Collection<Constant> inputTerms;
		private final Collection<Constant> outputTerms;
		private final Collection<PredicateFormula> inputAccessibleFacts;
		private final PredicateFormula inferredAccessibleFact;

		/**
		 * Constructor for Element.
		 * @param inputTerms Collection<Constant>
		 * @param outputTerms Collection<Constant>
		 * @param accessibleFacts Collection<PredicateFormula>
		 * @param inferredAccessibleFact PredicateFormula
		 */
		public Element(Collection<Constant> inputTerms,
				Collection<Constant> outputTerms,
				Collection<PredicateFormula> accessibleFacts,
				PredicateFormula inferredAccessibleFact) {
			Preconditions.checkArgument(inputTerms != null);
			Preconditions.checkArgument(outputTerms != null);
			Preconditions.checkArgument(accessibleFacts != null);
			Preconditions.checkArgument(inferredAccessibleFact != null);
			this.inputTerms = inputTerms;
			this.outputTerms = outputTerms;
			this.inputAccessibleFacts = accessibleFacts;
			this.inferredAccessibleFact = inferredAccessibleFact;
		}

		/**
		 * @return Collection<Constant>
		 */
		public Collection<Constant> getInputTerms() {
			return this.inputTerms;
		}

		/**
		 * @return Collection<Constant>
		 */
		public Collection<Constant> getOutputTerms() {
			return this.outputTerms;
		}

		/**
		 * @return Collection<PredicateFormula>
		 */
		public Collection<PredicateFormula> getInputAccessibleFacts() {
			return this.inputAccessibleFacts;
		}

		/**
		 * @return PredicateFormula
		 */
		public PredicateFormula getInferredAccessibleFact() {
			return this.inferredAccessibleFact;
		}

		/**
		 * @return String
		 */
		@Override
		public String toString() {
			return "Inputs " + this.inputTerms.toString() + "\n" +
					"Outputs " + this.outputTerms.toString() + "\n" +
					"Accessible facts " + this.inputAccessibleFacts.toString()  + "\n" +
					"Inferred Accessible facts " + this.inferredAccessibleFact.toString()  + "\n";
		}
	}
}
