package uk.ac.ox.cs.pdq.planner.parallel.threads;

import uk.ac.ox.cs.pdq.planner.dag.BinaryConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.cost.ConfigurationCostEstimator;
import uk.ac.ox.cs.pdq.planner.dag.explorer.Representative;
import uk.ac.ox.cs.pdq.planner.dag.explorer.Template;
import uk.ac.ox.cs.pdq.planner.homomorphism.DBHomomorphismManager;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismDetector;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Chaser;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.DatabaseChaseState;
import uk.ac.ox.cs.pdq.reasoning.Reasoner;

import com.google.common.base.Preconditions;

/**
 * Provides a function to speed up binary configuration creation either through the use of representative or template configurations.
 *
 * @author Efthymia Tsamoura
 *
 * @param <S>
 */
public abstract class DefaultThread<S extends ChaseState> extends ExecutionThread<S> {

	/** Performs reasoning. Closes newly created binary configurations*/
	protected final Reasoner reasoner;

	/** Detects homomorphisms*/
	protected final HomomorphismDetector detector;

	/** Estimates the cost of the plans */
	protected final ConfigurationCostEstimator costEstimator;

	/** Map of representatives. For each configuration c = BinConfiguration(c_1,c_2) we create a map from the
		equivalence classes of c and c' to c''. This map helps us reducing the chasing time, i.e.,
		if c'_1 and c'_2 are structurally equivalent to c_1 and c_2, respectively, and
		c = BinConfiguration(c_1,c_2) has already been fully chased,
		then we copy the state of c to the state of c' = BinConfiguration(c'_1,c'_2).*/
	protected final Representative<S> representatives;

	/**
	 * Maps each configuration to its constituting ApplyRule configurations. Used to speed up chasing, i.e.,
	 * when we are about to create a new binary configuration c''= BinaryConfiguration(c,c')
	 * from c and c' and there exists another configuration c^(3) with ApplyRules
	 * the ApplyRules of c and c' and c^(3) is already chased then we use c^(3)'s state as the state of c''
	 */
	protected final Template<S> templates;

	/**
	 *
	 * @param reasoner
	 * 		Performs reasoning. Closes newly created binary configurations
	 * @param detector
	 * 		Detects homomorphisms
	 * @param costEstimator
	 * 		Estimates the cost of the plans
	 * @param representatives
	 * 		Map of representatives. For each configuration c = BinConfiguration(c_1,c_2) we create a map from the
			equivalence classes of c and c' to c''. This map helps us reducing the chasing time, i.e.,
			if c'_1 and c'_2 are structurally equivalent to c_1 and c_2, respectively, and
			c = BinConfiguration(c_1,c_2) has already been fully chased,
			then we copy the state of c to the state of c' = BinConfiguration(c'_1,c'_2).
	 * @param templates
	 * 		Maps each configuration to its constituting ApplyRule configurations. Used to speed up chasing, i.e.,
	 * 		when we are about to create a new binary configuration c''= BinaryConfiguration(c,c')
	 * 		from c and c' and there exists another configuration c^(3) with ApplyRules
	 * 		the ApplyRules of c and c' and c^(3) is already chased then we use c^(3)'s state as the state of c''
	 */
	public DefaultThread(
			Reasoner reasoner,
			HomomorphismDetector detector,
			ConfigurationCostEstimator costEstimator,
			Representative<S> representatives,
			Template<S> templates
			) {
		Preconditions.checkNotNull(reasoner);
		Preconditions.checkNotNull(detector);
		Preconditions.checkNotNull(costEstimator);
		Preconditions.checkNotNull(representatives);

		this.reasoner = reasoner;
		this.detector = detector;
		this.costEstimator = costEstimator;
		this.representatives = representatives;
		this.templates = templates;
	}

	/**
	 * @param left
	 * @param right
	 * @return a new binary configuration BinConfiguration(left, right)
	 */
	protected DAGChaseConfiguration<S> merge(DAGChaseConfiguration<S> left, DAGChaseConfiguration<S> right) {
		DAGChaseConfiguration<S> configuration = null;
		//A configuration BinConfiguration(c,c'), where c and c' belong to the equivalence classes of
		//the left and right input configuration, respectively.
		DAGChaseConfiguration<S> representative = this.representatives.getRepresentative(left, right);
		
		DAGChaseConfiguration<S> template = null;
		if(this.templates != null) {
			template = this.templates.getTemplate(left, right);
		}
		
		if(representative == null) {
			representative = this.representatives.getRepresentative(right, left);
		}
		
		//If the representative is null or we do not use templates or we cannot find a template configuration that
		//consists of the corresponding ApplyRules, then create a binary configuration from scratch by fully chasing its state
		if(representative == null && (this.templates == null || template == null)) {
			configuration = new BinaryConfiguration<>(
					left.getAccessibleSchema(),
					left.getQuery(),
					(Chaser) this.reasoner,
					null,
					left.getDominanceDetectors(),
					left.getSuccessDominanceDetector(),
					this.costEstimator,
					left,
					right,
					false);
					
			if(configuration.getState() instanceof DatabaseChaseState) {
				((DatabaseChaseState)configuration.getState()).setManager((DBHomomorphismManager) this.detector);
			}
			((Chaser) this.reasoner).chase(configuration.getState(), left.getAccessibleSchema().getShortInferredAccessibilityAxioms());
	
			if(this.templates != null) {
				this.templates.put(configuration);
			}
			this.representatives.put(left, right, configuration);

		}
		//otherwise, re-use the state of the representative
		else if(representative != null) {
			configuration = new BinaryConfiguration<>(
					left.getAccessibleSchema(),
					left.getQuery(),
					(Chaser) this.reasoner,
					null,
					left.getDominanceDetectors(),
					left.getSuccessDominanceDetector(),
					this.costEstimator,
					left,
					right,
					false);

			if(this.templates != null && template == null) {
				this.templates.put(configuration);
			}
		}
		//or the template configuration
		else if(this.templates != null && template != null) {
			representative = template;
			configuration = new BinaryConfiguration<>(
					left.getAccessibleSchema(),
					left.getQuery(),
					(Chaser) this.reasoner,
					null,
					left.getDominanceDetectors(),
					left.getSuccessDominanceDetector(),
					this.costEstimator,
					left,
					right,
					false);

			this.representatives.put(left, right, configuration);
		}
		return configuration;
	}
}