package uk.ac.ox.cs.pdq.planner.parallel.threads;

import java.util.Collection;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.Callable;

import uk.ac.ox.cs.pdq.planner.dag.DAGChaseConfiguration;
import uk.ac.ox.cs.pdq.planner.dag.equivalence.DAGEquivalenceClasses;
import uk.ac.ox.cs.pdq.planner.homomorphism.DBHomomorphismManager;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismDetector;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.dominance.SuccessDominance;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.DatabaseChaseState;

/**
 * Iterates over the input collection of configurations to identify the minimum-cost one
 *
 * @author Efthymia Tsamoura
 *
 * @param <S>
 */
public class FinalIterationThread<S extends ChaseState> extends ExecutionThread<S> implements Callable<DAGChaseConfiguration<S>> {

	/** Performs success dominance checks*/
	private final SuccessDominance successDominance;

	/** Detects homomorphisms*/
	private final HomomorphismDetector detector;

	/** Input configurations*/
	private final Queue<DAGChaseConfiguration<S>> configurations;

	/** Classes of structurally equivalent configurations */
	private final DAGEquivalenceClasses<S> equivalenceClasses;

	/** The minimum cost closed and successful configuration found so far */
	private DAGChaseConfiguration<S> best = null;

	/** The output non-dominated and not successful configurations */
	private final Set<DAGChaseConfiguration<S>> output;

	/** The output non-dominated and successful (and not closed) configurations */
	private final Set<DAGChaseConfiguration<S>> successful;

	/**
	 *
	 * @param configurations
	 * 		Input configurations
	 * @param equivalenceClasses
	 * 		Classes of structurally equivalent configurations
	 * @param best
	 * 		The minimum cost closed and successful configuration found so far
	 * @param detector
	 * 		Detects homomorphisms
	 * @param successDominance
	 * 		Performs success dominance checks
	 * @param output
	 * 		The output non-dominated and not successful configurations
	 * @param successfulConfigurations
	 * 		The output non-dominated and successful (and not closed) configurations
	 */
	public FinalIterationThread(
			Queue<DAGChaseConfiguration<S>> configurations,
			DAGEquivalenceClasses<S> equivalenceClasses,
			DAGChaseConfiguration<S> best,
			HomomorphismDetector detector,
			SuccessDominance successDominance,
			Set<DAGChaseConfiguration<S>> output,
			Set<DAGChaseConfiguration<S>> successfulConfigurations
			) {
		this.best = best == null ? null : best.clone();
		this.detector = detector;
		this.configurations = configurations;
		this.equivalenceClasses = equivalenceClasses;
		this.output = output;
		this.successful = successfulConfigurations;
		this.successDominance = successDominance;
	}

	/**
	 * @return DAGChaseConfiguration<S>
	 * @throws Exception
	 * @see java.util.concurrent.Callable#call()
	 */
	@Override
	public DAGChaseConfiguration<S> call() throws Exception {
		DAGChaseConfiguration<S> configuration;
		//Poll the next configuration
		while((configuration = this.configurations.poll()) != null) {
			if(configuration.getState() instanceof DatabaseChaseState) {
				((DatabaseChaseState)configuration.getState()).setManager((DBHomomorphismManager) this.detector);
			}
			//If the configuration is not dominated
			DAGChaseConfiguration<S> dominator = this.equivalenceClasses.dominate(configuration);
			if (dominator != null
					//ExplorerUtils.isDominated(this.dag, binConfig) != null || ExplorerUtils.isDominated(binConfigs, binConfig) != null
					) {
			} else {
				//Assess its potential
				if (this.getPotential(configuration, this.best == null ? null : this.best.getPlan(), this.successDominance)) {
					//Find the configurations dominated by the current one and remove them
					Collection<DAGChaseConfiguration<S>> dominated = this.equivalenceClasses.dominatedBy(configuration);
					if(!dominated.isEmpty()) {
						this.output.removeAll(dominated);
						this.equivalenceClasses.removeAll(dominated);
						this.successful.removeAll(dominated);
					}
					//Check for query match
					boolean matchesQuery = configuration.isSuccessful();
					if (configuration.isClosed() && matchesQuery) {
						this.setBestConfiguration(configuration);
						//System.out.println(this.bestConfiguration + "\t" + this.bestConfiguration.getPlan().getCost());
					}
					//Update the best configuration
					else {
						this.equivalenceClasses.addEntry(configuration);
						this.output.add(configuration);
					}
					if(matchesQuery) {
						this.successful.add(configuration);
					}
				}
			}
		}
		return this.best;
	}

	/**
	 * @param configuration DAGChaseConfiguration<S>
	 */
	private void setBestConfiguration(DAGChaseConfiguration<S> configuration) {
		if (this.best == null
				|| (this.best != null && configuration != null
				&& this.best.getPlan().getCost().greaterThan(configuration.getPlan().getCost()))) {
			this.best = configuration;
		}
	}
}
