package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Objects;

import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Joiner;
import com.google.common.collect.Lists;

/**
 * Blocking chasers organise the facts into trees of bags.
 * This class provides a bag implementation. A bag can be either blocked or non-blocked.
 * Only non-blocked bags are expanded during chasing. Expansion consists of, first, 
 * finding an homomorphism of the left-hand side of an input dependency to a set of facts inside a bag and, second,
 * grounding the corresponding dependency based on the facts found and creating a new child bag using the newly created facts
 * (the consequence ones).
 *
 * @author Efthymia Tsamoura
 *
 */
public class Bag {

	/** Status of a Bag. */
	public static enum BagStatus { BLOCKED, NONBLOCKED }

	/** The status of this bag. */
	protected BagStatus type = BagStatus.NONBLOCKED;

	/** Bag's id. */
	protected int id;

	/** Global counter for assigning unique bag IDs. */
	protected static int globalId = 0;

	/** The facts of the bag*/
	protected final Collection<PredicateFormula> facts;

	/** The chase constants of the bag */
	protected final Collection<Term> constants;

	/**
	 *
	 * @param facts
	 * 		Input facts
	 * Creates a new unblocked bag using the input facts
	 */
	public Bag(Collection<PredicateFormula> facts) {
		this.facts = new LinkedHashSet<>(facts);
		this.id = globalId++;
		this.type = BagStatus.NONBLOCKED;
		this.constants = Utility.getTerms(facts);
	}

	/**
	 * Copy constructor
	 * @param bag
	 */
	public Bag(Bag bag) {
		this.facts = new LinkedHashSet<>(bag.facts);
		this.id = bag.id;
		this.type = bag.type;
		this.constants = Utility.getTerms(this.facts);
	}

	/**
	 * @return Collection<PredicateFormula>
	 */
	public Collection<PredicateFormula> getFacts() {
		return this.facts;
	}

	/**
	 * @return int
	 */
	public int getId() {
		return this.id;
	}

	/**
	 * @return BagStatus
	 */
	public BagStatus getType() {
		return this.type;
	}

	/**
	 * @param type BagStatus
	 */
	public void setType(BagStatus type) {
		this.type = type;
	}

	/**
	 * @return Collection<Term>
	 */
	public Collection<Term> getConstants() {
		return this.constants;
	}

	/**
	 * Copies the facts of the input bag which have constants also appearing in this bag
	 * @param bag
	 * 		Input bag
	 * @return
	 * 		The facts that have being copied to this bag
	 */
	public Collection<BagBoundPredicate> copyFacts(Bag bag) {
		Collection<BagBoundPredicate> copiedFacts = new LinkedHashSet<>();
		List<Term> c = Lists.newArrayList(this.constants);
		c.retainAll(bag.getConstants());
		if(!c.isEmpty()) {
			for(PredicateFormula fact:bag.getFacts()) {
				if(c.containsAll(fact.getTerms())) {
					this.addFacts(fact);
					copiedFacts.add(new BagBoundPredicate(fact, this.id));
				}
			}
		}
		return copiedFacts;
	}

	/**
	 * Function called during fact propagation
	 * (a fact is being propagated to a bag if the latter contains all of the terms of this fact)
	 * @param constants
	 * 		Input constants
	 * @return
	 * 		true if all of the input constants appear of the current bag.
	 */
	public boolean containsTerms(Collection<Term> constants) {
		return this.constants.containsAll(constants);
	}

	/**
	 *
	 * @param fact
	 * 		fact to be added to the bag
	 */
	public void addFacts(PredicateFormula fact) {
		this.facts.add(fact);
	}

	/**
	 * @param facts Collection<PredicateFormula>
	 */
	public void addFacts(Collection<PredicateFormula> facts) {
		this.facts.addAll(facts);
	}

	/**
	 * @return String
	 */
	@Override
	public String toString() {
		return '{' + Joiner.on(",").join(this.facts) + ',' + this.type.toString() + ',' + new Integer(this.id) + '}';
	}

	/**
	 * Method equals.
	 * @param o Object
	 * @return boolean
	 */
	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null) {
			return false;
		}
		return this.getClass().isInstance(o)
				&& this.id == ((Bag) o).id;
	}

	/**
	 * @return int
	 */
	@Override
	public int hashCode() {
		return Objects.hash(this.id);
	}

	/**
	 * @return an exact shallow copy of this object
	 */
	@Override
	public Bag clone() {
		return new Bag(this);
	}

	/**
	 * @return a shallow copy of this object with a fresh identifier
	 */
	public Bag replicate() {
		Bag replica = new Bag(this);
		replica.id = globalId++;
		return replica;
	}

}
