package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.db.Schema;
import uk.ac.ox.cs.pdq.fol.AcyclicQuery;
import uk.ac.ox.cs.pdq.fol.ConjunctiveQuery;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;

import com.google.common.base.Preconditions;


/**
 * Creates bags based on the input arguments
 *
 * @author Efthymia Tsamoura
 */
public final class BagFactory {

	protected static Logger log = Logger.getLogger(BagFactory.class);

	/** Types of the bags */
	public static enum BagTypes {
		/** Corresponds to a uk.ac.ox.cs.pdq.chase.structures.Bag object   */
		BAG,
		/** Corresponds to a uk.ac.ox.cs.pdq.chase.structures.ExtendedBag object */
		EXTENDEDBAG
	}

	/** The type of the next Bag the factory will create. */
	private BagTypes type = BagTypes.BAG;

	/** Input query*/
	private final Query<?> query;

	/** Accessible sub-queries of the input query*/
	private List<Query<?>> queries;

	/** The input schema */
	private final Schema schema;

	/** The accessible schema */
	private final AccessibleSchema accessibleSchema;


	/**
	 * Constructor for BagFactory.
	 * @param type BagTypes
	 * @param schema Schema
	 * @param accessibleSchema AccessibleSchema
	 * @param query Query<?>
	 */
	public BagFactory(BagTypes type, Schema schema, AccessibleSchema accessibleSchema, Query<?> query) {
		this(type, schema, accessibleSchema, query, createImportantQueries(query, accessibleSchema));
	}

	/**
	 * Constructor for BagFactory.
	 * @param type BagTypes
	 * @param schema Schema
	 * @param accessibleSchema AccessibleSchema
	 * @param query Query<?>
	 * @param queries List<Query<?>>
	 */
	private BagFactory(BagTypes type, Schema schema, AccessibleSchema accessibleSchema, Query<?> query, List<Query<?>> queries) {
		Preconditions.checkNotNull(schema);
		Preconditions.checkNotNull(accessibleSchema);
		Preconditions.checkNotNull(query);
		this.type = type;
		this.schema = schema;
		this.accessibleSchema = accessibleSchema;
		this.query = query;
		this.queries = queries;
	}

	/**
	 * @return BagFactory
	 */
	@Override
	public BagFactory clone() {
		return new BagFactory(this.type, this.schema, this.accessibleSchema, this.query, this.queries);
	}

	/**
	 *
	 * @param facts
	 * @return
	 * 		a bag initialised with the input set of atoms
	 */
	public Bag createBag(Collection<PredicateFormula> facts) {
		if (this.type == BagTypes.BAG) {
			return new Bag(facts);
		} else if (this.type == BagTypes.EXTENDEDBAG) {
			return new ExtendedBag(
					facts, this.schema.getDependencies(),
					this.accessibleSchema.getShortInferredAccessibilityAxioms(),
					this.queries);
		}
		throw new java.lang.IllegalStateException("Bag Factory type not among expected ones.");
	}


	/**
	 * Creates the accessible sub-queries that must be checked during blocking.
	 * If the input query is an acyclic one, then the sub-queries to be checked are the prefixes of the input query.
	 * Otherwise, every sub-query for every combination of free and existential variable must be checked
	 * @param query Query<?>
	 * @param accSchema AccessibleSchema
	 * @return List<Query<?>>
	 */
	private static List<Query<?>> createImportantQueries(Query<?> query, AccessibleSchema accSchema) {
		log.debug("Creating important sub-queries...");
		List<Query<?>> result = new LinkedList<>();
		if (query instanceof AcyclicQuery) {
			for (Query<?> q: ((AcyclicQuery) query).getSuffixQueries()) {
				result.add(accSchema.accessible(q));
			}
		}
		else if (query instanceof ConjunctiveQuery) {
			for (Query<?> q: query.getImportantSubqueries()) {
				result.add(accSchema.accessible(q));
			}
		}
		log.debug("Important sub-queries created. (" + result.size() + ")");
		return result;
	}
}
