package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import org.jgrapht.EdgeFactory;
import org.jgrapht.graph.DefaultEdge;

import com.google.common.base.Preconditions;

import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.util.IndexedDirectedGraph;
import uk.ac.ox.cs.pdq.util.TreeGraph;

/**
 * A tree of bags
 *
 * @author Efthymia Tsamoura
 *
 * @param <N>
 */

public class BagsTree<N extends Bag> extends IndexedDirectedGraph<N> implements TreeGraph<N>{

	/** The root of the tree*/
	private N root = null;

	/** Maps each tree node to its ID*/
	private Map<Integer, N> bags = new LinkedHashMap<>();

	/**
	 * Constructor for BagsTree.
	 * @param cl Class<DefaultEdge>
	 */
	public BagsTree(Class<DefaultEdge> cl) {
		super(cl);
	}

	/**
	 * Constructor for BagsTree.
	 * @param factory EdgeFactory<N,DefaultEdge>
	 */
	public BagsTree(EdgeFactory<N, DefaultEdge> factory) {
		super(factory);
	}

	/**
	 * @param n N
	 * @return boolean
	 */
	@Override
	public boolean addVertex(N n) {
		boolean added = super.addVertex(n);
		if (added) {
			this.bags.put(n.getId(), n);
			if(this.root == null) {
				this.root = n;
			}
		}
		return added;
	}

	/**
	 * @param fact
	 * @return a bag (randomly chosen) that contains the input fact
	 */
	public N getBag(PredicateFormula fact) {
		for (N bag:this.vertexSet()) {
			if(bag.getFacts().contains(fact)) {
				return bag;
			}
		}
		return null;
	}

	/**
	 * @param id Integer
	 * @return N
	 */
	@Override
	public N getVertex(Integer id) {
		return this.bags.get(id);
	}

	/**
	 * @return N
	 * @see uk.ac.ox.cs.pdq.planner.linear.node.TreeGraph#getRoot()
	 */
	@Override
	public N getRoot() {
		return this.root;
	}

	/**
	 * @return the facts of the bags
	 */
	public Collection<PredicateFormula> getFacts() {
		Collection<PredicateFormula> facts = new HashSet<>();
		for (N bag:this.vertexSet()) {
			for (PredicateFormula fact:bag.getFacts()) {
				facts.add(fact);
			}
		}
		return facts;
	}

	/**
	 * @return BagsTree<N>
	 */
	@Override
	public BagsTree<N> clone() {
		BagsTree<N> tree = new BagsTree<>(this.getEdgeFactory());

		Set<N> bags = this.vertexSet();
		Map<N,N> map =  new LinkedHashMap<>();

		for (N bag: bags) {
			N clone = (N) bag.clone();
			tree.addVertex(clone);
			map.put(bag, clone);
		}

		for (DefaultEdge edge: this.edgeSet()) {
			N source = this.getEdgeSource(edge);
			N target = this.getEdgeTarget(edge);
			N mySource = map.get(source);
			N myTarget = map.get(target);
			tree.addEdge(mySource, myTarget, this.getEdgeFactory().createEdge(mySource, myTarget));
		}
		return tree;
	}

	/**
	 * @return a tree with the same facts within a bag, but with different bag IDs
	 */
	public BagsTree<N> replicate() {
		BagsTree<N> tree = new BagsTree<>(this.getEdgeFactory());

		Set<N> bags = this.vertexSet();
		Map<N,N> map =  new LinkedHashMap<>();

		for (N bag: bags) {
			N clone = (N) bag.replicate();
			tree.addVertex(clone);
			map.put(bag, clone);
		}

		for (DefaultEdge edge: this.edgeSet()) {
			N source = this.getEdgeSource(edge);
			N target = this.getEdgeTarget(edge);
			N mySource = map.get(source);
			N myTarget = map.get(target);
			tree.addEdge(mySource, myTarget, this.getEdgeFactory().createEdge(mySource, myTarget));
		}
		return tree;
	}

	@Override
	public N getParent(N child) {
		Preconditions.checkArgument(this.vertexSet().contains(child));
		Set<DefaultEdge> edges = this.incomingEdgesOf(child);
		if(edges.isEmpty()) {
			return null;
		}
		else {
			return this.getEdgeSource(edges.iterator().next());
		}
	}
}
