package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import static uk.ac.ox.cs.pdq.logging.performance.StatKeys.GENERATED_FACTS;
import static uk.ac.ox.cs.pdq.logging.performance.StatKeys.MILLI_BLOCKING_CHECK;
import static uk.ac.ox.cs.pdq.logging.performance.StatKeys.MILLI_UPDATE;
import static uk.ac.ox.cs.pdq.logging.performance.StatKeys.MILLI_UPDATE_QUERY_DEPENDENCIES;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.db.Schema;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.Formula;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.planner.PlannerException;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismDetector;
import uk.ac.ox.cs.pdq.planner.logging.performance.ConstantsStatistics;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Bag.BagStatus;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.DatabaseTreeState;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.StateFactory;
import uk.ac.ox.cs.pdq.reasoning.Match;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.google.common.eventbus.EventBus;

/**
 * Blocking chase algorithm.
 * The facts that are being generated after each chase step are organised
 * into a tree of bags.
 *
 * @author Efthymia Tsamoura
 *
 */
public class BlockingChaser extends Chaser<DatabaseTreeState> {
	
	/** Detects which bags are blocked */
	protected final BlockingDetector blockingDetector;

	/** Produces new bags*/
	protected final BagFactory bagFactory;

	/** Defines how often will we perform blocking detection */
	protected final Integer blockingInterval;

	/**
	 * Constructor for BlockingChaser.
	 * @param eventBus EventBus
	 * @param collectStats boolean
	 * @param canonicalNames Boolean
	 * @param bagFactory BagFactory
	 * @param blockingDetector BlockingDetector
	 * @param blockingInterval Integer
	 */
	public BlockingChaser(
			EventBus eventBus, 
			boolean collectStats,
			Boolean canonicalNames,
			BagFactory bagFactory,
			BlockingDetector blockingDetector,
			Integer blockingInterval) {
		super(eventBus, collectStats, canonicalNames);
		Preconditions.checkNotNull(blockingDetector);
		Preconditions.checkNotNull(blockingInterval);
		Preconditions.checkNotNull(bagFactory);
		this.blockingDetector = blockingDetector;
		this.blockingInterval = blockingInterval;
		this.bagFactory = bagFactory;
	}

	/**
	 * @param detector HomomorphismDetector
	 * @param facts1 Collection<PredicateFormula>
	 * @param schema Schema
	 * @return DatabaseTreeState
	 */
	@Override
	public DatabaseTreeState initialize(
			HomomorphismDetector detector,
			Query<?> query,
			Schema schema) {
		DatabaseTreeState result = StateFactory.getInstance(this.bagFactory, detector, this.createInitialisationFormula(query, schema));
		this.chase(result, schema.getDependencies());
		return result;
	}

	/**
	 * @param state Input state
	 * @return true if all bags are being blocked
	 */
	protected boolean terminates(DatabaseTreeState state) {
		for (Bag b: state.getTree().vertexSet()) {
			if (b.getType() == BagStatus.NONBLOCKED) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Signals for blocking detection.
	 * It is called every this.blockingInterval steps
	 * @param state DatabaseTreeState
	 */
	protected void blockingCheck(DatabaseTreeState state) {
		this.stats.start(MILLI_BLOCKING_CHECK);
		this.blockingDetector.blockingCheck(state.getTree().vertexSet(), state.initialState());
		this.stats.stop(MILLI_BLOCKING_CHECK);
	}

	/**
	 * @param state DatabaseTreeState
	 * @param dependencies Collection<? extends IC>
	 */
	@Override
	public void chase(DatabaseTreeState state, Collection<? extends Constraint> dependencies) {
		state.updateDependenciesOrQueries();
		int rounds = 0;
		//True if at the end of the internal for loop at least one dependency has been fired
		boolean appliedStep = true;
		while (appliedStep && !this.terminates(state)) {
			appliedStep = false;
			//Find for each input dependency, the matches of its left-hand side to the facts of the input state
			Map<? extends Constraint, List<Match>> homomorphisms = state.getHomomorphisms(dependencies);
			for (Constraint dependency: homomorphisms.keySet()) {
				List<Match> matches = homomorphisms.get(dependency);
				if (matches != null) {
					for (Match match: matches) {
						//For each match, get the mapping of the dependency's variables to chase constants and
						//try to fire the corresponding dependency
						Map<Variable, Constant> mapping = match.getMapping();
						Constraint groundedFormula = dependency.fire(mapping, this.canonicalNames);
						this.eventBus.post(new ConstantsStatistics.Increment(groundedFormula.getRight().getTerms()));
						log.trace("Checking rule " + groundedFormula);
						//If the corresponding dependency has not being already fired, fire it and perform blocking detection
						if(!state.isFired(dependency, Sets.newHashSet(groundedFormula.getLeft().getPredicates()))){
							this.applyStep(state, match);
							appliedStep = true;
							++rounds;
							if (rounds % this.blockingInterval == 0) {
								this.stats.start(MILLI_UPDATE_QUERY_DEPENDENCIES);
								state.updateDependenciesOrQueries();
								state.resetBagsToUpdate();
								this.stats.stop(MILLI_UPDATE_QUERY_DEPENDENCIES);
								this.blockingCheck(state);
							}
						}
					}
				}
			}
		}
	}


	/**
	 * Updates the input state given the input match of dependency conjuncts to facts
	 * @param state Input state
	 * @param match Input match
	 * @return
	 * 		true if the update was successful
	 */
	public Boolean applyStep(DatabaseTreeState state, Match match) {
		Preconditions.checkArgument(match != null);
		return this.applySteps(state, Lists.newArrayList(match));
	}

	/**
	 * Updates the input state given the input matches of dependency conjuncts to facts.
	 * Each match maintains information w.r.t. the dependency that will be fired and
	 * the bag that stores the facts to fire the dependency.
	 * If the firing of an input dependency results in the creation of new chase constants, then
	 * the newly created facts are both stored in a new bag and propagated upwards the tree;
	 * otherwise, the newly created facts are only propagated upwards the tree.
	 *
	 * @param state DatabaseTreeState
	 * @param matches List<Matching>
	 * @return Boolean
	 * @throws PlannerException
	 */
	@Override
	public Boolean applySteps(DatabaseTreeState state, List<Match> matches) {
		Preconditions.checkArgument(matches != null);
		Collection<PredicateFormula> facts = new LinkedList<>();
		Collection<ExtendedBag> bagsToUpdate = new LinkedHashSet<>();
		for (Match match: matches) {
			Preconditions.checkState(match instanceof BagMatch);
			//For each match, get
			//(i) the corresponding dependency,
			//(ii) the mapping of dependency's variables to chase constants and
			//(iii) the bag where the match is found
			//and update the input state
			Constraint dependency = (Constraint) match.getQuery();
			Map<Variable, Constant> mapping = match.getMapping();
			Constraint groundedFormula = dependency.fire(mapping, this.canonicalNames);
			this.eventBus.post(new ConstantsStatistics.Increment(groundedFormula.getRight().getTerms()));
			log.debug("Fire rule " + groundedFormula + " on bag " + ((BagMatch) match).getBag());
			Formula parentFormula = groundedFormula.getLeft();
			Formula childFormula = groundedFormula.getRight();
			Bag parentBag = state.getTree().getVertex(((BagMatch) match).getBag());
			Preconditions.checkState(parentBag instanceof ExtendedBag);
			//Find to which bag each newly created fact should be propagated
			Collection<BagBoundPredicate> propagatedFacts = state.update((ExtendedBag) parentBag, dependency, parentFormula, childFormula);
			for(BagBoundPredicate propagatedFact:propagatedFacts) {
				bagsToUpdate.add(state.getTree().getVertex(propagatedFact.getBag()));
			}
			facts.addAll(propagatedFacts);

			log.trace("Updated facts: " + childFormula.getPredicates().size());
			this.stats.increase(GENERATED_FACTS, childFormula.getPredicates().size());
		}
		this.stats.start(MILLI_UPDATE);
		this.eventBus.post(facts);
		state.addFacts(facts);
		this.stats.stop(MILLI_UPDATE);
		state.setBagsToUpdate(bagsToUpdate);
		log.trace("Bags to update: " + bagsToUpdate.size());
		return true;
	}
}
