package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.List;

import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.planner.db.access.InferredAccessibleAxiom;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Bag.BagStatus;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;

/**
 * Blocking implementation which reconsiders the status of a bag every time new
 * facts are being propagated upwards the tree.
 * A bag B' blocks another bag B if
 *  -in the initial configuration, B' satisfies as many dependencies (particularly, the
 *   left-hand sides of these dependencies) as B does
 *  -in the next configurations, B' satisfies as many inferred accessible dependencies as B does
 *  -plus as many inferred accessible sub-queries of the original query as B does.
 *
 * The bags are expected to be of type ExtendedBag; otherwise, a runtime exception is thrown
 *
 * @author Efthymia Tsamoura
 *
 */
public class BlockingDetector {

	protected static Logger log = Logger.getLogger(BlockingDetector.class);

	/**
	 *
	 * @param bags
	 * 	 	Input bags.
	 * 	 	If a bag is being blocked then its status is set to BLOCKED; otherwise it is set to NONBLOCKED.
	 * 	 	This function is called during chasing (either every time we fire a dependency at specific time intervals).
	 * @param inInitialState
	 * 		True if we are in the initial configuration
	 */
	public void blockingCheck(Collection<? extends Bag> bags, boolean inInitialState) {
		List<? extends Bag> b = Lists.newArrayList(bags);
		for (int i = 0; i < b.size() - 1; ++i) {
			for (int j = i + 1; j < b.size(); ++j) {
				this.isHomomorphicToPreservesQueries(b.get(i), b.get(j), inInitialState);
			}
		}
	}

	/**
	 *
	 * @param parent
	 * 		Input bag
	 * @param child
	 * 		Input bag
	 * @param inInitialState
	 * 		True if we are in the initial configuration
	 */
	private void isHomomorphicToPreservesQueries(Bag parent, Bag child, boolean inInitialState) {
		Preconditions.checkArgument(parent instanceof ExtendedBag);
		Preconditions.checkArgument(child instanceof ExtendedBag);

		//If we are in the initial configuration
		if (inInitialState) {
			Collection<Constraint> parentICs = ((ExtendedBag) parent).getSatisfiedDependencies();
			Collection<Constraint> childICs = ((ExtendedBag) child).getSatisfiedDependencies();
			//Find the schema dependencies that are satisfied in the parent and child bags, respectively.
			if (parentICs.isEmpty() && childICs.isEmpty()) {
			} else {
				//If both bags satisfy the same dependencies, then both of them are blocked
				if (parentICs.equals(childICs)) {
					parent.setType(BagStatus.BLOCKED);
					child.setType(BagStatus.BLOCKED);
					//If the parent bag satisfies any dependency that the child bag also satisfies, then the child is blocked
				} else if (parentICs.containsAll(childICs)) {
					child.setType(BagStatus.BLOCKED);
					//If the child bag satisfies any dependency that the parent bag also satisfies, then the parent is blocked
				} else if (childICs.containsAll(parentICs)) {
					parent.setType(BagStatus.BLOCKED);
				}

			}
		} else {
			//Same as above; we do consider the accessible sub-queries though.
			Collection<InferredAccessibleAxiom> parentICs = ((ExtendedBag) parent).getSatisfiedAccessibleDependencies();
			Collection<InferredAccessibleAxiom> childICs = ((ExtendedBag) child).getSatisfiedAccessibleDependencies();
			Collection<Query<?>> parentQueries = ((ExtendedBag) parent).getSatisfiedQueries();
			Collection<Query<?>> childQueries = ((ExtendedBag) child).getSatisfiedQueries();
			if (!(parentICs.isEmpty() && childICs.isEmpty()
					&& parentQueries.isEmpty() && childQueries.isEmpty())) {
				if (parentICs.equals(childICs) && parentQueries.equals(childQueries)) {
					parent.setType(BagStatus.BLOCKED);
					child.setType(BagStatus.BLOCKED);
				} else if (parentICs.containsAll(childICs) && parentQueries.containsAll(childQueries)) {
					child.setType(BagStatus.BLOCKED);
				} else if (childICs.containsAll(parentICs) && childQueries.containsAll(parentQueries)) {
					parent.setType(BagStatus.BLOCKED);
				}
			}
		}
	}
}
