package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.db.Schema;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismDetector;
import uk.ac.ox.cs.pdq.planner.logging.performance.ConstantsStatistics;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.DatabaseListState;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.StateFactory;
import uk.ac.ox.cs.pdq.reasoning.Match;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Preconditions;
import com.google.common.base.Supplier;
import com.google.common.eventbus.EventBus;


/**
 * A bounded chase algorithm. According to this,
 * we find the maximum number of chase rounds k during chasing the input schema
 * and then perform only k rounds of chasing during chasing the accessible schema.
 *
 * @author Efthymia Tsamoura
 *
 */
public class BoundedChaser extends RestrictedChaser {

	/** If true, the initialisation phase is not bounded by k. */
	private final boolean fullInitialization;

	/** Factor of number of rounds to chase before stopping. */
	private final KSupplier k;

	/** The number of rounds performed during the last call of chase. */
	private Integer lastRound;


	/**
	 * Constructor for BoundedChaser.
	 * @param eventBus EventBus
	 * @param collectStats boolean
	 * @param canonicalNames Boolean
	 * @param k KSupplier
	 * @param fullInitialize boolean
	 */
	public BoundedChaser(
			EventBus eventBus, boolean collectStats,
			Boolean canonicalNames, KSupplier k, boolean fullInitialize) {
		super(eventBus, collectStats, canonicalNames);
		Preconditions.checkArgument(k != null);
		this.k = k;
		this.fullInitialization = fullInitialize;
	}

	/**
	 * @param detector HomomorphismDetector
	 * @param facts Collection<PredicateFormula>
	 * @param schema Schema
	 * @return DatabaseListState
	 */
	@Override
	public DatabaseListState initialize(
			HomomorphismDetector detector,
			Query<?> query, 
			Schema schema) {
		DatabaseListState result = StateFactory.getInstance(detector, this.createInitialisationFormula(query, schema));
		synchronized (this.k) {
			int oldK = this.k.get();
			if (this.fullInitialization) {
				this.k.set(Integer.MAX_VALUE);
			}
			this.chase(result, schema.getDependencies());
			this.k.set(Math.min(oldK, this.lastRound));
			this.k.freeze();
		}
		return result;
	}

	/**
	 * @param state DatabaseListState
	 * @param dependencies Collection<? extends IC>
	 */
	@Override
	public void chase(DatabaseListState state, Collection<? extends Constraint> dependencies) {
		int rounds = 0;
		boolean appliedStep = true;
		while (rounds < this.k.get() && appliedStep) {
			appliedStep = false;
			Map<? extends Constraint, List<Match>> homomorphisms = state.getHomomorphisms(dependencies);
			for (Constraint dependency: homomorphisms.keySet()) {
				List<Match> matches = homomorphisms.get(dependency);
				if (matches != null) {
					for (Match match: matches) {
						Map<Variable, Constant> mapping = match.getMapping();
						Constraint groundedFormula = dependency.fire(mapping, this.canonicalNames);
						this.eventBus.post(new ConstantsStatistics.Increment(groundedFormula.getRight().getTerms()));
						log.trace("Checking rule " + groundedFormula);
						// TODO: push the restricted chase condition as an
						// additional constraint in getHomomorphism().
						Map<Variable, Constant> inputMapping = Utility.retain(mapping, dependency.getBothSideVariables());
						if(!this.isApplied(state, dependency, inputMapping)){
							this.applyStep(state, match);
							appliedStep = true;
						}
					}
				}
			}
			if(appliedStep) {
				++rounds;
			}
		}
		this.lastRound = rounds;
	}

	/**
	 * Supplies the actual termination number to the bounded chase.
	 * This supplier allows delaying the determination of K, until after the
	 * Initialisation phase, and share the K value across distinct instances
	 * of BoundedChaser.
	 *
	 * @author Julien Leblay
	 */
	public static class KSupplier implements Supplier<Integer> {

		/** If true, any future modification of the supplied value will throw an exception */
		private boolean frozen = false;

		/** The supplied integer */
		private Integer k = null;

		/**
		 * Constructor for KSupplier.
		 * @param k int
		 */
		public KSupplier(int k) {
			Preconditions.checkArgument(k >= 0);
			this.k = k;
		}

		/**
		 * Forbids any further use of set().
		 */
		public void freeze() {
			this.frozen = true;
		}

		/**
		 * @param k Integer
		 */
		public void set(Integer k) {
			Preconditions.checkState(!this.frozen, "Cannot set a frozen KSupplier");
			Preconditions.checkArgument(k != null);
			this.k = k;
		}

		/**
		 * @return Integer
		 * @see com.google.common.base.Supplier#get()
		 */
		@Override
		public Integer get() {
			Preconditions.checkNotNull(this.k);
			return this.k;
		}
	}
}
