package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.List;

import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.db.Schema;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.logging.performance.StatisticsCollector;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessibleRelation;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismDetector;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;
import uk.ac.ox.cs.pdq.reasoning.Match;
import uk.ac.ox.cs.pdq.reasoning.Reasoner;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.eventbus.EventBus;

/**
 *
 * The chase algorithm
 *
 * @author Efthymia Tsamoura
 */
public abstract class Chaser<S extends ChaseState> implements Reasoner {

	protected static Logger log = Logger.getLogger(Chaser.class);

	/** True if Skolem terms are assigned to existentially quantified variables*/
	protected final Boolean canonicalNames;

	protected final EventBus eventBus;

	protected final StatisticsCollector stats;

	/**
	 * Constructor for Chaser.
	 * @param eventBus EventBus
	 * @param collectStats boolean
	 * @param canonicalNames Boolean
	 */
	public Chaser(EventBus eventBus, boolean collectStats, boolean canonicalNames) {
		Preconditions.checkNotNull(eventBus);
		this.eventBus = eventBus;
		this.stats = new StatisticsCollector(collectStats, eventBus);
		this.canonicalNames = canonicalNames;
	}

	/**
	 * Runs the chase algorithm given the input query and the schema's dependencies.
	 * @param detector Detects homomorphisms
	 * @param query 
	 * @param schema
	 * @return the initialised state
	 */
	public abstract S initialize(HomomorphismDetector detector, Query<?> query, Schema schema);
	

	/**
	 * 
	 * @param query
	 * @param schema
	 * @return the facts of the canonical query and the Accessible facts of the schema constants
	 */
	protected Collection<PredicateFormula> createInitialisationFormula(Query<?> query, Schema schema) {
		// Gets the canonical database of the query
		Collection<PredicateFormula> initFormula = query.getCanonicalQuery().getPredicates();
		// Create the Accessible(.) facts
		// One Accessible(.) is being created for every schema constant
		for (TypedConstant<?> constant : query.getSchemaConstants()) {
			initFormula.add(AccessibleRelation.getAccessibleFact(constant));
		}
		for (TypedConstant<?> constant : schema.getDependencyConstants()) {
			initFormula.add(AccessibleRelation.getAccessibleFact(constant));
		}
		log.debug("Canonical query: " + initFormula);
		return initFormula;
	}

	/**
	 * Runs the chase algorithm given the input state and the input set of dependencies
	 * @param state
	 * 		Input state
	 * @param dependencies
	 * 		Input dependencies
	 */
	public abstract void chase(S state, Collection<? extends Constraint> dependencies);

	/**
	 * Updates the input state given the input matches of dependency conjuncts to facts
	 * @param state Input state
	 * @param matches Input matches
	 * @return true if the update was successful
	 */
	public abstract Boolean applySteps(S state, List<Match> matches);

	/**
	 * @param s State
	 * @param dependencies Iterable<? extends IC>
	 * @see uk.ac.ox.cs.pdq.Reasoner#fixpoint(State, Iterable<? extends IC>)
	 */
	@Override
	public void fixpoint(State s, Iterable<? extends Constraint> dependencies) {
		Preconditions.checkArgument(s instanceof ChaseState);
		this.chase((S) s, Lists.newLinkedList(dependencies));
	}
}
