package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.planner.db.access.InferredAccessibleAxiom;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismConstraint;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismDetector;
import uk.ac.ox.cs.pdq.reasoning.Match;

import com.google.common.collect.Maps;
import com.google.common.collect.Sets;

/**
 * Extends Bag by maintaining the list of sub-queries and dependencies satisfied in the current bag. 
 * This implementation uses a database to detect homomorphisms.
 *
 * @author Efthymia Tsamoura
 *
 */
public class ExtendedBag extends Bag {

	/**
	 * A map of schema dependencies to boolean values. If the corresponding boolean value
	 * is true, then the left-hand side of the corresponding dependency is satisfied by the facts of this.bag
	 */
	private final Map<Constraint, Boolean> dependencies;

	/**
	 * A map of inferred accessible dependencies to boolean values
	 * with semantics similar to this.dependencies
	 */
	private final Map<InferredAccessibleAxiom, Boolean> accessibleDependencies;

	/**
	 * A map of the accessible sub-queries to boolean values with
	 * semantics similar to this.dependencies
	 */
	private final Map<Query<?>, Boolean> queries;


	/** True if we are updating the satisfied ICs for the first time */
	private Boolean initialConfiguration = true;

	/**
	 * True if we are updating the satisfied sub-queries and the inferred
	 * accessible dependencies for the first time
	 */
	private Boolean nextConfigurations = true;


	/**
	 * Constructor for ExtendedBag.
	 * @param atoms Collection<PredicateFormula>
	 * @param dep Collection<IC>
	 * @param accessibleDep Collection<InferredAccessibleAxiom>
	 * @param queries Collection<Query<?>>
	 */
	public ExtendedBag(
			Collection<PredicateFormula> atoms, Collection<Constraint> dep,
			Collection<InferredAccessibleAxiom> accessibleDep, Collection<Query<?>> queries) {
		super(atoms);

		this.dependencies = new LinkedHashMap<>();
		this.accessibleDependencies = new LinkedHashMap<>();
		this.queries = new LinkedHashMap<>();

		for (Constraint ic: dep) {
			this.dependencies.put(ic, false);
		}

		for (InferredAccessibleAxiom ic: accessibleDep) {
			this.accessibleDependencies.put(ic, false);
		}

		for (Query<?> query: queries) {
			this.queries.put(query, false);
		}
	}

	/**
	 * Constructor for ExtendedBag.
	 * @param bag ExtendedBag
	 */
	public ExtendedBag(ExtendedBag bag) {
		super(bag);
		this.dependencies = Maps.newLinkedHashMap(bag.dependencies);
		this.accessibleDependencies = Maps.newLinkedHashMap(bag.accessibleDependencies);
		this.queries = Maps.newLinkedHashMap(bag.queries);
		this.initialConfiguration = bag.initialConfiguration;
		this.nextConfigurations = bag.nextConfigurations;
	}

	/**
	 * @param s ExtendedBag
	 * @param t ExtendedBag
	 * @param detector HomomorphismDetector
	 * @param update boolean
	 * @return ExtendedBag
	 */
	public static ExtendedBag mergeBags(ExtendedBag s, ExtendedBag t, HomomorphismDetector detector, boolean update) {
		Collection<PredicateFormula> atoms =  Sets.newLinkedHashSet(s.getFacts());
		atoms.addAll(t.getFacts());
		ExtendedBag bag = new ExtendedBag(atoms, s.getDependencies(), s.getAccessibleDependencies(), s.getQueries());
		if(update) {
			bag.updateSatisfiedAccessibleDependencies(detector);
			bag.updateSatisfiedQueries(detector);
		}
		bag.initialConfiguration = false;
		bag.nextConfigurations = true;
		return bag;
	}

	/**
	 * It iterates over the dependencies. If a dependency's left-hand side conjunction is not
	 * already satisfied, then we search for an homomorphism of the latter to
	 * the current bag's facts.
	 * @param detector HomomorphismDetector
	 */
	public void updateSatisfiedDependencies(HomomorphismDetector detector) {
		Iterator<Entry<Constraint, Boolean>> it = this.dependencies.entrySet().iterator();
		while (it.hasNext()) {
			Entry<Constraint, Boolean> entry = it.next();
			if (!entry.getValue()) {
				List<Match> matchings = detector.getMatches(
						entry.getKey(),
						HomomorphismConstraint.topK(1),
						HomomorphismConstraint.scope(Conjunction.of(this.facts)),
						HomomorphismConstraint.scope(true, this));
				this.dependencies.put(entry.getKey(), !matchings.isEmpty());
			}
		}
		this.initialConfiguration = false;
	}

	/**
	 * It iterates over the map of inferred accessible dependencies. If a dependency's left-hand side conjunction
	 * is not already satisfied, then we search for an homomorphism of the
	 * latter to the current bag's facts.
	 * @param detector HomomorphismDetector
	 */
	public void updateSatisfiedAccessibleDependencies(HomomorphismDetector detector) {
		Iterator<Entry<InferredAccessibleAxiom, Boolean>> it = this.accessibleDependencies.entrySet().iterator();
		while (it.hasNext()) {
			Entry<InferredAccessibleAxiom, Boolean> entry = it.next();
			if (!entry.getValue()) {
				List<Match> matchings = detector.getMatches(
						entry.getKey(),
						HomomorphismConstraint.topK(1),
						HomomorphismConstraint.scope(Conjunction.of(this.facts)),
						HomomorphismConstraint.scope(true, this));
				this.accessibleDependencies.put(entry.getKey(), !matchings.isEmpty());
			}
		}
		this.nextConfigurations = false;
	}

	/**
	 * It iterates over the map of the accessible sub-queries. If a sub-query is not
	 * already satisfied, then we search for an homomorphism of the latter to
	 * the current bag's facts.
	 * @param detector HomomorphismDetector
	 */
	public void updateSatisfiedQueries(HomomorphismDetector detector) {
		Iterator<Entry<Query<?>, Boolean>> it = this.queries.entrySet().iterator();
		while (it.hasNext()) {
			Entry<Query<?>, Boolean> entry = it.next();
			if (!entry.getValue()) {
				List<Match> matchings = detector.getMatches(
						entry.getKey(),
						HomomorphismConstraint.topK(1),
						HomomorphismConstraint.scope(Conjunction.of(this.facts)),
						HomomorphismConstraint.scope(true, this));
				this.queries.put(entry.getKey(), !matchings.isEmpty());
			}
		}
		this.nextConfigurations = false;
	}

	/**
	 * @return the collection of satisfied schema dependencies
	 */
	public Collection<Constraint> getSatisfiedDependencies() {
		Collection<Constraint> satisfiedSchemaICs = new LinkedHashSet<>();
		Iterator<Entry<Constraint, Boolean>> it = this.dependencies.entrySet().iterator();
		while (it.hasNext()) {
			Entry<Constraint, Boolean> entry = it.next();
			if (entry.getValue()) {
				satisfiedSchemaICs.add(entry.getKey());
			}
		}
		return satisfiedSchemaICs;
	}

	/**
	 * @return the collection of satisfied inferred accessible dependencies
	 */
	public Collection<InferredAccessibleAxiom> getSatisfiedAccessibleDependencies() {
		Collection<InferredAccessibleAxiom> satisfiedAccessibleSchemaICs = new LinkedHashSet<>();
		Iterator<Entry<InferredAccessibleAxiom, Boolean>> it =
				this.accessibleDependencies.entrySet().iterator();
		while (it.hasNext()) {
			Entry<InferredAccessibleAxiom, Boolean> entry = it.next();

			if (entry.getValue()) {
				satisfiedAccessibleSchemaICs.add(entry.getKey());
			}
		}
		return satisfiedAccessibleSchemaICs;
	}

	/**
	 * @return the collection of satisfied accessible sub-queries
	 */
	public Collection<Query<?>> getSatisfiedQueries() {
		Collection<Query<?>> satisfiedQueries = new LinkedHashSet<>();
		Iterator<Entry<Query<?>, Boolean>> it =
				this.queries.entrySet().iterator();
		while (it.hasNext()) {
			Entry<Query<?>, Boolean> entry = it.next();
			if (entry.getValue()) {
				satisfiedQueries.add(entry.getKey());
			}
		}
		return satisfiedQueries;
	}

	/**
	 * @return Collection<IC>
	 */
	public Collection<Constraint> getDependencies() {
		return this.dependencies.keySet();
	}

	/**
	 * @return Collection<InferredAccessibleAxiom>
	 */
	public Collection<InferredAccessibleAxiom> getAccessibleDependencies() {
		return this.accessibleDependencies.keySet();
	}

	/**
	 * @return Collection<Query<?>>
	 */
	public Collection<Query<?>> getQueries() {
		return this.queries.keySet();
	}

	/**
	 * @return Boolean
	 */
	public Boolean initialConfiguration() {
		return this.initialConfiguration;
	}

	/**
	 * @return Boolean
	 */
	public Boolean nextConfigurations() {
		return this.nextConfigurations;
	}

	/**
	 * @return ExtendedBag
	 */
	@Override
	public ExtendedBag clone() {
		return new ExtendedBag(this);
	}

	/**
	 * @return ExtendedBag
	 */
	@Override
	public ExtendedBag replicate() {
		ExtendedBag replica = new ExtendedBag(this);
		replica.id = globalId++;
		return replica;
	}
}
