package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.planner.logging.performance.ConstantsStatistics;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.DatabaseListState;
import uk.ac.ox.cs.pdq.reasoning.Match;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Preconditions;
import com.google.common.eventbus.EventBus;


/**
 * A non-blocking chase algorithm which runs for kTermination chase steps.
 * The difference between this implementation and the RestrictedChaser one is
 * that the former one performs k chase rounds prior to stopping.
 *
 * This chase implementation is used only for testing purposes
 * (@see KStepBlockingTest in uk.ac.ox.cs.pdq.test.regression)
 *
 * @author Efthymia Tsamoura
 *
 */
public class KTerminationChaser extends RestrictedChaser {

	/** Factor of number of rounds to chase for before stop the chase. */
	private final Integer k;

	/**
	 *
	 * @param eventBus
	 * @param collectStats
	 * @param canonicalNames
	 * @param k int
	 */
	public KTerminationChaser(
			EventBus eventBus, boolean collectStats,
			Boolean canonicalNames, int k) {
		super(eventBus, collectStats, canonicalNames);
		Preconditions.checkArgument(k >= 0);
		this.k = k;
	}

	/**
	 * @param state DatabaseListState
	 * @param dependencies Collection<? extends IC>
	 */
	@Override
	public void chase(DatabaseListState state, Collection<? extends Constraint> dependencies) {
		int rounds = 0;
		boolean appliedStep = true;
		while (rounds < this.k && appliedStep) {
			appliedStep = false;
			Map<? extends Constraint, List<Match>> homomorphisms = state.getHomomorphisms(dependencies);
			for (Constraint dependency: homomorphisms.keySet()) {
				List<Match> matches = homomorphisms.get(dependency);
				if (matches != null) {
					for (Match match: matches) {
						Map<Variable, Constant> mapping = match.getMapping();
						Constraint groundedFormula = dependency.fire(mapping, this.canonicalNames);
						this.eventBus.post(new ConstantsStatistics.Increment(groundedFormula.getRight().getTerms()));
						log.trace("Checking rule " + groundedFormula);
						// TODO: push the restricted chase condition as an
						// additional constraint in getHomomorphism().
						Map<Variable, ? extends Term> inputMapping = Utility.retain(mapping, dependency.getBothSideVariables());
						if(!this.isApplied(state, dependency, inputMapping)){
							this.applyStep(state, match);
							appliedStep = true;
						}
					}
				}
			}
			++rounds;
		}
	}

}
