package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections4.map.MultiValueMap;
import org.apache.commons.lang3.tuple.Pair;
import org.jgrapht.Graph;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.SimpleGraph;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState;

import com.google.common.base.Preconditions;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;

/**
 * Implementation of a FiringGraph
 *
 * @author Efthymia Tsamoura
 *
 * @param <S>
 */

public class MapFiringGraph<S extends ChaseState> implements FiringGraph<S>{

	/**
	 * Keeps for each dependency, the facts that were used to fire it, as well as, the consequence facts
	 */
	private final MultiValueMap<Collection<PredicateFormula>, Collection<PredicateFormula>> map;

	/**
	 * Chase graph
	 */
	private final Graph<PredicateFormula, DefaultEdge> graph;

	/**
	 * Fact history map. Associates each chase fact with the facts and the dependency that were last fired to produce this fact
	 */
	private final Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> provenance;

	/**
	 * Keeps the set of facts that were used to fire each dependency
	 */
	private final Multimap<Collection<PredicateFormula>, Constraint> firings;

	public MapFiringGraph() {
		this.map = new MultiValueMap<>();
		this.provenance = new LinkedHashMap<>();
		this.firings = LinkedHashMultimap.create();
		this.graph = new SimpleGraph<PredicateFormula, DefaultEdge>(DefaultEdge.class);
	}

	/**
	 * Constructor for MapFiringGraph.
	 * @param map MultiValueMap<Collection<PredicateFormula>,Collection<PredicateFormula>>
	 * @param graph Graph<PredicateFormula,DefaultEdge>
	 * @param factProvenance Map<PredicateFormula,Pair<IC,Collection<PredicateFormula>>>
	 * @param firedDependencies Multimap<Collection<PredicateFormula>,IC>
	 */
	private MapFiringGraph(
			MultiValueMap<Collection<PredicateFormula>, Collection<PredicateFormula>> map,
			Graph<PredicateFormula, DefaultEdge> graph,
			Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> factProvenance,
			Multimap<Collection<PredicateFormula>, Constraint> firedDependencies) {
		this.map = map;
		this.graph = graph;
		this.provenance = factProvenance;
		this.firings = firedDependencies;
	}

	/**
	 * @param dependency IC
	 * @param sources Collection<PredicateFormula>
	 * @param targets Collection<PredicateFormula>
	 * @param initialState boolean
	 * @see uk.ac.ox.cs.pdq.chase.FiringGraph#put(IC, Collection<PredicateFormula>, Collection<PredicateFormula>, boolean)
	 */
	@Override
	public void put(Constraint dependency, Collection<PredicateFormula> sources, Collection<PredicateFormula> targets, boolean initialState) {
		for (PredicateFormula fact:targets) {
			if (!this.provenance.containsKey(fact)) {
				this.provenance.put(fact, Pair.of(dependency, sources));
			}
		}
		this.firings.put(sources, dependency);
		if(initialState) {
			this.map.put(sources, targets);
			this.updateGraph(sources, targets);
		}
	}

	/**
	 * @param dependency IC
	 * @param source PredicateFormula
	 * @param target PredicateFormula
	 * @param initialState boolean
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph#put(Constraint, PredicateFormula, PredicateFormula, boolean)
	 */
	@Override
	public void put(Constraint dependency, PredicateFormula source, PredicateFormula target, boolean initialState) {
		this.put(dependency, Sets.newHashSet(source), Sets.newHashSet(target), initialState);
	}

	/**
	 * @return MapFiringGraph<S>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph#clone()
	 */
	@Override
	public MapFiringGraph<S> clone() {
		Multimap<Collection<PredicateFormula>, Constraint> firedDependencies = LinkedHashMultimap.create();
		firedDependencies.putAll(this.firings);
		Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> factProvenance = Maps.newLinkedHashMap(this.getFactProvenance());
		return new MapFiringGraph<S>(this.map, this.graph, factProvenance, firedDependencies);
	}

	/**
	 * @return Graph<PredicateFormula,DefaultEdge>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph#getGraph()
	 */
	@Override
	public Graph<PredicateFormula, DefaultEdge> getGraph() {
		return this.graph;
	}

	/**
	 * @param sources Collection<PredicateFormula>
	 * @param targets Collection<PredicateFormula>
	 */
	private void updateGraph(Collection<PredicateFormula> sources, Collection<PredicateFormula> targets) {
		for(PredicateFormula source:sources) {
			this.graph.addVertex(source);
			for(PredicateFormula target:targets) {
				try {
					this.graph.addVertex(target);
					this.graph.addEdge(source, target); 
				}
				catch(IllegalArgumentException ex) {
					System.out.println();
				}
			}
		}
	}

	/**
	 * @param graph0 Graph<PredicateFormula,DefaultEdge>
	 * @return Graph<PredicateFormula,DefaultEdge>
	 */
	private Graph<PredicateFormula, DefaultEdge> clone(Graph<PredicateFormula, DefaultEdge> graph0) {
		Graph<PredicateFormula, DefaultEdge> graph = new SimpleGraph<PredicateFormula, DefaultEdge>(DefaultEdge.class);
		for(PredicateFormula vertex:graph0.vertexSet()) {
			graph.addVertex(vertex);
		}
		for(DefaultEdge edge:graph0.edgeSet()) {
			graph.addEdge(graph0.getEdgeSource(edge), graph0.getEdgeTarget(edge));
		}
		return graph;
	}

	/**
	 * @param graph0 Graph<PredicateFormula,DefaultEdge>
	 * @param graph1 Graph<PredicateFormula,DefaultEdge>
	 * @return Graph<PredicateFormula,DefaultEdge>
	 */
	private Graph<PredicateFormula, DefaultEdge> merge(Graph<PredicateFormula, DefaultEdge> graph0, Graph<PredicateFormula, DefaultEdge> graph1) {
		Graph<PredicateFormula, DefaultEdge> graph = new SimpleGraph<PredicateFormula, DefaultEdge>(DefaultEdge.class);
		for(PredicateFormula vertex:graph0.vertexSet()) {
			graph.addVertex(vertex);
		}
		for(PredicateFormula vertex:graph1.vertexSet()) {
			graph.addVertex(vertex);
		}
		for(DefaultEdge edge:graph0.edgeSet()) {
			graph.addEdge(graph0.getEdgeSource(edge), graph0.getEdgeTarget(edge));
		}
		for(DefaultEdge edge:graph1.edgeSet()) {
			graph.addEdge(graph1.getEdgeSource(edge), graph1.getEdgeTarget(edge));
		}
		return graph;
	}

	/**
	 * @param fact PredicateFormula
	 * @return Pair<IC,Collection<PredicateFormula>>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph#getFactProvenance(PredicateFormula)
	 */
	@Override
	public Pair<Constraint, Collection<PredicateFormula>> getFactProvenance(PredicateFormula fact) {
		return this.provenance.get(fact);
	}

	/**
	 * @param dependency IC
	 * @param facts Collection<PredicateFormula>
	 * @return boolean
	 * @see uk.ac.ox.cs.pdq.chase.FiringGraph#isFired(IC, Collection<PredicateFormula>)
	 */
	@Override
	public boolean isFired(Constraint dependency, Collection<PredicateFormula> facts) {
		return this.firings.get(facts).contains(dependency);
	}

	/**
	 * @param source FiringGraph<S>
	 * @return FiringGraph<S>
	 * @see uk.ac.ox.cs.pdq.chase.FiringGraph#merge(FiringGraph<S>)
	 */
	@Override
	public FiringGraph<S> merge(FiringGraph<S> source) {
		Preconditions.checkArgument(source instanceof MapFiringGraph);

		Multimap<Collection<PredicateFormula>, Constraint> firedDependencies = LinkedHashMultimap.create();
		firedDependencies.putAll(this.firings);
		firedDependencies.putAll(((MapFiringGraph<S>)source).getFiredDependencies());

		Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> factProvenance = Maps.newLinkedHashMap(this.getFactProvenance());
		factProvenance.putAll(((MapFiringGraph<S>)source).getFactProvenance());

		return new MapFiringGraph<S>(this.map, this.graph, factProvenance, firedDependencies);
	}

	/**
	 * @return MultiValueMap<Collection<PredicateFormula>,Collection<PredicateFormula>>
	 */
	public MultiValueMap<Collection<PredicateFormula>, Collection<PredicateFormula>> getMap() {
		return this.map;
	}

	/**
	 * @return Map<PredicateFormula,Pair<IC,Collection<PredicateFormula>>>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph#getFactProvenance()
	 */
	@Override
	public Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> getFactProvenance() {
		return this.provenance;
	}

	/**
	 * @return Multimap<Collection<PredicateFormula>,IC>
	 */
	public Multimap<Collection<PredicateFormula>, Constraint> getFiredDependencies() {
		return this.firings;
	}

	public Set<Collection<PredicateFormula>> getPreconditions() {
		return this.map.keySet();
	}

	public Collection<Collection<PredicateFormula>> getConsequences(Collection<PredicateFormula> key) {
		return this.map.getCollection(key);
	}

}
