package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.db.Schema;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.planner.PlannerParameters;
import uk.ac.ox.cs.pdq.planner.PlannerParameters.ReasoningTypes;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.BagFactory.BagTypes;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.BoundedChaser.KSupplier;
import uk.ac.ox.cs.pdq.reasoning.Reasoner;

import com.google.common.base.Preconditions;
import com.google.common.eventbus.EventBus;

/**
 * Creates chasers based on the input arguments
 *
 * @author Efthymia Tsamoura
 * @author Julien Leblay
 *
 */
public final class ReasonerFactory {

	protected static Logger log = Logger.getLogger(ReasonerFactory.class);

	/** True if Skolem terms are assigned to the existentially quantified variables when grounding an IC*/
	private final Boolean canonicalNames = true;

	/** The input schema */
	private final Schema schema;

	private final AccessibleSchema accessibleSchema;

	private final Query<?> query;

	/** The event bus that will be shared by all instances create by this factory */
	private final EventBus eventBus;

	/** If true, statistics are collected while the reasoner is used. */
	private final boolean collectStats;

	/** Type of reasoner */
	private ReasoningTypes type;

	/** K for the KTermination chase */
	private Integer terminationK;

	/** KSupplier to be shared across all BoundedChasers created by this factory */
	private KSupplier kSupplier = null;

	/** true, if the reasoner initialisation shall be unrestricted. */
	private Boolean fullInitialization;


	/**
	 *
	 * @param eventBus
	 * @param collectStats
	 * @param schema
	 * @param accessibleSchema
	 * @param query
	 * @param params
	 */
	public ReasonerFactory(
			EventBus eventBus,
			boolean collectStats,
			Schema schema,
			AccessibleSchema accessibleSchema,
			Query<?> query,
			PlannerParameters params) {
		this(eventBus, collectStats, schema, accessibleSchema, query,
				params.getReasoningType(), params.getTerminationK(),
				params.getFullInitialization());
	}

	/**
	 *
	 * @param eventBus
	 * @param collectStats
	 * @param schema
	 * @param accessibleSchema
	 * @param query
	 * @param type
	 * @param k
	 * @param fullInitialization
	 */
	private ReasonerFactory(
			EventBus eventBus,
			boolean collectStats,
			Schema schema,
			AccessibleSchema accessibleSchema,
			Query<?> query,
			ReasoningTypes type,
			Integer k,
			Boolean fullInitialization) {
		Preconditions.checkArgument(schema != null);
		Preconditions.checkArgument(eventBus != null);
		this.schema = schema;
		this.accessibleSchema = accessibleSchema;
		this.query = query;
		this.eventBus = eventBus;
		this.collectStats = collectStats;
		this.type = type;
		this.terminationK = k;
		this.fullInitialization = fullInitialization;
	}

	/**
	 * @return
	 * 		a fresh instance of reasoner. Currently, all reasoners
	 *      implement the Chaser interface.
	 */
	public Reasoner getInstance() {
		BlockingDetector blockingDetector = null;
		if (this.type == ReasoningTypes.BLOCKING_CHASE) {
			this.type = ReasoningTypes.BLOCKING_CHASE;
			blockingDetector = new BlockingDetector();

		} 
//		else  if (this.schema.isCyclic() && !this.schema.containsViews()) {
//			log.warn("Cycles detected in input schema. Forcing reasoning type to " + ReasoningTypes.BLOCKING_CHASE);
//			this.type = ReasoningTypes.BLOCKING_CHASE;
//			blockingDetector = new BlockingDetector();
//		}
		switch (this.type) {
		case RESTRICTED_CHASE:
			return new RestrictedChaser(
					this.eventBus,
					this.collectStats,
					this.canonicalNames);
		case KTERMINATION_CHASE:
			return new KTerminationChaser(
					this.eventBus, this.collectStats,
					this.canonicalNames, this.terminationK);
		case BLOCKING_CHASE:
			BagFactory bagFactory = null;
			bagFactory = new BagFactory(BagTypes.EXTENDEDBAG, this.schema, this.accessibleSchema, this.query);
			return new BlockingChaser(
					this.eventBus, this.collectStats,
					this.canonicalNames, bagFactory,
					blockingDetector,
					1);
		default:
			return null;
		}
	}

	/**
	 * @return the event bus associated with this factory
	 */
	public EventBus getEventBus() {
		return this.eventBus;
	}
}
