package uk.ac.ox.cs.pdq.planner.reasoning.chase;

import static uk.ac.ox.cs.pdq.logging.performance.StatKeys.GENERATED_FACTS;
import static uk.ac.ox.cs.pdq.logging.performance.StatKeys.MILLI_UPDATE;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.db.Schema;
import uk.ac.ox.cs.pdq.db.TGD;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.Constant;
import uk.ac.ox.cs.pdq.fol.Formula;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.fol.Variable;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismDetector;
import uk.ac.ox.cs.pdq.planner.logging.performance.ConstantsStatistics;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.DatabaseListState;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.state.StateFactory;
import uk.ac.ox.cs.pdq.reasoning.Match;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.eventbus.EventBus;


/**
 * The restricted chase algorithm.
 * According to this, a dependency is fired unless it is already satisfied in the database.
 * The facts that are being generated after each chase step are stored in a list.
 *
 * @author Efthymia Tsamoura
 *
 */
public class RestrictedChaser extends Chaser<DatabaseListState> {

	/**
	 * Constructor for RestrictedChaser.
	 * @param eventBus EventBus
	 * @param collectStats boolean
	 * @param canonicalNames Boolean
	 */
	public RestrictedChaser(
			EventBus eventBus, boolean collectStats,
			Boolean canonicalNames) {
		super(eventBus, collectStats, canonicalNames);
	}

	/**
	 * @param detector HomomorphismDetector
	 * @param facts Collection<PredicateFormula>
	 * @param schema Schema
	 * @return DatabaseListState
	 */
	@Override
	public DatabaseListState initialize(
			HomomorphismDetector detector,
			Query<?> query, 
			Schema schema) {
		DatabaseListState result = StateFactory.getInstance(detector, this.createInitialisationFormula(query, schema));
		this.chase(result, schema.getDependencies());
		return result;
	}

	/**
	 * @param state DatabaseListState
	 * @param dependencies Collection<? extends IC>
	 */
	@Override
	public void chase(DatabaseListState state, Collection<? extends Constraint> dependencies) {
		//True if at the end of the internal for loop at least one dependency has been fired
		boolean appliedStep = false;
		do {
			appliedStep = false;
			Map<? extends Constraint, List<Match>> homomorphisms = state.getHomomorphisms(dependencies);
			for (Constraint dependency: homomorphisms.keySet()) {
				List<Match> matches = homomorphisms.get(dependency);
				if (matches != null) {
					for (Match match: matches) {
						//For each match, get the mapping of the dependency's variables to chase constants and
						//try to fire the corresponding dependency
						Map<Variable, Constant> mapping = match.getMapping();
						Constraint groundedFormula = dependency.fire(mapping, this.canonicalNames);
						this.eventBus.post(new ConstantsStatistics.Increment(groundedFormula.getRight().getTerms()));
						log.trace("Checking rule " + groundedFormula);
						Map<Variable, ? extends Term> inputMapping = Utility.retain(mapping, dependency.getBothSideVariables());
						if(!this.isApplied(state, dependency, inputMapping)){
							this.applyStep(state, match);
							appliedStep = true;
						}
					}
				}
			}
		} while (appliedStep);
	}

	/**
	 *
	 * @param state
	 * 		Input state
	 * @param dependency
	 * 		Input dependency
	 * @param inputMapping
	 * 		Mapping of variables to chase constants
	 * @return
	 * 		true if the dependency is already satisfied
	 */
	protected boolean isApplied(DatabaseListState state, Constraint dependency, Map<Variable, ? extends Term> inputMapping) {
		Conjunction.Builder cb = Conjunction.builder();
		for (PredicateFormula p: dependency.getLeft().getPredicates()) {
			cb.and(p);
		}
		for (PredicateFormula p: dependency.getRight().getPredicates()) {
			cb.and(p);
		}
		TGD tgd = new TGD((Conjunction<PredicateFormula>) cb.build(), Conjunction.<PredicateFormula>of());
		List<Match> matchings = state.getHomomorphisms(tgd);
		Set<Variable> variables = dependency.getBothSideVariables();
		for(Match matching:matchings) {
			Map<Variable, Constant> mapping = Utility.retain(matching.getMapping(), variables);
			if (mapping.equals(inputMapping)) {
				return true;
			}
		}
		return false;
	}


	/**
	 * Updates the input state given the input match of dependency conjuncts to facts
	 * @param state
	 * 		Input state
	 * @param match
	 * 		Input match
	 * @return
	 * 		true if the update was successful
	 */
	public Boolean applyStep(DatabaseListState state, Match match) {
		Preconditions.checkArgument(match != null);
		return this.applySteps(state, Lists.newArrayList(match));
	}

	/**
	 * Method applySteps.
	 * @param state DatabaseListState
	 * @param matches List<Matching>
	 * @return Boolean
	 */
	@Override
	public Boolean applySteps(DatabaseListState state, List<Match> matches) {
		Preconditions.checkArgument(matches != null);
		Collection<PredicateFormula> createdFacts = new LinkedHashSet<>();
		for (Match match: matches) {
			//For each match, get (i) the corresponding dependency,
			//(ii) the mapping of dependency's variables to chase constants
			//and update the input state
			Constraint dependency = (Constraint) match.getQuery();
			Map<Variable, Constant> mapping = match.getMapping();
			Constraint groundedFormula = dependency.fire(mapping, this.canonicalNames);
			this.eventBus.post(new ConstantsStatistics.Increment(groundedFormula.getRight().getTerms()));
			Formula parentFormula = groundedFormula.getLeft();
			Formula childFormula = groundedFormula.getRight();
			createdFacts.addAll(childFormula.getPredicates());
			log.debug(groundedFormula);
			this.stats.increase(GENERATED_FACTS, childFormula.getPredicates().size());
			state.update(dependency, parentFormula, childFormula);
		}
		this.stats.start(MILLI_UPDATE);
		this.eventBus.post(createdFacts);
		state.addFacts(createdFacts);
		this.stats.stop(MILLI_UPDATE);
		return true;
	}
}
