package uk.ac.ox.cs.pdq.planner.reasoning.chase.state;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibilityAxiom;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph;
import uk.ac.ox.cs.pdq.reasoning.Match;
import uk.ac.ox.cs.pdq.reasoning.Reasoner;

/**
 *
 * The state of a configuration. This can be either a list of atoms (implemented by the DatabaseListState class) or
 * a tree of bags (implemented by the DatabaseTreeState class)
 *
 * @author Efthymia Tsamoura
 *
 */
public interface ChaseState extends Reasoner.State {

	/**
	 *
	 * @param query
	 * 		The input query
	 * @param constraint
	 * 		The mapping constraint that should be satisfied by any of the homomorphisms found
	 * 		(e.g., the chase constants to which the query free variables must be mapped to)
	 * @return
	 * 		the list of matches found.
	 */
	List<Match> matchesQuery(Query<?> query);

	/**
	 *
	 * @param dependency
	 * 		The dependency for which we want to detect
	 * 		homomorphisms of its left-hand side to the state's facts
	 * @return
	 * 		the list of matches found
	 */
	List<Match> getHomomorphisms(Constraint dependency);

	/**
	 *
	 * @param dependencies
	 * 		The dependencies for which we want to detect homomorphisms of their left-hand sides to the state's facts
	 * @return
	 * 		the list of matches found
	 */
	<D extends Constraint> Map<D, List<Match>> getHomomorphisms(Collection<D> dependencies);

	/**
	 *
	 * @param facts
	 * 		A list of facts
	 *
	 * Adds the input facts to the state
	 */
	void addFacts(Collection<? extends PredicateFormula> facts);

	/**
	 *
	 * @return
	 * 		the dependency firings that have leaded to the derivation of each fact in the state
	 */
	Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> getFactProvenance();

	/**
	 *
	 * @return
	 * 		the firing that has leaded to the derivation of the input fact
	 */
	Pair<Constraint, Collection<PredicateFormula>> getFactProvenance(PredicateFormula fact);

	/**
	 *
	 * @return
	 * 		the facts of the state
	 * @see uk.ac.ox.cs.pdq.Reasoner$State#getFacts()
	 */
	@Override
	Collection<PredicateFormula> getFacts();

	/**
	 *
	 * @return
	 * 		the inferred accessible facts of the state
	 */
	Collection<String> getInferred();

	/**
	 *
	 * @return
	 * 		the inferred accessible facts that were derived when chasing the state
	 */
	Collection<PredicateFormula> getDerivedInferred();

	/**
	 *
	 * @return
	 * 		the firing graph associated with the facts of the state
	 */
	FiringGraph<ChaseState>  getFiringGraph();
	/**
	 *
	 * @param s
	 * 		The state of another configuration
	 * @return
	 * 		a state that is the merge of this.state and the input one
	 */
	<S extends ChaseState> S merge(S s);

	/**
	 * For each input accessibility axiom, it
	 * groups the corresponding state facts based on the chase constants assigned to their input positions.
	 * This function is called when creating the initial ApplyRule configurations
	 * @param axioms
	 *
	 * @return
	 * 		pairs of accessibility axioms to chase facts
	 */
	Collection<Pair<AccessibilityAxiom, Collection<PredicateFormula>>> groupByBinding(Collection<AccessibilityAxiom> axioms);

	/**
	 * Adds the accessed, accessible and inferred accessible facts associated to each input fact.
	 *
	 * @param schema
	 * 			Input accessible schema
	 * @param axiom
	 * 			Input accessibility axiom
	 * @param facts
	 * 			Input facts
	 */
	void generate(AccessibleSchema schema, AccessibilityAxiom axiom, Collection<PredicateFormula> facts);

	/**
	 *
	 * @param accessibleSchema
	 * @param axioms
	 * @return
	 * 		the unexposed facts and information to expose them
	 */
	Map<AccessibilityAxiom, List<Match>> getUnExposedFacts(AccessibleSchema accessibleSchema);

	/**
	 * @return ChaseState
	 */
	ChaseState clone();
}
