package uk.ac.ox.cs.pdq.planner.reasoning.chase.state;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.Formula;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.fol.Signature;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibilityAxiom;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessedRelation;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.AccessibleRelation;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema.InferredAccessibleRelation;
import uk.ac.ox.cs.pdq.planner.homomorphism.DBHomomorphismManager;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismConstraint;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.MapFiringGraph;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.MatchFactory;
import uk.ac.ox.cs.pdq.reasoning.Match;

import com.google.common.base.Preconditions;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;

/**
 *
 * Top-level class for database-backed chase states.
 *
 * @author Efthymia Tsamoura
 *
 */
public abstract class DatabaseChaseState implements ChaseState {

	/** Queries and updates the facts' database**/
	private DBHomomorphismManager manager;

	/** The state's facts*/
	private final Collection<PredicateFormula> facts;

	/** String signatures of the inferred accessible facts*/
	private final Collection<String> inferred = new LinkedHashSet<>();

	/** The inferred accessible facts that were derived when chasing this.state **/
	protected final Collection<PredicateFormula> derivedInferred = new LinkedHashSet<>();

	/** The firing graph associated with the facts of the state*/
	private FiringGraph<ChaseState> firingGraph = new MapFiringGraph<>();

	/** The chase constants that appear in the facts */
	private final Collection<Term> terms = new LinkedHashSet<>();

	/** Maps each schema signature (relation) to its chase facts*/
	private final Multimap<Signature, PredicateFormula> signatureToFact = LinkedHashMultimap.create();

	/** Maps each chase constant the Accessed facts it appears*/
	private final Multimap<Term,PredicateFormula> termToAccessible  = LinkedHashMultimap.create();

	/** True if the current state is associated to the initial configuration
	 * (the one obtained after chasing with the dependencies of the input schema)*/
	private final boolean initialState;


	/**
	 * Constructor for DatabaseChaseState.
	 * @param manager DBHomomorphismManager
	 * @param facts Collection<PredicateFormula>
	 * @param initialState boolean
	 */
	protected DatabaseChaseState(
			DBHomomorphismManager manager,
			Collection<PredicateFormula> facts,
			boolean initialState) {
		Preconditions.checkNotNull(manager);
		Preconditions.checkNotNull(facts);
		this.manager = manager;
		this.initialState = initialState;

		this.facts = Sets.newLinkedHashSet(facts);
		for(PredicateFormula fact:facts) {
			this.terms.addAll(fact.getTerms());
			if (!(fact.getSignature() instanceof AccessibleRelation) && 
					!(fact.getSignature() instanceof AccessedRelation) && 
					!(fact.getSignature() instanceof InferredAccessibleRelation)) {
				this.signatureToFact.put(fact.getSignature(), fact);
			}
			if (fact.getSignature() instanceof AccessibleRelation) {
				this.termToAccessible.put(fact.getTerm(0), fact);
			}
			if (fact.getSignature() instanceof InferredAccessibleRelation) {
				this.inferred.add(fact.toString());
			}
		}
	}

	/**
	 * Constructor for DatabaseChaseState.
	 * @param manager DBHomomorphismManager
	 * @param facts Collection<PredicateFormula>
	 * @param initialState boolean
	 * @param firingGraph FiringGraph<ChaseState>
	 */
	protected DatabaseChaseState(
			DBHomomorphismManager manager,
			Collection<PredicateFormula> facts,
			boolean initialState,
			FiringGraph<ChaseState> firingGraph) {
		this(manager, facts, initialState);
		Preconditions.checkNotNull(firingGraph);
		this.firingGraph = firingGraph;
	}

	/**
	 * @param facts Collection<? extends PredicateFormula>
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#addFacts(Collection<? extends PredicateFormula>)
	 */
	@Override
	public void addFacts(Collection<? extends PredicateFormula> facts) {
		this.manager.addFacts(facts);
	}

	/**
	 * @return Map<PredicateFormula,Pair<IC,Collection<PredicateFormula>>>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#getFactProvenance()
	 */
	@Override
	public Map<PredicateFormula, Pair<Constraint, Collection<PredicateFormula>>> getFactProvenance() {
		return this.firingGraph.getFactProvenance();
	}


	/**
	 * @param fact PredicateFormula
	 * @return Pair<IC,Collection<PredicateFormula>>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#getFactProvenance(PredicateFormula)
	 */
	@Override
	public Pair<Constraint, Collection<PredicateFormula>> getFactProvenance(PredicateFormula fact) {
		return this.firingGraph.getFactProvenance(fact);
	}

	/**
	 * @return Collection<PredicateFormula>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#getFacts()
	 */
	@Override
	public Collection<PredicateFormula> getFacts() {
		return this.facts;
	}

	/**
	 * @return Collection<String>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#getInferred()
	 */
	@Override
	public Collection<String> getInferred() {
		return this.inferred;
	}

	/**
	 * @return Collection<PredicateFormula>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#getDerivedInferred()
	 */
	@Override
	public Collection<PredicateFormula> getDerivedInferred() {
		return this.derivedInferred;
	}

	/**
	 * @return FiringGraph<ChaseState>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#getFiringGraph()
	 */
	@Override
	public FiringGraph<ChaseState> getFiringGraph() {
		return this.firingGraph;
	}

	/**
	 * @return DBHomomorphismManager
	 */
	public DBHomomorphismManager getManager() {
		return this.manager;
	}

	/**
	 * @param manager DBHomomorphismManager
	 */
	public void setManager(DBHomomorphismManager manager) {
		this.manager = manager;
	}

	/**
	 * @return Collection<Term>
	 */
	protected Collection<Term> getTerms() {
		return this.terms;
	}

	/**
	 * @return Multimap<Signature,PredicateFormula>
	 */
	protected Multimap<Signature, PredicateFormula> getSignatureToSchemaFact() {
		return this.signatureToFact;
	}

	/**
	 * @return Multimap<Term,PredicateFormula>
	 */
	protected Multimap<Term,PredicateFormula> getTermToAccessibleFact() {
		return this.termToAccessible;
	}

	/**
	 * @return boolean
	 */
	public boolean initialState() {
		return this.initialState;
	}

	/**
	 * Calls the manager to detect matches of the accessible query to facts.
	 * The manager detects homomorphisms using a database backend.
	 * @param query Query
	 * @param mappingConstraint Map<Variable,Term>
	 * @return List<Matching>
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#matchesQuery(Query, Map<Variable,Term>)
	 */
	@Override
	public List<Match> matchesQuery(Query<?> query) {
		return this.manager.getMatches(
				query,
				HomomorphismConstraint.topK(1),
				HomomorphismConstraint.scope(Conjunction.of(this.facts)),
				HomomorphismConstraint.satisfy(query.getFreeToCanonical()));
	}

	/**
	 * Calls the manager to detect matches of the
	 * left-hand side conjunction of the input dependency to the facts in this.state.
	 * The manager detects homomorphisms using a database backend.
	 * @param dependency IC
	 * @return List<Matching>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#getHomomorphisms(Constraint)
	 */
	@Override
	public List<Match> getHomomorphisms(Constraint dependency) {
		return this.manager.getMatches(
				dependency, HomomorphismConstraint.scope(
						Conjunction.of(this.facts)));
	}

	/**
	 * @param dependencies Collection<D>
	 * @return Map<D,List<Matching>>
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#getHomomorphisms(Collection<D>)
	 */
	@Override
	public <D extends Constraint> Map<D, List<Match>> getHomomorphisms(Collection<D> dependencies) {
		return this.manager.getMatches(
				dependencies, HomomorphismConstraint.scope(
						Conjunction.of(this.facts)));
	}

	/**
	 * @param dependency
	 *            Input dependency
	 * @param facts
	 *            The facts used to fire the dependency
	 * @return true if the dependency has already been fired using the input facts
	 */
	public Boolean isFired(Constraint dependency, Collection<PredicateFormula> facts) {
		return this.firingGraph.isFired(dependency, facts);
	}

	/**
	 * Updates this.state
	 * @param dependency
	 * 		The dependency to fire
	 * @param parentFormula
	 * 		The grounded left-hand side of the input dependency
	 * @param childFormula
	 * 		The grounded right-hand side of the input dependency
	 */
	public void update(Constraint dependency, Formula parentFormula, Formula childFormula) {
		Preconditions.checkArgument(dependency != null);
		Preconditions.checkArgument(parentFormula != null);
		Preconditions.checkArgument(childFormula != null);

		// Update the list of facts
		this.facts.addAll(childFormula.getPredicates());
		for (PredicateFormula fact: childFormula.getPredicates()) {
			this.terms.addAll(fact.getTerms());
			if (!(fact.getSignature() instanceof AccessibleRelation) && 
					!(fact.getSignature() instanceof AccessedRelation) && 
					!(fact.getSignature() instanceof InferredAccessibleRelation)) {
				this.signatureToFact.put(fact.getSignature(), fact);
			}
			if (fact.getSignature() instanceof AccessibleRelation) {
				this.termToAccessible.put(fact.getTerm(0), fact);
			}
			if (fact.getSignature() instanceof InferredAccessibleRelation) {
				this.inferred.add(fact.toString());
			}
		}

		for(PredicateFormula predicate:childFormula.getPredicates()) {
			if(predicate.getSignature() instanceof InferredAccessibleRelation) {
				this.derivedInferred.add(predicate);
			}
		}
		this.firingGraph.put(dependency, Sets.newHashSet(parentFormula.getPredicates()), Sets.newHashSet(childFormula.getPredicates()), this.initialState());
	}

	/**
	 *
	 * @param axioms
	 * 		Input accessibility axioms
	 * @param signatureToSchemaFact
	 * 		Maps each schema signature (relation) to its chase facts
	 * @return
	 * 		pairs of accessibility axioms to chase facts.
	 */
	protected List<Pair<AccessibilityAxiom, Collection<PredicateFormula>>> groupByBinding(
			Collection<AccessibilityAxiom> axioms, Multimap<Signature, PredicateFormula> signatureToSchemaFact) {
		List<Pair<AccessibilityAxiom, Collection<PredicateFormula>>> ret = new ArrayList<>();
		for(AccessibilityAxiom axiom: axioms) {
			//For each axiom, we get the relevant facts
			//and group them based on the constants of their input positions
			Collection<PredicateFormula> facts = signatureToSchemaFact.get(axiom.getBaseRelation());
			Multimap<Collection<Term>, PredicateFormula> groupsOfFacts = LinkedHashMultimap.create();
			for(PredicateFormula fact: facts) {
				groupsOfFacts.put(fact.getTerms(axiom.getAccessMethod().getZeroBasedInputs()), fact);
			}
			Iterator<Collection<Term>> keyIterator = groupsOfFacts.keySet().iterator();
			while(keyIterator.hasNext()) {
				Collection<PredicateFormula> collection = Sets.newLinkedHashSet(groupsOfFacts.get(keyIterator.next()));
				ret.add(Pair.of(axiom, collection));
			}
		}
		return ret;
	}

	/**
	 *
	 * @param accessibleSchema
	 * @param axioms
	 * @param signatureToSchemaFact
	 * 		Maps each schema signature (relation) to its chase facts
	 * @param termToAccessibleFact
	 * 		Maps each chase constant the Accessed facts it appears
	 * @return
	 * 		the non-fired accessibility axioms of the state and information to fire them
	 */
	protected Map<AccessibilityAxiom, List<Match>> getUnExposedFacts(
			AccessibleSchema accessibleSchema,
			Multimap<Signature, PredicateFormula> signatureToSchemaFact,
			Multimap<Term,PredicateFormula> termToAccessibleFact) {
		Map<AccessibilityAxiom, List<Match>> ret = new LinkedHashMap<>();
		List<Pair<AccessibilityAxiom,Collection<PredicateFormula>>> groups = this.groupByBinding(accessibleSchema.getAccessibilityAxioms(), signatureToSchemaFact);
		for(Pair<AccessibilityAxiom, Collection<PredicateFormula>> pair: groups) {
			AccessibilityAxiom axiom = pair.getLeft();
			Iterator<PredicateFormula> iterator = pair.getRight().iterator();
			while (iterator.hasNext()) {
				PredicateFormula fact = iterator.next();
				PredicateFormula accessedFact = new PredicateFormula(accessibleSchema.getAccessedRelation((Relation) fact.getSignature()), fact.getTerms());
				Collection<Term> inputTerms = accessedFact.getTerms(axiom.getAccessMethod().getZeroBasedInputs());
				if(this.firingGraph.getFactProvenance(accessedFact) == null && termToAccessibleFact.keySet().containsAll(inputTerms)) {
					Match matching = MatchFactory.getMatching(pair.getLeft(), fact);
					List<Match> matchings = ret.get(pair.getLeft());
					if(matchings == null) {
						ret.put(axiom, Lists.newArrayList(matching));
					}
					else {
						matchings.add(matching);
					}
				}
			}
		}
		return ret;
	}

	//TODO Update the firing history of the accessed and accessible facts
	/**
	 * @param schema
	 * @param axiom
	 * @param facts
	 * @return
	 * 		the corresponding accessed, accessible and inferred accessible facts for each fact in the input collection
	 */
	protected Collection<PredicateFormula> generateFacts(AccessibleSchema schema, AccessibilityAxiom axiom, Collection<PredicateFormula> facts) {
		Collection<PredicateFormula> createdFacts = new LinkedHashSet<>();
		for(PredicateFormula fact:facts) {
			PredicateFormula accessedFact = new PredicateFormula(schema.getAccessedRelation((Relation) fact.getSignature()), fact.getTerms());
			createdFacts.add(accessedFact);
			PredicateFormula infAccFact = new PredicateFormula(schema.getInferredAccessibleRelation((Relation) fact.getSignature()), fact.getTerms());
			createdFacts.add(infAccFact);
			this.inferred.add(infAccFact.toString());
			this.derivedInferred.add(infAccFact);
			this.firingGraph.put(axiom, accessedFact, infAccFact, this.initialState);
			for(Term term:fact.getTerms()) {
				createdFacts.add(new PredicateFormula(schema.getAccessibleRelation(), term));
			}
			this.getFacts().addAll(createdFacts);
		}
		return createdFacts;
	}

	/**
	 * @param schema AccessibleSchema
	 * @param axiom AccessibilityAxiom
	 * @param facts Collection<PredicateFormula>
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#generate(AccessibleSchema, AccessibilityAxiom, Collection<PredicateFormula>)
	 */
	@Override
	public void generate(AccessibleSchema schema, AccessibilityAxiom axiom, Collection<PredicateFormula> facts) {
		Collection<PredicateFormula> generatedFacts = this.generateFacts(schema, axiom, facts);
		this.addFacts(generatedFacts);
	}

	/**
	 * @return DatabaseChaseState
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#clone()
	 */
	@Override
	public abstract DatabaseChaseState clone();
}
