package uk.ac.ox.cs.pdq.planner.reasoning.chase.state;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.lang3.tuple.Pair;

import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibilityAxiom;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.homomorphism.DBHomomorphismManager;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph;
import uk.ac.ox.cs.pdq.reasoning.Match;

import com.google.common.base.Preconditions;

/**
 *
 * @author Efthymia Tsamoura
 *
 *         The state of a configuration organised as a list of atoms. This type of
 *         state is used in non-blocking chase implementations.
 */
public class DatabaseListState extends DatabaseChaseState {

	/**
	 * Constructor for DatabaseListState.
	 * @param manager DBHomomorphismManager
	 * @param facts Collection<PredicateFormula>
	 */
	public DatabaseListState(DBHomomorphismManager manager, Collection<PredicateFormula> facts) {
		super(manager, facts, true);
		this.addFacts(facts);
	}

	/**
	 * Constructor for DatabaseListState.
	 * @param manager DBHomomorphismManager
	 * @param facts Collection<PredicateFormula>
	 * @param initialState boolean
	 * @param firingGraph FiringGraph<ChaseState>
	 */
	private DatabaseListState(
			DBHomomorphismManager manager,
			Collection<PredicateFormula> facts,
			boolean initialState,
			FiringGraph<ChaseState> firingGraph) {
		super(manager, facts,
				initialState,
				firingGraph);
	}

	/**
	 * @param s S
	 * @return S
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#merge(S)
	 */
	@Override
	public <S extends ChaseState> S merge(S s) {
		Preconditions.checkState(s instanceof DatabaseListState);
		Collection<PredicateFormula> facts =  new LinkedHashSet<>(this.getFacts());
		facts.addAll(s.getFacts());
		S state = (S) new DatabaseListState(
				this.getManager(),
				facts,
				false,
				this.getFiringGraph().merge(s.getFiringGraph())
				);

		return state;
	}

	/**
	 * For each input accessibility axiom, it
	 * groups the corresponding state facts based on the chase constants assigned to their input positions.
	 * This function is called when creating the initial ApplyRule configurations
	 * @param axioms
	 *
	 * @return
	 * 		pairs of accessibility axioms to chase facts
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#groupByBinding(Collection<AccessibilityAxiom>)
	 */
	@Override
	public List<Pair<AccessibilityAxiom, Collection<PredicateFormula>>> groupByBinding(Collection<AccessibilityAxiom> axioms) {
		return this.groupByBinding(axioms, this.getSignatureToSchemaFact());
	}

	/**
	 *
	 * @param accessibleSchema
	 * @param axioms
	 * @return
	 * 		the unexposed facts and information to expose them
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#getUnExposedFacts(AccessibleSchema, Collection<AccessibilityAxiom>)
	 */
	@Override
	public Map<AccessibilityAxiom, List<Match>> getUnExposedFacts(AccessibleSchema accessibleSchema) {
		return this.getUnExposedFacts(accessibleSchema, this.getSignatureToSchemaFact(), this.getTermToAccessibleFact());
	}

	/**
	 * @param o Object
	 * @return boolean
	 */
	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null) {
			return false;
		}
		return this.getClass().isInstance(o)
				&& this.getFacts().equals(((DatabaseListState) o).getFacts());
	}

	/**
	 * @return int
	 */
	@Override
	public int hashCode() {
		return Objects.hash(this.getFacts());
	}

	/**
	 * @return DatabaseListState
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#clone()
	 */
	@Override
	public DatabaseListState clone() {
		return new DatabaseListState(this.getManager(), this.getFacts(), false,
				this.getFiringGraph().clone());
	}
}
