package uk.ac.ox.cs.pdq.planner.reasoning.chase.state;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;

import org.apache.commons.lang3.tuple.Pair;
import org.jgrapht.graph.DefaultEdge;

import uk.ac.ox.cs.pdq.db.Constraint;
import uk.ac.ox.cs.pdq.fol.Conjunction;
import uk.ac.ox.cs.pdq.fol.Formula;
import uk.ac.ox.cs.pdq.fol.PredicateFormula;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.fol.Signature;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibilityAxiom;
import uk.ac.ox.cs.pdq.planner.db.access.AccessibleSchema;
import uk.ac.ox.cs.pdq.planner.homomorphism.DBHomomorphismManager;
import uk.ac.ox.cs.pdq.planner.homomorphism.HomomorphismConstraint;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Bag;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.BagBoundPredicate;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.BagFactory;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.BagsTree;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.BagsTreeEdge;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.ExtendedBag;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.FiringGraph;
import uk.ac.ox.cs.pdq.planner.reasoning.chase.Bag.BagStatus;
import uk.ac.ox.cs.pdq.reasoning.Match;

import com.google.common.base.Preconditions;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Multimap;

/**
 * The state of a configuration organised as a tree of bags. This type of
 * state is used in blocking chase implementations.
 *
 * @author Efthymia Tsamoura
 *
 */
public class DatabaseTreeState extends DatabaseChaseState {

	/** A tree of bags */
	private final BagsTree<ExtendedBag> tree;

	/** The bags that we have to update with new facts */
	private Collection<ExtendedBag> bagsToUpdate = new LinkedHashSet<>();

	/** Produces new bags*/
	private final BagFactory bagFactory;


	/**
	 * Constructor for DatabaseTreeState.
	 * @param manager DBHomomorphismManager
	 * @param facts Collection<PredicateFormula>
	 * @param bagFactory BagFactory
	 */
	public DatabaseTreeState(DBHomomorphismManager manager, Collection<PredicateFormula> facts, BagFactory bagFactory) {
		this(manager,
				initialiseTree(facts, bagFactory),
				bagFactory,
				true);
		Collection<PredicateFormula> propagatedFacts = new LinkedList<>();
		for (PredicateFormula fact:facts) {
			propagatedFacts.add(new BagBoundPredicate(fact, this.tree.getRoot().getId()));
		}
		this.addFacts(propagatedFacts);
	}


	/**
	 * Constructor for DatabaseTreeState.
	 * @param manager DBHomomorphismManager
	 * @param tree BagsTree<ExtendedBag>
	 * @param bagFactory BagFactory
	 * @param initialState boolean
	 */
	private DatabaseTreeState(
			DBHomomorphismManager manager,
			BagsTree<ExtendedBag> tree,
			BagFactory bagFactory,
			boolean initialState) {
		super(manager, tree.getFacts(), initialState);
		Preconditions.checkNotNull(tree);
		this.tree = tree;
		this.bagFactory = bagFactory;
	}

	/**
	 * Constructor for DatabaseTreeState.
	 * @param manager DBHomomorphismManager
	 * @param tree BagsTree<ExtendedBag>
	 * @param bagFactory BagFactory
	 * @param initialState boolean
	 * @param firingGraph FiringGraph<ChaseState>
	 * @param bagsToUpdate Collection<ExtendedBag>
	 */
	private DatabaseTreeState(
			DBHomomorphismManager manager,
			BagsTree<ExtendedBag> tree,
			BagFactory bagFactory,
			boolean initialState,
			FiringGraph<ChaseState> firingGraph,
			Collection<ExtendedBag> bagsToUpdate) {
		super(manager, tree.getFacts(), initialState, firingGraph);
		Preconditions.checkNotNull(tree);
		Preconditions.checkNotNull(bagsToUpdate);
		this.tree = tree;
		this.bagsToUpdate.addAll(bagsToUpdate);
		this.bagFactory = bagFactory;
	}

	/**
	 *
	 * @param facts
	 * @param bagFactory
	 * @return
	 * 		a tree with root a bag containing the input facts
	 */
	private static BagsTree<ExtendedBag> initialiseTree(
			Collection<PredicateFormula> facts, BagFactory bagFactory) {
		BagsTree<ExtendedBag> bagsTree = new BagsTree<>(DefaultEdge.class);
		ExtendedBag root = (ExtendedBag) bagFactory.createBag(facts);
		bagsTree.addVertex(root);
		return bagsTree;
	}

	/**
	 * @return DatabaseTreeState
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#clone()
	 */
	@Override
	public DatabaseTreeState clone() {
		BagsTree<ExtendedBag> tree = this.tree.clone();
		return new DatabaseTreeState(this.getManager(),
				tree,
				this.bagFactory.clone(),
				false,
				this.getFiringGraph().clone(),
				this.bagsToUpdate);
	}


	/**
	 * @return
	 * 		a clone of this state having the same bags with different identifiers though
	 */
	public DatabaseTreeState replicate() {
		BagsTree<ExtendedBag> tree = this.tree.replicate();
		this.materialiseTree(this.getManager(), tree);
		return new DatabaseTreeState(
				this.getManager(),
				tree,
				this.bagFactory.clone(),
				false,
				this.getFiringGraph().clone(),
				this.bagsToUpdate);
	}

	/**
	 * Materialises the input tree in a database
	 * @param manager
	 * @param tree
	 */
	private void materialiseTree(DBHomomorphismManager manager, BagsTree<ExtendedBag> tree) {
		Collection<PredicateFormula> facts = new ArrayList<>();
		for(ExtendedBag bag:tree.vertexSet()) {
			for(PredicateFormula fact:bag.getFacts()) {
				facts.add(new BagBoundPredicate(fact,bag.getId()));
			}
		}
		manager.addFacts(facts);
	}

	/**
	 * @param facts Collection<PredicateFormula>
	 * @return Collection<BagBoundPredicate>
	 */
	private Collection<BagBoundPredicate> toPropagate(Collection<PredicateFormula> facts) {
		Collection<BagBoundPredicate> propagatedFacts = new ArrayList<>();
		for (PredicateFormula fact: facts) {
			for (ExtendedBag bag:this.tree.vertexSet()) {
				if (bag.containsTerms(fact.getTerms()) && !bag.getFacts().contains(fact)) {
					propagatedFacts.add(new BagBoundPredicate(fact, bag.getId()));
					bag.addFacts(fact);
				}
			}
		}
		return propagatedFacts;
	}

	/**
	 *
	 * @param parentBag
	 * 			The bag over which the input dependency has been fired
	 * @param dependency
	 * 			The input dependency
	 * @param parentFormula
	 * 			The grounded left-hand side of the input dependency
	 * @param childFormula
	 * 			The grounded right-hand side of the input dependency
	 * @return
	 * 			a mapping of atoms to the bags the latter should be propagated to
	 */
	public Collection<BagBoundPredicate> update(
			ExtendedBag parentBag,
			Constraint dependency,
			Formula parentFormula,
			Formula childFormula) {
		Preconditions.checkArgument(parentBag != null);
		Preconditions.checkArgument(parentFormula != null);
		Preconditions.checkArgument(childFormula != null);
		Preconditions.checkArgument(dependency != null);

		Collection<BagBoundPredicate> propagatedFacts = this.toPropagate(childFormula.getPredicates());

		if (this.freshTermsExist(childFormula)) {
			Bag childBag = this.getBagFactory().createBag(childFormula.getPredicates());
			Preconditions.checkState(childBag instanceof ExtendedBag);
			this.tree.addVertex((ExtendedBag) childBag);
			this.tree.addEdge(parentBag, (ExtendedBag) childBag, new BagsTreeEdge(parentBag, childBag, dependency));
			for(PredicateFormula fact: childFormula.getPredicates()) {
				propagatedFacts.add(new BagBoundPredicate(fact, childBag.getId()));
			}
		}
		this.update(dependency, parentFormula, childFormula);
		return propagatedFacts;
	}

	/**
	 * @param formula
	 * @return
	 * 		true if the input formula contains fresh terms
	 */
	private Boolean freshTermsExist(Formula formula) {
		return !this.getTerms().containsAll(formula.getTerms());
	}

	/**
	 * Updates the list of satisfied queries/dependencies of each bag
	 */
	public void updateDependenciesOrQueries() {
		/*
		For each bag in the tree we update (i) the collection of the
		dependencies whose left-hand side is satisfied, as well as,
		(ii) the satisfied sub-queries of the input query. A bag must
 		be updated when new facts are propagated to it
		 */
		for (ExtendedBag bag:this.tree.vertexSet()) {
			//If true then we consider updating only (i)
			if (this.initialState()) {
				if (this.bagsToUpdate.contains(bag) || bag.initialConfiguration()) {
					bag.updateSatisfiedDependencies(this.getManager());
				}
			} else {
				if (this.bagsToUpdate.contains(bag) || bag.nextConfigurations()) {
					bag.updateSatisfiedAccessibleDependencies(this.getManager());
					bag.updateSatisfiedQueries(this.getManager());
				}
			}
		}
	}

	/**
	 * Calls the manager to detect matches of the
	 * left-hand side conjunction of the input dependency to the facts in this.state.
	 * The manager detects homomorphisms using a database backend.
	 * We search for homomorphisms only in non-blocked bags.
	 * The match found must not be constraint in a single bag.
	 * @param dependency IC
	 * @return List<Matching>
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#getHomomorphisms(Constraint)
	 */
	@Override
	public List<Match> getHomomorphisms(Constraint dependency) {
		Collection<Bag> targetBags = new ArrayList<>();
		for (Bag bag : this.getTree().vertexSet()) {
			if (bag.getType() == BagStatus.NONBLOCKED) {
				targetBags.add(bag);
			}
		}
		return this.getManager().getMatches(
				dependency,
				HomomorphismConstraint.scope(Conjunction.of(this.getFacts())),
				HomomorphismConstraint.scope(true, targetBags.toArray(new Bag[targetBags.size()])));
	}

	/**
	 * @param dependencies Collection<D>
	 * @return Map<D,List<Matching>>
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#getHomomorphisms(Collection<D>)
	 */
	@Override
	public <D extends Constraint> Map<D, List<Match>> getHomomorphisms(Collection<D> dependencies) {
		Collection<Bag> targetBags = new ArrayList<>();
		for (Bag bag : this.getTree().vertexSet()) {
			if (bag.getType() == BagStatus.NONBLOCKED) {
				targetBags.add(bag);
			}
		}
		return this.getManager().getMatches(
				dependencies,
				HomomorphismConstraint.scope(Conjunction.of(this.getFacts())),
				HomomorphismConstraint.scope(true, targetBags.toArray(new Bag[targetBags.size()])));
	}


	/**
	 * Calls the manager to detect matches of the accessible query to facts.
	 * The manager detects homomorphisms using a database backend.
	 * @param query Query
	 * @param constraints Map<Variable,Term>
	 * @return List<Matching>
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#matchesQuery(Query, Map<Variable,Term>)
	 */
	@Override
	public List<Match> matchesQuery(Query<?> query) {
		Collection<Bag> targetBags = new ArrayList<>();
		for (Bag bag:this.getTree().vertexSet()) {
			targetBags.add(bag);
		}
		return this.getManager().getMatches(
				query,
				HomomorphismConstraint.topK(1),
				HomomorphismConstraint.scope(Conjunction.of(this.getFacts())),
				HomomorphismConstraint.scope(false, targetBags.toArray(new Bag[targetBags.size()])),
				HomomorphismConstraint.satisfy(query.getFreeToCanonical()));
	}

	/**
	 * @return BagsTree<ExtendedBag>
	 */
	public BagsTree<ExtendedBag> getTree() {
		return this.tree;
	}

	/**
	 * @param bagsToUpdate Collection<ExtendedBag>
	 */
	public void setBagsToUpdate(Collection<ExtendedBag> bagsToUpdate) {
		this.bagsToUpdate.addAll(bagsToUpdate);
	}

	public void resetBagsToUpdate() {
		this.bagsToUpdate.clear();
	}

	/**
	 * @return BagFactory
	 */
	public BagFactory getBagFactory() {
		return this.bagFactory;
	}

	/**
	 * @param s S
	 * @return S
	 * @see uk.ac.ox.cs.pdq.planner.reasoning.chase.state.ChaseState#merge(S)
	 */
	@Override
	public <S extends ChaseState> S merge(S s) {
		if(s instanceof DatabaseTreeState) {
			MergeTreesImplementation merger = new MergeTreesImplementation(this.getManager());
			Pair<BagsTree<ExtendedBag>, ExtendedBag> treeRoot =
					merger.merge(this.tree,
							((DatabaseTreeState) s).getTree()
							);
			Collection<PredicateFormula> copiedFacts = merger.getCopiedFacts();
			S state = (S) new DatabaseTreeState(
					this.getManager(),
					treeRoot.getLeft(),
					this.bagFactory.clone(),
					false,
					this.getFiringGraph().merge(s.getFiringGraph()),
					new LinkedHashSet<ExtendedBag>());

			state.addFacts(copiedFacts);
			((DatabaseTreeState)state).setBagsToUpdate(merger.getUpdatedBags());
			return state;
		}
		throw new java.lang.IllegalArgumentException();
	}

	/**
	 * For each input accessibility axiom, it
	 * groups the corresponding state facts based on the chase constants assigned to their input positions.
	 * This function is called when creating the initial ApplyRule configurations
	 * @param axioms
	 *
	 * @return
	 * 		pairs of accessibility axioms to chase facts
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#groupByBinding(Collection<AccessibilityAxiom>)
	 */
	@Override
	public List<Pair<AccessibilityAxiom, Collection<PredicateFormula>>> groupByBinding(Collection<AccessibilityAxiom> axioms) {
		Multimap<Signature, PredicateFormula> signatureToSchemaFact = LinkedHashMultimap.create();
		for (Entry<Signature, PredicateFormula> entry: this.getSignatureToSchemaFact().entries()) {
			Signature signature = entry.getKey();
			PredicateFormula fact = entry.getValue();
			Preconditions.checkArgument(this.tree.getBag(fact) != null);
			signatureToSchemaFact.put(signature, new BagBoundPredicate(fact, this.tree.getBag(fact).getId()));
		}
		return this.groupByBinding(axioms, signatureToSchemaFact);
	}

	/**
	 * @param accessibleSchema
	 * @param axioms
	 * @return
	 * 		the unexposed facts and information to expose them
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#getUnExposedFacts(AccessibleSchema, Collection<AccessibilityAxiom>)
	 */
	@Override
	public Map<AccessibilityAxiom, List<Match>> getUnExposedFacts(AccessibleSchema accessibleSchema) {
		Multimap<Signature, PredicateFormula> signatureToSchemaFact = LinkedHashMultimap.create();
		for (Entry<Signature, PredicateFormula> entry: this.getSignatureToSchemaFact().entries()) {
			Signature signature = entry.getKey();
			PredicateFormula fact = entry.getValue();
			Preconditions.checkArgument(this.tree.getBag(fact) != null);
			signatureToSchemaFact.put(signature, new BagBoundPredicate(fact, this.tree.getBag(fact).getId()));
		}
		return this.getUnExposedFacts(accessibleSchema, signatureToSchemaFact, this.getTermToAccessibleFact());
	}

	/**
	 * @param schema AccessibleSchema
	 * @param axiom AccessibilityAxiom
	 * @param facts Collection<PredicateFormula>
	 * @see uk.ac.ox.cs.pdq.chase.state.ChaseState#generate(AccessibleSchema, AccessibilityAxiom, Collection<PredicateFormula>)
	 */
	@Override
	public void generate(AccessibleSchema schema, AccessibilityAxiom axiom, Collection<PredicateFormula> facts) {
		Collection<PredicateFormula> generatedFacts = this.generateFacts(schema, axiom, facts);
		Collection<BagBoundPredicate> toPropagate = this.toPropagate(generatedFacts);
		this.addFacts(toPropagate);
	}

	/**
	 * @param o Object
	 * @return boolean
	 */
	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null) {
			return false;
		}
		return this.getClass().isInstance(o)
				&& this.tree.equals(((DatabaseTreeState) o).tree);
	}

	/**
	 * @return int
	 */
	@Override
	public int hashCode() {
		return Objects.hash(this.tree);
	}
}
