package uk.ac.ox.cs.pdq.runtime.exec;

import java.util.Properties;

import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.plan.AccessOperator;
import uk.ac.ox.cs.pdq.plan.Plan;
import uk.ac.ox.cs.pdq.plan.Plan.ControlFlows;
import uk.ac.ox.cs.pdq.runtime.RuntimeParameters;

import com.google.common.base.Preconditions;

/**
 * Go to class for executing plan.
 * This class can be seen as a PlanExecutor factory, i.e. it will choose the
 * right PlanExecutor implementation to use for a given plan and query, and 
 * properly instantiate it.
 * 
 * @author Efi Tsamoura
 * @author Julien Leblay
 */
public final class Middleware {
	
	/**
	 * @param p
	 * @param q
	 * @param params RuntimeParameters
	 * @return a plan execution executor that is appropriate for the given plan
	 * type.
	 * @throws MiddlewareException
	 */
	public static PlanExecutor newExecutor(RuntimeParameters params, Plan p, Query<?> q) throws MiddlewareException {
		Preconditions.checkArgument(p != null, "Cannot execute null plan");
		Properties properties = findRelationalProperties(p);
		switch(params.getExecutorType()) {
		case SQL_TREE: 
			if ((p.getControlFlow() != ControlFlows.BOTTOM_UP)
				|| (properties == null || properties.isEmpty())) {
				throw new MiddlewareException("SQL plan executor cannot be used for plan " + p);
			}
			return new SQLPlanExecutor(p, q, params.getSemantics(), properties);
		case SQL_STEP: 
			if ((p.getControlFlow() != ControlFlows.BOTTOM_UP)
				|| (properties == null || properties.isEmpty())) {
				throw new MiddlewareException("SQL-step plan executor cannot be used for plan " + p);
			}
			return new SQLStepPlanExecutor(p, q, params.getSemantics(), properties);
		case SQL_WITH: 
			if ((p.getControlFlow() != ControlFlows.BOTTOM_UP)
				|| (properties == null || properties.isEmpty())
				|| !properties.containsKey("url")
				|| !properties.getProperty("url").contains("postgres")) {
				throw new MiddlewareException("SQL-with plan executor cannot be used for plan " + p);
			}
			return new SQLWithPlanExecutor(p, q, params.getSemantics(), properties);
		default: 
			return new PipelinedPlanExecutor(p, q, params.getSemantics(), params.getTimeout().longValue());
		}
	}

	/**
	 * @param plan Plan
	 * @return the schema underlying relational's properties if all of the
	 * relations in the schema have the same properties, null otherwise.
	 */
	private static Properties findRelationalProperties(Plan plan) {
		Properties result = null;
		for (AccessOperator access: plan.getAccesses()) {
			Properties properties = new Properties();
			properties.putAll(access.getRelation().getProperties());
			properties.remove(Relation.PropertyKeys.METADATA);
			if (result == null) {
				result = properties;
			} else if (!(properties.equals(result))) {
				return null;
			}
		}
		return result;
	}

}
