package uk.ac.ox.cs.pdq.runtime.exec;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.algebra.RelationalOperator;
import uk.ac.ox.cs.pdq.fol.Query;
import uk.ac.ox.cs.pdq.plan.Plan;
import uk.ac.ox.cs.pdq.plan.Plan.ControlFlows;
import uk.ac.ox.cs.pdq.runtime.EvaluationException;
import uk.ac.ox.cs.pdq.runtime.RuntimeParameters.Semantics;
import uk.ac.ox.cs.pdq.runtime.TimeoutException;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Distinct;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.SurrogateIterator;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.TupleIterator;
import uk.ac.ox.cs.pdq.util.BooleanResult;
import uk.ac.ox.cs.pdq.util.Result;
import uk.ac.ox.cs.pdq.util.Table;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.eventbus.EventBus;

/**
 * Alternate executor for Linear plan, which first translates a linear plan into
 * a logical relation tree, then translates the logical tree plan into a
 * physical operator tree, and runs it in a pipelined fashion.
 * 
 * @author Julien Leblay
 * 
 */
public class PipelinedPlanExecutor implements PlanExecutor {

	/** Logger. */
	private static Logger log = Logger.getLogger(PipelinedPlanExecutor.class);

	private Table universalTable = null;
	private final Plan plan;
	private final Query<?> query;
	private final Semantics semantics;
	private EventBus eventBus;
	private Long timeout;
	private Long responseTime = null;
	private Long startTime = null;
	private int tuplesLimit = Integer.MAX_VALUE;

	Map<RelationalOperator, SurrogateIterator> accessCache = new LinkedHashMap<>();

	/**
	 * Default constructor
	 * 
	 * @param plan
	 * @param query
	 * @param sem Semantics
	 * @param timeout Long
	 */
	public PipelinedPlanExecutor(Plan plan, Query<?> query, Semantics sem, Long timeout) {
//		Preconditions.checkArgument(plan instanceof LinearPlan ?
//				(query.isBoolean() || ((LinearPlan) plan).getLast().getOperator() instanceof Projection):
//				true, "Linear plan does not end with a Projection");
		this.plan = plan;
		this.query = query;
		this.semantics = sem;
		this.timeout = timeout;
	}

	/**
	 * Default constructor
	 * 
	 * @param plan
	 * @param query
	 * @param sem Semantics
	 */
	public PipelinedPlanExecutor(Plan plan, Query<?> query, Semantics sem) {
		this(plan, query, sem, Long.MAX_VALUE);
	}

	/**
	 * @param eventBus EventBus
	 * @see uk.ac.ox.cs.pdq.runtime.exec.PlanExecutor#setEventBus(EventBus)
	 */
	@Override
	public void setEventBus(EventBus eventBus) {
		this.eventBus = eventBus;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see uk.ac.ox.cs.pdq.runtime.plan.PlanExecutor#execute()
	 */
	@Override
	public Result execute() throws EvaluationException {
		return this.execute(ExecutionModes.DEFAULT);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see uk.ac.ox.cs.pdq.runtime.plan.PlanExecutor#execute(ExecutionModes mode)
	 */
	@Override
	public Result execute(ExecutionModes mode) throws EvaluationException {
		this.startTime = System.currentTimeMillis();
		RelationalOperator logOp = this.plan.getEffectiveOperator();
		ControlFlows controlFlow = this.plan.getControlFlow();
		// Boolean query
		if (this.query.isBoolean()) {
			logOp = new uk.ac.ox.cs.pdq.algebra.IsEmpty(logOp);
			try (TupleIterator phyOp = PlanTranslator.translate(logOp, controlFlow)) {
				phyOp.open();
				return new BooleanResult(!((boolean) phyOp.next().getValue(0)));
			}
		}
		// Non-boolean query
		try (TupleIterator phyOp = PlanTranslator.translate(logOp, controlFlow);
				TupleIterator top = (this.semantics == Semantics.SET
						? new Distinct(phyOp)
						: phyOp)) {
			this.universalTable = new Table(top.getColumns());
			ExecutorService execService = Executors.newFixedThreadPool(1);
			execService.execute(new TimeoutChecker(this.timeout, top));
			switch (mode) {
			case PROFILE:
				top.setEventBus(this.eventBus);
				break;
			}
			
			top.open();
			while (top.hasNext() && this.universalTable.size() < this.tuplesLimit) {
				Tuple t = top.next();
				this.universalTable.appendRow(t);
				this.attemptPost(t);
				this.estimateResponseTime();
				log.info(t + "\t" + this.universalTable.size());
			}
			this.estimateResponseTime();
			execService.shutdownNow();
			if (top.isInterrupt()) {
				throw new TimeoutException();
			}
			this.universalTable.setHeader(Utility.termsToAttributes(
					this.query.getFree(), this.universalTable.getType()));
		}
		return this.universalTable;
	}

	/**
	 * @param t Tuple
	 */
	private void attemptPost(Tuple t) {
		assert t != null : "RuntimeParameters of attemptPost cannot be null.";
		if (this.eventBus != null) {
			this.eventBus.post(t);
		}
	}
	
	private void estimateResponseTime() {
		if(this.responseTime == null) {
			this.responseTime = System.currentTimeMillis() - this.startTime;
		}
	}
	
	/**
	 * Internal evaluation timeout mechanism.
	 * 
	 * @author Julien Leblay
	 *
	 */
	public class TimeoutChecker extends Thread {

		private final long timeout;
		private final TupleIterator iterator;
		
		/**
		 * Constructor for TimeoutChecker.
		 * @param timeout long
		 * @param i TupleIterator
		 */
		public TimeoutChecker(long timeout, TupleIterator i) {
			super("TimeoutChecker");
			this.setDaemon(true);
			this.timeout = timeout;
			this.iterator = i;
		}
		
		/**
		 * @see java.lang.Runnable#run()
		 */
		@Override
		public void run() {
			if (0L < this.timeout
				&& this.timeout < Long.MAX_VALUE) {
				try {
					Thread.sleep(this.timeout);
					this.iterator.interrupt();
				} catch (InterruptedException e) {
					log.debug(e);
					// Nothing to do.
				}
			}
		}
	}

	@Override
	public Long getResponseTime() {
		return this.responseTime;
	}

	@Override
	public void setTuplesLimit(int tuples) {
		this.tuplesLimit = tuples;
	}
}
