package uk.ac.ox.cs.pdq.runtime.exec;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.pdq.algebra.DependentAccess;
import uk.ac.ox.cs.pdq.algebra.DependentJoin;
import uk.ac.ox.cs.pdq.algebra.NaryOperator;
import uk.ac.ox.cs.pdq.algebra.RelationalOperator;
import uk.ac.ox.cs.pdq.algebra.UnaryOperator;
import uk.ac.ox.cs.pdq.algebra.predicates.AttributeEqualityPredicate;
import uk.ac.ox.cs.pdq.algebra.predicates.ConjunctivePredicate;
import uk.ac.ox.cs.pdq.db.AccessMethod;
import uk.ac.ox.cs.pdq.db.AccessMethod.AccessMethodTypes;
import uk.ac.ox.cs.pdq.db.Relation;
import uk.ac.ox.cs.pdq.db.wrappers.RelationAccessWrapper;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.plan.Plan.ControlFlows;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Access;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.AsymmetricMemoryHashJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Count;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.CrossProduct;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Distinct;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.IsEmpty;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.LasyCachedIterator;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.MemoryScan;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.NestedLoopJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Projection;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Scan;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Selection;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.SurrogateIterator;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.SymmetricMemoryHashJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.TupleIterator;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Union;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenAccess;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenBindJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenCrossProduct;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenDistinct;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenIsEmpty;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenMemoryScan;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenNestedLoopJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenProjection;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenScan;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenSelection;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenSymmetricMemoryHashJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.open.OpenTupleIterator;
import uk.ac.ox.cs.pdq.util.Pipelineable;
import uk.ac.ox.cs.pdq.util.Typed;
import uk.ac.ox.cs.pdq.util.Utility;

import com.google.common.base.Preconditions;

/**
 * Translate logical plans to physical relation plans.
 * 
 * @author Julien Leblay
 *
 */
public class PlanTranslator {
	
	/**
	 * @param logOp LogicalOperator
	 * @return TupleIterator
	 */
	public static TupleIterator translate(RelationalOperator logOp) {
		return translate(logOp, ControlFlows.BOTTOM_UP);
	}
	
	/**
	 * @param logOp LogicalOperator
	 * @param variant ControlFlows
	 * @return TupleIterator
	 */
	public static TupleIterator translate(RelationalOperator logOp, ControlFlows variant) {
		Preconditions.checkArgument(variant != null);
		switch (variant) {
		case BOTTOM_UP:
			return toBottomUpPlan(logOp);
		default:
			return toTopDownPlan(logOp);
		}
	}

	/**
	 * Translate a logical plan to a bottom-up physical plan.
	 * @param logOp
	 * @return a physical that corresponds exactly to the given logical plan.
	 * In particular, not further optimization is applied to the resulting plan.
	 * It uses hash join as default for equijoins and OpenNestedLoopJoin for 
	 * joins with arbitrary predicates.
	 */
	private static OpenTupleIterator toTopDownPlan(RelationalOperator logOp) {
		Preconditions.checkArgument(logOp != null);
		OpenTupleIterator[] children = null;
		
		if (logOp instanceof DependentAccess) {
			Relation r = ((DependentAccess) logOp).getRelation(); 
			AccessMethod b = ((DependentAccess) logOp).getBinding();
			if (b.getType() == AccessMethodTypes.FREE) {
				return new OpenScan((RelationAccessWrapper) r);
			}
			return new OpenAccess((RelationAccessWrapper) r, b, ((DependentAccess) logOp).getStaticInputs());

		} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Scan) {
			Relation r = ((uk.ac.ox.cs.pdq.algebra.Scan) logOp).getRelation(); 
			return new OpenScan((RelationAccessWrapper) r);

		} else if (logOp instanceof UnaryOperator) {
			children = new OpenTupleIterator[1];
			children[0] = toTopDownPlan(((UnaryOperator) logOp).getChild());

			if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Projection) {
				List<Typed> columns = Utility.termsToTyped(((uk.ac.ox.cs.pdq.algebra.Projection) logOp).getProjected(), logOp.getType());
				Map<Integer, Term> logicalRenaming = ((uk.ac.ox.cs.pdq.algebra.Projection) logOp).getRenaming();
				if (logicalRenaming != null) {
					Map<Integer, Typed> renaming = new LinkedHashMap<>();
					for (Integer i: logicalRenaming.keySet()) {
						renaming.put(i, Utility.termToTyped(logicalRenaming.get(i), ((UnaryOperator) logOp).getChild().getType().getType(i)));
					}
					return new OpenProjection(columns, renaming, children[0]);
				}
				return new OpenProjection(columns, children[0]);

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Selection) {
				return new OpenSelection(
						children[0],
						((uk.ac.ox.cs.pdq.algebra.Selection) logOp).getPredicate());

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Distinct) {
				return new OpenDistinct(children[0]);

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.IsEmpty) {
				return new OpenIsEmpty(children[0]);

			}			
		} else if (logOp instanceof NaryOperator) {
			List<RelationalOperator> c = ((NaryOperator) logOp).getChildren();
			children = new OpenTupleIterator[c.size()];
			for (int i = 0, l = children.length; i < l; i++) {
				children[i] = toTopDownPlan(c.get(i));
			}

			if (logOp instanceof DependentJoin && ((DependentJoin) logOp).hasSideWaysInputs()) {
				return new OpenBindJoin(children[0], children[1],
						((DependentJoin) logOp).getPredicate(), ((DependentJoin) logOp).getSidewaysInput());

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Join) {
				uk.ac.ox.cs.pdq.algebra.Join join = (uk.ac.ox.cs.pdq.algebra.Join) logOp;
				List<Typed> inputsColumns = Utility.termsToTyped(logOp.getInputTerms(), logOp.getInputType());
				switch(join.getVariant()) {
				case NESTED_LOOP:
					return new OpenNestedLoopJoin(
							inputsColumns,
							join.getPredicate(), children);
				default:
					return new OpenSymmetricMemoryHashJoin(
							inputsColumns,
							(ConjunctivePredicate<AttributeEqualityPredicate>) join.getPredicate(),
							children[0], children[1]);
				}

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.CrossProduct) {
				List<Typed> inputsColumns = Utility.termsToTyped(logOp.getInputTerms(), logOp.getInputType());
				return new OpenCrossProduct(
						inputsColumns,
						children);

			}

		} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.StaticInput) {
			List<Typed> columns = Utility.termsToTyped(logOp.getColumns(), logOp.getType());
			return new OpenMemoryScan(
					columns, ((uk.ac.ox.cs.pdq.algebra.StaticInput) logOp).getTuples());



		} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.SubPlanAlias) {
			return toTopDownPlan((RelationalOperator) ((uk.ac.ox.cs.pdq.algebra.SubPlanAlias) logOp).getPlan().getOperator());

		}  
		throw new IllegalArgumentException("Unsupported logical operator " + logOp);
	}
	
	/**
	 * Translate a logical plan to a bottom-up physical plan.
	 * @param logOp
	 * @return a physical that corresponds exactly to the given logical plan.
	 * In particular, not further optimization is applied to the resulting plan.
	 * It uses hash join as default for equijoins and OpenNestedLoopJoin for 
	 * joins with arbitrary predicates.
	 */
	private static TupleIterator toBottomUpPlan(RelationalOperator logOp) {
		Preconditions.checkArgument(logOp != null);
		TupleIterator[] children = null;
		
		if (logOp instanceof UnaryOperator) {
			children = new TupleIterator[1];
			children[0] = toBottomUpPlan(((UnaryOperator) logOp).getChild());

			if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Access) {
				Relation r = ((uk.ac.ox.cs.pdq.algebra.Access) logOp).getRelation(); 
				AccessMethod b = ((uk.ac.ox.cs.pdq.algebra.Access) logOp).getBinding();
				Preconditions.checkState(r instanceof RelationAccessWrapper, "Relation '" + r + "' cannot be used to produce an access physical operator.");
				return new Access((RelationAccessWrapper) r, b, children[0]);

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Projection) {
				List<Typed> columns = Utility.termsToTyped(((uk.ac.ox.cs.pdq.algebra.Projection) logOp).getProjected(), logOp.getType());
				Map<Integer, Term> logicalRenaming = ((uk.ac.ox.cs.pdq.algebra.Projection) logOp).getRenaming();
				if (logicalRenaming != null) {
					Map<Integer, Typed> renaming = new LinkedHashMap<>();
					for (Integer i: logicalRenaming.keySet()) {
						renaming.put(i, Utility.termToTyped(logicalRenaming.get(i), ((UnaryOperator) logOp).getChild().getType().getType(i)));
					}
					return new Projection(columns, renaming, children[0]);
				}
				return new Projection(columns, children[0]);

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Selection) {
				return new Selection(
						children[0],
						((uk.ac.ox.cs.pdq.algebra.Selection) logOp).getPredicate());

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Distinct) {
				return new Distinct(children[0]);

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Count) {
				return new Count(children[0]);

			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.IsEmpty) {
				return new IsEmpty(children[0]);
			}
			
		} else if (logOp instanceof NaryOperator) {
			List<RelationalOperator> c = ((NaryOperator) logOp).getChildren();
			children = new TupleIterator[c.size()];
			for (int i = 0, l = children.length; i < l; i++) {
				children[i] = toBottomUpPlan(c.get(i));
			}

			if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Join) {
				uk.ac.ox.cs.pdq.algebra.Join join = (uk.ac.ox.cs.pdq.algebra.Join) logOp;
				switch(join.getVariant()) {
				case NESTED_LOOP:
					return new NestedLoopJoin(join.getPredicate(), children);
				case ASYMMETRIC_HASH:
					return new AsymmetricMemoryHashJoin(
							(ConjunctivePredicate) join.getPredicate(),
							children[0], children[1]);
				default:
					return new SymmetricMemoryHashJoin(
							(ConjunctivePredicate) join.getPredicate(),
							children[0], children[1]);
				}
			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.CrossProduct) {
				return new CrossProduct(children);
			} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Union) {
				return new Union(children);
			}

		} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.Scan) {
			Relation r = ((uk.ac.ox.cs.pdq.algebra.Scan) logOp).getRelation();
			Preconditions.checkState(r instanceof RelationAccessWrapper, "Relation '" + r + "' cannot be used to produce a scan physical operator.");
			return new Scan((RelationAccessWrapper) r);

		} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.StaticInput) {
			List<Typed> columns = Utility.termsToTyped(logOp.getColumns(), logOp.getType());
			return new MemoryScan(
					columns,
					((uk.ac.ox.cs.pdq.algebra.StaticInput) logOp).getTuples());

		} else if (logOp instanceof uk.ac.ox.cs.pdq.algebra.SubPlanAlias) {
			return toBottomUpPlan((RelationalOperator) ((uk.ac.ox.cs.pdq.algebra.SubPlanAlias) logOp).getPlan().getOperator());

		}
		throw new IllegalArgumentException("Unsupported logical operator " + logOp);
	}
}
