package uk.ac.ox.cs.pdq.runtime.exec.iterator;

import java.util.ArrayList;
import java.util.List;

import uk.ac.ox.cs.pdq.db.AccessMethod;
import uk.ac.ox.cs.pdq.db.Attribute;
import uk.ac.ox.cs.pdq.db.wrappers.RelationAccessWrapper;
import uk.ac.ox.cs.pdq.util.ResetableIterator;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;

/**
 * Access over a relation
 * 
 * @author Julien Leblay
 */
public class Access extends UnaryIterator {


	/** The input table of the access. */
	private final RelationAccessWrapper relation;
	
	/** The access method to use. */
	private final AccessMethod binding;
	
	/** Iterator over the output tuples. */
	protected ResetableIterator<Tuple> outputs = null;
	
	/**
	 * Instantiates a new join.
	 * 
	 * @param relation RelationAccessWrapper
	 * @param binding AccessMethod
	 * @param child TupleIterator
	 */
	public Access(RelationAccessWrapper relation, AccessMethod binding, TupleIterator child) {
		super(Lists.<Typed>newArrayList(relation.getAttributes()), child);
		Preconditions.checkArgument(relation != null);
		Preconditions.checkArgument(binding != null);
		this.relation = relation;
		this.binding = binding;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName());
		result.append('[').append(this.relation.getName()).append('/');
		result.append(this.binding).append(']');
		result.append('(').append(this.child).append(')');
		return result.toString();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#open()
	 */
	@Override
	public void open() {
		super.open();
		this.outputs = null;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#reset()
	 */
	@Override
	public void reset() {
		super.reset();
		this.outputs = null;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#hasNext()
	 */
	@Override
	public boolean hasNext() {
		Preconditions.checkArgument(this.open);
		if (this.outputs == null) {
			this.init();
		}
		return this.outputs.hasNext();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#next()
	 */
	@Override
	public Tuple next() {
		Preconditions.checkArgument(this.open);
		if (this.outputs == null) {
			this.init();
		}
		return this.outputs.next();
	}

	/**
	 * Initializes the operator by performing the access and preloading the
	 * results into iteratorCache.
	 */
	private void init() {
		List<Attribute> inputAttributes = new ArrayList<>();
		List<Typed> outputAttributes = this.getColumns();
		for (Integer i: this.binding.getInputs()) {
			Typed att = outputAttributes.get(i - 1);
			if (!(att instanceof Attribute)) {
				throw new IllegalStateException();
			}
			inputAttributes.add((Attribute) att);
		}
		this.outputs = this.relation.iterator(inputAttributes, this.child);
		this.outputs.open();
	}
	
	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#deepCopy()
	 */
	@Override
	public Access deepCopy() {
		return new Access(this.relation, this.binding, this.child.deepCopy());
	}
}