package uk.ac.ox.cs.pdq.runtime.exec.iterator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.log4j.Logger;

import uk.ac.ox.cs.pdq.algebra.predicates.AttributeEqualityPredicate;
import uk.ac.ox.cs.pdq.algebra.predicates.ConjunctivePredicate;
import uk.ac.ox.cs.pdq.fol.Term;
import uk.ac.ox.cs.pdq.util.Tuple;

/**
 * AsymmetricMemoryHashJoin inspired from the ViP2P implementation. 
 * The buckets are build on the left operator, and the iteration run over the 
 * right one.
 * 
 * @author Julien Leblay
 *
 */
public class AsymmetricMemoryHashJoin extends Join {

	/** The logger */
	private static Logger log = Logger.getLogger(AsymmetricMemoryHashJoin.class);
	
	/** The hash table. */
	protected Map<JoinKey, List<Tuple>> hashTable = new LinkedHashMap<>();

	/** Tuple iterator on the current bucket. */
	protected Iterator<Tuple> bucketIterator = null;
	
	/** The right part of the next tuple to return */
	protected Tuple rightTuple = null;

	/** The join keys for the left child. */
	protected Integer[] leftKeys;

	/** The join keys for the right child. */
	protected Integer[] rightKeys;
	
	/** Index of the last child */
	protected int lastChild = 1;
	
	/** true, if this built phase has already completed. */
	protected boolean isBuilt = false;
	
	/**
	 * Constructor an unbound array of children
	 * @param left TupleIterator
	 * @param right TupleIterator
	 */
	public AsymmetricMemoryHashJoin(
			TupleIterator left, TupleIterator right) {
		super(left, right);
		this.leftKeys = this.makeLeftKey();
		this.rightKeys = this.makeRightKey(left.getType().size());
	}
	
	/**
	 * Constructor an unbound array of children
	 * @param predicate AttributeEqualityPredicate
	 * @param left TupleIterator
	 * @param right TupleIterator
	 */
	public AsymmetricMemoryHashJoin(AttributeEqualityPredicate predicate,
			TupleIterator left, TupleIterator right) {
		this(new ConjunctivePredicate<>(predicate), left, right);
	}
	
	/**
	 * Constructor an unbound array of children
	 * @param predicate ConjunctivePredicate<AttributeEqualityPredicate>
	 * @param left TupleIterator
	 * @param right TupleIterator
	 */
	public AsymmetricMemoryHashJoin(ConjunctivePredicate<AttributeEqualityPredicate> predicate,
			TupleIterator left, TupleIterator right) {
		super(predicate, left, right);
		this.leftKeys = this.makeLeftKey();
		this.rightKeys = this.makeRightKey(left.getType().size());
	}
	
	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.NaryIterator#open()
	 */
	@Override
	public void open() {
		super.open();
		this.build();
		this.nextTuple();
	}

	/**
	 * Build phase of the hash join. Load in a hash table all tuple by join key
	 */
	protected void build() {
		int nbTuples = 0;
		try (TupleIterator left = this.children.get(0)) {
			while (left.hasNext()) {
				Tuple t = left.next();
				JoinKey joinKey = this.joinKeys(this.leftKeys, t);
				if (joinKey != null) {
					List<Tuple> tuples = this.hashTable.get(joinKey);
					if (tuples == null) {
						this.hashTable.put(joinKey, (tuples = new ArrayList<>()));
					}
					tuples.add(t);
				}
				nbTuples++;
			}
		}
		if (this.hashTable.size() == 0) {
			this.isEmpty = true;
			this.nextTuple = null;
		}
		log.debug(nbTuples + " tuples used during build phase.");
		log.debug("HashTable has " + this.hashTable.size() + " entries.");
		this.isBuilt = true;
	}
	
	/**
	 * @return true, if the built phase has already occurred.
	 */
	public boolean isBuilt() {
		return this.isBuilt;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.Join#nextTuple()
	 */
	@Override
	protected void nextTuple() {
		TupleIterator right = this.children.get(this.lastChild);
		this.nextTuple = null;
		if (this.interrupted) {
			return;
		}
		int rightTuples = 0;
		while (this.nextTuple == null) {
			if (this.bucketIterator != null 
					&& this.bucketIterator.hasNext() && this.rightTuple != null) {
				Tuple t = this.outputType.appendTuples(this.bucketIterator.next(), this.rightTuple);

				this.nextTuple = t;
				return;
			} else if (right.hasNext()) {
				do {
					this.rightTuple = right.next();
					rightTuples++;
					JoinKey joinKey = this.joinKeys(this.rightKeys, this.rightTuple);
					if (joinKey != null) {
						List<Tuple> leftCandidates = this.hashTable.get(joinKey);
						if (leftCandidates != null && !leftCandidates.isEmpty()) {
							this.bucketIterator = leftCandidates.iterator();
							break;
						}
					}
				} while (right.hasNext());
			} else {
				break;
			}
		}
		log.debug("Right tuples: " + rightTuples);
		this.nextTuple = null;
	}

	/**
	 * Build a variableName-to-value binding from the given join key and tuple
	 * @param keys
	 * @param tuple
	 * @return JoinKey
	 */
	protected JoinKey joinKeys(Integer[] keys, Tuple tuple) {
		List<Object> result = new ArrayList<>(keys.length);
		for (Integer i: keys) {
			result.add(tuple.getValue(i));
		}
		return new JoinKey(result);
	}
	
	/**
	 * Returns an integer to Variable maps representing the position in arg1
	 * of each variable appearing in both arg1 and arg2.
	 * 
	 * @param left List<? extends Term>
	 * @param right Collection<? extends Term>
	 * @return the position in arg1 of variable appearing in both arg1 and arg2
	 */
	protected Map<Integer, Term> joinColumns(
			List<? extends Term> left, Collection<? extends Term> right) {
		Map<Integer, Term> result = new LinkedHashMap<>();
		
		List<Term> inter = new ArrayList<>(right);
		inter.retainAll(left);
		for (Term t : inter) {
//			if (t.isVariable()) {
				for (int i = 0, l = left.size(); i < l; i++) {
					if (left.get(i).equals(t)) {
						result.put(i, t);
					}
				}
//			}
		}
		return result;
	}
	
	/**
	 * @return Integer[]
	 */
	protected Integer[] makeLeftKey() {
		List<Integer> result = new ArrayList<>();
		for (AttributeEqualityPredicate p: ((ConjunctivePredicate<AttributeEqualityPredicate>) this.predicate)) {
			result.add(p.getPosition());
		}
		return result.toArray(new Integer[result.size()]);
	}

	/**
	 * @param offset int
	 * @return Integer[]
	 */
	protected Integer[] makeRightKey(int offset) {
		List<Integer> result = new ArrayList<>();
		for (AttributeEqualityPredicate p: ((ConjunctivePredicate<AttributeEqualityPredicate>) this.predicate)) {
			result.add(p.getOther() - offset);
		}
		return result.toArray(new Integer[result.size()]);
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#deepCopy()
	 */
	@Override
	public AsymmetricMemoryHashJoin deepCopy() {
		List<TupleIterator> clones = new ArrayList<>();
		for (TupleIterator child: this.children) {
			clones.add(child.deepCopy());
		}
		return new AsymmetricMemoryHashJoin(
				(ConjunctivePredicate<AttributeEqualityPredicate>) this.predicate,
				this.children.get(0), this.children.get(1));
	}
	
	/**
	 * JoinKey is the representation of a key during the execution of a hash
	 * join. A key is typically a mapping from column names to values.
	 * @author Julien LEBLAY
	 */
	protected class JoinKey {

		/** Mapping between columns names and values */
		private final List<Object> joinValues;
		private final int hashCode;
		
		/**
		 * Instantiates a JoinKey form a tuple of bindings.
		 * @param joinValues
		 */
		JoinKey(List<Object> joinValues) {
			this.joinValues = joinValues;
			this.hashCode = Objects.hash(joinValues);
		}
		
		/**
		 * @return int
		 */
		@Override
		public int hashCode() {
			return this.hashCode;
		}
		
		/**
		 * @param o Object
		 * @return boolean
		 */
		@Override
		public boolean equals(Object o ) {
			if (this == o) {
				return true;
			}
			return this.getClass().isInstance(o)
					&& ((JoinKey) o).joinValues.equals(this.joinValues);
		}
		
		/**
		 * @return String
		 */
		@Override
		public String toString() {
			return this.joinValues.toString();
		}
	}
}
