package uk.ac.ox.cs.pdq.runtime.exec.iterator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import uk.ac.ox.cs.pdq.algebra.predicates.AttributeEqualityPredicate;
import uk.ac.ox.cs.pdq.algebra.predicates.ConjunctivePredicate;
import uk.ac.ox.cs.pdq.algebra.predicates.Predicate;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;

/**
 * Join is a top-level class for all join operators.
 * 
 * @author Julien Leblay
 */
public abstract class Join extends NaryIterator {

	protected Predicate predicate;
	
	/** Determines whether the operator is known to have an empty result. */
	protected boolean isEmpty = false;
	
	/** The next tuple to return. */
	protected Tuple nextTuple = null;

	/**
	 * Instantiates a new join.
	 * 
	 * @param children
	 *            the children
	 */
	protected Join(TupleIterator... children) {
		this(Lists.newArrayList(children));
	}
	
	/**
	 * Instantiates a new join.
	 * 
	 * @param children
	 *            the children
	 */
	protected Join(List<TupleIterator> children) {
		super(children);
		this.initNaturalJoin();
	}
	
	/**
	 * Instantiates a new join.
	 * 
	 * @param children
	 *            the children
	 * @param predicate Predicate
	 */
	protected Join(
			Predicate predicate,
			TupleIterator... children) {
		super(children);
		this.predicate = predicate;
	}
	
	/**
	 * Initialises the join variables.
	 */
	protected void initNaturalJoin() {
		Map<Typed, SortedSet<Integer>> joinVariables = new LinkedHashMap<>();
		int totalCol = 0;
		// Cluster patterns by variables
		for (TupleIterator child : this.children) {
			for (int i = 0; i < child.columns.size(); i++) {
				Typed col = child.getColumns().get(i);
				SortedSet<Integer> joined = joinVariables.get(col);
				if (joined == null) {
					joined = new TreeSet<>();
					joinVariables.put(col, joined);
				}
				joined.add(totalCol);
				totalCol++;
			}
		}
		
		Collection<AttributeEqualityPredicate> equalities = new ArrayList<>();
		// Remove clusters containing only one pattern
		for (Iterator<Typed> keys = joinVariables.keySet().iterator(); keys.hasNext();) {
			Set<Integer> cluster = joinVariables.get(keys.next());
			if (cluster.size() < 2) {
				keys.remove();
			} else {
				Iterator<Integer> i = cluster.iterator();
				Integer left = i.next();
				while (i.hasNext()) {
					Integer right = i.next();
					equalities.add(new AttributeEqualityPredicate(left, right));
				}
			}
		}

		this.predicate = new ConjunctivePredicate<>(equalities);
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.NaryIterator#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName());
		result.append(this.predicate).append('(');
		if (this.children != null) {
			for (TupleIterator child: this.children) {
				result.append(child.toString()).append(',');
			}
			result.deleteCharAt(result.length() - 1);
		}
		result.append(')');
		return result.toString();
	}

	/**
	 * Move the iterator forward and prepares the next tuple to be returned.
	 * 
	 */
	protected abstract void nextTuple() ;

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#hasNext()
	 */
	@Override
	public boolean hasNext() {
		Preconditions.checkArgument(this.open);
		if (this.nextTuple != null) {
			return true;
		}
		if (this.isEmpty) {
			return false;
		}
		this.nextTuple();
		return this.nextTuple != null;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#next()
	 */
	@Override
	public Tuple next() {
		Preconditions.checkArgument(this.open);
		Tuple result = this.nextTuple;
		this.nextTuple = null;
		if ((result == null && !this.hasNext()) || this.isEmpty) {
			throw new NoSuchElementException("End of operator reached.");
		}
		return result;
	}

	/**
	 * @return Predicate
	 */
	public Predicate getPredicate() {
		return this.predicate;
	}
}