package uk.ac.ox.cs.pdq.runtime.exec.iterator;

import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;

import uk.ac.ox.cs.pdq.util.Tuple;

/**
 * An iterator that wraps another iterator. Caches and return tuples from the
 * underlying iterator upon the first traversal. Any subsequent traversal, 
 * occurring after a reset of performed on the cached items. 
 * 
 * @author Julien Leblay
 */
public class LasyCachedIterator extends TupleIterator {

	/** Cached items. */
	protected final Deque<Tuple> cache = new LinkedList<>();

	/** Inner iterator. */
	protected final TupleIterator inner;

	/** True if the iterator has been reset. */
	protected boolean isReset = false;

	/** Cached items. */
	protected Iterator<Tuple> iterator;
	
	/**
	 * Instantiates a new join.
	 * 
	 * @param inner TupleIterator
	 */
	public LasyCachedIterator(TupleIterator inner) {
		super(inner.getType(), inner.getColumns());
		this.inner = inner;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName());
		result.append('(').append(this.inner).append(')');
		return result.toString();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#open()
	 */
	@Override
	public void open() {
		this.isReset = false;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#reset()
	 */
	@Override
	public void reset() {
		this.isReset = true;
		this.iterator = this.cache.iterator();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#hasNext()
	 */
	@Override
	public boolean hasNext() {
		if (this.interrupted) {
			return false;
		}
		if (this.isReset) {
			return this.iterator.hasNext();
		}
		return this.inner.hasNext();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#next()
	 */
	@Override
	public Tuple next() {
		if (this.isReset) {
			return this.iterator.next();
		}
		Tuple result = this.inner.next();
		this.cache.add(result);
		return result;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#remove()
	 */
	@Override
	public void remove() {
		if (!this.isReset) {
			this.cache.removeLast();
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#deepCopy()
	 */
	@Override
	public LasyCachedIterator deepCopy() {
		return new LasyCachedIterator(this.inner);
	}

	/**
	 * @see java.lang.AutoCloseable#close()
	 */
	@Override
	public void close() {
		// nothing to do
	}

	@Override
	public void interrupt() {
		this.interrupted = true;
	}
}