package uk.ac.ox.cs.pdq.runtime.exec.iterator;

import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import uk.ac.ox.cs.pdq.util.TupleType;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.common.eventbus.EventBus;


/**
 * Superclass to all nary logical operator.
 * Factorises the open, close, reset methods and the default constructor.
 * 
 * @author Julien Leblay
 */
public abstract class NaryIterator extends TupleIterator {

	/** The children. */
	protected List<TupleIterator> children;

	/**
	 * Instantiates a new operator.
	 * 
	 * @param children
	 *            the children
	 * @param typeOverride TupleType
	 */
	public NaryIterator(TupleType typeOverride, TupleIterator... children) {
		this(typeOverride, Lists.newArrayList(children));
	}

	/**
	 * Instantiates a new operator.
	 * 
	 * @param children
	 *            the children
	 */
	public NaryIterator(TupleIterator... children) {
		this(Lists.newArrayList(children));
	}
	
	/**
	 * Instantiates a new operator.
	 * 
	 * @param children
	 *            the children
	 */
	public NaryIterator(Collection<TupleIterator> children) {
		this(null, children);
	}
	
	/**
	 * Instantiates a new operator.
	 * 
	 * @param children
	 *            the children
	 * @param typeOverride TupleType
	 */
	public NaryIterator(TupleType typeOverride, Collection<TupleIterator> children) {
		super(typeOverride, null);
		Preconditions.checkArgument(children != null);
		Preconditions.checkArgument(children.size() > 0);
		this.children = ImmutableList.copyOf(children);
		this.columns = ImmutableList.copyOf(this.updateColumns());
		if (typeOverride == null) {
			List<Type> types = new LinkedList<>();
			for (TupleIterator child: children) {
				for (int i = 0, l = child.outputType.size(); i < l; i++) {
					types.add(child.outputType.getType(i));
				}
			}
			this.outputType = TupleType.DefaultFactory.create(types.toArray(new Type[types.size()]));
		}
	}

	/**
	 * Gets the children.
	 * 
	 * @return the children */
	public List<TupleIterator> getChildren() {
		return this.children;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName()).append('(');
		if (this.children != null && !this.children.isEmpty()) {
			for (TupleIterator child: this.children) {
				result.append(child.toString()).append(',');
			}
			result.deleteCharAt(result.length() - 1);
		}
		result.append(')');
		return result.toString();
	}
	
	/**
	 * Ensures the current operator's list of columns is consistent with the
	 * concatenation of lists of columns of its children.
	 * @return List<Typed>
	 */
	protected List<Typed> updateColumns() {
		List<Typed> result = new ArrayList<>();
		for (TupleIterator child: this.children) {
			result.addAll(child.getColumns());
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#open()
	 */
	@Override
	public void open() {
		for (TupleIterator child: this.children) {
			child.open();
		}
		this.open = true;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#close()
	 */
	@Override
	public void close() {
		for (TupleIterator child: this.children) {
			child.close();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#reset()
	 */
	@Override
	public void reset() {
		for (TupleIterator child: this.children) {
			child.reset();
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#setEventBus(com.google.common.eventbus.EventBus)
	 */
	@Override
	public void setEventBus(EventBus eb) {
		super.setEventBus(eb);
		for (TupleIterator child: this.children) {
			child.setEventBus(eb);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#interrupt()
	 */
	@Override
	public void interrupt() {
		this.interrupted = true;
		for (TupleIterator child: this.children) {
			child.interrupt();
		}
	}
}
