package uk.ac.ox.cs.pdq.runtime.exec.iterator;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import uk.ac.ox.cs.pdq.db.Attribute;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.base.Preconditions;

/**
 * Projection operator.
 * 
 * @author Julien Leblay
 */
public class Projection extends UnaryIterator {

	/** The head, the list of term projected by this operator. */
	public List<Typed> projected;

	/** Maps each variable is the head to a position in the children. */
	public Map<Typed, Integer> positions;

	/** Maps each child term position to its (optional renaming). */
	public Map<Integer, Typed> renaming;

	/**
	 * Instantiates a new projection.
	 * 
	 * @param head
	 *            the head
	
	 * @param child TupleIterator
	 */
	public Projection(List<Typed> head, TupleIterator child) {
		this(head, new LinkedHashMap<Integer, Typed>(), child);
	}

	/**
	 * Instantiates a new projection.
	 * 
	 * @param projected
	 *            the head
	
	 * @param renaming Map<Integer,Typed>
	 * @param child TupleIterator
	 */
	public Projection(List<Typed> projected, Map<Integer, Typed> renaming, TupleIterator child) {
		super(applyRenaming(projected, renaming, child.getColumns()), child);
		this.projected = projected != null ? new ArrayList<>(projected) : null;
		this.renaming = renaming;
		this.positions = new LinkedHashMap<>();
		int i = 0;
		for (Typed c : this.child.getColumns()) {
			if (c instanceof Attribute && this.projected.contains(c) && !this.positions.containsKey(c)) {
				this.positions.put(c, i);
			}
			i++;
		}
	}

	/**
	 * Method applyRenaming.
	 * @param projected List<Typed>
	 * @param renaming Map<Integer,Typed>
	 * @param columns List<Typed>
	 * @return List<Typed>
	 */
	private static List<Typed> applyRenaming(List<Typed> projected, Map<Integer, Typed> renaming, List<Typed> columns) {
		if (renaming != null && !renaming.isEmpty()) {
			List<Typed> result = new ArrayList<>(projected.size());
			for (Typed t : projected) {
				if (t instanceof Attribute) {
					int pos = columns.indexOf(t);
					if (renaming.containsKey(pos)) {
						result.add(renaming.get(pos));
					} else {
						throw new IllegalStateException(t + " not found in child attributes list.");
					}
				} else {
					result.add(t);
				}
			}
			return result;
		}
		return projected;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#hasNext()
	 */
	@Override
	public boolean hasNext() {
		Preconditions.checkArgument(this.open);
		return this.child.hasNext();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#next()
	 */
	@Override
	public Tuple next() {
		Preconditions.checkArgument(this.open);
		Tuple next = this.child.next();
		if (next == null) {
			throw new NoSuchElementException("End of projection operator reached.");
		}
		Object[] result = new Object[this.projected.size()];
		int i = 0;
		for (Typed t : this.projected) {
			if (t instanceof Attribute) {
				result[i++] = next.getValue(this.positions.get(t));
			} else if (t instanceof TypedConstant) {
				result[i++] = ((TypedConstant) t).getValue();
			}
		}
		return this.outputType.createTuple(result);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#deepCopy()
	 */
	@Override
	public Projection deepCopy() {
		return new Projection(
				this.projected,
				this.renaming,
				this.child.deepCopy());
	}

	/**
	
	 * @return a list of human readable column headers. */
	@Override
	public List<String> getColumnsDisplay() {
		List<String> result = new ArrayList<>();
		int i = 0;
		for (Typed t : this.projected) {
			String s = t.toString();
			if (this.renaming.containsKey(i)) {
				s += "/" + this.renaming.get(i);
			}
			result.add(s);
			i++;
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName());
		result.append(this.getColumnsDisplay());
		result.append('(').append(this.child.toString()).append(')');
		return result.toString();
	}
}
