package uk.ac.ox.cs.pdq.runtime.exec.iterator;

import java.util.ArrayList;
import java.util.List;

import uk.ac.ox.cs.pdq.util.ResetableIterator;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.TupleType;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.eventbus.EventBus;

/**
 * TupleIterator defines a top-class for all physical operators.
 * 
 * @author Julien Leblay
 */
public abstract class TupleIterator implements AutoCloseable, ResetableIterator<Tuple> {
	
	/** The columns. */
	protected List<Typed> columns;
	
	/** The iterator's type. */
	protected TupleType outputType;
	
	/** Tells whether the operator was voluntarily interrupted */
	protected boolean interrupted = false;
	
	/** Tells whether the operator has been open */
	protected boolean open = false;

	protected EventBus eventBus;
	
	/**
	 * Instantiates a new operator.
	 * 
	 * @param columns
	 *            the columns
	 */
	public TupleIterator(List<Typed> columns) {
		this(TupleType.DefaultFactory.createFromTyped(columns), columns);
	}

	/**
	 * Instantiates a new operator.
	 * 
	 * @param type TupleType
	 * @param columns List<Typed>
	 */
	public TupleIterator(TupleType type, List<Typed> columns) {
		this.outputType = type;
		this.columns = columns;
	}

	/**
	 * Deep copy.
	 * @return a deep copy of the operator.
	 * @throws TupleIteratorException
	 * @see uk.ac.ox.cs.pdq.util.ResetableIterator#deepCopy()
	 */
	@Override
	public abstract TupleIterator deepCopy();

	/**
	 * Gets the column at index i.
	 * 
	 * @param i the index of the column to return.
	 * @return the column at index i.
	 */
	public Typed getColumn(int i) {
		return this.columns.get(i);
	}

	/**
	 * Gets the columns.
	 * 
	 * @return the columns
	 */
	public List<Typed> getColumns() {
		return this.columns;
	}

	/**
	 * Gets the tuple type.
	 * 
	 * @return the type
	 */
	public TupleType getType() {
		return this.outputType;
	}
	
	/**
	 * Sets the given event bus to this operator and all its childrem. 
	 * @param eb EventBus
	 */ 
	public void setEventBus(EventBus eb) {
		this.eventBus = eb;
	}

	/**
	 * Opens the operator. An operator must be opened prior to calling any of
	 * the 'reset', 'hasNext', 'next' and 'close' methods. This method throws an 
	 * exception if called when already opened. 
	 * @see uk.ac.ox.cs.pdq.util.ResetableIterator#open()
	 */ 
	@Override
	public abstract void open();
	
	/**
	 * Places the result iterator back to the first items. This method may not
	 * be implemented by all subclasses, in which case it throws a 
	 * MethodNotSupportedException.
	 * @see uk.ac.ox.cs.pdq.util.ResetableIterator#reset()
	 */
	@Override
	public abstract void reset();
	
	/**
	 * Checks whether there is a next tuple in the result.
	 * 
	 * @return true, if the operator has a next result.
	 * @see java.util.Iterator#hasNext()
	 */
	@Override
	public abstract boolean hasNext();
	
	/**
	 * Returns the next tuple in the result.
	 * 
	 * @return the next tuple in the result.
	 * @see java.util.Iterator#next()
	 */
	@Override
	public abstract Tuple next();
	
	/**
	 * @throws UnsupportedOperationException
	 * @see java.util.Iterator#remove()
	 */
	@Override
	public void remove() {
		throw new UnsupportedOperationException();
	}
	
	/**
	 * Closes the operator. This method throws an exception if called when 
	 * already closed. 
	 * @see java.lang.AutoCloseable#close()
	 */
	@Override
	public abstract void close();
	
	
	/**
	 * Interrupted the operator. This method throws an exception if called when 
	 * already closed.
	 */
	public abstract void interrupt();
	
	/**
	 * Tells whether the operator was voluntarily interrupted 
	 * @return boolean
	 */
	public boolean isInterrupt() {
		return this.interrupted;
	}
	
	/**
	 * @return a list of human readable column headers.
	 */
	public List<String> getColumnsDisplay() {
		List<String> result = new ArrayList<>();
		for (Typed t: this.getColumns()) {
			result.add(t.toString());
		}
		return result;
	}
}
