package uk.ac.ox.cs.pdq.runtime.exec.iterator;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;

import uk.ac.ox.cs.pdq.util.Tuple;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;

/**
 * Union the results of its children
 * 
 * @author Julien Leblay
 */
public class Union extends NaryIterator {

	/** The iteratorCache */
	private Set<Tuple> cache = new LinkedHashSet<>();

	/** The next result to return */
	private Tuple nextResult = null;

	/** The iterator of the current operator's children */
	private Iterator<TupleIterator> childIterator = null;

	/** The iterator of the current children */
	private TupleIterator currentChild = null;
	
	/**
	 * Instantiates a new operator.
	 * 
	 * @param children
	 *            the children
	 */
	public Union(List<TupleIterator> children) {
		super(children);
		if (this.children.isEmpty()) {
			throw new IllegalArgumentException("Attempting to instantiate union operator with an empty list of children.");
		}
	}

	/**
	 * Instantiates a new union.
	 * 
	 * @param children
	 *            the children
	 */
	public Union(TupleIterator... children) {
		this(Lists.newArrayList(children));
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#deepCopy()
	 */
	@Override
	public Union deepCopy() {
		List<TupleIterator> clones = new ArrayList<>();
		for (TupleIterator child: this.children) {
			clones.add(child.deepCopy());
		}
		return new Union(clones);
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.NaryIterator#open()
	 */
	@Override
	public void open() {
		this.childIterator = this.children.iterator();
		this.currentChild = this.childIterator.next();
		this.currentChild.open();
		this.nextTuple();
	}

	/**
	 * Prepares the next tuple to return. If the end of the operator is reached,
	 * nextResult shall be null.
	 */
	private void nextTuple() {
		while (this.currentChild.hasNext()) {
			this.nextResult = this.currentChild.next();
			if (!this.cache.contains(this.nextResult)) {
				this.cache.add(this.nextResult);
				return;
			}
		}
		while (this.childIterator.hasNext()) {
			this.currentChild = this.childIterator.next();
			this.currentChild.open();
			if (this.currentChild.hasNext()) {
				this.nextResult = this.currentChild.next();
				if (!this.cache.contains(this.nextResult)) {
					this.cache.add(this.nextResult);
					return;
				}
			}
		}
		this.nextResult = null;
	}
	
	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#hasNext()
	 */
	@Override
	public boolean hasNext() {
		Preconditions.checkArgument(this.open);
		return this.nextResult != null;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#next()
	 */
	@Override
	public Tuple next() {
		Preconditions.checkArgument(this.open);
		if (this.nextResult == null) {
			throw new NoSuchElementException("End of operator reached.");
		}
		Tuple result = this.nextResult;
		this.nextTuple();
		return result;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.NaryIterator#reset()
	 */
	@Override
	public void reset() {
		super.reset();
		this.cache.clear();
	}
}
