package uk.ac.ox.cs.pdq.runtime.exec.iterator.open;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

import uk.ac.ox.cs.pdq.db.AccessMethod;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.db.wrappers.RelationAccessWrapper;
import uk.ac.ox.cs.pdq.util.ResetableIterator;
import uk.ac.ox.cs.pdq.util.Table;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.TupleType;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

/**
 * Open access over a relation
 * 
 * @author Julien Leblay
 */
public class OpenAccess extends OpenTupleIterator {


	/** The input table of the access. */
	private final RelationAccessWrapper relation;
	
	/** The access method to use. */
	private final AccessMethod binding;
	
	private final TupleType inputBindingType;

	/** The map of static input from their position within input tuples. */
	private final Map<Integer, TypedConstant<?>> staticInputs;

	/** Iterator over the output tuples. */
	protected Map<Tuple, ResetableIterator<Tuple>> outputs = null;
	
	/** Iterator over the output tuples. */
	protected ResetableIterator<Tuple> iterator = null;
	
	protected Tuple nextTuple = null;

	/**
	 * Instantiates a new join.
	 * 
	 * @param relation RelationAccessWrapper
	 * @param binding AccessMethod
	 */
	public OpenAccess(RelationAccessWrapper relation, AccessMethod binding) {
		this(relation, binding, ImmutableMap.<Integer, TypedConstant<?>>of());
	}

	/**
	 * Instantiates a new join.
	 * 
	 * @param relation RelationAccessWrapper
	 * @param binding AccessMethod
	 * @param staticInputs Map<Integer,TypedConstant<?>>
	 */
	public OpenAccess(RelationAccessWrapper relation, AccessMethod binding, Map<Integer, TypedConstant<?>> staticInputs) {
		super(inferInput(relation, binding, staticInputs.keySet()),
				Lists.<Typed>newArrayList(relation.getAttributes()));
		this.relation = relation;
		this.binding = binding;
		this.staticInputs = staticInputs;
		this.inputBindingType = TupleType.DefaultFactory.createFromTyped(completeInput(relation, binding));
	}
	
	/**
	 * @param relation RelationAccessWrapper
	 * @param binding AccessMethod
	 * @param staticInputs Set<Integer>
	 * @return List<Typed>
	 */
	private static List<Typed> inferInput(RelationAccessWrapper relation, AccessMethod binding, Set<Integer> staticInputs) {
		List<Typed> result = new ArrayList<>();
		for (Integer i: binding.getInputs()) {
			if (!staticInputs.contains(i - 1)) {
				result.add(relation.getAttributes().get(i - 1));
			}
		}
		return result;
	}
	
	
	/**
	 * @param relation RelationAccessWrapper
	 * @param binding AccessMethod
	 * @return List<Typed>
	 */
	private static List<Typed> completeInput(RelationAccessWrapper relation, AccessMethod binding) {
		List<Typed> result = new ArrayList<>();
		for (Integer i: binding.getInputs()) {
			result.add(relation.getAttributes().get(i - 1));
		}
		return result;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName());
		result.append('[').append(this.relation.getName()).append('/');
		result.append(this.binding).append(']');
//		result.append('(').append(this.child).append(')');
		return result.toString();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#open()
	 */
	@Override
	public void open() {
		this.outputs = Maps.newLinkedHashMap();
		this.open = true;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#reset()
	 */
	@Override
	public void reset() {
//		this.outputs = Maps.newLinkedHashMap();
		this.iterator.reset();
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.UnaryIterator#hasNext()
	 */
	@Override
	public boolean hasNext() {
		if (this.interrupted) {
			return false;
		}
		if (this.nextTuple != null) {
			return true;
		}
		this.nextTuple();
		return this.nextTuple != null;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#next()
	 */
	@Override
	public Tuple next() {
		if (this.eventBus != null) {
			this.eventBus.post(this);
		}
		Tuple result = this.nextTuple;
		this.nextTuple = null;
		if ((result == null && !this.hasNext())) {
			throw new NoSuchElementException("End of operator reached.");
		}
		return result;
	}

	public void nextTuple() {
		Preconditions.checkState(this.open);
		if (this.interrupted) {
			this.nextTuple = null;
		}
		this.nextTuple = null;
		if (this.iterator == null) {
			// If iterator has not been set at this stage, it implies all 
			// inputs this access are statically defined.
//			Preconditions.checkState(this.inputType.size() == 0);
			Tuple staticInput = this.makeInput(Tuple.EmptyTuple);
			Table inputs = new Table(this.relation.getInputAttributes(this.binding));
			inputs.appendRow(staticInput);
			this.iterator = this.relation.iterator(this.relation.getInputAttributes(this.binding), inputs.iterator());
			this.iterator.open();
			this.outputs.put(staticInput, this.iterator);
		}
		if (this.iterator.hasNext()) {
			this.nextTuple = this.iterator.next();
		}
	}
	
	/**
	 * @param input Tuple
	 */
	@Override
	public void bind(Tuple input) {
		Tuple combinedInputs = this.makeInput(input);
		this.iterator = this.outputs.get(combinedInputs);
		if (this.iterator == null) {
			Table inputs = new Table(this.relation.getInputAttributes(this.binding));
			inputs.appendRow(combinedInputs);
			this.iterator = this.relation.iterator(
					this.relation.getInputAttributes(this.binding),
					inputs.iterator());
			this.iterator.open();
			this.outputs.put(combinedInputs, this.iterator);
		} else {
			this.iterator.reset();
		}
	}

	/**
	 * @param dynamicInput
	
	 * @return an tuple obtained by mixing input from dynamicInput with inputs
	 * defined statically for this access.
	 */
	private Tuple makeInput(Tuple dynamicInput) {
		Object[] result = new Object[this.inputBindingType.size()];
		int j = 0, k = 0;
		for (int i : this.binding.getZeroBasedInputs()) {
			TypedConstant<?> staticInput = this.staticInputs.get(i);
			if (staticInput != null) {
				result[k++] = staticInput.getValue();
			} else {
				result[k++] = dynamicInput.getValue(j++);
			}
		}
		return this.inputBindingType.createTuple(result);
	}
	
	/**
	 * @return RelationAccessWrapper
	 */
	public RelationAccessWrapper getRelation() {
		return this.relation;
	}

	/**
	 * @return AccessMethod
	 */
	public AccessMethod getBinding() {
		return this.binding;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#deepCopy()
	 */
	@Override
	public OpenAccess deepCopy() {
		return new OpenAccess(this.relation, this.binding, this.staticInputs);
	}

	/**
	 * @see java.lang.AutoCloseable#close()
	 */
	@Override
	public void close() {
		this.outputs = null;
	}

	@Override
	public void interrupt() {
		this.interrupted = true;
	}
}