package uk.ac.ox.cs.pdq.runtime.exec.iterator.open;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.pdq.algebra.predicates.Predicate;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.LasyCachedIterator;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.TupleIterator;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.collect.Lists;

/**
 * OpenBindJoin is an implementation of a physical dependent open join.
 * 
 * @author Julien Leblay
 */
public class OpenBindJoin extends OpenJoin {

	protected final List<Integer> inputSidewaysMapping;

	protected final OpenTupleIterator left;
	protected final OpenTupleIterator right;
	protected TupleIterator cached;
	protected Map<Tuple, TupleIterator> iteratorCache = new LinkedHashMap<>();

	protected Tuple leftTuple = null;
	
	protected Tuple currentInput = null;
	
	/**
	 * Instantiates a new join.
	 * 
	 * @param left OpenTupleIterator
	 * @param right OpenTupleIterator
	 */
	public OpenBindJoin(OpenTupleIterator left, OpenTupleIterator right) {
		super(inferInputColumns(left, right),
				inferNaturalJoin(left, right), left, right);
		this.left = left;
		this.right = right;
		this.inputSidewaysMapping = inferInputMappings(left.getColumns(), right.getInputColumns());
	}
	
	/**
	 * Instantiates a new join.
	 * 
	 * @param left OpenTupleIterator
	 * @param right OpenTupleIterator
	 * @param pred Predicate
	 */
	public OpenBindJoin(OpenTupleIterator left, OpenTupleIterator right, Predicate pred) {
		super(inferInputColumns(left, right),
				pred, left, right);
		this.left = left;
		this.right = right;
		this.inputSidewaysMapping = inferInputMappings(left.getColumns(), right.getInputColumns());
	}
	
	/**
	 * Instantiates a new join.
	 * 
	 * @param left OpenTupleIterator
	 * @param right OpenTupleIterator
	 * @param pred Predicate
	 * @param sideWays List<Integer>
	 */
	public OpenBindJoin(OpenTupleIterator left, OpenTupleIterator right, Predicate pred, List<Integer> sideWays) {
		super(inferInputColumns(left, right),
				pred, left, right);
		this.left = left;
		this.right = right;
		this.inputSidewaysMapping = sideWays;
	}

	/**
	 * @param leftOutput
	 * @param rightInput
	 * @return a mapping from right input position to their position on the left output.
	 * Right input positions that comes from the parent are distinguished as negative numbers.
	*/
	private static List<Integer> inferInputMappings(
			List<Typed> leftOutput, List<? extends Typed> rightInput) {
		List<Integer> result = new ArrayList<>(rightInput.size());
		for (Typed t: rightInput) {
			result.add(leftOutput.indexOf(t));
		}
		return result;
	}
	
	/**
	 * Method inferInputColumns.
	 * @param left OpenTupleIterator
	 * @param right OpenTupleIterator
	 * @return List<Typed>
	 */
	protected static List<Typed> inferInputColumns(OpenTupleIterator left, OpenTupleIterator right) {
		List<Typed> result = Lists.newArrayList(left.getInputColumns());
		List<Typed> rInputs = right.getInputColumns();
		for (int i = 0, l = rInputs.size(); i < l; i++) {
			Typed t = rInputs.get(i);
			if (!left.getColumns().contains(t)) {
				result.add(rInputs.get(i));
			}
		}
		return result;
	}
	
	/**
	 * @param currentInput
	 * @param leftInput	
	 * @return an input tuple obtained by mixing inputs coming from the parent
	 * (currentInput) and the LHS (leftInput).
	 */
	protected Tuple project(Tuple currentInput, Tuple leftInput) {
		Object[] result = new Object[this.right.getInputType().size()];
		int j = 0;
		for (int i = 0, l = this.inputSidewaysMapping.size(); i < l; i++) {
			int position = this.inputSidewaysMapping.get(i);
			if (position >= 0) {
				result[i] = leftInput.getValue(this.inputSidewaysMapping.get(i));
			} else {
				result[i] = currentInput.getValue(j++);
			}
		}
		return this.right.getInputType().createTuple(result);
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.NaryIterator#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName());
		result.append(this.predicate).append('(');
		if (this.children != null) {
			for (TupleIterator child: this.children) {
				result.append(child.toString()).append(',');
			}
			result.deleteCharAt(result.length() - 1);
		}
		result.append(')');
		return result.toString();
	}

	/**
	 * Move the iterator forward and prepares the next tuple to be returned.
	 * 
	 */
	@Override
	protected void nextTuple() {
		if (this.cached != null) {
			while (this.cached.hasNext()) {
				Tuple t = this.leftTuple.appendTuple(this.cached.next());
				if (this.predicate.isSatisfied(t)) {
					this.nextTuple = t;
					return;
				}
			}
		}
		this.nextTuple = null;
		do {
			if (!this.left.hasNext()) {
				return;
			}
			this.leftTuple = this.left.next();
			
			Tuple rInput = this.project(this.currentInput, this.leftTuple);
			this.cached = this.iteratorCache.get(rInput);
			if (this.cached == null) {
				this.right.bind(rInput);
				this.cached = new LasyCachedIterator(this.right);
				this.iteratorCache.put(rInput, this.cached);
			} else {
				this.cached.reset();
			}
			while (this.cached.hasNext()) {
				Tuple t = this.leftTuple.appendTuple(this.cached.next());
				if (this.predicate.isSatisfied(t)) {
					this.nextTuple = t;
					return;
				}
			}
		} while (!this.cached.hasNext());
	}

	/**
	 * @return OpenBindJoin
	 * @see uk.ac.ox.cs.pdq.util.ResetableIterator#deepCopy()
	 */
	@Override
	public OpenBindJoin deepCopy() {
		return new OpenBindJoin(this.left.deepCopy(), this.right.deepCopy(), this.predicate);
	}
	
	/**
	 * @param t Tuple
	 */
	@Override
	public void bind(Tuple t) {
		this.left.bind(this.project(this.left, t));
		this.currentInput = t;
	}
}