package uk.ac.ox.cs.pdq.runtime.exec.iterator.open;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.base.Preconditions;

/**
 * OpenCrossProduct implements a n-ary cartesian product.
 * 
 * @author Julien Leblay
 */
public class OpenCrossProduct extends OpenNaryIterator {

	/** A stack of tuple fragment, used in the incremental computation of the cross product. */
	protected Deque<Tuple> tupleStack = new ArrayDeque<>();
	
	/** Determines whether the operator is known to have an empty result. */
	protected boolean isEmpty = false;
	
	/** The next tuple to return. */
	protected Tuple nextTuple = null;
	
	/**
	 * Instantiates a new cross product.
	 * 
	 * @param children
	 *            the children
	 * @param inputs List<Typed>
	 */
	public OpenCrossProduct(List<Typed> inputs, OpenTupleIterator... children) {
		super(inputs, children);
	}

	/**
	 * Instantiates a new cross product.
	 * 
	 * @param children
	 *            the children
	 * @param inputs List<Typed>
	 */
	public OpenCrossProduct(List<Typed> inputs, List<OpenTupleIterator> children) {
		super(inputs, children);
	}

	/* (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.Join#open()
	 */
	@Override
	public void open() {
		for (OpenTupleIterator child: this.children) {
			child.open();
		}
		this.nextTuple = null;
		this.open = true;
	}

	/**
	 * Computes the next tuple in the cross product.
	 * 
	 * @param i
	 *            the level at which to start the computation
	 * @return true, if the tuple fragment at the given level could be completed.
	 */
	protected boolean nextInCrossProduct(int i) {
		if (i < this.children.size()) {
			OpenTupleIterator child = this.children.get(i);
			if (this.tupleStack.size() > i) {
				if (this.nextInCrossProduct(i + 1)) {
					return true;
				}
			}
			while (child.hasNext()) {
				this.tupleStack.push(child.next());
				if (this.nextInCrossProduct(i + 1)) {
					return true;
				}
			}
			if (i > 0) {
				child.reset();
			}
			if (!this.tupleStack.isEmpty()) {
				this.tupleStack.pop();
			}
			return false;
		}
		return true;
	}

	/**
	 * Finds the next tuple to return. If new tuple can be obtain for this 
	 * operator, the nextTuple attribute will be set to null.
	 * 
	 */
	protected void nextTuple() {
		if (this.interrupted) {
			this.nextTuple = null;
			return;
		}
		if (this.nextInCrossProduct(0)) {
			this.nextTuple = Tuple.EmptyTuple;
			for (Iterator<Tuple> it = this.tupleStack.descendingIterator(); it.hasNext();) {
				this.nextTuple = this.nextTuple.appendTuple(it.next());
			}
			this.tupleStack.pop();
		} else {
			this.nextTuple = null;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#deepCopy()
	 */
	@Override
	public OpenCrossProduct deepCopy() {
		List<OpenTupleIterator> clones = new ArrayList<>();
		for (OpenTupleIterator child: this.children) {
			clones.add(child.deepCopy());
		}
		return new OpenCrossProduct(this.inputColumns, clones);
	}

	/* (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#hasNext()
	 */
	@Override
	public boolean hasNext() {
		Preconditions.checkArgument(this.open);
		if (this.nextTuple != null) {
			return true;
		}
		if (this.isEmpty) {
			return false;
		}
		this.nextTuple();
		return this.nextTuple != null;
	}

	/* (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#next()
	 */
	@Override
	public Tuple next() {
		Preconditions.checkArgument(this.open);
		if (this.eventBus != null) {
			this.eventBus.post(this);
		}
		Tuple result = this.nextTuple;
		this.nextTuple = null;
		if ((result == null && !this.hasNext()) || this.isEmpty) {
			throw new NoSuchElementException("End of operator reached.");
		}
		return result;
	}
}