package uk.ac.ox.cs.pdq.runtime.exec.iterator.open;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.pdq.runtime.exec.iterator.TupleIterator;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.TupleType;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.eventbus.EventBus;


/**
 * Superclass to all n-ary physical open operator.
 * Factorises the open, close, reset methods and the default constructor.
 * 
 * @author Julien Leblay
 */
public abstract class OpenNaryIterator extends OpenTupleIterator {

	/** The children. */
	protected List<OpenTupleIterator> children;

	/** For each child, a mapping from this operator's input positions to the child's. */
	protected Map<OpenTupleIterator, List<Integer>> relativeInputPositions;

	/**
	 * Instantiates a new operator.
	 * 
	 * @param children
	 *            the children
	 * @param inputs List<Typed>
	 */
	public OpenNaryIterator(List<Typed> inputs, OpenTupleIterator... children) {
		this(inputs, Lists.newArrayList(children));
	}
	
	/**
	 * Instantiates a new operator.
	 * 
	 * @param children
	 *            the children
	 * @param inputs List<Typed>
	 */
	public OpenNaryIterator(List<Typed> inputs,
			Collection<OpenTupleIterator> children) {
		this(inputs, inferType(children), inferColumns(children), children);
	}
	
	/**
	 * Instantiates a new operator.
	 * 
	 * @param children
	 *            the children
	 * @param inputs List<Typed>
	 * @param outputType TupleType
	 * @param outputTerms List<Typed>
	 */
	public OpenNaryIterator(List<Typed> inputs,
			TupleType outputType, List<Typed> outputTerms,
			Collection<OpenTupleIterator> children) {
		super(inputs, outputType, outputTerms);
		Preconditions.checkArgument(children != null);
		Preconditions.checkArgument(children.size() > 0);
		this.children = ImmutableList.copyOf(children);
		this.relativeInputPositions = ImmutableMap.copyOf(inferInputMappings(inputs, children));
	}
	
	/**
	 * @param children Collection<OpenTupleIterator>
	 * @return TupleType
	 */
	protected static TupleType inferType(Collection<OpenTupleIterator> children) {
		TupleType result = null;
		for (TupleIterator child: children) {
			if (result == null) {
				result = child.getType();
			} else {
				result = result.append(child.getType());
			}
		}
		return result;
	}
	
	/**
	 * @param children Collection<OpenTupleIterator>
	 * @return List<Typed>
	 */
	protected static List<Typed> inferColumns(Collection<OpenTupleIterator> children) {
		List<Typed> result = Lists.newArrayList();
		for (TupleIterator child: children) {
			result.addAll(child.getColumns());
		}
		return result;
	}
	
	/**
	 * @param inputTerms List<? extends Typed>
	 * @param operators Collection<OpenTupleIterator>
	 * @return Map<OpenTupleIterator,List<Integer>>
	 */
	private static Map<OpenTupleIterator, List<Integer>> inferInputMappings(
			List<? extends Typed> inputTerms, Collection<OpenTupleIterator> operators) {
		Map<OpenTupleIterator, List<Integer>> result = Maps.newLinkedHashMap();
		List<Typed> unassigned = Lists.newArrayList(inputTerms);
//		int cumul = 0;
		for (OpenTupleIterator op: operators) {
			List<Typed> inputs = op.getInputColumns();
			List<Integer> positions = new ArrayList<>(inputs.size());
			for (Typed t: inputs) {
				int position = inputTerms.indexOf(t);
				if (position >= 0) {
					positions.add(position);
					unassigned.remove(t);
				}
			}
			result.put(op, positions);
//			cumul += op.getColumns().size();
		}
		if (!unassigned.isEmpty()) {
			throw new IllegalStateException("Inconsistent input mapping");
		}
		return result;
	}
	
	/**
	 * @param op OpenTupleIterator
	 * @param input Tuple
	 * @return Tuple
	 */
	protected Tuple project(OpenTupleIterator op, Tuple input) {
		Object[] result = new Object[op.getInputType().size()];
		List<Integer> positions = this.relativeInputPositions.get(op);
		for (int i = 0, l = positions.size(); i < l; i++) {
			result[i] = input.getValue(positions.get(i));
		}
		return op.getInputType().createTuple(result);
	}

	/**
	 * Gets the children.
	 * 
	 * @return the children
	 */
	public List<OpenTupleIterator> getChildren() {
		return this.children;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(this.getClass().getSimpleName()).append('(');
		if (this.children != null && !this.children.isEmpty()) {
			for (OpenTupleIterator child: this.children) {
				result.append(child.toString()).append(',');
			}
			result.deleteCharAt(result.length() - 1);
		}
		result.append(')');
		return result.toString();
	}
	
	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#setEventBus(com.google.common.eventbus.EventBus)
	 */
	@Override
	public void setEventBus(EventBus eb) {
		super.setEventBus(eb);
		for (TupleIterator child: this.children) {
			child.setEventBus(eb);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#open()
	 */
	@Override
	public void open() {
		for (OpenTupleIterator child: this.children) {
			child.open();
		}
		this.open = true;
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#close()
	 */
	@Override
	public void close() {
		for (OpenTupleIterator child: this.children) {
			child.close();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#reset()
	 */
	@Override
	public void reset() {
		for (OpenTupleIterator child: this.children) {
			child.reset();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#interrupt()
	 */
	@Override
	public void interrupt() {
		this.interrupted = true;
		for (OpenTupleIterator child: this.children) {
			child.interrupt();
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.b2.TupleIterator#bind(uk.ac.ox.cs.pdq.structures.Tuple)
	 */
	@Override
	public void bind(Tuple t) {
		for (OpenTupleIterator child: this.children) {
			Tuple in = this.project(child, t);
			child.bind(in);
		}
	}
}
