package uk.ac.ox.cs.pdq.runtime.exec.iterator.open;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.Iterator;
import java.util.List;

import uk.ac.ox.cs.pdq.algebra.predicates.Predicate;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.Typed;

/**
 * Open nested loop implementation for n-ary joins.
 * 
 * @author Julien Leblay
 */
public class OpenNestedLoopJoin extends OpenJoin {

	/**
	 * A stack of tuple fragment, used in the incremental computation of the 
	 * cross product.
	 */
	protected Deque<Tuple> tupleStack = new ArrayDeque<>();
	
	protected boolean hasNext = true;
	
	/**
	 * Instantiates a new nested loop multi-join.
	 * 
	 * @param children
	 *            the children
	 * @param inputs List<Typed>
	 */
	public OpenNestedLoopJoin(List<Typed> inputs, OpenTupleIterator... children) {
		this(inputs, inferNaturalJoin(children), children);
	}

	/**
	 * Instantiates a new nested loop multi-join.
	 * 
	 * @param children
	 *            the children
	 * @param inputs List<Typed>
	 */
	public OpenNestedLoopJoin(List<Typed> inputs, List<OpenTupleIterator> children) {
		this(inputs,
				inferNaturalJoin(children.toArray(new OpenTupleIterator[children.size()])),
				children.toArray(new OpenTupleIterator[children.size()]));
	}
	
	/**
	 * Instantiates a new nested loop multi-join.
	 * 
	 * @param children
	 *            the children
	 * @param inputs List<Typed>
	 * @param predicate Predicate
	 */
	public OpenNestedLoopJoin(
			List<Typed> inputs, Predicate predicate,
			OpenTupleIterator... children) {
		super(inputs, predicate, children);
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.NaryIterator#open()
	 */
	@Override
	public void open() {
		super.open();
		this.nextTuple = null;
	}

	/**
	 * Computes the next tuple in the cross product.
	 * 
	 * @param i
	 *            the level at which to start the computation
	 * @return true, if the tuple fragment at the given level could be completed.
	 */
	protected boolean nextInCrossProduct(int i) {
		if (i < this.children.size()) {
			OpenTupleIterator child = this.children.get(i);
			if (this.tupleStack.size() > i) {
				if (this.nextInCrossProduct(i + 1)) {
					return true;
				}
			}
			while (child.hasNext()) {
				this.tupleStack.push(child.next());
				if (this.nextInCrossProduct(i + 1)) {
					return true;
				}
			}
			child.reset();
			if (!this.tupleStack.isEmpty()) {
				this.tupleStack.pop();
			}
			return false;
		}
		return true;
	}

	/**
	 * Finds the next tuple to return. If new tuple can be obtain for this 
	 * operator, the nextTuple attribute will be set to null.
	 * 
	 */
	@Override
	protected void nextTuple() {
		if (this.interrupted) {
			this.nextTuple = null;
			return;
		}
		do  {
			if (this.nextInCrossProduct(0)) {
				this.nextTuple = Tuple.EmptyTuple;
				for (Iterator<Tuple> it = this.tupleStack.descendingIterator(); it.hasNext();) {
					this.nextTuple = this.nextTuple.appendTuple(it.next());
				}
				this.tupleStack.pop();
			} else {
				this.nextTuple = null;
				break;
			}
		} while (!this.predicate.isSatisfied(this.nextTuple));
	}

	/*
	 * (non-Javadoc)
	 * @see uk.ac.ox.cs.pdq.runtime.plan.physical.iterator.TupleIterator#deepCopy()
	 */
	@Override
	public OpenNestedLoopJoin deepCopy() {
		List<OpenTupleIterator> clones = new ArrayList<>();
		for (OpenTupleIterator child: this.children) {
			clones.add(child.deepCopy());
		}
		return new OpenNestedLoopJoin(this.getInputColumns(), clones);
	}
}