package uk.ac.ox.cs.pdq.test.runtime.plan.iterator;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import uk.ac.ox.cs.pdq.algebra.predicates.ConjunctivePredicate;
import uk.ac.ox.cs.pdq.algebra.predicates.ConstantEqualityPredicate;
import uk.ac.ox.cs.pdq.db.Attribute;
import uk.ac.ox.cs.pdq.db.TypedConstant;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.AsymmetricMemoryHashJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Count;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.CrossProduct;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.EmptyIterator;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.IsEmpty;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.MemoryScan;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.NestedLoopJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Projection;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.Selection;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.SymmetricMemoryHashJoin;
import uk.ac.ox.cs.pdq.runtime.exec.iterator.TupleIterator;
import uk.ac.ox.cs.pdq.util.Tuple;
import uk.ac.ox.cs.pdq.util.TupleType;
import uk.ac.ox.cs.pdq.util.Typed;

import com.google.common.collect.Lists;

public class IteratorTest {

	final Collection<Tuple> empty = new LinkedList<>();
	final TupleType booleanTuple = TupleType.DefaultFactory.create(Boolean.class);
	final TupleType intTuple = TupleType.DefaultFactory.create(Integer.class);

	List<Typed> columns;
	TupleType type;
	Collection<Tuple> tuples;
	
	@Before
	public void prepare() {
		this.columns = Lists.<Typed>newArrayList(new Attribute(String.class, "Col1"), new Attribute(Integer.class, "Col2"), new Attribute(Double.class, "Col3"));
		this.type = TupleType.DefaultFactory.create(String.class, Integer.class, Double.class);
		this.tuples = new LinkedList<>();
		this.tuples.add(this.type.createTuple("A", 1, 10.0));
		this.tuples.add(this.type.createTuple("B", -2, 8.0));
		this.tuples.add(this.type.createTuple("C", 3, 6.0));
		this.tuples.add(this.type.createTuple("D", -4, 4.0));
		this.tuples.add(this.type.createTuple("E", 5, 2.0));
		this.tuples.add(this.type.createTuple("F", -6, 0.0));
		this.tuples.add(this.type.createTuple("G", 7, -1.0));
		this.tuples.add(this.type.createTuple("I", -8, -3.0));
		this.tuples.add(this.type.createTuple("J", 9, -5.0));
		this.tuples.add(this.type.createTuple("K", -10, -7.0));
	}
	
	@Test
	public void testScan() {
		try (TupleIterator it = new MemoryScan(this.columns, this.tuples)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				Assert.assertTrue(this.tuples.contains(it.next()));
				i++;
			}
			Assert.assertEquals(this.tuples.size(), i);
		}
	}
	
	@Test
	public void testEmptyIterator() {
		try (TupleIterator it = new EmptyIterator()) {
			it.open();
			Assert.assertFalse(it.hasNext());
		}
	}
	
	@Test
	public void testIsEmptyFalse() {
		try (TupleIterator it = new IsEmpty(new MemoryScan(this.columns, this.tuples))) {
			it.open();
			Assert.assertTrue(it.hasNext());
			Assert.assertEquals(this.booleanTuple.createTuple(false), it.next());
		}
	}
	
	@Test
	public void testIsEmptyTrue() {
		try (TupleIterator it = new IsEmpty(new MemoryScan(this.columns, this.empty))) {
			it.open();
			Assert.assertTrue(it.hasNext());
			Assert.assertEquals(this.booleanTuple.createTuple(true), it.next());
		}
	}
	
	@Test
	public void testCount() {
		try (TupleIterator it = new Count(new MemoryScan(this.columns, this.tuples))) {
			it.open();
			Assert.assertTrue(it.hasNext());
			Assert.assertEquals(this.intTuple.createTuple(this.tuples.size()), it.next());
		}
	}
	
	@Test
	public void testCountEmpty() {
		try (TupleIterator it = new Count(new MemoryScan(this.columns, this.empty))) {
			it.open();
			Assert.assertTrue(it.hasNext());
			Assert.assertEquals(this.intTuple.createTuple(this.empty.size()), it.next());
		}
	}
	
	@Test
	public void testCartesianProduct3Empty() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.empty); 
				TupleIterator c2 = new MemoryScan(this.columns, this.empty); 
				TupleIterator c3 = new MemoryScan(this.columns, this.empty); 
				TupleIterator it = new CrossProduct(c1, c2, c3)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testCartesianProductEmpty1() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.empty); 
				TupleIterator c2 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c3 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new CrossProduct(c1, c2, c3)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testCartesianProductEmpty2() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c2 = new MemoryScan(this.columns, this.empty); 
				TupleIterator c3 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new CrossProduct(c1, c2, c3)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testCartesianProductEmpty3() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c2 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c3 = new MemoryScan(this.columns, this.empty); 
				TupleIterator it = new CrossProduct(c1, c2, c3)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testCartesianProductNonEmpty() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c2 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c3 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new CrossProduct(c1, c2, c3)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			int size = this.tuples.size();
			Assert.assertEquals(size * size * size, i);
		}
	}
	
	@Test
	public void testNestLoopEmpty1() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.empty); 
				TupleIterator c2 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new NestedLoopJoin(c1, c2)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testNestLoopEmpty2() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c2 = new MemoryScan(this.columns, this.empty); 
				TupleIterator it = new NestedLoopJoin(c1, c2)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testNestLoop() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c2 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new NestedLoopJoin(c1, c2)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			int size = this.tuples.size();
			Assert.assertEquals(size, i);
		}
	}
	
	@Test
	public void testAssymetricHashJoinEmpty1() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.empty); 
				TupleIterator c2 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new AsymmetricMemoryHashJoin(c1, c2)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testAssymetricHashJoinEmpty2() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c2 = new MemoryScan(this.columns, this.empty); 
				TupleIterator it = new AsymmetricMemoryHashJoin(c1, c2)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testAssymetricHashJoin() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c2 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new AsymmetricMemoryHashJoin(c1, c2)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(this.tuples.size(), i);
		}
	}
	
	@Test
	public void testSymmetricHashJoin() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator c2 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new SymmetricMemoryHashJoin(c1, c2)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(this.tuples.size(), i);
		}
	}

	@Test
	public void testSelectionOneResult() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new Selection(c1, new ConstantEqualityPredicate(0, new TypedConstant<>("C")))) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(1, i);
		}
	}
	
	@Test
	public void testSelectionNoPredicate() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new Selection(c1, new ConjunctivePredicate<>())) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(this.tuples.size(), i);
		}
	}
	
	@Test
	public void testSelectionEmptyResult() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new Selection(c1, new ConstantEqualityPredicate(1, new TypedConstant<>("a")))) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testSelectionEmptyChild() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.empty); 
				TupleIterator it = new Selection(c1, new ConstantEqualityPredicate(1, new TypedConstant<>("A")))) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				it.next();
				i++;
			}
			Assert.assertEquals(0, i);
		}
	}
	
	@Test
	public void testProjectionRetainAll() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new Projection(Lists.newArrayList(this.columns), c1)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				Assert.assertTrue(this.tuples.contains(it.next()));
				i++;
			}
			Assert.assertEquals(this.tuples.size(), i);
		}
	}
	
	@Test
	public void testProjectionRemoveAll() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new Projection(Lists.<Typed>newArrayList(), c1)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				Assert.assertEquals(Tuple.EmptyTuple, it.next());
				i++;
			}
			Assert.assertEquals(this.tuples.size(), i);
		}
	}
	
	@Test
	public void testProjectionFirstColumn() {
		try (	TupleIterator c1 = new MemoryScan(this.columns, this.tuples); 
				TupleIterator it = new Projection(Lists.<Typed>newArrayList(new Attribute(String.class, "Col1")), c1)) {
			int i = 0;
			it.open();
			while (it.hasNext()) {
				Assert.assertTrue("ABCDEFGHIJK".contains((String) it.next().getValue(0)));
				i++;
			}
			Assert.assertEquals(this.tuples.size(), i);
		}
	}
}
